Package: CytoPipeline
Title: Automation and visualization of flow cytometry data analysis
        pipelines
Version: 1.13.0
Authors@R: 
    c(person(given = "Philippe",
             family = "Hauchamps",
             role = c("aut", "cre"),
             email = "philippe.hauchamps@uclouvain.be",
             comment = c(ORCID = "0000-0003-2865-1852")),
      person(given = "Laurent", family = "Gatto",
             email = "laurent.gatto@uclouvain.be",
             role = "aut",
             comment = c(ORCID = "0000-0002-1520-2268")),
      person(given = "Dan",
             family = "Lin",
             role = "ctb",
             email = "dan.8.lin@gsk.com"))
Description: This package provides support for automation and visualization of
 flow cytometry data analysis pipelines. In the current state, the package
 focuses on the preprocessing and quality control part. The framework is based
 on two main S4 classes, i.e. CytoPipeline and CytoProcessingStep. The pipeline
 steps are linked to corresponding R functions - that are either provided in
 the CytoPipeline package itself, or exported from a third party package,
 or coded by the user her/himself. The processing steps need to be specified
 centrally and explicitly using either a json input file or through step by step
 creation of a CytoPipeline object with dedicated methods. After having run the
 pipeline, obtained results at all steps can be retrieved and visualized thanks
 to file caching (the running facility uses a BiocFileCache implementation).
 The package provides also specific visualization tools like
 pipeline workflow summary display, and 1D/2D comparison plots of obtained
 flowFrames at various steps of the pipeline.
License: GPL-3
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
BugReports: https://github.com/UCLouvain-CBIO/CytoPipeline/issues
URL: https://uclouvain-cbio.github.io/CytoPipeline
biocViews: FlowCytometry, Preprocessing, QualityControl, WorkflowStep,
        ImmunoOncology, Software, Visualization
Collate: 'CytoPipeline-functions.R' 'CytoPipeline-package.R'
        'CytoPipelineClass.R' 'CytoProcessingStep.R'
        'CytoProcessingStepImplementations.R' 'data.R' 'gating.R'
        'utils.R' 'ggplots.R'
Depends: R (>= 4.4)
Imports: methods, stats, utils, withr, rlang, ggplot2 (>= 3.4.1),
        ggcyto, BiocFileCache, BiocParallel, flowCore, PeacoQC, flowAI,
        diagram, jsonlite, scales
Suggests: testthat (>= 3.0.0), vdiffr, diffviewer, knitr, rmarkdown,
        BiocStyle, reshape2, dplyr, CytoPipelineGUI
VignetteBuilder: knitr
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/CytoPipeline
git_branch: devel
git_last_commit: 723595f
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:08:13 UTC; biocbuild
Author: Philippe Hauchamps [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2865-1852>),
  Laurent Gatto [aut] (ORCID: <https://orcid.org/0000-0002-1520-2268>),
  Dan Lin [ctb]
Maintainer: Philippe Hauchamps <philippe.hauchamps@uclouvain.be>

# Generated by roxygen2: do not edit by hand

S3method(as.list,CytoPipeline)
S3method(as.list,CytoProcessingStep)
export("experimentName<-")
export("pData<-")
export("sampleFiles<-")
export(CytoProcessingStep)
export(addProcessingStep)
export(aggregateAndSample)
export(appendCellID)
export(applyScaleTransforms)
export(areFluoCols)
export(areSignalCols)
export(as.json.CytoProcessingStep)
export(buildCytoPipelineFromCache)
export(checkCytoPipelineConsistencyWithCache)
export(cleanProcessingSteps)
export(collectNbOfRetainedEvents)
export(compensateFromMatrix)
export(computeScatterChannelsLinearScale)
export(deleteCytoPipelineCache)
export(estimateScaleTransforms)
export(execute)
export(executeProcessingStep)
export(experimentName)
export(export2JSONFile)
export(findTimeChannel)
export(from.json.CytoProcessingStep)
export(getAcquiredCompensationMatrix)
export(getCPSARGS)
export(getCPSFUN)
export(getCPSName)
export(getChannelNamesFromMarkers)
export(getCytoPipelineExperimentNames)
export(getCytoPipelineFlowFrame)
export(getCytoPipelineObjectFromCache)
export(getCytoPipelineObjectInfos)
export(getCytoPipelineScaleTransform)
export(getFCSFileName)
export(getNbProcessingSteps)
export(getProcessingStep)
export(getProcessingStepNames)
export(getTransfoParams)
export(ggplotEvents)
export(ggplotFilterEvents)
export(ggplotFlowRate)
export(pData)
export(plotCytoPipelineProcessingQueue)
export(qualityControlFlowAI)
export(qualityControlPeacoQC)
export(readRDSObject)
export(readSampleFiles)
export(removeChannels)
export(removeDeadCellsManualGate)
export(removeDebrisManualGate)
export(removeDoubletsCytoPipeline)
export(removeMarginsPeacoQC)
export(removeProcessingStep)
export(resetCellIDs)
export(runCompensation)
export(sampleDisplayNames)
export(sampleFiles)
export(sampleNameFromDisplayName)
export(showProcessingSteps)
export(singletsGate)
export(subsample)
export(updateMarkerName)
export(writeFlowFrame)
exportClasses(CytoPipeline)
exportClasses(CytoProcessingStep)
exportMethods(CytoPipeline)
exportMethods(show)
import(ggplot2)
importFrom(ggcyto,scale_x_flowjo_biexp)
importFrom(ggcyto,scale_x_logicle)
importFrom(ggcyto,scale_y_flowjo_biexp)
importFrom(ggcyto,scale_y_logicle)
importFrom(rlang,.data)
importFrom(utils,head)
importMethodsFrom(methods,show)

## Automation and visualization of flow cytometry data analysis pipelines

[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check-bioc](https://github.com/UCLouvain-CBIO/CytoPipeline/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/UCLouvain-CBIO/CytoPipeline/actions?query=workflow%3AR-CMD-check-bioc)
[![codecov.io](https://codecov.io/github/UCLouvain-CBIO/CytoPipeline/coverage.svg?branch=main)](https://codecov.io/github/UCLouvain-CBIO/CytoPipeline?branch=main)
[![license](https://img.shields.io/badge/license-GPL3.0-blue)](https://opensource.org/licenses/GPL-3.0)

### What is CytoPipeline?

`CytoPipeline` is a package that provides support for automation and 
visualization of flow cytometry data analysis pipelines. In the current 
state, the package focuses on the preprocessing and quality control part.   

The framework is based on two main S4 classes, i.e. `CytoPipeline` and 
`CytoProcessingStep`. The `CytoProcessingStep` defines the link between
pipeline step names and corresponding R functions that are either provided in
the CytoPipeline package itself, or exported from a third party package,
or coded by the user her/himself. The processing steps need to be specified 
centrally and explicitly using either a json input file or through step by step 
creation of a `CytoPipeline` object with dedicated methods.  

After having run the pipeline, obtained results at all steps can be retrieved 
and visualized thanks to file caching (the running facility uses a 
`BiocFileCache` implementation). The package provides also specific 
visualization tools like pipeline workflow summary display, and 1D/2D comparison 
plots of obtained flowFrames at various steps of the pipeline.

### License

The `CytoPipeline` code is provided under [GPL license version 3.0 or 
higher](https://opensource.org/licenses/GPL-3.0). The documentation, 
including the manual pages and the vignettes, are distributed under a [CC BY-SA 
4.0 license](https://creativecommons.org/licenses/by-sa/4.0/).

### Citation

If you use `Cytopipeline` in your research, please use the following citation:

>Hauchamps P, Bayat B, Delandre S, Hamrouni M, Toussaint M, Temmerman S, 
> Lin D, Gatto L (2024).
> “CytoPipeline and CytoPipelineGUI: a Bioconductor R package suite 
> for building and visualizing automated pre-processing pipelines 
> for flow cytometry data.” _BMC Bioinformatics_, *25*(1), 80. 
> doi:10.1186/s12859-024-05691-z <https://doi.org/10.1186/s12859-024-05691-z>.

or run `citation("CytoPipeline")` to get the bibtex entry.

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).


#' @title Clean doublet events from flow cytometry data
#' @description will adjust a polygon gate aimed at cleaning doublet events
#' from the flowFrame.
#' The main idea is to use the ratio between the two indicated channel as an
#' indicator and select only the events for which this ratio is 'not too far'
#' from the median ratio. More specifically, the computed ratio is ch1/(1+ch2).
#' However, instead of looking at a constant range of this ratio, as is done in
#' PeacoQC::removeDoublets(), which leads to a semi-conic gate,
#' we apply a parallelogram shaped gate, by keeping a constant range of channel
#' 2 intensity, based on the target ratio range at the mid value of channel 1.
#' @param ff A flowCore::flowframe that contains flow cytometry data.
#' @param filterId the name for the filter that is returned
#' @param channel1 The first channel that will be used to determine the doublet
#' events. Default is "FSC-A"
#' @param channel2 The second channels that will be used to determine the
#' doublet events. Default is "FSC-H"
#' @param nmad Bandwidth above the ratio allowed (cells are kept if their ratio
#' is smaller than the median ratio + nmad times the median absolute deviation
#' of the ratios). Default is 4.
#' @param verbose If set to TRUE, the median ratio and width will be printed.
#' Default is FALSE.
#'
#' @return This function returns a flowCore::polygonGate.
#' @export
#'
#' @examples
#' 
#' data(OMIP021Samples)
#'
#' # simple example with one single singlets gate filter 
#' # FSC-A and FSC-H channels are used by default
#' 
#' mySingletsGate <- singletsGate(OMIP021Samples[[1]], nmad = 3)
#' 
#' selectedSinglets <- flowCore::filter(
#'     OMIP021Samples[[1]],
#'     mySingletsGate)
#' 
#' ff_l <- flowCore::Subset(OMIP021Samples[[1]], selectedSinglets)
#' 
#' linRange <- c(0, 250000)
#' 
#' ggplotFilterEvents(
#'     ffPre = OMIP021Samples[[1]],
#'     ffPost = ff_l,
#'     xChannel = "FSC-A", xLinearRange = linRange,
#'     yChannel = "FSC-H", yLinearRange = linRange)
#' 
#' # application of two singlets gates one after the other
#' 
#' singletsGate1 <- singletsGate(OMIP021Samples[[1]], nmad = 3)
#' singletsGate2 <- singletsGate(OMIP021Samples[[1]],
#'                               channel1 = "SSC-A",
#'                               channel2 = "SSC-H",
#'                               filterId = "Singlets2")
#' 
#' singletCombinedGate <- singletsGate1 & singletsGate2
#' 
#' selectedSinglets <- flowCore::filter(
#'     OMIP021Samples[[1]],
#'     singletCombinedGate)
#' 
#' ff_l <- flowCore::Subset(OMIP021Samples[[1]], selectedSinglets)
#' 
#' ggplotFilterEvents(
#'     ffPre = OMIP021Samples[[1]],
#'     ffPost = ff_l,
#'     xChannel = "FSC-A", xLinearRange = linRange,
#'     yChannel = "FSC-H", yLinearRange = linRange)
#' 
#' ggplotFilterEvents(
#'     ffPre = OMIP021Samples[[1]],
#'     ffPost = ff_l,
#'     xChannel = "SSC-A", xLinearRange = linRange,
#'     yChannel = "SSC-H", yLinearRange = linRange)
#' 
singletsGate <- function(ff, filterId = "Singlets",
                         channel1 = "FSC-A", channel2 = "FSC-H", nmad = 4,
                         verbose = FALSE) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }

    ratio <- flowCore::exprs(ff)[, channel1] / (1 + flowCore::exprs(ff)[
        ,
        channel2
    ])

    ratioMedian <- stats::median(ratio)
    ratioMad <- stats::mad(ratio)
    ch1Median <- stats::median(flowCore::exprs(ff)[, channel1])
    ch1Min <- min(flowCore::exprs(ff)[, channel1])
    ch1Max <- max(flowCore::exprs(ff)[, channel1])
    ch2RefPoint <- ch1Median / ratioMedian - 1
    ch2Min <- ch1Median / (ratioMedian + nmad * ratioMad) - 1
    ch2Range <- ch2RefPoint - ch2Min
    if (verbose) {
        message(
            "Median ratio: ", ratioMedian, ", width: ", nmad * ratioMad,
            ", Ch1 median : ", ch1Median, ", Ch 2 range : ",
            ch2Range
        )
    }

    ch2TargetAtCh1Min <- ch1Min / ratioMedian - 1
    ch2TargetAtCh1Max <- ch1Max / ratioMedian - 1

    # browser()

    # parallelogram gate with vertical edges
    polygonGateMatrix <- matrix(
        data = c(
            ch1Min, ch1Min, ch1Max, ch1Max,
            ch2TargetAtCh1Min - ch2Range,
            ch2TargetAtCh1Min + ch2Range,
            ch2TargetAtCh1Max + ch2Range,
            ch2TargetAtCh1Max - ch2Range
        ),
        ncol = 2,
        dimnames = list(
            c(),
            c(channel1, channel2)
        )
    )


    singletsGate <- flowCore::polygonGate(
        filterId = filterId,
        .gate = polygonGateMatrix
    )

    return(singletsGate)
}

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

# obtain OMIP021UTSamples, light-weight version used specifically for these
# unit tests
path <- system.file("scripts",
                    package = "CytoPipeline"
)

source(file.path(path,"MakeOMIP021UTSamples.R"))

test_that("singletsGate works", {
    mySingletsGate <- singletsGate(OMIP021UTSamples[[1]])

    selectedSinglets <- flowCore::filter(
        OMIP021UTSamples[[1]],
        mySingletsGate
    )
    ff_l <- flowCore::Subset(OMIP021UTSamples[[1]], selectedSinglets)

    linRange <- c(0, 250000)
    p <- ggplotFilterEvents(
        ffPre = OMIP021UTSamples[[1]],
        ffPost = ff_l,
        seed = 1,
        xChannel = "FSC-A", xLinearRange = linRange,
        yChannel = "FSC-H", yLinearRange = linRange
    )

    vdiffr::expect_doppelganger("singletsGate default channels", fig = p)

    mySingletsGate <- singletsGate(OMIP021UTSamples[[1]], nmad = 3)

    selectedSinglets <- flowCore::filter(
        OMIP021UTSamples[[1]],
        mySingletsGate
    )
    ff_l <- flowCore::Subset(OMIP021UTSamples[[1]], selectedSinglets)

    p <- ggplotFilterEvents(
        ffPre = OMIP021UTSamples[[1]],
        ffPost = ff_l,
        seed = 1,
        xChannel = "FSC-A", xLinearRange = linRange,
        yChannel = "FSC-H", yLinearRange = linRange
    )

    vdiffr::expect_doppelganger(
        "singletsGate default channels with fixed nmad",
        fig = p
    )

    mySingletsGate <- singletsGate(OMIP021UTSamples[[1]],
        channel1 = "SSC-A",
        channel2 = "SSC-H"
    )

    selectedSinglets <- flowCore::filter(
        OMIP021UTSamples[[1]],
        mySingletsGate
    )
    ff_l <- flowCore::Subset(OMIP021UTSamples[[1]], selectedSinglets)

    p <- ggplotFilterEvents(
        ffPre = OMIP021UTSamples[[1]],
        ffPost = ff_l,
        seed = 1,
        xChannel = "SSC-A", xLinearRange = linRange,
        yChannel = "SSC-H", yLinearRange = linRange
    )

    vdiffr::expect_doppelganger("singletsGate selected channels", fig = p)

    # test application of two gates one after the other
    singletsGate1 <- singletsGate(OMIP021UTSamples[[1]], nmad = 3)
    singletsGate2 <- singletsGate(OMIP021UTSamples[[1]],
        channel1 = "SSC-A",
        channel2 = "SSC-H",
        filterId = "Singlets2"
    )

    singletCombinedGate <- singletsGate1 & singletsGate2

    selectedSinglets <- flowCore::filter(
        OMIP021UTSamples[[1]],
        singletCombinedGate
    )

    ff_l <- flowCore::Subset(OMIP021UTSamples[[1]], selectedSinglets)

    p1 <- ggplotFilterEvents(
        ffPre = OMIP021UTSamples[[1]],
        ffPost = ff_l,
        seed = 1,
        xChannel = "FSC-A", xLinearRange = linRange,
        yChannel = "FSC-H", yLinearRange = linRange
    )

    p2 <- ggplotFilterEvents(
        ffPre = OMIP021UTSamples[[1]],
        ffPost = ff_l,
        seed = 1,
        xChannel = "SSC-A", xLinearRange = linRange,
        yChannel = "SSC-H", yLinearRange = linRange
    )


    vdiffr::expect_doppelganger("singletsGates one after the other - fig1",
        fig = p1
    )
    vdiffr::expect_doppelganger("singletsGates one after the other - fig2",
        fig = p2
    )
})

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

##' @name handlingProcessingSteps
##' @title handling processing steps in CytoPipeline objects
##' @description functions to manipulate processing steps in processing queues
##' of CytoPipeline objects
##' @param x a CytoPipeline object
##' @param whichQueue selects the processing queue for which we manage the
##' processing steps
##' @examples 
##' 
##' rawDataDir <-
##'     system.file("extdata", package = "CytoPipeline")
##' experimentName <- "OMIP021_PeacoQC"
##' sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
##'                                              pattern = "Donor"))
##' transListPath <- 
##'     file.path(system.file("extdata", package = "CytoPipeline"), 
##'               "OMIP021_TransList.rds")
##' 
##' # main parameters : sample files and experiment name
##' pipelineParams <- list()
##' pipelineParams$experimentName <- experimentName
##' pipelineParams$sampleFiles <- sampleFiles
##' 
##' # create CytoPipeline object (no step defined yet)
##' pipL <- CytoPipeline(pipelineParams)
##' 
##' # add a processing step in scale tranformation queue
##' pipL <- addProcessingStep(pipL,
##'                           whichQueue = "scale transform",
##'                           CytoProcessingStep(
##'                               name = "scale_transform_read",
##'                               FUN = "readRDS",
##'                               ARGS = list(file = transListPath)
##'                           ))
##' 
##' getNbProcessingSteps(pipL, "scale transform") # returns 1
##' 
##' # add another processing step in scale transformation queue
##' pipL <- addProcessingStep(pipL,
##'                           whichQueue = "scale transform",
##'                           CytoProcessingStep(
##'                               name = "scale_transform_sum",
##'                               FUN = "sum",
##'                               ARGS = list()
##'                           )
##' )
##' 
##' getNbProcessingSteps(pipL, "scale transform") # returns 2
##' 
##' getProcessingStepNames(pipL, whichQueue = "scale transform")
##' 
##' # removes second processing step in scale transformation queue
##' pipL <- removeProcessingStep(pipL,
##'                              whichQueue = "scale transform",
##'                              index = 2)
##' 
##' # get processing step object
##' pS <- getProcessingStep(pipL, whichQueue = "scale transform", index = 1)
##' getCPSName(pS) #"scale_transform_read"
##' 
##' # add a processing step in pre-processing queue
##' pipL <- addProcessingStep(pipL,
##'                           whichQueue = "pre-processing",
##'                           CytoProcessingStep(
##'                               name = "pre-processing_sum",
##'                               FUN = "sum",
##'                               ARGS = list()
##'                           ))
##' getNbProcessingSteps(pipL, "scale transform") # returns 1
##' getNbProcessingSteps(pipL, "pre-processing") # returns also 1
##' 
##' showProcessingSteps(pipL, whichQueue = "scale transform")
##' showProcessingSteps(pipL, whichQueue = "pre-processing")
##' 
##' # cleans both processing queues
##' pipL <- cleanProcessingSteps(pipL)
##' pipL
NULL

##' @title handling processing steps in CytoPipeline objects
##' @param newPS the new processing step to be added (CytoProcessingStep object)
##' @describeIn handlingProcessingSteps adds a processing step in one of the
##' processing queues (at the end), returns the modified CytoPipeline object
##' @returns - for `addProcessingStep`: the updated CytoPipeline object
##'
##' @export
#'
addProcessingStep <- function(x,
                              whichQueue = c(
                                  "scale transform",
                                  "pre-processing"
                              ),
                              newPS) {
    stopifnot(c(
        inherits(x, "CytoPipeline"),
        inherits(newPS, "CytoProcessingStep")
    ))
    whichQueue <- match.arg(whichQueue)

    if (whichQueue == "scale transform") {
        ll <- length(x@scaleTransformProcessingQueue)
        for (j in seq_along(x@scaleTransformProcessingQueue)) {
            pS <- getProcessingStep(x, whichQueue = whichQueue, index = j)
            if (pS@name == newPS@name) {
                stop(
                    "There already exist a step with name '", newPS@name,
                    "' in ", whichQueue, " queue!"
                )
            }
        }
        x@scaleTransformProcessingQueue[[ll + 1]] <- newPS
    } else {
        ll <- length(x@flowFramesPreProcessingQueue)
        for (j in seq_along(x@flowFramesPreProcessingQueue)) {
            pS <- getProcessingStep(x, whichQueue = whichQueue, index = j)
            if (pS@name == newPS@name) {
                stop(
                    "There already exist a step with name '", newPS@name,
                    "' in ", whichQueue, " queue!"
                )
            }
        }
        x@flowFramesPreProcessingQueue[[ll + 1]] <- newPS
    }
    return(x)
}




##' @param index index of the processing step to remove
##' @describeIn handlingProcessingSteps removes a processing step from one of
##' the processing queues, returns the modified CytoPipeline object
##' @returns - for `removeProcessingStep`: the updated CytoPipeline object
##'
##' @export
##'
##'
removeProcessingStep <- function(x,
                                 whichQueue = c(
                                     "scale transform",
                                     "pre-processing"
                                 ),
                                 index) {
    if (!is.numeric(index)) {
        stop("index should be a numeric")
    }
    stopifnot(inherits(x, "CytoPipeline"))
    whichQueue <- match.arg(whichQueue)
    if (whichQueue == "scale transform") {
        ll <- length(x@scaleTransformProcessingQueue)
        if (index <= 0 || index > ll) {
            stop("index out of bound")
        }
        x@scaleTransformProcessingQueue <-
            x@scaleTransformProcessingQueue[-index]
    } else {
        ll <- length(x@flowFramesPreProcessingQueue)
        if (index <= 0 || index > ll) {
            stop("index out of bound")
        }
        x@flowFramesPreProcessingQueue <-
            x@flowFramesPreProcessingQueue[-index]
    }
    return(x)
}

##' @describeIn handlingProcessingSteps gets the number of processing
##' steps in a processing queue
##' @return - for `getNbProcessingSteps`: the number of processing steps present
##' in the target queue
##' @export
##'
getNbProcessingSteps <- function(x,
                                 whichQueue = c(
                                     "scale transform",
                                     "pre-processing"
                                 )) {
    stopifnot(inherits(x, "CytoPipeline"))
    whichQueue <- match.arg(whichQueue)

    if (whichQueue == "scale transform") {
        return(length(x@scaleTransformProcessingQueue))
    } else {
        return(length(x@flowFramesPreProcessingQueue))
    }
}

##' @describeIn handlingProcessingSteps gets a processing step at a
##' specific index of a processing queue
##' @returns - for `getProcessingStep`: the obtained CytoProcessingStep object
##' @export
##'
getProcessingStep <- function(x,
                              whichQueue = c(
                                  "scale transform",
                                  "pre-processing"
                              ),
                              index) {
    if (!is.numeric(index)) {
        stop("index should be a numeric")
    }
    stopifnot(inherits(x, "CytoPipeline"))
    whichQueue <- match.arg(whichQueue)
    if (whichQueue == "scale transform") {
        ll <- length(x@scaleTransformProcessingQueue)
        if (index <= 0 || index > ll) {
            stop("index out of bound")
        }
        return(x@scaleTransformProcessingQueue[[index]])
    } else {
        ll <- length(x@flowFramesPreProcessingQueue)
        if (index <= 0 || index > ll) {
            stop("index out of bound")
        }
        return(x@flowFramesPreProcessingQueue[[index]])
    }
    return(x)
}

##' @describeIn handlingProcessingSteps gets a character vector of all
##' processing step names of a specific processing queue
##' @returns - for `getProcessingStepNames`: the vector of step names
##' @export
##'
getProcessingStepNames <- function(x, whichQueue = c(
                                       "scale transform",
                                       "pre-processing"
                                   )) {
    nSteps <- getNbProcessingSteps(x, whichQueue)
    stepNames <- character()
    if (nSteps > 0) {
        stepNames <- vapply(seq_len(nSteps),
            FUN.VALUE = character(1),
            FUN = function(i) {
                pS <- getProcessingStep(x, whichQueue, i)
                pS@name
            }
        )
    }

    return(stepNames)
}

##' @describeIn handlingProcessingSteps deletes all processing steps in one
##' or both processing queues, returns the modified CytoPipeline object
##' @returns - for `cleanProcessingSteps`: the updated CytoPipeline object
##' @export
##'
cleanProcessingSteps <- function(x,
                                 whichQueue = c(
                                     "both", "scale transform",
                                     "pre-processing"
                                 )) {
    stopifnot(inherits(x, "CytoPipeline"))
    whichQueue <- match.arg(whichQueue)
    if (whichQueue == "scale transform" || whichQueue == "both") {
        x@scaleTransformProcessingQueue <- list()
    }
    if (whichQueue == "pre-processing" || whichQueue == "both") {
        x@flowFramesPreProcessingQueue <- list()
    }
    return(x)
}

##' @describeIn handlingProcessingSteps shows all processing steps in a
##' processing queue
##' @returns - for `showProcessingSteps`: nothing (only console display side
##' effect is required)
##' @export
##'
showProcessingSteps <- function(x,
                                whichQueue = c(
                                    "scale transform",
                                    "pre-processing"
                                )) {
    stopifnot(inherits(x, "CytoPipeline"))
    whichQueue <- match.arg(whichQueue)

    if (whichQueue == "scale transform") {
        queue <- x@scaleTransformProcessingQueue
        queueName <- "Scale transformations evaluation queue"
    } else {
        queue <- x@flowFramesPreProcessingQueue
        queueName <- "Flow frames pre-processing evaluation queue"
    }

    if (length(queue)) {
        cat(queueName, ":", length(queue), "processing step(s)\n")
        for (i in seq_along(queue)) {
            ps <- queue[[i]]
            cat(i, ":")
            show(ps)
        }
    } else {
        cat(queueName, "has no processing step\n")
    }
}

# internal function to remove ill-defined bfc entries ("type" col is NA)
# which could have been generated when a previous pipeline run 
# either crashed or was manually forced to stop in the middle of a step
makeBFCClean <- function(bfc){
    cacheInfo <- BiocFileCache::bfcinfo(bfc)
    if (nrow(cacheInfo) > 0) {
        toRemoveRids <- cacheInfo[is.na(cacheInfo$type),"rid"]
        if (nrow(toRemoveRids) > 0) {
            bfc <- BiocFileCache::bfcremove(bfc, toRemoveRids$rid)    
        }
    }
    invisible(bfc)
}


#' @title executing CytoPipeline object
#' @description this function triggers the execution of the processing queues of
#' a CytoPipeline object.
#' First, the scale tranform processing queue is run, taking
#' the set of sample names as an implicit first input. At the end of the queue,
#' a scale transform List is assumed to be created.
#' Second, the flowFrame pre-processing queue, reapeatedly for each sample file.
#' The scale transform list generated in the previous step is taken as implicit
#' input, together with the initial sample file. At the end of the queue run, a
#' pre-processed flowFrame is assumed to be generated.
#' No change is made on the input CytoPipeline object, all results are stored in
#' the cache.
#' @param x CytoPipeline object
#' @param path base path, a subdirectory with name equal to the experiment will
#' be created to store the output data, in particular the experiment cache
#' @param rmCache if TRUE, starts by removing the already existing cache
#' directory corresponding to the experiment
#' @param useBiocParallel if TRUE, use BiocParallel for computation of the
#' sample file pre-processing in parallel (one file per worker at a time).
#' Note the BiocParallel function used is `bplapply()`
#' @param BPPARAM if `useBiocParallel` is TRUE, sets the BPPARAM back-end to
#' be used for the computation. If not provided, will use the top back-end on 
#' the `BiocParallel::registered()` stack.
#' @param BPOPTIONS if `useBiocParallel` is TRUE, sets the BPOPTIONS to be 
#' passed to `bplapply()` function.   
#' Note that if you use a `SnowParams` back-end, you need to specify all   
#' the packages that need to be loaded for the different CytoProcessingStep   
#' to work properly (visibility of functions). As a minimum,    
#' the `flowCore` package needs to be loaded.  
#' (hence the default `BPOPTIONS = bpoptions(packages = c("flowCore"))` )
#' @param saveLastStepFF if TRUE, save the final result of the pre-processing, 
#' for each file. 
#' By convention, these output files are stored in 
#' `path`/`x@experimentName`/output/, the file names used are the same 
#' as the initial fcs file basenames, concatenated with `saveFFSuffix`, 
#' and with file extension corresponding to `saveFFFormat`.
#' @param saveFFSuffix FF file name suffix
#' @param saveFFFormat either `fcs` or `csv` 
#' @param saveFFCsvUseChannelMarker if TRUE (default), converts the channels   
#' to the corresponding marker names (where the Marker is not NA).  
#' This setting is only applicable to export in csv format.
#' @param saveScaleTransforms if TRUE (default FALSE), save on disk 
#' (in RDS format) the `flowCore::transformList` object obtained after running  
#' the scaleTransform processing queue. The file name is hardcoded to
#' `path`/`experimentName`/`RDS`/`scaleTransformList.rds`
#' @returns nothing
#' @export
#' 
#' @examples
#' 
#' ### *** EXAMPLE 1: building CytoPipeline step by step *** ###
#' 
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' experimentName <- "OMIP021_PeacoQC"
#' sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
#'                                              pattern = "Donor"))
#'                                              
#' outputDir <- base::tempdir()
#' 
#' # main parameters : sample files and output files
#' pipelineParams <- list()
#' pipelineParams$experimentName <- experimentName
#' pipelineParams$sampleFiles <- sampleFiles
#' pipL <- CytoPipeline(pipelineParams)
#' 
#' ### SCALE TRANSFORMATION STEPS ###
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "flowframe_read",
#'                           FUN = "readSampleFiles",
#'                           ARGS = list(
#'                               whichSamples = "all",
#'                               truncate_max_range = FALSE,
#'                               min.limit = NULL
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "remove_margins",
#'                           FUN = "removeMarginsPeacoQC",
#'                           ARGS = list()
#'                      )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "compensate",
#'                           FUN = "compensateFromMatrix",
#'                           ARGS = list(matrixSource = "fcs")
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "flowframe_aggregate",
#'                           FUN = "aggregateAndSample",
#'                           ARGS = list(
#'                               nTotalEvents = 10000,
#'                               seed = 0
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "scale_transform_estimate",
#'                           FUN = "estimateScaleTransforms",
#'                           ARGS = list(
#'                               fluoMethod = "estimateLogicle",
#'                               scatterMethod = "linear",
#'                               scatterRefMarker = "BV785 - CD3"
#'                           )
#'                       )
#'     )
#' 
#' ### PRE-PROCESSING STEPS ###
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "flowframe_read",
#'                           FUN = "readSampleFiles",
#'                           ARGS = list(
#'                               truncate_max_range = FALSE,
#'                               min.limit = NULL
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "remove_margins",
#'                           FUN = "removeMarginsPeacoQC",
#'                           ARGS = list()
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "compensate",
#'                           FUN = "compensateFromMatrix",
#'                           ARGS = list(matrixSource = "fcs")
#'                       )
#'     )
#' 
#' pipL <-
#' addProcessingStep(
#'     pipL,
#'     whichQueue = "pre-processing",
#'     CytoProcessingStep(
#'         name = "remove_debris",
#'         FUN = "removeDebrisManualGate",
#'         ARGS = list(
#'             FSCChannel = "FSC-A",
#'             SSCChannel = "SSC-A",
#'             gateData =  c(73615, 110174, 213000, 201000, 126000,
#'                           47679, 260500, 260500, 113000, 35000)
#'                    )
#'    )
#' )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "remove_dead_cells",
#'                           FUN = "removeDeadCellsManualGate",
#'                           ARGS = list(
#'                               FSCChannel = "FSC-A",
#'                               LDMarker = "L/D Aqua - Viability",
#'                               gateData = c(0, 0, 250000, 250000,
#'                                            0, 650, 650, 0)
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(
#'         pipL,
#'         whichQueue = "pre-processing",
#'         CytoProcessingStep(
#'             name = "perform_QC",
#'             FUN = "qualityControlPeacoQC",
#'             ARGS = list(
#'                 preTransform = TRUE,
#'                 min_cells = 150, # default
#'                 max_bins = 500, # default
#'                 step = 500, # default,
#'                 MAD = 6, # default
#'                 IT_limit = 0.55, # default
#'                 force_IT = 150, # default
#'                 peak_removal = 0.3333, # default
#'                 min_nr_bins_peakdetection = 10 # default
#'             )
#'         )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "transform",
#'                           FUN = "applyScaleTransforms",
#'                           ARGS = list()
#'                       )
#'     )
#' 
#' # execute pipeline, remove cache if existing with the same experiment name
#' suppressWarnings(execute(pipL, rmCache = TRUE, path = outputDir))
#' 
#' # re-execute as is without removing cache => all results found in cache!
#' suppressWarnings(execute(pipL, rmCache = FALSE, path = outputDir))
#' 
#' ### *** EXAMPLE 2: building CytoPipeline from JSON file *** ###
#' 
#' jsonDir <- system.file("extdata", package = "CytoPipeline")
#' jsonPath <- file.path(jsonDir, "pipelineParams.json")
#' 
#' pipL2 <- CytoPipeline(jsonPath, 
#'                       experimentName = experimentName,
#'                       sampleFiles = sampleFiles)
#' 
#' # note we temporarily set working directory into package root directory
#' # needed as json path mentions "./" path for sample files
#' suppressWarnings(execute(pipL2, rmCache = TRUE, path = outputDir))
#' 
#' ### *** EXAMPLE 3: building CytoPipeline from cache (previously run) *** ###
#' 
#' experimentName <- "OMIP021_PeacoQC"
#' pipL3 <- buildCytoPipelineFromCache(
#'     experimentName = experimentName,
#'     path = outputDir)
#' 
#' suppressWarnings(execute(pipL3,
#'         rmCache = FALSE,
#'         path = outputDir))
#'
execute <- function(x,
                    path = ".",
                    rmCache = FALSE,
                    useBiocParallel = FALSE,
                    BPPARAM = BiocParallel::bpparam(),
                    BPOPTIONS = BiocParallel::bpoptions(
                        packages = c("flowCore")),
                    saveLastStepFF = TRUE,
                    saveFFSuffix = "_preprocessed",
                    saveFFFormat = c("fcs", "csv"), 
                    saveFFCsvUseChannelMarker = TRUE,
                    saveScaleTransforms = FALSE) {
    stopifnot(inherits(x, "CytoPipeline"))
    
    if (length(sampleFiles(x)) == 0)
        stop("Can't execute CytoPipeline object with no sample file")

    #browser()

    outputDir <- file.path(path, x@experimentName, "output")
    # qualityControlDir <- file.path(outputDir, "QC")
    rdsOutputDir <- file.path(outputDir, "RDS")
    
    newDirs <- c(outputDir, rdsOutputDir)
    createDirs <- c(saveLastStepFF, saveScaleTransforms)
    

    for (i in seq_along(newDirs)) {
        if (createDirs[i] && !dir.exists(newDirs[i])) {
            dir.create(newDirs[i], recursive = TRUE)
        }
    }

    # create or use local cache
    localCacheDir <- file.path(path, x@experimentName, ".cache")
    bfc <- BiocFileCache::BiocFileCache(localCacheDir, ask = FALSE)

    # remove cache if 'rmCache' is set to
    if (rmCache) {
        bfci <- BiocFileCache::bfcinfo(bfc)
        if (nrow(bfci) > 0) {
            warning(
                "Found a previous cache with experiment name ",
                x@experimentName, " => removed it"
            )
            BiocFileCache::removebfc(bfc, ask = FALSE)
            # recreate cache
            bfc <- BiocFileCache::BiocFileCache(localCacheDir, ask = FALSE)
        }
    }
    
    # for robustness-sake, clean bfc from potentially interrupted previous run 
    # (crash/manually forced interruption)
    bfc <- makeBFCClean(bfc)

    # checking consistency between to-be-run processing steps and cache
    res <- checkCytoPipelineConsistencyWithCache(x, path = path)
    if (!res$isConsistent) {
        stop(res$inconsistencyMsg)
    }
    
    # store phenoData if present in CytoPipeline object
    if (!is.null(pData(x))) {
        cacheResourceName <- "phenoData"
        msg <- paste0("Proceeding with phenoData")
        if (cacheResourceName %in% BiocFileCache::bfcinfo(bfc)$rname) {
            #message(msg, ": found in cache => no need to store again!")
            message(msg, ": found in cache => updating it!")
            cacheResourceFile <- BiocFileCache::bfcrpath(
                x = bfc,
                rnames = cacheResourceName
            )
            saveRDS(pData(x), unname(cacheResourceFile))
            # no need to update meta data in this case...
        } else {
            message(msg, " ...")
            # browser()
            # store file in cache
            cacheResourceFile <- BiocFileCache::bfcnew(bfc, cacheResourceName)
            saveRDS(pData(x), unname(cacheResourceFile))
            
            # add one entry in cache meta data tables
            outputClass <- class(pData(x))
            outputObjectName <- "pData"
            genericMeta <-
                data.frame(list(
                    rid = names(cacheResourceFile),
                    type = "phenoData",
                    stepNb = 0,
                    stepName = "",
                    stepJsonSerialize = "",
                    outputClass = outputClass,
                    outputObjectName = outputObjectName
                ))
            
            BiocFileCache::bfcmeta(bfc, name = "generic", append = TRUE) <-
                genericMeta
        }
    } 

    ### first part is always to compute common transformation list ###
    message("#####################################################")
    message("### running SCALE TRANSFORMATION processing steps ###")
    message("#####################################################")

    currentTransList <- NULL

    for (s in seq_along(x@scaleTransformProcessingQueue)) {
        cacheResourceName <- getCPSName(x@scaleTransformProcessingQueue[[s]])
        stepName <- getCPSName(x@scaleTransformProcessingQueue[[s]])
        cacheResourceName <- paste0(
            "scaleTransform_step", s, "_",
            stepName
        )

        msg <- paste0("Proceeding with step ", s, " [", stepName, "]")
        if (cacheResourceName %in% BiocFileCache::bfcinfo(bfc)$rname) {
            message(msg, ": found in cache!")
            cacheResourceFile <-
                BiocFileCache::bfcrpath(
                    x = bfc,
                    rnames = cacheResourceName
                )
            res <- readRDS(cacheResourceFile)
        } else {
            message(msg, " ...")
            # browser()
            if (s == 1) {
                res <- executeProcessingStep(
                    x@scaleTransformProcessingQueue[[s]],
                    pData = x@pData,
                    sampleFiles = sampleFiles(x))
                
            } else {
                currentPData <- x@pData
                if (!is.null(currentPData) && inherits(res, "flowSet")) {
                    currentPData <- 
                        currentPData[flowCore::sampleNames(res), , drop=FALSE]
                }
                res <-
                    executeProcessingStep(
                        x@scaleTransformProcessingQueue[[s]],
                        pData = currentPData,
                        res
                    )
            }
            # store file in cache
            cacheResourceFile <- BiocFileCache::bfcnew(bfc, cacheResourceName)
            saveRDS(res, unname(cacheResourceFile))

            # add one entry in cache meta data tables
            outputClass <- class(res)
            outputObjectName <- paste0(stepName, "_obj")
            stepJsonSerialize <-
                as.json.CytoProcessingStep(x@scaleTransformProcessingQueue[[s]])
            genericMeta <-
                data.frame(list(
                    rid = names(cacheResourceFile),
                    type = "scale transform",
                    stepNb = s,
                    stepName = stepName,
                    stepJsonSerialize = stepJsonSerialize,
                    outputClass = outputClass,
                    outputObjectName = outputObjectName
                ))

            BiocFileCache::bfcmeta(bfc, name = "generic", append = TRUE) <-
                genericMeta
        }
    } # end loop on steps

    # store transformList if needed
    currentTransList <- NULL
    if (inherits(res, "transformList")) {
        currentTransList <- res
        if (saveScaleTransforms) {
            scaleTransformFile <- "scaleTransformList.rds"
            saveRDS(res,
                    file = file.path(
                        rdsOutputDir,
                        scaleTransformFile
                    )
            )
        }
    } else {
        if (saveScaleTransforms) {
            warning("could not store last step tranformList", 
                    "result object was no tranformList")
        }
    }
    
    preProcessOneFile <- function(i, files) {
        # browser()
        message("#####################################################")
        message("### NOW PRE-PROCESSING FILE ", files[i], "...")
        message("#####################################################")
        
        file_basenames <- basename(files)
        file_name_in_cache <- file_basenames[i]
        if (length(unique(file_basenames)) < length(file_basenames)){
            file_name_in_cache <- files[i]
        }
        
        for (s in seq_along(x@flowFramesPreProcessingQueue)) {
            stepName <- getCPSName(x@flowFramesPreProcessingQueue[[s]])
            cacheResourceName <- paste0(
                "preprocessing_",
                file_name_in_cache,
                "_step", s, "_",
                stepName
            )
            
            msg <- paste0("Proceeding with step ", s, " [", stepName, "]")
            
            if (cacheResourceName %in% BiocFileCache::bfcinfo(bfc)$rname) {
                message(msg, ": found in cache!")
                cacheResourceFile <-
                    BiocFileCache::bfcrpath(
                        x = bfc,
                        rnames = cacheResourceName
                    )
                res <- readRDS(cacheResourceFile)
            } else {
                message(msg, " ...")
                # browser()
                samplePhenoData <- if(is.null(x@pData)) NULL else 
                    x@pData[i, , drop = FALSE]
                if (s == 1) {
                    res <-
                        executeProcessingStep(
                            x@flowFramesPreProcessingQueue[[s]],
                            pData = samplePhenoData,
                            sampleFiles = files[i],
                            transList = currentTransList
                        )
                } else {
                    res <-
                        executeProcessingStep(
                            x@flowFramesPreProcessingQueue[[s]],
                            pData = samplePhenoData,
                            # ff = res,
                            res,
                            transList = currentTransList
                        )
                }
                # store file in cache
                cacheResourceFile <-
                    BiocFileCache::bfcnew(bfc, cacheResourceName)
                saveRDS(res, unname(cacheResourceFile))
                
                # add one entry in cache meta data tables
                outputClass <- class(res)
                outputObjectName <- paste0(stepName, "_obj")
                stepJsonSerialize <-
                    as.json.CytoProcessingStep(
                        x@flowFramesPreProcessingQueue[[s]]
                    )
                # specific for pre-processing step generating flow frames:
                # store nb of events at step
                nEvents <- 0
                if (inherits(res, "flowFrame")) {
                    nEvents <- flowCore::nrow(res)
                }
                genericMeta <-
                    data.frame(list(
                        rid = names(cacheResourceFile),
                        type = "pre-processing",
                        stepNb = s,
                        stepName = stepName,
                        stepJsonSerialize = stepJsonSerialize,
                        outputClass = outputClass,
                        outputObjectName = outputObjectName
                    ))
                preprocessingMeta <-
                    data.frame(list(
                        rid = names(cacheResourceFile),
                        fcsfile = file_name_in_cache,
                        nEvents = nEvents
                    ))
                
                BiocFileCache::bfcmeta(bfc, name = "generic", append = TRUE) <-
                    genericMeta
                BiocFileCache::bfcmeta(
                    bfc,
                    name = "preprocessing", append = TRUE
                ) <- preprocessingMeta
            } # if (newResource)
            # if last step and fcs saving is configured => do it!
            if (s == length(x@flowFramesPreProcessingQueue) &&
                saveLastStepFF) {
                if (!inherits(res, "flowFrame")) {
                    warning("could not store last step flowFrame", 
                            "result object was no flowFrame")
                } else {
                    CytoPipeline::writeFlowFrame(
                        ff = res,
                        dir = outputDir,
                        useFCSFileName = TRUE,
                        prefix = "",
                        suffix = saveFFSuffix,
                        format = saveFFFormat,
                        csvUseChannelMarker = saveFFCsvUseChannelMarker)
                }
            }
        } # end loop on steps
        
        
    }
    
    if (useBiocParallel) {
        #browser()
        invisible(BiocParallel::bplapply(seq_along(sampleFiles(x)),
                                         sampleFiles(x),
                                         BPPARAM = BPPARAM,
                                         BPOPTIONS = BPOPTIONS, 
                                         FUN = preProcessOneFile))
    } else {
        invisible(lapply(seq_along(sampleFiles(x)),
                         sampleFiles(x), 
                         FUN = preProcessOneFile))
    }
}

#' @name interactingWithCytoPipelineCache
#' @title interaction between CytoPipeline object and disk cache
#' @description functions supporting the interaction between a CytoPipeline
#' object and the file cache on disk
#' @param x a CytoPipeline object
#' @param path the full path to the experiment storage on disk
#' (without the /.cache)
#' @param experimentName the experimentName used to select the file cache on
#' disk
#' @param whichQueue which processing queue to check the consistency of
#' @param sampleFile if whichQueue == "pre-processing" or "both": which sample
#' file(s) to check on the disk cache
#' 
#' @return
#' for `deleteCytoPipelineCache`: TRUE if successfully removed\cr
#' for `buildCytoPipelineFromCache`: the built CytoPipeline object\cr
#' for `checkCytoPipelineConsistencyWithCache`: a list with the following
#' values:
#' - `isConsistent` (TRUE/FALSE)
#' - `inconsistencyMsg`: character filled in     
#' by an inconsistency message in case the cache and     
#' CytoPipeline object are not consistent with each other
#' - `scaleTransformStepStatus`: a character vector,     
#' containing, for each scale transform step, a status     
#' from c("run", "not run", "inconsistent")
#' - `preProcessingStepStatus`: a character matrix,     
#' containing, for each pre-processing step (rows), 
#' for each sample file (columns), a status from
#' c("run", "not run", "inconsistent")
#' @examples
#' 
#' # preliminary run:
#' # build CytoPipeline object using json input, run and store results in cache
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' experimentName <- "OMIP021_PeacoQC"
#' sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
#'                                              pattern = "Donor"))
#'                                              
#' jsonDir <- system.file("extdata", package = "CytoPipeline")
#' jsonPath <- file.path(jsonDir, "pipelineParams.json")
#' outputDir <- base::tempdir()
#' pipL <- CytoPipeline(jsonPath,
#'                      experimentName = experimentName,
#'                      sampleFiles = sampleFiles)
#' 
#' # note we temporarily set working directory into package root directory
#' # needed as json path mentions "./" path for sample files
#' suppressWarnings(execute(pipL, rmCache = TRUE, path = outputDir))
#'      
#' 
#' # rebuild CytoPipeline from stored results in cache, for a specific 
#' # experiment
#' 
#' experimentName <- "OMIP021_PeacoQC"
#' pipL2 <- buildCytoPipelineFromCache(
#'     experimentName = experimentName,
#'     path = outputDir)
#' 
#' 
#' # checking consistency between CytoPipeline object and cache
#' res <- checkCytoPipelineConsistencyWithCache(pipL2)
#' #res
#' 
#' suppressWarnings(execute(pipL2, rmCache = FALSE, path = outputDir))
#' # (everything is already stored in cache)
#' 
#' # deleting cache related to a specific experiment
#' pipL3 <- CytoPipeline(experimentName = experimentName)
#' deleteCytoPipelineCache(pipL3, path = outputDir)
#'

NULL


##' @describeIn interactingWithCytoPipelineCache delete the whole disk cache
##' corresponding to the experiment of a CytoPipeline object
##' @export
##'
deleteCytoPipelineCache <- function(x, path = ".") {
    stopifnot(inherits(x, "CytoPipeline"))
    localCacheDir <- file.path(path, x@experimentName, ".cache")
    bfc <- BiocFileCache::BiocFileCache(localCacheDir, ask = FALSE)
    BiocFileCache::removebfc(bfc, ask = FALSE)
}


##' @describeIn interactingWithCytoPipelineCache builds a new CytoPipeline
##' object,     
##' based on the information stored in the file cache
##' @export
##'
buildCytoPipelineFromCache <- function(experimentName, path = ".") {

    # browser()

    x <- CytoPipeline(experimentName = experimentName)

    # find cache corresponding to experiment name
    experimentNames <-
        getCytoPipelineExperimentNames(
            path = path,
            pattern = x@experimentName,
            fixed = TRUE
        )
    if (length(experimentNames) == 0) {
        warning(
            "no cache directory found for [", x@experimentName,
            "], did not build any processing step"
        )
        return(x)
    } else {
        phenoData <- NULL
        cacheDir <- file.path(path, x@experimentName, ".cache")

        bfc <- BiocFileCache::BiocFileCache(cacheDir, ask = FALSE)
        cacheInfo <- BiocFileCache::bfcinfo(bfc)
        
        # getting phenoData if there
        stepsInfos <- 
            cacheInfo[
                cacheInfo$type == "phenoData",
            ]
        nSteps <- nrow(stepsInfos)
        if (nSteps > 1)
            stop("more than one line for phenoData => inconsistency!")
        else if (nSteps == 1) {
            cacheResourceName <- "phenoData"
            cacheResourceFile <-
                BiocFileCache::bfcrpath(
                    x = bfc,
                    rnames = cacheResourceName
                )
            phenoData <- readRDS(file = cacheResourceFile)
        }

        # now building scale transform processing queue
        # take only steps with scale transform
        stepsInfos <-
            cacheInfo[
                cacheInfo$type == "scale transform",
                c("stepNb", "stepName", "stepJsonSerialize")
            ]
        stepsInfos <- unique(stepsInfos)
        nSteps <- nrow(stepsInfos)
        if (nSteps > 0) {
            uniqueStepNbs <- unique(stepsInfos$stepNb)

            if (length(uniqueStepNbs) != nrow(stepsInfos)) {
                stop(
                    "more than one step having the same step nb. ",
                    "Cache is inconsistent, deleting it manually is advised"
                )
            }

            stepsInfos <- stepsInfos[order(stepsInfos$stepNb), ]

            # x <- cleanProcessingSteps(x, whichQueue = "scale transform")

            for (j in seq_len(nSteps)) {
                pS <- from.json.CytoProcessingStep(
                    as.character(stepsInfos[j, "stepJsonSerialize"])
                )
                x <- addProcessingStep(x, whichQueue = "scale transform", pS)
            }
        }

        # now building pre-processing queue
        # take unique steps with pre-processing type

        stepsInfos <-
            cacheInfo[
                cacheInfo$type == "pre-processing",
                c("stepNb", "stepName", "stepJsonSerialize")
            ]
        stepsInfos <- unique(stepsInfos)
        nSteps <- nrow(stepsInfos)
        if (nSteps > 0) {
            uniqueStepNbs <- unique(stepsInfos$stepNb)
            if (length(uniqueStepNbs) != nrow(stepsInfos)) {
                stop(
                    "more than one step having the same step nb. ",
                    "Cache is inconsistent, deleting it manually is advised"
                )
            }
            stepsInfos <- stepsInfos[order(stepsInfos$stepNb), ]
            # x <- cleanProcessingSteps(x, whichQueue = "pre-processing")

            for (j in seq_len(nSteps)) {
                pS <- from.json.CytoProcessingStep(
                    as.character(stepsInfos[j, "stepJsonSerialize"])
                )
                x <- addProcessingStep(x, whichQueue = "pre-processing", pS)
            }

            # updating sample files
            # follow pData order if any, otherwise alphabetical order of 
            # file basenames 
            sampleFiles(x) <- sort(unique(stats::na.omit(cacheInfo$fcsfile)))
        } else {
            # no pre-processing step found in cache
            # => no sample file can be updated
            sampleFiles(x) <- character()
        }
        
        # finally: set pheno data (can only be done when sample files are known)
        if (!is.null(phenoData) && length(sampleFiles(x)) > 0) {
            pData(x) <- phenoData    
        }
        return(x)
    }
}

##' @describeIn interactingWithCytoPipelineCache check the consistency
##' between the processing steps described in a CytoPipeline object,
##' and what is stored in the file cache
##' @export
##'
checkCytoPipelineConsistencyWithCache <- function(
        x, path = ".",
        whichQueue = c("both", "scale transform", "pre-processing"),
        sampleFile = NULL) {
    
    stopifnot(inherits(x, "CytoPipeline"))
    
    useBaseName <- TRUE
    basenames <- basename(sampleFiles(x))
    if (length(unique(basenames)) < length(basenames)){
        useBaseName <- FALSE
    }
        
    whichQueue <- match.arg(whichQueue)    
        
    #browser()
    ret <- list(isConsistent = TRUE, inconsistencyMsg = character(0))

    if (whichQueue %in% c("both", "scale transform")) {
        nScaleTransformSteps <- length(x@scaleTransformProcessingQueue)
        ret$scaleTransformStepStatus <- rep("not_run", nScaleTransformSteps)
        ret$scaleTransformStepOutputObjNames <-
            rep("unknown", nScaleTransformSteps)
        ret$scaleTransformStepOutputClasses <-
            rep("unknown", nScaleTransformSteps)
        
        if (nScaleTransformSteps > 0) {
            names(ret$scaleTransformStepStatus) <-
                getProcessingStepNames(x, whichQueue = "scale transform")
        }
    }
    
    if (whichQueue %in% c("both", "pre-processing")) {
        nPreProcessingSteps <- length(x@flowFramesPreProcessingQueue)
        nSampleFiles <- length(sampleFiles(x))
        ret$preProcessingStepStatus <-
            matrix(rep("not_run", nPreProcessingSteps * nSampleFiles),
                   nrow = nPreProcessingSteps,
                   ncol = nSampleFiles
            )
        ret$preProcessingStepOutputObjNames <-
            rep("unknown", nPreProcessingSteps)
        ret$preProcessingStepOutputClasses <-
            rep("unknown", nPreProcessingSteps)
        
        if (nPreProcessingSteps > 0 && nSampleFiles > 0) {
            rownames(ret$preProcessingStepStatus) <-
                getProcessingStepNames(x, whichQueue = "pre-processing")

            if (useBaseName) {
                colnames(ret$preProcessingStepStatus) <- 
                    basename(sampleFiles(x))
            } else {
                colnames(ret$preProcessingStepStatus) <- 
                    sampleFiles(x)
            }
            
        }
    }
    

    # find cache corresponding to experiment name
    experimentNames <-
        getCytoPipelineExperimentNames(
            path = path,
            pattern = x@experimentName,
            fixed = TRUE
        )
    if (length(experimentNames) == 0) {
        return(ret)
    }

    # cache found => check consistency
    cacheDir <- file.path(path, x@experimentName, ".cache")

    bfc <- BiocFileCache::BiocFileCache(cacheDir, ask = FALSE)
    cacheInfo <- BiocFileCache::bfcinfo(bfc)

    if (nrow(cacheInfo) == 0) {
        return(ret)
    }
    
    # now check scale transform steps consistency
    # take only steps with scale transform type
    
    if (whichQueue %in% c("both", "scale transform")) {
        stepsInfos <-
            cacheInfo[
                cacheInfo$type == "scale transform",
                c(
                    "stepNb", "stepName", "stepJsonSerialize",
                    "outputClass", "outputObjectName"
                )
            ]
        
        stepsInfos <- unique(stepsInfos)
        nStepsInCache <- nrow(stepsInfos)
        if (nStepsInCache > 0) {
            uniqueStepNbs <- unique(stepsInfos$stepNb)
            
            if (length(uniqueStepNbs) != nrow(stepsInfos)) {
                stop(
                    "more than one step having the same step nb. ",
                    "Cache is inconsistent, deleting it manually is advised"
                )
            }
            
            stepsInfos <- stepsInfos[order(stepsInfos$stepNb), ]
            if (nStepsInCache > nScaleTransformSteps) {
                ret$isConsistent <- FALSE
                ret$inconsistencyMsg <-
                    "more scale transform steps in cache than "
                "in CytoPipeline object"
                return(ret)
            }
            
            for (j in seq_len(nStepsInCache)) {
                pS <- from.json.CytoProcessingStep(
                    as.character(stepsInfos[j, "stepJsonSerialize"])
                )
                pS2 <- getProcessingStep(x,
                                         whichQueue = "scale transform",
                                         index = j
                )
                argsComparison <- all.equal(getCPSARGS(pS), getCPSARGS(pS2))
                if (identical(getCPSName(pS), getCPSName(pS2)) &&
                    identical(getCPSFUN(pS), getCPSFUN(pS2)) &&
                    is.logical(argsComparison) && argsComparison) {
                    ret$scaleTransformStepStatus[j] <- "run"
                    ret$scaleTransformStepOutputClasses[j] <-
                        as.character(stepsInfos[j, "outputClass"])
                    ret$scaleTransformStepOutputObjNames[j] <-
                        as.character(stepsInfos[j, "outputObjectName"])
                    # and continue...
                } else {
                    ret$scaleTransformStepStatus[j] <- "inconsistent"
                    ret$isConsistent <- FALSE
                    ret$inconsistencyMsg <-
                        paste0(
                            "inconsistent scale transform step #", j,
                            " (different in cache)"
                        )
                    return(ret)
                }
            }
        }
    }
    
    #browser()

    # now check pre-processing steps consistency
    # take only steps with pre-processing type
    if (whichQueue %in% c("both", "pre-processing")) {
        stepsInfos <-
            cacheInfo[
                cacheInfo$type == "pre-processing",
                c(
                    "stepNb", "stepName", "stepJsonSerialize",
                    "outputClass", "outputObjectName"
                )
            ]
        stepsInfos <- unique(stepsInfos)
        nStepsInCache <- nrow(stepsInfos)
        if (nStepsInCache > 0) {
            uniqueStepNbs <- unique(stepsInfos$stepNb)
            
            if (length(uniqueStepNbs) != nrow(stepsInfos)) {
                stop(
                    "more than one step having the same step nb. ",
                    "Cache is inconsistent, deleting it manually is advised"
                )
            }
            
            stepsInfos <- stepsInfos[order(stepsInfos$stepNb), ]
            if (nStepsInCache > nPreProcessingSteps) {
                ret$isConsistent <- FALSE
                ret$inconsistencyMsg <-
                    paste0("more pre-processing steps in cache than in ",
                           "CytoPipeline object")
                return(ret)
            }
        }
        
        if (nrow(cacheInfo[cacheInfo$type == "pre-processing", ]) == 0) {
            return(ret)
        }
        
        if (is.null(sampleFile)) {
            sampleFileIndices <- seq_along(sampleFiles(x))
        } else if (is.numeric(sampleFile)) {
            if (all(sampleFile > 0 && sampleFile <= length(sampleFiles(x)))) {
                sampleFileIndices <- sampleFile
            } else {
                stop("sampleFile out of bounds")
            }
        } else {
            sampleFileIndices <-
                which(sampleFiles(x) == sampleFile)
            if (length(sampleFileIndices) == 0) {
                stop("sampleFile not found in CytoPipeline")
            } 
        }
        
        
        for (s in sampleFileIndices) {
            # take only steps with the target sample file
            if (useBaseName) {
                sampleFile <- basename(sampleFiles(x)[s])    
            } else {
                sampleFile <- sampleFiles(x)[s]
            }
            
            stepsInfos <-
                cacheInfo[
                    !is.na(cacheInfo$fcsfile) & cacheInfo$fcsfile == sampleFile,
                    c(
                        "stepNb", "stepName", "stepJsonSerialize",
                        "outputClass", "outputObjectName"
                    )
                ]
            stepsInfos <- unique(stepsInfos)
            nStepsInCache <- nrow(stepsInfos)
            if (nStepsInCache > 0) {
                uniqueStepNbs <- unique(stepsInfos$stepNb)
                
                if (length(uniqueStepNbs) != nrow(stepsInfos)) {
                    stop(
                        "more than one step having the same step nb. ",
                        "Cache is inconsistent, deleting it manually is advised"
                    )
                }
                
                stepsInfos <- stepsInfos[order(stepsInfos$stepNb), ]
                
                for (j in seq_len(nStepsInCache)) {
                    #browser()
                    pS <- from.json.CytoProcessingStep(
                        as.character(stepsInfos[j, "stepJsonSerialize"])
                    )
                    pS2 <- getProcessingStep(x,
                                             whichQueue = "pre-processing",
                                             index = j
                    )
                    argsComparison <- all.equal(getCPSARGS(pS), getCPSARGS(pS2))
                    if (identical(getCPSName(pS), getCPSName(pS2)) &&
                        identical(getCPSFUN(pS), getCPSFUN(pS2)) &&
                        is.logical(argsComparison) && argsComparison) {
                        ret$preProcessingStepStatus[j, sampleFile] <- "run"
                        ret$preProcessingStepOutputClasses[j] <-
                            as.character(stepsInfos[j, "outputClass"])
                        ret$preProcessingStepOutputObjNames[j] <-
                            as.character(stepsInfos[j, "outputObjectName"])
                        # and continue...
                    } else {
                        ret$preProcessingStepStatus[j, sampleFile] <- 
                            "inconsistent"
                        ret$isConsistent <- FALSE
                        ret$inconsistencyMsg <-
                            paste0(
                                "inconsistent pre-processing step #", j,
                                " for sample file ", sampleFile,
                                " (different in cache)"
                            )
                        return(ret)
                    }
                }
            }
        } # end loop on sample files
    }
    


    return(ret)
}

#' @name exportCytoPipeline
#' @title exporting CytoPipeline objects
#' @description functions to export CytoPipeline objects in various
#' formats
#' @param x a CytoPipeline object
#' @param path the full path to the name of the file to be created
#' @return - for `export2JSONFile`: nothing
#' @examples
#'
#' outputDir <- base::tempdir()
#' 
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' experimentName <- "OMIP021_PeacoQC"
#' sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
#'                                              pattern = "Donor"))
#' 
#' # build CytoPipeline object using json input
#' jsonPath <- file.path(system.file("extdata", package = "CytoPipeline"), 
#'                       "pipelineParams.json")
#'   
#' pipL <- CytoPipeline(jsonPath,
#'                      experimentName = experimentName,
#'                      sampleFiles = sampleFiles)
#' 
#' # remove the last pre-processing step
#' nPreProcessing <- getNbProcessingSteps(pipL, whichQueue = "pre-processing")
#' pipL <- removeProcessingStep(pipL, whichQueue = "pre-processing", 
#'                                    index = nPreProcessing)
#'
#' # export back to json file    
#' export2JSONFile(pipL, path = file.path(outputDir, "newFile.json")) 
NULL


#'
#' @describeIn exportCytoPipeline exports a CytoPipeline object
#' to a JSON file (writing the file = side effect)
#' @export
#'
export2JSONFile <- function(x, path) {
    myList <- as.list.CytoPipeline(x)
    jsonlite::write_json(myList, path = path, pretty = TRUE, null = "null")
}

#' @name inspectCytoPipelineObjects
#' @title inspect CytoPipeline results objects
#' @description functions to obtain results objects
#' formats
#' @param x a CytoPipeline object
#' @param experimentName the experimentName used to select the file cache on
#' disk
#' @param path root path to locate the search for file caches
#' @param whichQueue which queue to look into
#' @param sampleFile which sampleFile is looked for:
#' - if whichQueue == "scale transform", the sampleFile is ignored
#' - if NULL and whichQueue == "pre-processing", the sampleFile is
#' defaulted to the first one belonging to the experiment
#' @param whichSampleFiles indicates for which sample files the number
#' of retained events are to be collected. 
#' If missing, all sample files will be used.
#' @param objectName (character) which object name to look for
#' @param pattern optional pattern limiting the search for experiment
#' names
#' @param ignore.case (TRUE/FALSE) used in pattern matching (grepl)
#' @param fixed (TRUE/FALSE) used in pattern matching (grepl)
#' @param title if TRUE, adds a title to the plot
#' @param purpose purpose of the workflow plot
#' - if "run status" (default), the disk cache will be inspected and the
#' box colours will be set according to run status (green = run, 
#' orange = not run, red = definition not consistent with cache). Moreover, the 
#' object classes and names will be filled in if found in the cache.
#' - if "description", the workflow will be obtained from the step definition 
#' in the `x` object, not from the disk cache. As a result, all boxes will be
#' coloured in black, and no object class and name will be provided.
#' @param box.type shape of label box (rect, ellipse, diamond,
#' round, hexa, multi)
#' @param lwd default line width of arrow and box (one numeric value)
#' @param box.prop length/width ratio of label box (one numeric value)
#' @param box.cex relative size of text in boxes (one numeric value)
#' @param cex.txt relative size of arrow text (one numeric value)
#' @param box.size size of label box (one numeric value)
#' @param dtext controls the position of arrow text relative to arrowhead
#' (one numeric value)
#' @param ... other arguments passed to diagram::plotmat()  
#' @examples
#'
#' 
#' # preliminary run:
#' # build CytoPipeline object using json input, run and store results in cache
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' experimentName <- "OMIP021_PeacoQC"
#' sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
#'                                              pattern = "Donor"))
#' 
#' jsonDir <- system.file("extdata", package = "CytoPipeline")
#' jsonPath <- file.path(jsonDir, "pipelineParams.json")
#' outputDir <- base::tempdir()
#' pipL <- CytoPipeline(jsonPath,
#'                      experimentName = experimentName,
#'                      sampleFiles = sampleFiles)
#' 
#' # note we temporarily set working directory into package root directory
#' # needed as json path mentions "./" path for sample files
#' suppressWarnings(execute(pipL, rmCache = TRUE, path = outputDir))
#'      
#' 
#' # get a list of all stored experiments in a specific path taken as root dir
#' experimentNames <- getCytoPipelineExperimentNames(path = outputDir)
#' 
#' # rebuilding Cytopipeline object from cache
#' pipL2 <- buildCytoPipelineFromCache(experimentName = experimentNames[1],
#'                                     path = outputDir)
#' 
#' # plot scale transformation queue
#' plotCytoPipelineProcessingQueue(pipL2, whichQueue = "pre-processing",
#'                                 path = outputDir)
#' 
#' # plot pre-processing queue
#' plotCytoPipelineProcessingQueue(pipL2, whichQueue = "scale transform",
#'                                 path = outputDir)
#'                                 
#' # get object infos for a specific queue
#' df <- getCytoPipelineObjectInfos(pipL2, whichQueue = "pre-processing",
#'                                  path = outputDir,
#'                                  sampleFile = sampleFiles(pipL2)[1]) 
#'                                 
#' # get transform list (output of one step)
#' trans <-
#'     getCytoPipelineScaleTransform(pipL2, whichQueue = "scale transform",
#'                                   objectName =
#'                                       "scale_transform_estimate_obj",
#'                                   path = outputDir)
#' 
#' # get flowFrame (output of one step)
#' ff <- getCytoPipelineFlowFrame(pipL2, whichQueue = "pre-processing",
#'                                objectName = "remove_doublets_obj",
#'                                path = outputDir,
#'                                sampleFile = sampleFiles(pipL2)[1])
#' 
#' # get any object (output of one step)
#' obj <-
#'     getCytoPipelineObjectFromCache(pipL2, whichQueue = "scale transform",
#'                                    objectName = "compensate_obj",
#'                                    path = outputDir)
#' class(obj) # flowCore::flowSet 
#' 
#' # collect number of retained events at each step
#' nbEventsDF <- collectNbOfRetainedEvents( 
#'         experimentName = experimentNames[1],
#'         path = outputDir) 
#'
NULL

#' @title Find CytoPipeline experiments stored in a file cache
#' @describeIn inspectCytoPipelineObjects 
#'   This function     
#'   looks into a path for stored file caches     
#'   and gets the corresponding experiment names
#' @return - for `getCytoPipelineExperimentNames`: 
#'   a vector of character containing found experiment names
#' @export
#'
getCytoPipelineExperimentNames <-
    function(path = ".",
             pattern = NULL,
             ignore.case = FALSE,
             fixed = FALSE) {
        subdirs <- list.dirs(path, full.names = FALSE, recursive = FALSE)
        if (!is.null(pattern)) {
            subdirs <- subdirs[grepl(
                pattern = pattern,
                x = subdirs,
                ignore.case = ignore.case,
                fixed = fixed
            )]
        }
        experimentNames <- vector(mode = "character", length = 0)
        for (d in subdirs) {
            tentative <- file.path(path, d, ".cache")
            if (dir.exists(tentative)) {
                experimentNames <- c(experimentNames, d)
            }
        }
        return(experimentNames)
    }


#' @title Retrieves an object from a file cache
#' @describeIn inspectCytoPipelineObjects 
#'   Given a CytoPipeline object,     
#'   this function retrieves    
#' a specific object in the corresponding file cache
#'
#' @returns - for `getCytoPipelineObjectFromCache`: 
#'   the found object (or stops with an error message    
#'   if the target object is not found)  
#' @export
#'
getCytoPipelineObjectFromCache <-
    function(x,
             path = ".",
             whichQueue = c("scale transform", "pre-processing"),
             sampleFile = NULL,
             objectName) {
        # browser()
        stopifnot(inherits(x, "CytoPipeline"))
        whichQueue <- match.arg(whichQueue)
        if (whichQueue == "scale transform") {
            sampleFile <- NULL
        } else {
            if (is.numeric(sampleFile)) {
                if (sampleFile > 0 && sampleFile <= length(sampleFiles(x))) {
                    sampleFile <- sampleFiles(x)[sampleFile]
                } else {
                    stop("sampleFile out of bounds")
                }
            } else {
                sampleFileIndex <-
                    which(sampleFiles(x) == sampleFile)
                if (length(sampleFileIndex) == 0) {
                    stop("sampleFile not found in CytoPipeline")
                } else if (length(sampleFileIndex) > 1) {
                    stop(
                        "sampleFile found multiple times in CytoPipeline ",
                        "(unexpected inconsistency)"
                    )
                }
            }
        }
        
        #browser()

        # checking consistency between to-be-run processing steps and cache
        res <- checkCytoPipelineConsistencyWithCache(x, path = path,
                                                     whichQueue = whichQueue,
                                                     sampleFile = sampleFile)
        if (!res$isConsistent) {
            stop(res$inconsistencyMsg)
        }


        cacheDir <- file.path(path, x@experimentName, ".cache")

        bfc <- BiocFileCache::BiocFileCache(cacheDir, ask = FALSE)
        cacheInfo <- BiocFileCache::bfcinfo(bfc)
        if (!is.null(sampleFile)) {
            sampleFiles <- unique(stats::na.omit(cacheInfo$fcsfile))
            if (length(unique(basename(sampleFiles(x)))) == 
                length(basename(sampleFiles(x)))) {
                sampleFile <- basename(sampleFile)
            }
            if (!(sampleFile %in% sampleFiles)) {
                stop(sampleFile, " not found in experiment run!")
            }
            indexInCache <- which(cacheInfo$outputObjectName == objectName &
                cacheInfo$fcsfile == sampleFile)
            if (length(indexInCache) > 1) {
                stop(
                    "More than one line in cache corresponding to object ",
                    "name and sample file => unexpected inconsistency"
                )
            }

            if (length(indexInCache) == 0) {
                stop(
                    "Combination (sample file, object name) ",
                    "not found in cache"
                )
            }
        } else {
            if (!("fcsfile" %in% colnames(cacheInfo))) {
                # special case appearing when no pre-processing steps
                # have been run
                indexInCache <-
                    which(cacheInfo$outputObjectName == objectName)
            } else {
                indexInCache <-
                    which(cacheInfo$outputObjectName == objectName &
                        is.na(cacheInfo$fcsfile))
            }

            if (length(indexInCache) > 1) {
                stop(
                    "More than one line in cache corresponding to object ",
                    "name and w/o sample file => unexpected inconsistency"
                )
            }

            if (length(indexInCache) == 0) {
                stop("Object name not found in cache")
            }
        }

        cacheResourceFile <-
            BiocFileCache::bfcpath(bfc, rids = cacheInfo$rid[indexInCache])

        ret <- readRDS(cacheResourceFile)

        return(ret)
    }


#' @title File cache objects information
#' @describeIn inspectCytoPipelineObjects 
#'   Given a CytoPipeline object,     
#'   this function retrieves    
#' the information related to a specific object name,    
#' i.e. object name and object class
#' @returns - for `getCytoPipelineObjectInfos`:
#'   a dataframe with the collected information about the
#'   found objects (or stops with an error message if no target object
#'   was found) 
#
#' @export
#
getCytoPipelineObjectInfos <-
    function(x,
             path = ".",
             whichQueue = c("scale transform", "pre-processing"),
             sampleFile = NULL) {
        # browser()
        stopifnot(inherits(x, "CytoPipeline"))
        whichQueue <- match.arg(whichQueue)
        if (whichQueue == "scale transform") {
            sampleFile <- NULL
        }

        # checking consistency between to-be-run processing steps and cache
        res <- checkCytoPipelineConsistencyWithCache(
            x, path = path,
            whichQueue = whichQueue,
            sampleFile = sampleFile)
        
        if (!res$isConsistent) {
            stop(res$inconsistencyMsg)
        }

        cacheDir <- file.path(path, x@experimentName, ".cache")

        bfc <- BiocFileCache::BiocFileCache(cacheDir, ask = FALSE)
        cacheInfo <- BiocFileCache::bfcinfo(bfc)
        indexesInCache <- numeric()
        if (!is.null(sampleFile)) {
            sampleFiles <- unique(stats::na.omit(cacheInfo$fcsfile))
            if (is.numeric(sampleFile)) {
                sampleFileIndex <- sampleFile
                if (sampleFileIndex < 1 || 
                    sampleFileIndex > length(sampleFiles(x))) {
                    stop("sampleFile (provided as index) out of bounds!")
                }
                sampleFile <- sampleFiles(x)[sampleFileIndex]
            }
            if (length(unique(basename(sampleFiles(x)))) == 
                       length(basename(sampleFiles(x)))) {
                sampleFile <- basename(sampleFile)
            }
            if (!(sampleFile %in% sampleFiles)) {
                stop(sampleFile, " not found in experiment run!")
            }
            indexesInCache <- which(cacheInfo$type == whichQueue &
                cacheInfo$fcsfile == sampleFile)

            if (length(indexesInCache) == 0) {
                # stop(
                #     "Combination ('", whichQueue, "'", sampleFile,
                #     ") not found in cache"
                # )
                # return empty data frame
                retDF <- data.frame(outputObjectName = character(),
                                    outputClass = character())
                return (retDF)
            }
        } else {
            if (!("fcsfile" %in% colnames(cacheInfo))) {
                # special case appearing when no pre-processing steps
                # have been run
                indexesInCache <- which(cacheInfo$type == whichQueue)
            } else {
                indexesInCache <- which(cacheInfo$type == whichQueue &
                    is.na(cacheInfo$fcsfile))
            }

            if (length(indexesInCache) == 0) {
                # stop(
                #     "Combination ('", whichQueue,
                #     "', no sample file) not found in cache"
                # )
                # return empty data frame
                retDF <- data.frame(outputObjectName = character(),
                                    outputClass = character())
                return (retDF)
            }
        }

        objectInfos <- as.data.frame(cacheInfo[
            indexesInCache,
            c(
                "outputObjectName",
                "outputClass"
            )
        ])
        colnames(objectInfos) <- c("ObjectName", "ObjectClass")
        return(objectInfos)
    }

#' @title Retrieves a flowFrame object from a file cache
#' @describeIn inspectCytoPipelineObjects 
#'   Given a CytoPipeline object,     
#'   this function retrieves    
#'   a specific flowCore::flowFrame object in the corresponding     
#'   file cache object name and object class
#' @returns - for `getCytoPipelineFlowFrame`: 
#'   the found flowFrame (or stops with an error message if the
#'   target object is not found, or if the object is no flowFrame) 
#' @export
#'
getCytoPipelineFlowFrame <-
    function(x,
             path = ".",
             whichQueue = c("scale transform", "pre-processing"),
             sampleFile,
             objectName) {
        ret <-
            getCytoPipelineObjectFromCache(
                x = x,
                whichQueue = whichQueue,
                sampleFile = sampleFile,
                objectName = objectName,
                path = path
            )

        if (!inherits(ret, "flowFrame")) {
            stop(
                "Object '", objectName,
                "' does not appear to be a flowFrame"
            )
        }
        return(ret)
    }

#' @title Retrieves a transformList object from a file cache
#' @describeIn inspectCytoPipelineObjects 
#'   Given a CytoPipeline object,     
#'   this function retrieves    
#'   a specific flowCore::transformList object in the corresponding    
#'   file cache
#' @returns - for `getCytoPipelineScaleTransform`: the found flowFrame 
#'   (or stops with an error message if the
#'   target object is not found, or if the object is no transformList)
#'
#' @export
#'
getCytoPipelineScaleTransform <-
    function(x,
             path = ".",
             whichQueue = c("scale transform", "pre-processing"),
             sampleFile = NULL,
             objectName) {
        ret <-
            getCytoPipelineObjectFromCache(
                x = x,
                whichQueue = whichQueue,
                sampleFile = sampleFile,
                objectName = objectName,
                path = path
            )

        if (!inherits(ret, "transformList")) {
            stop(
                "Object '", objectName,
                "' does not appear to be a transformList"
            )
        }

        return(ret)
    }

#' @title Plots a processing queue
#' @describeIn inspectCytoPipelineObjects 
#' This functions displays     
#' a plot of a processing queue of a CytoPipeline object,     
#' using diagram::plotmat().    
#' 
#' - If a step is in run state for all sample files, the    
#' corresponding box appears in green
#' - If a step is in non run state for at least one sample file,    
#' the corresponding box appears in orange
#' - If at least one step is not consistent with cache, the whole    
#' set of boxes appears in red
#' @returns - for `plotCytoPipelineProcessingQueue`:
#'   nothing  
#'
#' @export
#'
plotCytoPipelineProcessingQueue <-
    function(x,
             whichQueue = c("pre-processing", "scale transform"),
             purpose = c("run status", "description"),
             sampleFile = NULL,
             path = ".",
             title = TRUE,
             box.type = "ellipse",
             lwd = 1,
             box.prop = 0.5,
             box.cex = 0.7,
             cex.txt = 0.7,
             box.size = 0.1,
             dtext = 0.15,
             ...) {

        # browser()

        stopifnot(inherits(x, "CytoPipeline"))

        whichQueue <- match.arg(whichQueue)
        purpose <- match.arg(purpose)
        
        nSteps <- getNbProcessingSteps(x, whichQueue = whichQueue)
        steps <- getProcessingStepNames(x, whichQueue = whichQueue)
        names <- c("raw data", steps)
        
        sampleFileIndex <- 0
        
        if (purpose == "description") {
            box.lcol <- rep("black", nSteps+1)
            names <- c("raw data", paste0("output ", seq_len(nSteps)))
            
        } else {
            if (whichQueue == "pre-processing") {
                if (is.null(sampleFile) && length(sampleFiles(x)) > 0) {
                    message(
                        "no sample file passed as argument ",
                        "=> defaulting to first sample file"
                    )
                    sampleFileIndex <- 1
                } else if (is.numeric(sampleFile)) {
                    if (sampleFile > 0 && sampleFile <= length(sampleFiles(x))) {
                        sampleFileIndex <- sampleFile
                    } else {
                        stop("sampleFile out of bounds")
                    }
                } else {
                    sampleFileIndex <-
                        which(sampleFiles(x) == sampleFile)
                    if (length(sampleFileIndex) == 0) {
                        stop("sampleFile not found in CytoPipeline")
                    } else if (length(sampleFileIndex) > 1) {
                        stop(
                            "sampleFile found multiple times in CytoPipeline ",
                            "(unexpected inconsistency)"
                        )
                    }
                }
            } else {
                if (!is.null(sampleFile)) {
                    message(
                        "sample file passed but not taken into account as the ",
                        "processing queue is 'scale transform'"
                    )
                }
            }
            
            
            box.lcol <- c("black", rep("orange", nSteps))
            
            res <- 
                checkCytoPipelineConsistencyWithCache(
                    x, path = path,
                    whichQueue = whichQueue,
                    sampleFile = sampleFileIndex)
            
            if (!res$isConsistent) {
                warning(
                    "CytoPipeline object not consistent with cache: ",
                    res$inconsistencyMsg
                )
                box.lcol <- c("black", rep("red", nSteps))
            } else if (sampleFileIndex == 0 && whichQueue == "pre-processing") {
                warning("No sample file in CytoPipeline object")
                box.lcol <- c("black", rep("red", nSteps))
            } else if (nSteps > 0) {
                for (i in seq_len(nSteps)) {
                    stepStatus <- ""
                    if (whichQueue == "pre-processing") {
                        if (unname(
                            res$preProcessingStepStatus[i, sampleFileIndex])
                            == "run") {
                            box.lcol[i + 1] <- "green"
                        }
                        if (res$preProcessingStepOutputObjNames[i] 
                            != "unknown") {
                            names[i + 1] <- 
                                res$preProcessingStepOutputObjNames[i]
                            if (res$preProcessingStepOutputClasses[i] !=
                                "unknown") {
                                names[i + 1] <- paste0(
                                    names[i + 1], "\n(",
                                    res$preProcessingStepOutputClasses[i], ")"
                                )
                            }
                        }
                    } else {
                        if (unname(res$scaleTransformStepStatus[i]) == "run") {
                            box.lcol[i + 1] <- "green"
                        }
                        if (res$scaleTransformStepOutputObjNames[i] 
                            != "unknown") {
                            names[i + 1] <- 
                                res$scaleTransformStepOutputObjNames[i]
                            if (res$scaleTransformStepOutputClasses[i] !=
                                "unknown") {
                                names[i + 1] <- paste0(
                                    names[i + 1], "\n(",
                                    res$scaleTransformStepOutputClasses[i], ")"
                                )
                            }
                        }
                    }
                }
            }
        }

        

        # browser()

        M <- matrix(nrow = nSteps + 1, 
                    ncol = nSteps + 1, 
                    byrow = TRUE, 
                    data = 0)

        if (nSteps >= 1) {
            for (j in seq_along(steps)) {
                M[j + 1, j] <- steps[j]
            }
        }

        if (nSteps %% 2 == 0) {
            pos <- c(rep(2, nSteps / 2), 1)
        } else {
            pos <- rep(2, (nSteps + 1) / 2)
        }

        diagram::plotmat(M,
            pos = pos, curve = 0, name = names,
            box.type = box.type,
            box.lcol = box.lcol,
            lwd = lwd,
            box.prop = box.prop,
            box.cex = box.cex,
            cex.txt = cex.txt,
            box.size = box.size,
            dtext = dtext,
            ...
        )


        if (title) {
            theTitle <- paste0(
                "Experiment: ", x@experimentName,
                "\nProcessing queue: ", whichQueue
            )
            
            if (sampleFileIndex != 0) {
                displayFileName <- sampleDisplayNames(x)[sampleFileIndex]
                theTitle <- paste0(
                    theTitle,
                    "\nSample:",
                    displayFileName)
            }
            graphics::title(main = theTitle)
        }
    }

#' @title Nb of retained events at each preprocessing step
#' @describeIn inspectCytoPipelineObjects 
#'   Given a CytoPipeline object,     
#'   this function retrieves, for all pre-processing steps,
#'   given the output is a flowFrame,
#'   the number of retained event.    
#' @returns - for `collectNbOfRetainedEvents`:
#'   a dataframe with the collected number of events
#'   columns refer to pre-processing steps
#'   rows refer to samples
#' @export
#
collectNbOfRetainedEvents <- function( 
        experimentName,
        path = ".",
        whichSampleFiles) {
    
    pipL <- CytoPipeline::buildCytoPipelineFromCache(
        experimentName = experimentName,
        path = path
    )
    
    if (missing(whichSampleFiles)) {
        whichSampleFiles <- CytoPipeline::sampleFiles(pipL)
    } else if (is.numeric(whichSampleFiles)) {
        sF <- CytoPipeline::sampleFiles(pipL)
        if (min(whichSampleFiles) < 1 || max(whichSampleFiles) > length(sF)) {
            stop("whichSampleFiles out of bounds")
        }
        whichSampleFiles <- CytoPipeline::sampleFiles(pipL)[whichSampleFiles]
    } else if (is.character(whichSampleFiles)) {
        #whichSampleFiles <- basename(whichSampleFiles)
    } else {
        stop("whichSampleFiles should be either character or numeric")
    }
    
    cacheDir <- file.path(path,  experimentName, ".cache")
    bfc <- BiocFileCache::BiocFileCache(cacheDir, ask = FALSE)
    
    metaPrepDF <- BiocFileCache::bfcmeta(bfc, name = "preprocessing")
    
    if ("nEvents" %in% colnames(metaPrepDF)) {
        # as of CytoPipeline version 1.3.6,
        # nb of events are stored for each step
        cacheInfo <- BiocFileCache::bfcinfo(bfc)
        metaPrepDF <- metaPrepDF[metaPrepDF$fcsfile %in% whichSampleFiles,]
        DFM <- merge(x = metaPrepDF, y = cacheInfo, by = "rid")
        #browser()
        DFM <- 
            DFM[order(DFM$fcsfile.x, DFM$stepNb),][
                ,c("rid","fcsfile.x","stepNb", "stepName", "nEvents.x")]
        allStepNames <- unique(DFM$stepName)
        nSampleFiles <- length(whichSampleFiles)
        nAllSteps <- length(allStepNames)
        eventNbs <- matrix(rep(NA, nSampleFiles * nAllSteps),
                           nrow = nSampleFiles)
        rownames(eventNbs) <- as.character(whichSampleFiles)
        colnames(eventNbs) <- allStepNames
        
        for (i in seq_len(nrow(DFM))) {
            sampleName <- DFM[i, "fcsfile.x"]
            stepName <- DFM[i, "stepName"]
            eventNbs[sampleName, stepName] <- DFM[i, "nEvents.x"]
        }
    }
    else {
        # version < 1.3.6 => collect nb of events from reading the flowFrames
        # from cache
        
        nEventPerSampleList <- list()
        allStepNames <- c()
        for (s in seq_along(whichSampleFiles)) {
            message("Collecting nb of events for sample file ", 
                    whichSampleFiles[s], 
                    "...")
            objInfos <- CytoPipeline::getCytoPipelineObjectInfos(
                pipL,
                whichQueue = "pre-processing",
                sampleFile = whichSampleFiles[s],
                path = path)
            objInfos <- objInfos[objInfos[,"ObjectClass"] == "flowFrame",]
            
            nEventPerSampleList[[s]] <- lapply(
                objInfos[,"ObjectName"],
                FUN = function(objName) {
                    message("Reading object ", objName, "...")
                    ff <- CytoPipeline::getCytoPipelineFlowFrame(
                        pipL,
                        path = path,
                        whichQueue = "pre-processing",
                        sampleFile = whichSampleFiles[s],
                        objectName = objName)
                    flowCore::nrow(ff)})
            
            stepNames <- vapply(objInfos[,"ObjectName"],
                                FUN = function(str){
                                    gsub(x = str,
                                         pattern = "_obj",
                                         replacement = "")
                                },
                                FUN.VALUE = character(length = 1))   
            
            names(nEventPerSampleList[[s]]) <- stepNames
            
            allStepNames <- union(allStepNames, stepNames)
        }
        
        nSampleFiles <- length(whichSampleFiles)
        nAllSteps <- length(allStepNames)
        eventNbs <- matrix(rep(NA, nSampleFiles * nAllSteps),
                           nrow = nSampleFiles)
        rownames(eventNbs) <- as.character(whichSampleFiles)
        colnames(eventNbs) <- allStepNames
        
        for (s in seq_along(whichSampleFiles)) {
            stepNames <- names(nEventPerSampleList[[s]])
            for (st in seq_along(stepNames)) {
                eventNbs[as.character(whichSampleFiles)[s],
                         stepNames[st]] <- 
                    nEventPerSampleList[[s]][[st]]
            }
        }
    }
    as.data.frame(eventNbs)
}

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).


outputDir <- base::tempdir()

if (!interactive()) pdf(NULL)

test_that("CytoPipeline default creation raises no error", {
    expect_error(pipL0 <- CytoPipeline(), NA)
})

test_that("Cytopipeline add/remove/clean processing step works", {
    rawDataDir <- system.file("extdata", package = "CytoPipeline")
    experimentName <- "OMIP021_PeacoQC"
    sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
        pattern = "Donor"
    ))
    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList.rds")


    # main parameters : sample files and experiment name
    pipelineParams <- list()
    pipelineParams$experimentName <- experimentName

    pipL <- CytoPipeline(pipelineParams)
    expect_error(show(pipL), NA)

    pipelineParams$sampleFiles <- sampleFiles

    pipL <- CytoPipeline(pipelineParams)
    expect_error(show(pipL), NA)

    pipL <- addProcessingStep(pipL,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "scale_transform_read",
            FUN = "readRDS",
            ARGS = list(file = transListPath)
        )
    )

    expect_equal(getNbProcessingSteps(pipL, "scale transform"), 1)

    pipL <- addProcessingStep(pipL,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "scale_transform_sum",
            FUN = "sum",
            ARGS = list()
        )
    )

    expect_equal(getNbProcessingSteps(pipL, "scale transform"), 2)

    pipL <- removeProcessingStep(pipL,
        whichQueue = "scale transform",
        index = 2
    )
    expect_equal(getNbProcessingSteps(pipL, "scale transform"), 1)
    pS <- getProcessingStep(pipL, whichQueue = "scale transform", index = 1)
    expect_equal(pS@FUN, "readRDS")

    pipL <- addProcessingStep(pipL,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "pre-processing_sum",
            FUN = "sum",
            ARGS = list()
        )
    )
    expect_equal(getNbProcessingSteps(pipL, "scale transform"), 1)
    expect_equal(getNbProcessingSteps(pipL, "pre-processing"), 1)

    expect_error(pipL <- addProcessingStep(pipL,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "pre-processing_sum",
            FUN = "mean",
            ARGS = list()
        )
    ), regexp = "There already exist a step")

    pipL <- cleanProcessingSteps(pipL)
    expect_equal(getNbProcessingSteps(pipL, "scale transform"), 0)
    expect_equal(getNbProcessingSteps(pipL, "pre-processing"), 0)

    newExp <- "newExperiment"
    experimentName(pipL) <- newExp
    expect_equal(experimentName(pipL), newExp)

    newPhenoData <- data.frame(name = c("Donor1", "Donor2"),
                               donor = c(1,2))
    rownames(newPhenoData) <- newPhenoData$name
    
    expect_error(pData(pipL) <- "invalidCharacterType",
                 regexp = "is not TRUE")

    expect_error(pData(pipL) <- newPhenoData,
                 regexp = "should correspond to sample file names")

})

test_that("CytoPipeline with reading scale transfo only raises no error", {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            experimentName <- "OMIP021_PeacoQC"
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                pattern = "Donor"
            ))
            transListPath <- file.path(system.file("extdata",
                                                   package = "CytoPipeline"),
                                       "OMIP021_TransList.rds")

            # main parameters : sample files and output files
            pipelineParams <- list()
            pipelineParams$experimentName <- experimentName
            pipelineParams$sampleFiles <- sampleFiles

            pipL <- CytoPipeline(pipelineParams)

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "scale transform",
                    CytoProcessingStep(
                        name = "scale_transform_read",
                        FUN = "readRDSObject",
                        ARGS = list(RDSFile = transListPath)
                    )
                )

            suppressWarnings(execute(pipL,
                rmCache = TRUE,
                path = outputDir
            ))
        },
        NA
    )
})

test_that("CytoPipeline with no sample raises an execution error", {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            experimentName <- "OMIP021_PeacoQC"
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                            pattern = "NotGood"
            ))

            # main parameters : sample files and output files
            pipelineParams <- list()
            pipelineParams$experimentName <- experimentName
            pipelineParams$sampleFiles <- sampleFiles

            pipL <- CytoPipeline(pipelineParams)

            pipL <-
                addProcessingStep(pipL,
                                  whichQueue = "scale transform",
                                  CytoProcessingStep(
                                      name = "flowframe_read",
                                      FUN = "readSampleFiles",
                                      ARGS = list(
                                          whichSamples = "all",
                                          truncate_max_range = FALSE,
                                          min.limit = NULL
                                      )
                                  )
                )

            suppressWarnings(execute(pipL,
                                     rmCache = TRUE,
                                     path = outputDir
            ))
        },
        "Can't execute CytoPipeline object with no sample file"
    )
})

test_that("Creation of CytoPipeline with wrong phenoData raises an error", {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            experimentName <- "OMIP021_PeacoQC"
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                            pattern = "Donor"
            ))

            phenoData <- data.frame(row.names = c("wrong1", "wrong2"),
                                    donor = c(1,2),
                                    group = c("G1", "G1"))

            pipL <- CytoPipeline(experimentName = experimentName,
                                 sampleFiles = sampleFiles,
                                 pData = phenoData)
        }, "Row names of non-null @pData slot should correspond")
})

test_that("Execution of CytoPipeline with correct phenoData raises no error", {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            experimentName <- "OMIP021_PeacoQC"
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                            pattern = "Donor"
            ))

            phenoData <- data.frame(row.names = basename(sampleFiles),
                                    donor = c(1,2),
                                    group = c("G1", "G1"))

            # main parameters : sample files and output files
            pipelineParams <- list()
            pipelineParams$experimentName <- experimentName
            pipelineParams$sampleFiles <- sampleFiles
            pipelineParams$pData <- phenoData

            pipL <- CytoPipeline(pipelineParams)

            pipL <-
                addProcessingStep(pipL,
                                  whichQueue = "scale transform",
                                  CytoProcessingStep(
                                      name = "flowframe_read",
                                      FUN = "readSampleFiles",
                                      ARGS = list(
                                          whichSamples = "all",
                                          truncate_max_range = FALSE,
                                          min.limit = NULL
                                      )
                                  )
                )
            
            pipL <- 
                addProcessingStep(pipL,
                                  whichQueue = "pre-processing",
                                  CytoProcessingStep(
                                    name = "flowFrame_read",
                                    FUN = "readSampleFiles",
                                    ARGS = list(
                                          truncate_max_range = FALSE,
                                        min.limit = NULL
                                    )
                                  )
                )

            suppressWarnings(execute(pipL,
                                     rmCache = TRUE,
                                     path = outputDir
            ))

            # re-execute un second time to test behaviour with pData
            execute(pipL,
                    rmCache = FALSE,
                    path = outputDir
            )
        },
        NA
    )

    newPipL <- buildCytoPipelineFromCache(experimentName,
                                          path = outputDir)

    newPData <- pData(newPipL)
    expect_true(all.equal(phenoData, newPData))
    
    # test whether pheno data row names are set by default
    phenoData2 <- phenoData
    rownames(phenoData2) <- NULL
    
    expect_message(
        pData(newPipL) <- phenoData2,
        "pData row names has been set by default")
    
    newPData2 <- pData(newPipL)
    expect_true(all.equal(phenoData, newPData2))
    
    # test whether pheno data subset can be taken
    phenoData3 <- data.frame(
        row.names = c("Donor0.fcs", "Donor1.fcs", "Donor2.fcs", "Donor3.fcs"),
                      donor = c(0,1,2,3),
                      group = c("G0", "G1", "G1", "G2"))
    
    expect_message(
        pData(newPipL) <- phenoData3,
        "pData row subset corresponding to existing sample names")           
    newPData3 <- pData(newPipL)
    expect_true(all.equal(phenoData, newPData3))
    
    
    # test whether updating sample files can work
    
    expect_error({
        sampleFiles(newPipL) <- c("Donor1.fcs", "Donor2.fcs")}, NA)
    
    expect_error({
        sampleFiles(newPipL) <- c("./Donor1.fcs", "./Donor2.fcs")}, NA)
    
    expect_error({
        sampleFiles(newPipL) <- c("Donor2.fcs", "Donor1.fcs")}, NA)
    
    expect_error({
        sampleFiles(newPipL) <- c("Donor1.fcs")}, 
        "number of rows should be equal to number of samples")
    
    expect_error({
        sampleFiles(newPipL) <- c("Donor1.fcs", "Donor3.fcs")}, 
        "should correspond to sample file names")
    
    expect_error({
        sampleFiles(newPipL) <- c("Donor1.fcs", "Donor1.fcs")}, 
        "does not contain unique values")
    
    expect_error({
        sampleFiles(newPipL) <- c(1,2)},
        "not a character vector")
    
    # test that order is driven by pData row names
    sampleFiles(newPipL) <- c("Donor2.fcs", "Donor1.fcs")
    expect_equal(rownames(pData(newPipL)), c("Donor1.fcs", "Donor2.fcs"))
})


test_that("CytoPipeline with complex flows raises no error", {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            experimentName <- "OMIP021_PeacoQC"
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                pattern = "Donor"
            ))

            # main parameters : sample files and output files
            pipL <- CytoPipeline(experimentName = experimentName,
                                 sampleFiles = sampleFiles)

            ### SCALE TRANSFORMATION STEPS ###

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "scale transform",
                    CytoProcessingStep(
                        name = "flowframe_read",
                        FUN = "readSampleFiles",
                        ARGS = list(
                            whichSamples = "all",
                            truncate_max_range = FALSE,
                            min.limit = NULL
                        )
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "scale transform",
                    CytoProcessingStep(
                        name = "remove_margins",
                        FUN = "removeMarginsPeacoQC",
                        ARGS = list()
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "scale transform",
                    CytoProcessingStep(
                        name = "compensate",
                        FUN = "compensateFromMatrix",
                        ARGS = list(matrixSource = "fcs")
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "scale transform",
                    CytoProcessingStep(
                        name = "flowframe_aggregate",
                        FUN = "aggregateAndSample",
                        ARGS = list(
                            nTotalEvents = 10000,
                            seed = 0
                        )
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "scale transform",
                    CytoProcessingStep(
                        name = "scale_transform_estimate",
                        FUN = "estimateScaleTransforms",
                        ARGS = list(
                            fluoMethod = "estimateLogicle",
                            scatterMethod = "linear",
                            scatterRefMarker = "BV785 - CD3"
                        )
                    )
                )

            ### PRE-PROCESSING STEPS ###

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "flowframe_read",
                        FUN = "readSampleFiles",
                        ARGS = list(
                            truncate_max_range = FALSE,
                            min.limit = NULL
                        )
                    )
                )


            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "remove_margins",
                        FUN = "removeMarginsPeacoQC",
                        ARGS = list()
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "compensate",
                        FUN = "compensateFromMatrix",
                        ARGS = list(matrixSource = "fcs")
                    )
                )

            pipL <-
                addProcessingStep(
                    pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "remove_doublets",
                        FUN = "removeDoubletsCytoPipeline",
                        ARGS = list(
                            areaChannels = c("FSC-A", "SSC-A"),
                            heightChannels = c("FSC-H", "SSC-H"),
                            nmads = c(3, 5))
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "remove_debris",
                        FUN = "removeDebrisManualGate",
                        ARGS = list(
                            FSCChannel = "FSC-A",
                            SSCChannel = "SSC-A",
                            gateData =  c(73615, 110174, 213000, 201000, 126000,
                                          47679, 260500, 260500, 113000, 35000)
                        )
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "remove_dead_cells",
                        FUN = "removeDeadCellsManualGate",
                        ARGS = list(
                            FSCChannel = "FSC-A",
                            LDMarker = "L/D Aqua - Viability",
                            gateData = c(0, 0, 250000, 250000,
                                         0, 650, 650, 0)
                        )
                    )
                )

            pipL <-
                addProcessingStep(
                    pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "perform_QC",
                        FUN = "qualityControlPeacoQC",
                        ARGS = list(
                            preTransform = TRUE,
                            min_cells = 150, # default
                            max_bins = 500, # default
                            step = 500, # default,
                            MAD = 6, # default
                            IT_limit = 0.55, # default
                            force_IT = 150, # default
                            peak_removal = 0.3333, # default
                            min_nr_bins_peakdetection = 10 # default
                        )
                    )
                )

            pipL <-
                addProcessingStep(pipL,
                    whichQueue = "pre-processing",
                    CytoProcessingStep(
                        name = "transform",
                        FUN = "applyScaleTransforms",
                        ARGS = list()
                    )
                )

            suppressWarnings(execute(pipL,
                rmCache = TRUE,
                path = outputDir
            ))

            suppressWarnings(execute(pipL,
                                     rmCache = FALSE,
                                     path = outputDir,
                                     saveLastStepFF = FALSE,
                                     saveScaleTransforms = TRUE
            ))
        },
        NA
    )
})

test_that("CytoPipeline with json input raises no error", {
    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    experimentName <- "OMIP021_PeacoQC"
    sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                    pattern = "Donor"))
    jsonDir <- system.file("extdata", package = "CytoPipeline")
    jsonPath <- file.path(jsonDir, "pipelineParams.json")
    
    expect_error(
        {
            pipL2 <- CytoPipeline(jsonPath,
                                  experimentName = experimentName,
                                  sampleFiles = sampleFiles)

            suppressWarnings(execute(pipL2,
                                     rmCache = TRUE,
                                     path = outputDir))
        },
        NA
    )
    # test if pheno data is correctly set with json parameters
    phenoData <- data.frame(row.names = basename(sampleFiles),
                            donor = c(1,2),
                            group = c("G1", "G1"))
    
    pipL3 <- CytoPipeline(jsonPath,
                          experimentName = experimentName,
                          sampleFiles = sampleFiles,
                          pData = phenoData)
    
    expect_true(all.equal(phenoData, pData(pipL3)))
    
    # now without row names
    phenoData2 <- data.frame(donor = c(1,2),
                             group = c("G1", "G1"))
    
    pipL4 <- CytoPipeline(jsonPath,
                          experimentName = experimentName,
                          sampleFiles = sampleFiles,
                          pData = phenoData2)
    
    expect_true(all.equal(phenoData, pData(pipL4)))
    
})

test_that("CytoPipeline with Biocparallel::Serial (by default) raises no error",
          {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                            pattern = "Donor"))

            jsonDir <- system.file("extdata", package = "CytoPipeline")
            jsonPath <- file.path(jsonDir, "pipelineParams.json")

            pipL2 <- CytoPipeline(jsonPath,
                                  sampleFiles = sampleFiles)

            # testing changing the experiment name on the fly
            experimentName(pipL2) <- "BPSerial_Experiment"

            bp <- BiocParallel::SerialParam()
            BiocParallel::register(bp, default = TRUE)
            suppressWarnings(execute(pipL2, path = outputDir,
                                     useBiocParallel = TRUE))
        },
        NA
    )
})

test_that("CytoPipeline with Biocparallel::SnowParam raises no error", {
    expect_error(
        {
            rawDataDir <-
                system.file("extdata", package = "CytoPipeline")
            sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                            pattern = "Donor"))

            jsonDir <- system.file("extdata", package = "CytoPipeline")
            jsonPath <- file.path(jsonDir, "pipelineParams.json")

            pipL2 <- CytoPipeline(jsonPath,
                                  experimentName = "BPSNOW_Experiment",
                                  sampleFiles = sampleFiles)

            logDir <- file.path(outputDir, "BiocParallel", "log")

            suppressWarnings(dir.create(logDir, recursive = TRUE))
            bp <- BiocParallel::SnowParam(workers = 2, log = TRUE,
                                          logdir = logDir,
                                          progressbar = TRUE)
            suppressWarnings(execute(pipL2, path = outputDir,
                                     useBiocParallel = TRUE,
                                     BPPARAM = bp, rmCache = TRUE))

        },
        NA
    )
})

test_that("CytoPipeline export as list works", {
    jsonDir <- system.file("extdata", package = "CytoPipeline")
    jsonPath <- file.path(jsonDir, "pipelineParams.json")

    pipL1 <- CytoPipeline(jsonPath)
    pipList <- as.list(pipL1)

    pipL2 <- CytoPipeline(pipList)
    expect_identical(pipL1, pipL2)

})

test_that("CytoPipeline rebuilt from cache raises no error", {
    expect_error(
        {
            experimentName <- "OMIP021_PeacoQC"
            pipL3 <- buildCytoPipelineFromCache(
                experimentName = experimentName,
                path = outputDir
            )
            suppressWarnings(execute(pipL3,
                rmCache = FALSE,
                path = outputDir
            ))
        },
        NA
    )
})


test_that("CytoPipeline not in cache with warning", {
    expect_warning(
        pipL4 <- buildCytoPipelineFromCache(
            experimentName = "non_existent",
            path = outputDir
        ),
        regexp = "no cache directory found"
    )
})


test_that("Check consistency with cache works", {
    rawDataDir <- system.file("extdata", package = "CytoPipeline")
    sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
        pattern = "Donor"
    ))

    pipL5 <- CytoPipeline(experimentName = "DummyExperiment")

    sampleFiles(pipL5) <- sampleFiles
    deleteCytoPipelineCache(pipL5, path = outputDir)

    pipL5 <- addProcessingStep(
        pipL5,
        "scale transform",
        CytoProcessingStep(
            name = "flowframe_read",
            FUN = "readSampleFiles",
            ARGS = list(
                whichSamples = "all",
                truncate_max_range = FALSE,
                min.limit = NULL
            )
        )
    )
    res <- checkCytoPipelineConsistencyWithCache(pipL5, path = outputDir)

    expect_error(suppressWarnings(execute(pipL5,
        rmCache = TRUE,
        path = outputDir
    )), NA)

    res <- checkCytoPipelineConsistencyWithCache(pipL5,
        path = outputDir
    )
    expect_equal(res$isConsistent, TRUE)
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")

    pipL5 <-
        addProcessingStep(pipL5,
            whichQueue = "pre-processing",
            CytoProcessingStep(
                name = "flowframe_read",
                FUN = "readSampleFiles",
                ARGS = list(
                    truncate_max_range = FALSE,
                    min.limit = NULL
                )
            )
        )
    res <- checkCytoPipelineConsistencyWithCache(pipL5,
        path = outputDir
    )
    expect_equal(res$isConsistent, TRUE)
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 1]), "not_run")
    expect_equal(unname(res$preProcessingStepStatus[1, 2]), "not_run")
    expect_equal(res$scaleTransformStepOutputObjNames, c("flowframe_read_obj"))
    expect_equal(res$scaleTransformStepOutputClasses, c("flowSet"))
    expect_equal(res$preProcessingStepOutputObjNames, c("unknown"))
    expect_equal(res$preProcessingStepOutputClasses, c("unknown"))

    expect_error(execute(pipL5,
        rmCache = FALSE,
        path = outputDir
    ), NA)
    res <- checkCytoPipelineConsistencyWithCache(pipL5,
        path = outputDir
    )
    expect_equal(res$isConsistent, TRUE)
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 2]), "run")
    expect_equal(res$scaleTransformStepOutputObjNames, c("flowframe_read_obj"))
    expect_equal(res$scaleTransformStepOutputClasses, c("flowSet"))
    expect_equal(res$preProcessingStepOutputObjNames, c("flowframe_read_obj"))
    expect_equal(res$preProcessingStepOutputClasses, c("flowFrame"))

    pipL5 <-
        addProcessingStep(pipL5,
            whichQueue = "pre-processing",
            CytoProcessingStep(
                name = "remove_margins",
                FUN = "removeMarginsPeacoQC",
                ARGS = list()
            )
        )

    res <- checkCytoPipelineConsistencyWithCache(pipL5,
        path = outputDir
    )
    expect_equal(res$isConsistent, TRUE)
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 2]), "run")
    expect_equal(unname(res$preProcessingStepStatus[2, 1]), "not_run")
    expect_equal(unname(res$preProcessingStepStatus[2, 2]), "not_run")
    expect_equal(res$scaleTransformStepOutputObjNames, c("flowframe_read_obj"))
    expect_equal(res$scaleTransformStepOutputClasses, c("flowSet"))
    expect_equal(
        res$preProcessingStepOutputObjNames,
        c("flowframe_read_obj", "unknown")
    )
    expect_equal(
        res$preProcessingStepOutputClasses,
        c("flowFrame", "unknown")
    )


    expect_error(suppressWarnings(execute(pipL5,
        rmCache = FALSE,
        path = outputDir
    )), NA)
    res <- checkCytoPipelineConsistencyWithCache(pipL5,
        path = outputDir
    )
    expect_equal(res$isConsistent, TRUE)
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 2]), "run")
    expect_equal(unname(res$preProcessingStepStatus[2, 1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[2, 2]), "run")
    expect_equal(res$scaleTransformStepOutputObjNames, c("flowframe_read_obj"))
    expect_equal(res$scaleTransformStepOutputClasses, c("flowSet"))
    expect_equal(
        res$preProcessingStepOutputObjNames,
        c("flowframe_read_obj", "remove_margins_obj")
    )
    expect_equal(
        res$preProcessingStepOutputClasses,
        c("flowFrame", "flowFrame")
    )

    pipL5_bad <- pipL5

    pipL5_bad@flowFramesPreProcessingQueue[[2]]@name <- "aaaaaa"
    res <- checkCytoPipelineConsistencyWithCache(pipL5_bad,
        path = outputDir
    )
    expect_equal(res$isConsistent, FALSE)
    expect_equal(
        res$inconsistencyMsg,
        paste0(
            "inconsistent pre-processing step #2 for sample file ",
            "Donor1.fcs (different in cache)"
        )
    )
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[2, 1]), "inconsistent")

    expect_error(suppressWarnings(execute(pipL5_bad,
        rmCache = FALSE,
        path = outputDir
    )),
    regexp = "inconsistent pre-processing step"
    )

    pipL5_bad2 <- pipL5
    pipL5_bad2@flowFramesPreProcessingQueue[[1]]@ARGS$truncate_max_range <-
        TRUE
    res <- checkCytoPipelineConsistencyWithCache(pipL5_bad2,
                                                 path = outputDir
    )
    expect_equal(res$isConsistent, FALSE)
    expect_equal(
        res$inconsistencyMsg,
        paste0(
            "inconsistent pre-processing step #1 for sample file ",
            "Donor1.fcs (different in cache)"
        )
    )
    expect_equal(unname(res$scaleTransformStepStatus[1]), "run")
    expect_equal(unname(res$preProcessingStepStatus[1, 1]), "inconsistent")
    expect_equal(unname(res$preProcessingStepStatus[2, 1]), "not_run")


    pipL5 <- removeProcessingStep(pipL5,
        whichQueue = "pre-processing",
        index = 2
    )

    res <- checkCytoPipelineConsistencyWithCache(pipL5,
                                                 path = outputDir
    )

    expect_equal(res$isConsistent, FALSE)
    expect_equal(
        res$inconsistencyMsg,
        paste0(
            "more pre-processing steps in cache than in CytoPipeline object"
        )
    )
})


test_that("plotCytoPipelineProcessingQueue works", {
    pipL6 <- CytoPipeline(experimentName = "DummyExperiment")

    deleteCytoPipelineCache(pipL6, path = outputDir)

    rawDataDir <- system.file("extdata", package = "CytoPipeline")
    sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
        pattern = "Donor"
    ))

    # put only second sample file for the time being
    sampleFiles(pipL6) <- sampleFiles[2]
    pipL6 <- addProcessingStep(
        pipL6,
        "scale transform",
        CytoProcessingStep(
            name = "flowframe_read",
            FUN = "readSampleFiles",
            ARGS = list(
                whichSamples = "all",
                truncate_max_range = FALSE,
                min.limit = NULL
            )
        )
    )

    expect_error(plotCytoPipelineProcessingQueue(
        pipL6,
        whichQueue = "scale transform",
        path = outputDir
    ), NA)

    expect_error(suppressWarnings(execute(pipL6,
        rmCache = TRUE,
        path = outputDir
    )), NA)

    expect_error(plotCytoPipelineProcessingQueue(
        pipL6,
        whichQueue = "scale transform",
        path = outputDir
    ), NA)
    expect_message(plotCytoPipelineProcessingQueue(
        pipL6,
        whichQueue = "pre-processing",
        path = outputDir
    ),
    regexp = "no sample file passed"
    )

    pipL6 <-
        addProcessingStep(pipL6,
            whichQueue = "pre-processing",
            CytoProcessingStep(
                name = "flowframe_read",
                FUN = "readSampleFiles",
                ARGS = list(
                    truncate_max_range = FALSE,
                    min.limit = NULL
                )
            )
        )

    expect_error(plotCytoPipelineProcessingQueue(
        pipL6,
        whichQueue = "scale transform",
        path = outputDir
    ), NA)
    expect_message(plotCytoPipelineProcessingQueue(
        pipL6,
        whichQueue = "pre-processing",
        path = outputDir
    ),
    regexp = "no sample file passed"
    )

    execute(pipL6, rmCache = FALSE, path = outputDir)

    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "scale transform",
        path = outputDir
    )
    expect_error(plotCytoPipelineProcessingQueue(pipL6,
        sampleFile = 1,
        whichQueue = "pre-processing",
        path = outputDir
    ), NA)
    expect_error(plotCytoPipelineProcessingQueue(pipL6,
        sampleFile = 2,
        whichQueue = "pre-processing",
        path = outputDir
    ),
    regexp = "out of bounds"
    )


    pipL6 <-
        addProcessingStep(pipL6,
            whichQueue = "pre-processing",
            CytoProcessingStep(
                name = "remove_margins",
                FUN = "removeMarginsPeacoQC",
                ARGS = list()
            )
        )
    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "scale transform",
        path = outputDir
    )
    plotCytoPipelineProcessingQueue(pipL6,
        sampleFile = 1,
        whichQueue = "pre-processing",
        path = outputDir
    )

    suppressWarnings(execute(pipL6,
        rmCache = FALSE,
        path = outputDir
    ))

    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "scale transform",
        path = outputDir
    )
    plotCytoPipelineProcessingQueue(pipL6,
        sampleFile = 1,
        whichQueue = "pre-processing",
        path = outputDir
    )

    # add first sample file to see the impact
    sampleFiles(pipL6) <- sampleFiles

    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "scale transform",
        path = outputDir
    )
    # following should show yellow boxes
    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "pre-processing",
        path = outputDir
    )
    # following should show bow in green
    plotCytoPipelineProcessingQueue(pipL6,
        sampleFile = 2,
        whichQueue = "pre-processing",
        path = outputDir
    )

    suppressWarnings(execute(pipL6,
        rmCache = FALSE,
        path = outputDir
    ))

    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "scale transform",
        path = outputDir
    )
    # following should now show the green box
    plotCytoPipelineProcessingQueue(pipL6,
        whichQueue = "pre-processing",
        path = outputDir
    )
    # following as well
    plotCytoPipelineProcessingQueue(pipL6,
        sampleFile = sampleFiles[2],
        whichQueue = "pre-processing",
        path = outputDir
    )


    pipL6@flowFramesPreProcessingQueue[[2]]@name <- "aaaaaa"

    expect_error(plotCytoPipelineProcessingQueue(
        pipL6,
        whichQueue = "scale transform",
        path = outputDir
    ),
    NA
    )

    expect_warning(plotCytoPipelineProcessingQueue(
        pipL6,
        sampleFile = 1,
        whichQueue = "pre-processing",
        path = outputDir
    ),
    regexp = "CytoPipeline object not consistent with cache"
    )


    expect_error(execute(pipL6,
        rmCache = FALSE,
        path = outputDir
    ), regexp = "inconsistent")
})

test_that("getCytoPipelineObject works", {

    expect_error(
        {
            experimentName <- "OMIP021_PeacoQC"

            pipL7 <- buildCytoPipelineFromCache(
                experimentName = experimentName,
                path = outputDir
            )

            plotCytoPipelineProcessingQueue(pipL7,
                sampleFile = 1,
                whichQueue = "pre-processing",
                path = outputDir
            )

            getCytoPipelineObjectInfos(pipL7,
                whichQueue = "pre-processing",
                sampleFile = sampleFiles(pipL7)[1],
                path = outputDir
            )
            getCytoPipelineObjectInfos(pipL7,
                                       whichQueue = "pre-processing",
                                       sampleFile = 1,
                                       path = outputDir
            )
            getCytoPipelineObjectInfos(pipL7,
                whichQueue = "scale transform",
                sampleFile = sampleFiles(pipL7)[1],
                path = outputDir
            )

            ffFrom <- getCytoPipelineFlowFrame(pipL7,
                whichQueue = "pre-processing",
                sampleFile = sampleFiles(pipL7)[1],
                objectName = "compensate_obj",
                path = outputDir
            )

            ffTo <- getCytoPipelineFlowFrame(pipL7,
                whichQueue = "pre-processing",
                sampleFile = sampleFiles(pipL7)[1],
                objectName = "remove_doublets_obj",
                path = outputDir
            )

            ggplotFilterEvents(ffFrom, ffTo,
                xChannel = "FSC-A", yChannel = "FSC-H"
            )

            plotCytoPipelineProcessingQueue(pipL7,
                whichQueue = "scale transform",
                path = outputDir
            )
        },
        NA
    )

    expect_error(getCytoPipelineObjectInfos(pipL7,
                                            whichQueue = "pre-processing",
                                            sampleFile = 3,
                                            path = outputDir),
                 "out of bounds")


    expect_error(
        getCytoPipelineScaleTransform(
            pipL7,
            whichQueue = "scale transform",
            objectName = "flowframe_aggregate_obj",
            path = outputDir
        ),
        regexp = "does not appear to be a transformList"
    )


    expect_error(
        transList <-
            getCytoPipelineScaleTransform(
                pipL7,
                whichQueue = "scale transform",
                objectName = "scale_transform_estimate_obj",
                path = outputDir
            ),
        NA
    )

    expect_error(
        ffAgg <-
            getCytoPipelineFlowFrame(pipL7,
                whichQueue = "scale transform",
                objectName = "flowframe_aggregate_obj",
                path = outputDir
            ),
        NA
    )


})

test_that("collectNbOfRetainedEvents works", {

    experimentName <- "OMIP021_PeacoQC"

    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                    pattern = "Donor"
    ))

    stepNames <- c("flowframe_read", "remove_margins", "compensate",
                   "remove_doublets", "remove_debris", "remove_dead_cells",
                   "perform_QC", "transform")

    # with missing whichSampleFiles argument
    nbEventsDF1 <- collectNbOfRetainedEvents(
        experimentName = experimentName,
        path = outputDir
    )

    expect_equal(unname(colnames(nbEventsDF1)), stepNames)
    expect_equal(unname(rownames(nbEventsDF1)), basename(sampleFiles))
    expect_equal(nbEventsDF1[1,1], 5000)
    expect_equal(nbEventsDF1[1,2], 4494)
    expect_equal(nbEventsDF1[1,3], 4494)
    expect_equal(nbEventsDF1[1,4], 3541)
    expect_equal(nbEventsDF1[1,5], 2983)
    expect_equal(nbEventsDF1[1,6], 2888)
    expect_equal(nbEventsDF1[1,7], 2000)
    expect_equal(nbEventsDF1[1,8], 2000)
    expect_equal(nbEventsDF1[2,1], 5000)
    expect_equal(nbEventsDF1[2,2], 4700)
    expect_equal(nbEventsDF1[2,3], 4700)
    expect_equal(nbEventsDF1[2,4], 3809)
    expect_equal(nbEventsDF1[2,5], 3347)
    expect_equal(nbEventsDF1[2,6], 3306)
    expect_equal(nbEventsDF1[2,7], 2500)
    expect_equal(nbEventsDF1[2,8], 2500)

    # with explicit whichSampleFiles argument
    nbEventsDF2 <- collectNbOfRetainedEvents(
        experimentName = experimentName,
        path = outputDir,
        whichSampleFiles = basename(sampleFiles[1])
    )

    expect_equal(unname(colnames(nbEventsDF2)), stepNames)
    expect_equal(unname(rownames(nbEventsDF2)), basename(sampleFiles[1]))
    expect_equal(nbEventsDF2[1,1], 5000)
    expect_equal(nbEventsDF2[1,2], 4494)
    expect_equal(nbEventsDF2[1,3], 4494)
    expect_equal(nbEventsDF2[1,4], 3541)
    expect_equal(nbEventsDF2[1,5], 2983)
    expect_equal(nbEventsDF2[1,6], 2888)
    expect_equal(nbEventsDF2[1,7], 2000)
    expect_equal(nbEventsDF2[1,8], 2000)

    # with wrong whichSampleFiles argument
    expect_error(
        collectNbOfRetainedEvents(
            experimentName = experimentName,
            path = outputDir,
            whichSampleFiles = 3
        ), regexp = "whichSampleFiles out of bounds")

})

test_that("getMinimumUniqueFileNames works", {
    expect_error(.getMinimumUniqueFileNames(1),
                 "is not TRUE")
    
    fileNames <- "C:Users/folder1/file1.csv"
    ret <- .getMinimumUniqueFileNames(fileNames)
    expect_equal(ret, basename(fileNames))
    
    fileNames <- c(
        "C:Users/folder1/file1.csv",
        "C:Users/folder1/file2.csv")
    
    ret <- .getMinimumUniqueFileNames(fileNames)
    expect_equal(ret, basename(fileNames))
    
    fileNames <- c(
        "../../folder1/subfolder1/file1.csv",
        "../../folder1/subfolder2/file1.csv",
        "../../folder1/subfolder2/file2.csv")
    
    expected_ret <- c(
        "subfolder1/file1.csv",
        "subfolder2/file1.csv",
        "subfolder2/file2.csv")
    
    ret <- .getMinimumUniqueFileNames(fileNames)
    expect_equal(ret, expected_ret)    
    
    fileNames <- c(
        "../../folder1/subfolder1/file1.csv",
        "../../folder1/file1.csv",
        "../../folder1/subfolder2/file2.csv")
    
    expected_ret <- c(
        "subfolder1/file1.csv",
        "folder1/file1.csv",
        "subfolder2/file2.csv")
    
    ret <- .getMinimumUniqueFileNames(fileNames)
    expect_equal(ret, expected_ret)    

    fileNames <- c(
        "/folder1/subfolder1/file1.csv",
        "/folder2/subfolder1/file1.csv")    
    
    expected_ret <- c(
        "folder1/subfolder1/file1.csv",
        "folder2/subfolder1/file1.csv")
    
    ret <- .getMinimumUniqueFileNames(fileNames)
    expect_equal(ret, expected_ret)   
    
    fileNames <- c(
        "/file1.csv",
        "/folder1/subfolder1/file1.csv")    
    
    expected_ret <- c(
        "/file1.csv",
        "subfolder1/file1.csv")
    
    ret <- .getMinimumUniqueFileNames(fileNames)
    expect_equal(ret, expected_ret)   

    fileNames <- c("/folder1/file1.csv",
                   "/folder1/file1.csv")    
    
    expect_error(.getMinimumUniqueFileNames(fileNames),
                 "provided file names are not unique")
    
})


test_that("display names do work", {

    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    experimentName <- "OMIP021_PeacoQC"
    sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                    pattern = "Donor"
    ))
    
    phenoData <- data.frame(donor = c(1,2),
                            group = c("G1", "G1"))
    
    # main parameters : sample files and output files
    pipelineParams <- list()
    pipelineParams$experimentName <- experimentName
    pipelineParams$sampleFiles <- sampleFiles
    pipelineParams$pData <- phenoData
    
    pipL <- CytoPipeline(pipelineParams)
    
    ret <- sampleDisplayNames(pipL)
    
    expected_ret <- c("Donor1.fcs", "Donor2.fcs")
    
    expect_equal(ret, expected_ret)
    
    ret <- sampleDisplayNames(pipL, c(1,2))
    
    expect_equal(ret, expected_ret)
    
    ret <- sampleDisplayNames(pipL, 2)
    
    expect_equal(ret, expected_ret[2])
    
    ret <- sampleDisplayNames(pipL, sampleFiles(pipL)[1])
    
    expect_equal(ret, expected_ret[1])
    
    phenoData <- data.frame(displayName = c("S1", "S2"),
                            donor = c(1,2),
                            group = c("G1", "G1"))
    
    pData(pipL) <- phenoData
    
    ret <- sampleDisplayNames(pipL)
    
    expected_ret <- c("S1", "S2")
    
    expect_equal(ret, expected_ret)
    
    ret <- sampleNameFromDisplayName(pipL, "S2")
    
    expect_equal(ret, sampleFiles(pipL)[2])
    
})
                   
    

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

##' @include utils.R
NULL


#' @title plot flow rate as a function of time, using ggplot2
#'
#' @param obj       a flowCore::flowFrame or flowCore::flowSet
#' @param title     a title for the graph
#' @param timeUnit  which time interval is used to calculate "instant" flow rate
#'                  (default = 100 ms)
#'
#' @return a ggplot graph
#' @import ggplot2
#' @export
#'
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' # single flowFrame plot
#' ggplotFlowRate(OMIP021Samples[[1]])
#' 
#' # two flowFrames plot 
#' ggplotFlowRate(OMIP021Samples)
#' 
#' # single plot with title
#' ggplotFlowRate(OMIP021Samples[[1]], title = "Test Flow Rate plot")
#' 
#' # explicit time unit
#' ggplotFlowRate(OMIP021Samples[[1]], timeUnit = 50)
#' 
ggplotFlowRate <- function(obj, title = "Flow Rate", timeUnit = 100) {
    isFlowSet <- FALSE
    if (inherits(obj, "flowSet")) {
        isFlowSet <- TRUE
    } else if (inherits(obj, "flowFrame")) {
    } else {
        stop("obj type not recognized, should be a flowFrame or flowSet")
    }

    flowRateDataFrame <- function(ff, timeUnit) {
        Y <- flowCore::exprs(ff)
        timeChName <- findTimeChannel(ff)
        if (is.null(timeChName)) {
            stop(
                "No Time columns in flow frame ",
                "=> impossible to calculate flow rate"
            )
        }

        h <- graphics::hist(
            flowCore::exprs(ff)[, timeChName],
            breaks = seq(min(flowCore::exprs(ff)[, timeChName]),
                max(flowCore::exprs(ff)[, timeChName]) +
                    timeUnit,
                by = timeUnit
            ),
            plot = FALSE
        )

        df <- data.frame(
            time = h$mids,
            nbEvents = h$counts * 1000 / timeUnit,
            name = flowCore::identifier(ff)
        )

        df
    }


    if (isFlowSet) {
        df <- flowCore::fsApply(obj,
            FUN = flowRateDataFrame,
            timeUnit = timeUnit
        )
        df <- do.call(rbind.data.frame, df)
    } else {
        df <- flowRateDataFrame(obj, timeUnit = timeUnit)
    }

    # following 2 statements just to allow R cmd CHECK w/o note
    time <- NULL
    nbEvents <- NULL
    pTime <- ggplot(df) +
        theme_gray() +
        geom_point(aes(x = time, y = nbEvents)) +
        geom_line(aes(x = time, y = nbEvents)) +
        ggtitle(title) +
        facet_wrap(~name) +
        xlab("Time") +
        ylab("Nr of events per second")
    return(pTime)
}

# internal function to select the appropriate logicle scale to use from ggCyto
.my_scale_logicle <- function(axis = c("x", "y"),
                              usedScale = c("logicle", "flowjo_biexp"),
                              logicleParams,
                              axisRange,
                              interactive = FALSE) {

    axis <- match.arg(axis)
    usedScale <- match.arg(usedScale)
    
    if (usedScale == "logicle") {
        theArgs <- logicleParams
    } else {
        theArgs <- list(maxValue = logicleParams$t,
                        widthBasis = -10^logicleParams$w,
                        pos = logicleParams$m - logicleParams$w,
                        neg = logicleParams$a)
    }
    theArgs$limits <- axisRange
    if (interactive) {
        theArgs$label <- scales::scientific_format()
    }
    
    if (axis == "x") {
        if (usedScale == "logicle") {
            do.call(scale_x_logicle,
                    args = theArgs)
        } else {
            do.call(scale_x_flowjo_biexp,
                    args = theArgs)
        }
    } else {
        if (usedScale == "logicle") {
            do.call(scale_y_logicle,
                    args = theArgs)
        } else {
            do.call(scale_y_flowjo_biexp,
                    args = theArgs)
        } 
    }
}

#' @title plot events in 1D or 2D, using ggplot2
#' @description  plot events of specific channels of either :
#' flowCore::flowFrame, or flowCore::flowSet
#' in 2D or 1D, mimicking FlowJo type of graph. \cr
#' if 1D : geom_density will be used \cr
#' if 2D : geom_hex will be used \cr
#'
#' @param obj a flowCore::flowFrame or flowCore::flowSet
#' @param xChannel channel (name or index) or marker name to be displayed
#' on x axis
#' @param yChannel channel (name or index) or marker name to be displayed
#' on y axis
#' @param nDisplayCells maximum number of events that will be plotted. If
#' the number of events exceed this number, a sub-sampling will be performed
#' @param seed seed used for sub-sampling (if any)
#' @param bins used in geom_hex
#' @param fill used in geom_density
#' @param alpha used in geom_density
#' @param xScale scale to be used for the x axis 
#' (note "linear" corresponds to no transformation)
#' @param yScale scale to be used for the y axis
#' (note "linear" corresponds to no transformation)
#' @param xLogicleParams if (xScale == "logicle"), the parameters of the logicle
#' transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...).
#' If NULL, these parameters will be estimated by flowCore::estimateLogicle()
#' @param yLogicleParams if (yScale == "logicle"), the parameters of the logicle
#' transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...).
#' If NULL, these parameters will be estimated by flowCore::estimateLogicle()
#' @param xLinearRange if (xScale == "linear"), the x axis range to be used
#' @param yLinearRange if (yScale == "linear"), the y axis range to be used
#' @param transList optional list of scale transformations to be applied to each
#' channel. If it is non null, 'x/yScale', 'x/yLogicleParams' and
#' 'x/yLinear_range' will be discarded.
#' @param runTransforms (TRUE/FALSE) Will the application of non linear scale
#' result in data being effectively transformed ?
#' - If TRUE, than the data will undergo transformations prior to
#' visualization.
#' - If FALSE, the axis will be scaled but the data themselves will not be
#' transformed.
#' @return a list of ggplot objects
#' @import ggplot2
#' @importFrom ggcyto scale_x_logicle
#' @importFrom ggcyto scale_y_logicle
#' @importFrom ggcyto scale_x_flowjo_biexp
#' @importFrom ggcyto scale_y_flowjo_biexp
#' @importFrom rlang .data
#' @export
#'
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' ### 1D Examples ###
#' 
#' # simple linear scale example
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "FSC-A",
#'              xScale = "linear")
#' 
#' # with explicit linear range
#' ggplotEvents(OMIP021Samples[[1]],
#'                   xChannel = "FSC-A",
#'                   xScale = "linear",
#'                   xLinearRange = c(0, 250000))
#' 
#' # with linear scale, several flow frames
#' ggplotEvents(OMIP021Samples, xChannel = "FSC-A", xScale = "linear")
#' 
#' # simple logicle scale example
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "logicle")
#' 
#' # logicle scale, explicit parameters
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "logicle", xLogicleParams = list(
#'                  a = 1,
#'                  w = 2,
#'                  m = 7,
#'                  t = 270000))
#' 
#' # with sub-sampling
#' ggplotEvents(OMIP021Samples[[2]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "logicle", nDisplayCells = 5000)
#' 
#' # tuning some plot parameters
#' ggplotEvents(OMIP021Samples[[2]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "logicle", alpha = 0.5, fill = "red")
#' 
#' # examples that use a transformation list, estimated after compensation
#' compensationMatrix <- flowCore::spillover(OMIP021Samples[[1]])$SPILL
#' 
#' ffC <- runCompensation(OMIP021Samples[[1]],
#'                        spillover = compensationMatrix,
#'                        updateChannelNames = FALSE)
#' 
#' transList <- flowCore::estimateLogicle(
#'     ffC,
#'     colnames(compensationMatrix))
#' 
#' transList <-
#'     c(transList,
#'       flowCore::transformList(
#'           "FSC-A",
#'           flowCore::linearTransform(a = 0.00001)))
#' 
#' # linear example, without running the transformations on data
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "linear", 
#'              transList = transList,
#'              runTransforms = FALSE)
#' 
#' # linear example, now running the transformations on data
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "linear", 
#'              transList = transList,
#'              runTransforms = TRUE)
#' 
#' # logicle example, without running the transformations on data
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "FSC-A",
#'              xScale = "logicle", 
#'              transList = transList,
#'              runTransforms = FALSE)
#' 
#' # logicle example, now running the transformations on data
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "FSC-A",
#'              xScale = "logicle", 
#'              transList = transList,
#'              runTransforms = TRUE)
#' 
#' ### 2D examples ###
#' 
#' 
#' # simple linear example
#' ggplotEvents(OMIP021Samples[[1]],
#'                   xChannel = "FSC-A",
#'                   xScale = "linear",
#'                   yChannel = "610/20Violet-A",
#'                   yScale = "logicle")
#' 
#' # simple linear example, 2 flow frames
#' ggplotEvents(OMIP021Samples,
#'              xChannel = "FSC-A",
#'              xScale = "linear",
#'              yChannel = "SSC-A",
#'              yScale = "linear")
#' 
#' # logicle vs linear example
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "450/50Violet-A",
#'              xScale = "logicle",
#'              yChannel = "SSC-A",
#'              yScale = "linear")
#' 
#' # 2X logicle example
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "TETaGC",
#'              xScale = "logicle",
#'              yChannel = "CD27",
#'              yScale = "logicle")
#' 
#' # tuning nb of bins
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "TETaGC",
#'              xScale = "logicle",
#'              yChannel = "CD27",
#'              yScale = "logicle",
#'              bins = 128)
#' 
#' # using transformation list, not run on data
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "TETaGC",
#'              xScale = "logicle",
#'              yChannel = "CD27",
#'              yScale = "logicle",
#'              transList = transList,
#'              runTransforms = FALSE)
#' 
#' # using transformation list, run on data                  
#' ggplotEvents(OMIP021Samples[[1]],
#'              xChannel = "TETaGC",
#'              xScale = "logicle",
#'              yChannel = "CD27",
#'              yScale = "logicle",
#'              transList = transList,
#'              runTransforms = TRUE)
#' 
ggplotEvents <- function(obj,
                         xChannel,
                         yChannel = NULL,
                         nDisplayCells = Inf,
                         seed = NULL,
                         bins = 216,
                         fill = "lightblue",
                         alpha = 0.2,
                         xScale = c("linear", "logicle"),
                         yScale = c("linear", "logicle"),
                         xLogicleParams = NULL,
                         yLogicleParams = NULL,
                         xLinearRange = NULL,
                         yLinearRange = NULL,
                         transList = NULL,
                         runTransforms = FALSE) {


    #browser()
    
    xScale <- match.arg(xScale)
    yScale <- match.arg(yScale)

    isFlowSet <- FALSE
    if (inherits(obj, "flowSet")) {
        fr <- obj[[1]]
        isFlowSet <- TRUE
    } else if (inherits(obj, "flowFrame")) {
        fr <- obj
    } else {
        stop("obj type not recognized, should be a flowFrame or flowSet")
    }
    
    # selection of scale_logical function to use (temporary fix due to ggcyto)
    used_logical_scale <- "logicle"
    #used_logical_scale <- "flowjo_biexp"

    # find channel and marker names to specify axis labels

    xChMk <- flowCore::getChannelMarker(fr, xChannel)
    xChannel <- xChMk$name
    xLabel <- xChannel
    if (!is.na(xChMk$desc)) {
        xLabel <- paste0(xLabel, " : ", xChMk$desc)
    }

    if (!is.null(yChannel)) {
        yChMk <- flowCore::getChannelMarker(fr, yChannel)
        yChannel <- yChMk$name
        yLabel <- yChannel
        if (!is.na(yChMk$desc)) {
            yLabel <- paste0(yLabel, " : ", yChMk$desc)
        }
    }
    
    # reduce flow frame or flow set to the displayed channels
    selectedCols <- xChannel
    if (!is.null(yChannel)){
        selectedCols <- c(selectedCols, yChannel)
    }
    if (isFlowSet) {
        obj <- flowCore::fsApply(obj,
                                 FUN = function(ff, selectedCols){
                                     ff <- ff[, selectedCols]
                                 },
                                 selectedCols = selectedCols
        )    
    } else {
        obj <- obj[, selectedCols]
    }
    
    # build scale transforms for not linear scales,
    # when these are not directly provided by the user
    if (is.null(transList) && (xScale != "linear" || yScale != "linear")) {
        estimatedCols <- c()
        if (xScale == "logicle" && is.null(xLogicleParams)) {
            estimatedCols <- xChannel
        }
        if (!is.null(yChannel) && yScale == "logicle" && 
            is.null(yLogicleParams)) {
            estimatedCols <- c(estimatedCols, yChannel)
        }
        if (!is.null(estimatedCols)) {
            if (isFlowSet){
                transList <- flowCore::estimateLogicle(obj[[1]], estimatedCols)
            } else {
                transList <- flowCore::estimateLogicle(obj, estimatedCols)
            }
        }
        
        # add the self provided logicle parameters
        if (xScale == "logicle" && !is.null(xLogicleParams)){
            xLogicleTrans <- flowCore::logicleTransform(
                w = xLogicleParams$w,
                t = xLogicleParams$t,
                m = xLogicleParams$m,
                a = xLogicleParams$a
            )
            if (is.null(transList)) {
                transList <- flowCore::transformList(from = xChannel,
                                                     tfun = xLogicleTrans)
            } else {
                transList@transforms[[xChannel]] <- xLogicleTrans
            }
        }
        if (!is.null(yChannel) && yScale == "logicle" &&
            !is.null(yLogicleParams)) {
            yLogicleTrans <- flowCore::logicleTransform(
                w = yLogicleParams$w,
                t = yLogicleParams$t,
                m = yLogicleParams$m,
                a = yLogicleParams$a
            )
            if (is.null(transList)) {
                transList <- flowCore::transformList(from = yChannel,
                                                     tfun = yLogicleTrans)
            } else {
                transList@transforms[[yChannel]] <- yLogicleTrans
            }
        }
    }

    # perform sub-sampling if necessary

    if (nDisplayCells < Inf) {
        if (nDisplayCells < 1) {
            stop("n_display_cells should be strictly positive!")
        }
        if (isFlowSet) {
            obj <- flowCore::fsApply(obj,
                FUN = subsample,
                nEvents = nDisplayCells,
                seed = seed
            )
        } else {
            obj <- subsample(obj, nEvents = nDisplayCells, seed = seed)
        }
    }

    xTransformed <- FALSE
    yTransformed <- FALSE
    if (!is.null(transList)) {
        res <- getTransfoParams(transList, xChannel)
        if (!is.null(res)) {
            if (runTransforms) {
                xScale <- "linear"
                xTransformed <- TRUE
                if (res$type == "logicle") {
                    xLinearRange <- c(0, res$paramsList$m)
                }
            } else {
                xScale <- res$type
                if (res$type == "logicle") {
                    xLogicleParams <- res$paramsList
                }
            }
        }

        if (!is.null(yChannel)) {
            res <- getTransfoParams(transList, yChannel)
            if (!is.null(res)) {
                if (runTransforms) {
                    yScale <- "linear"
                    yTransformed <- TRUE
                    if (res$type == "logicle") {
                        yLinearRange <- c(0, res$paramsList$m)
                    }
                } else {
                    yScale <- res$type
                    if (res$type == "logicle") {
                        yLogicleParams <- res$paramsList
                    }
                }
            }
        }

        if (runTransforms) {
            # adapt scale to linear and linear_range to null
            # if trans_list is passed and is effectively run
            appliedTransList <- c()

            # remove not mentioned channels from trans_list
            appliedTransList <- transList
            transChannels <- names(transList@transforms)
            for (n in transChannels) {
                if (!(n %in% c(xChannel, yChannel))) {
                    appliedTransList@transforms[[n]] <- NULL
                }
            }
            if (isFlowSet) {
                obj <- flowCore::fsApply(obj, FUN = function(ff, transList) {
                    flowCore::transform(obj, translist = transList)
                }, transList = appliedTransList)
            } else {
                obj <- flowCore::transform(obj, translist = appliedTransList)
            }
        }
    }

    # find axis ranges depending on scales

    xScale <- match.arg(xScale)
    yScale <- match.arg(yScale)

    if (xScale == "logicle") {
        if (is.null(xLogicleParams)) {
            stop("xLogicleParams can't be NULL if xScale == logicle")
        } else {
            xAxesLimits <- c(0, xLogicleParams$m)
            myTrans <-
                do.call(flowCore::logicleTransform, args = xLogicleParams)
            myInverseTrans <- flowCore::inverseLogicleTransform(myTrans)
            myXRange <- myInverseTrans@.Data(xAxesLimits)
        }
    }
    if (yScale == "logicle") {
        if (is.null(yLogicleParams)) {
            stop("yLogicleParams can't be NULL if yScale == logicle")
        } else {
            yAxesLimits <- c(0, yLogicleParams$m)
            myTrans <-
                do.call(flowCore::logicleTransform, args = yLogicleParams)
            myInverseTrans <- flowCore::inverseLogicleTransform(myTrans)
            myYRange <- myInverseTrans@.Data(yAxesLimits)
        }
    }

    # main plot layer

    if (xTransformed) {
        xLabel <- paste0(xLabel, " (transformed)")
    }
    if (yTransformed) {
        yLabel <- paste0(yLabel, " (transformed)")
    }
    
    # browser()

    if (is.null(yChannel)) {
        p <- ggplot(
            data = obj,
            mapping = aes(x = .data[[xChannel]])
        ) +
            geom_density(fill = fill, alpha = alpha, na.rm = TRUE) +
            xlab(xLabel)
    } else {
        p <- ggplot(
            data = obj,
            mapping = aes(
                x = .data[[xChannel]],
                y = .data[[yChannel]])) +
            geom_hex(bins = bins, na.rm = TRUE) +
            scale_fill_gradientn(
            colours = grDevices::rainbow(
                n = 5, end = 0.65,
                rev = TRUE,
                s = 0.8, v = 0.8)) + 
            xlab(xLabel) +
            ylab(yLabel)
    }


    # add faceting

    p <- p + facet_wrap(~name) +
        theme(legend.position = "none")

    # add scales for x and y, as well as axis limits

    if (xScale == "logicle") {
        if (is.null(yChannel)) {
            p <- p +
                .my_scale_logicle(axis = "x", 
                                  usedScale = used_logical_scale,
                                  logicleParams = xLogicleParams,
                                  axisRange = myXRange) #+
                #coord_cartesian(xlim = myXRange)
        } else if (yScale == "logicle") {
            p <- p + 
                .my_scale_logicle(axis = "x", 
                                  usedScale = used_logical_scale,
                                  logicleParams = xLogicleParams,
                                  axisRange = myXRange) + 
                .my_scale_logicle(axis = "y", 
                                  usedScale = used_logical_scale,
                                  logicleParams = yLogicleParams,
                                  axisRange = myYRange)  #+
                # coord_cartesian(
                #     xlim = myXRange,
                #     ylim = myYRange
                # )
        } else {
            p <- p +
                .my_scale_logicle(axis = "x", 
                                  usedScale = used_logical_scale,
                                  logicleParams = xLogicleParams,
                                  axisRange = myXRange) + 
                scale_y_continuous(limits = yLinearRange)
                # coord_cartesian(
                #     xlim = myXRange,
                #     ylim = yLinearRange
                # )
        }
    } else {
        if (is.null(yChannel)) {
            p <- p +
                scale_x_continuous(limits = xLinearRange)
                #coord_cartesian(xlim = xLinearRange)
        } else if (yScale == "logicle") {
            p <- p +
                scale_x_continuous(limits = xLinearRange) + 
                .my_scale_logicle(axis = "y", 
                                  usedScale = used_logical_scale,
                                  logicleParams = yLogicleParams,
                                  axisRange = myYRange) #+
                # coord_cartesian(
                #     xlim = xLinearRange,
                #     ylim = myYRange
                # )
        } else {
            p <- p +
                scale_x_continuous(limits = xLinearRange) + 
                scale_y_continuous(limits = yLinearRange) # +
                # coord_cartesian(
                #     xlim = xLinearRange,
                #     ylim = yLinearRange
                # )
        }
    }

    p
}


#' @title plot filtered events in 2D, using ggplot
#' @description  plot events of specific channels of either :
#' flowCore::flowFrame, or
#' flowCore::flowSet
#' in 2D, showing the impact of applying a filter between :
#' - a 'pre' flowframe
#  - a 'post' flowframe
#'
#' @param ffPre a flowCore::flowFrame, before applying filter
#' @param ffPost a flowCore::flowFrame, after applying filter
#' @param xChannel channel (name or index) or marker name to be displayed
#' on x axis
#' @param yChannel channel (name or index) or marker name to be displayed
#' on y axis
#' @param nDisplayCells maximum number of events that will be plotted. If
#' the number of events exceed this number, a subsampling will be performed
#' @param seed seed used for sub-sampling (if any)
#' @param size used by geom_point()
#' @param xScale scale to be used for the x axis
#' (note "linear" corresponds to no transformation)
#' @param yScale scale to be used for the y axis
#' (note "linear" corresponds to no transformation)
#' @param xLogicleParams if (xScale == "logicle"), the parameters of the logicle
#' transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...)
#' If NULL, these parameters will be estimated by flowCore::estimateLogicle()
#' @param yLogicleParams if (yScale == "logicle"), the parameters of the logicle
#' transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...)
#' If NULL, these parameters will be estimated by flowCore::estimateLogicle()
#' @param xLinearRange if (xScale == "linear"), linear range to be used
#' @param yLinearRange if (yScale == "linear"), linear range to be used
#' @param transList optional list of scale transformations to be applied to each
#' channel. If it is non null, 'x/yScale', 'x/yLogicleParams' and
#' 'x/yLinear_range' will be discarded.
#' @param runTransforms (TRUE/FALSE) Will the application of non linear scale
#' result in data being effectively transformed ?
#' - If TRUE, than the data will undergo transformations prior to
#' visualization.
#' - If FALSE, the axis will be scaled but the data themselves are
#' not transformed.
#' @param interactive if TRUE, transform the scaling formats such that the
#' ggcyto::x_scale_logicle() and ggcyto::y_scale_logicle() do work with
#' plotly::ggplotly()
#'
#' @return a ggplot object
#' @import ggplot2
#' @importFrom ggcyto scale_x_logicle
#' @importFrom ggcyto scale_y_logicle
#' @importFrom ggcyto scale_x_flowjo_biexp
#' @importFrom ggcyto scale_y_flowjo_biexp
#' @importFrom rlang .data
#' @export
#'
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' ffPre <- OMIP021Samples[[1]]
#' 
#' # creating a manual polygon gate filter based on channels L/D and FSC-A
#' 
#' LDMarker <- "L/D Aqua - Viability"
#' 
#' LDChannel <- getChannelNamesFromMarkers(ffPre, markers = LDMarker)
#' liveGateMatrix <- matrix(
#'     data = c(
#'         50000, 50000, 100000, 200000, 200000,
#'         100, 1000, 2000, 2000, 1
#'     ),
#'     ncol = 2,
#'     dimnames = list(
#'         c(),
#'         c("FSC-A", LDChannel)
#'     )
#' )
#' 
#' liveGate <- flowCore::polygonGate(
#'     filterId = "Live",
#'     .gate = liveGateMatrix
#' )
#' 
#' selectedLive <- flowCore::filter(ffPre, liveGate)
#' ffL <- flowCore::Subset(ffPre, selectedLive)
#' 
#' 
#' # show the results
#' 
#' # subsample 5000 points    
#' ggplotFilterEvents(
#'     ffPre = ffPre,
#'     ffPost = ffL,
#'     nDisplayCells = 5000,
#'     xChannel = "FSC-A", xScale = "linear",
#'     yChannel = LDMarker, yScale = "logicle") +
#'     ggplot2::ggtitle("Live gate filter - 5000 points")
#' 
#' # with all points
#' ggplotFilterEvents(
#'     ffPre = ffPre,
#'     ffPost = ffL,
#'     nDisplayCells = Inf,
#'     xChannel = "FSC-A", xScale = "linear",
#'     yChannel = LDMarker, yScale = "logicle") +
#'     ggplot2::ggtitle("Live gate filter - all points")
#'
#' 
ggplotFilterEvents <- function(ffPre, ffPost,
                               xChannel,
                               yChannel,
                               nDisplayCells = 10000,
                               seed = NULL,
                               size = 0.5,
                               xScale = c("linear", "logicle"),
                               yScale = c("linear", "logicle"),
                               xLogicleParams = NULL,
                               yLogicleParams = NULL,
                               xLinearRange = NULL,
                               yLinearRange = NULL,
                               transList = NULL,
                               runTransforms = FALSE,
                               interactive = FALSE) {
    if (!inherits(ffPre, "flowFrame")) {
        stop("ffPre type not recognized, should be a flowFrame")
    }

    if (!inherits(ffPost, "flowFrame")) {
        stop("ffPost type not recognized, should be a flowFrame")
    }
    
    xScale <- match.arg(xScale)
    yScale <- match.arg(yScale)
    
    # selection of scale_logical function to use (temporary fix due to ggcyto)
    used_logical_scale <- "logicle"
    #used_logical_scale <- "flowjo_biexp"

    # find channel and marker names to specify axis labels

    # browser()

    xChMk <- flowCore::getChannelMarker(ffPre, xChannel)
    xChannel <- xChMk$name
    xLabel <- xChannel
    if (!is.na(xChMk$desc)) {
        xLabel <- paste0(xLabel, " : ", xChMk$desc)
    }

    if (!is.null(yChannel)) {
        yChMk <- flowCore::getChannelMarker(ffPre, yChannel)
        yChannel <- yChMk$name
        yLabel <- yChannel
        if (!is.na(yChMk$desc)) {
            yLabel <- paste0(yLabel, " : ", yChMk$desc)
        }
    }

    if (!"Original_ID" %in% flowCore::colnames(ffPre)) {
        ffPre <- appendCellID(ffPre)
    }

    if (!"Original_ID" %in% flowCore::colnames(ffPost)) {
        stop(
            "ffPost should have Original_ID column in expression matrix ",
            "for the filter display to work!"
        )
    }

    # reduce flow frame or flow set to the displayed channels
    selectedCols <- xChannel
    if (!is.null(yChannel)){
        selectedCols <- c(selectedCols, yChannel)
    }
    selectedCols <- c(selectedCols, "Original_ID")

    ffPre <- ffPre[, selectedCols]
    ffPost <- ffPost[, selectedCols]


    # build scale transforms for not linear scales,
    # when these are not directly provided by the user
    if (is.null(transList) && (xScale != "linear" || yScale != "linear")) {
        estimatedCols <- c()
        if (xScale == "logicle" && is.null(xLogicleParams)) {
            estimatedCols <- xChannel
        }
        if (!is.null(yChannel) && yScale == "logicle" &&
            is.null(yLogicleParams)) {
            estimatedCols <- c(estimatedCols, yChannel)
        }
        if (!is.null(estimatedCols)) {
            transList <- flowCore::estimateLogicle(ffPre, estimatedCols)
        }

        # add the self provided logicle parameters
        if (xScale == "logicle" && !is.null(xLogicleParams)){
            xLogicleTrans <- flowCore::logicleTransform(
                w = xLogicleParams$w,
                t = xLogicleParams$t,
                m = xLogicleParams$m,
                a = xLogicleParams$a
            )
            if (is.null(transList)) {
                transList <- flowCore::transformList(from = xChannel,
                                                     tfun = xLogicleTrans)
            } else {
                transList@transforms[[xChannel]] <- xLogicleTrans
            }
        }
        if (!is.null(yChannel) && yScale == "logicle" &&
            !is.null(yLogicleParams)) {
            yLogicleTrans <- flowCore::logicleTransform(
                w = yLogicleParams$w,
                t = yLogicleParams$t,
                m = yLogicleParams$m,
                a = yLogicleParams$a
            )
            if (is.null(transList)) {
                transList <- flowCore::transformList(from = yChannel,
                                                     tfun = yLogicleTrans)
            } else {
                transList@transforms[[yChannel]] <- yLogicleTrans
            }
        }
    }

    # perform sub-sampling if necessary

    nEvents <- nrow(flowCore::exprs(ffPre))
    if (nDisplayCells < 1) stop("n_display_cells should be strictly positive!")
    i <- 0
    if (nDisplayCells < nEvents) {
        if (!is.null(seed)) {
            withr::with_seed(
                seed,
                i <- sample(nEvents, nDisplayCells)
            )
        } else {
            i <- sample(nEvents, nDisplayCells)
        }
    } else {
        i <- seq(nEvents)
    }
    
    xTransformed <- FALSE
    yTransformed <- FALSE
    if (!is.null(transList)) {
        res <- getTransfoParams(transList, xChannel)
        if (!is.null(res)) {
            if (runTransforms) {
                xScale <- "linear"
                xTransformed <- TRUE
                if (res$type == "logicle") {
                    xLinearRange <- c(0, res$paramsList$m)
                }
            } else {
                xScale <- res$type
                if (res$type == "logicle") {
                    xLogicleParams <- res$paramsList
                }
            }
        }
        
        if (!is.null(yChannel)) {
            res <- getTransfoParams(transList, yChannel)
            if (!is.null(res)) {
                if (runTransforms) {
                    yScale <- "linear"
                    yTransformed <- TRUE
                    if (res$type == "logicle") {
                        yLinearRange <- c(0, res$paramsList$m)
                    }
                } else {
                    yScale <- res$type
                    if (res$type == "logicle") {
                        yLogicleParams <- res$paramsList
                    }
                }
            }
        }
        
        if (runTransforms) {
            # adapt scale to linear and linear_range to null
            # if trans_list is passed and is effectively run
            appliedTransList <- c()
            
            # remove not mentioned channels from trans_list
            appliedTransList <- transList
            transChannels <- names(transList@transforms)
            for (n in transChannels) {
                if (!(n %in% c(xChannel, yChannel))) {
                    appliedTransList@transforms[[n]] <- NULL
                }
            }
            
            ffPre <- flowCore::transform(ffPre, translist = appliedTransList)
            ffPost <- flowCore::transform(ffPost, translist = appliedTransList)
        }
    }

    # find axis ranges depending on scales

    xScale <- match.arg(xScale)
    yScale <- match.arg(yScale)

    if (xScale == "logicle") {
        if (is.null(xLogicleParams)) {
            stop("x_logicle_params can't be NULL if xScale == logicle")
        } else {
            xAxesLimits <- c(0, xLogicleParams$m)
            myTrans <-
                do.call(flowCore::logicleTransform, args = xLogicleParams)
            myInverseTrans <- flowCore::inverseLogicleTransform(myTrans)
            myXRange <- myInverseTrans@.Data(xAxesLimits)
        }
    }
    if (yScale == "logicle") {
        if (is.null(yLogicleParams)) {
            stop("yLogicleParams can't be NULL if yScale == logicle")
        } else {
            yAxesLimits <- c(0, yLogicleParams$m)
            myTrans <-
                do.call(flowCore::logicleTransform, args = yLogicleParams)
            myInverseTrans <- flowCore::inverseLogicleTransform(myTrans)
            myYRange <- myInverseTrans@.Data(yAxesLimits)
        }
    }

    # main plot layer
    
    if (xTransformed) {
        xLabel <- paste0(xLabel, " (transformed)")
    }
    if (yTransformed) {
        yLabel <- paste0(yLabel, " (transformed)")
    }
    
    df <- data.frame(
        x = flowCore::exprs(ffPre)[, xChannel],
        y = flowCore::exprs(ffPre)[, yChannel]
    )
    
    # following 2 statements just to allow R cmd CHECK w/o note
    # x <- NULL
    # y <- NULL
    p <- ggplot(df[i, ], aes(x = .data$x, y = .data$y)) +
        geom_point(
            size = size,
            color = ifelse(flowCore::exprs(ffPre)[i, "Original_ID"] %in%
                flowCore::exprs(ffPost)[, "Original_ID"],
            "blue",
            "red"
            )
        ) +
        xlab(xLabel) +
        ylab(yLabel) +
        # theme_minimal() +
        theme(legend.position = "none")
    # ggtitle("my title")

    # add faceting
    # p <- p + facet_wrap(~name) +


    # add scales for x and y, as well as axis limits

    if (xScale == "logicle") {
        if (is.null(yChannel)) {
            p <- p +
                .my_scale_logicle(axis = "x", 
                                  usedScale = used_logical_scale,
                                  logicleParams = xLogicleParams,
                                  axisRange = myXRange,
                                  interactive = interactive) +
                coord_cartesian(xlim = myXRange)
        } else if (yScale == "logicle") {
            p <- p +
                .my_scale_logicle(axis = "x", 
                                  usedScale = used_logical_scale,
                                  logicleParams = xLogicleParams,
                                  axisRange = myXRange,
                                  interactive = interactive) +
                .my_scale_logicle(axis = "y", 
                                  usedScale = used_logical_scale,
                                  logicleParams = yLogicleParams,
                                  axisRange = myYRange,
                                  interactive = interactive) +
                coord_cartesian(
                    xlim = myXRange,
                    ylim = myYRange
                )
        } else {
            p <- p +
                .my_scale_logicle(axis = "x", 
                                  usedScale = used_logical_scale,
                                  logicleParams = xLogicleParams,
                                  axisRange = myXRange,
                                  interactive = interactive) +
                coord_cartesian(
                    xlim = myXRange,
                    ylim = yLinearRange
                )
        }
    } else {
        if (is.null(yChannel)) {
            p <- p +
                coord_cartesian(xlim = xLinearRange)
        } else if (yScale == "logicle") {
            p <- p +
                .my_scale_logicle(axis = "y", 
                                  usedScale = used_logical_scale,
                                  logicleParams = yLogicleParams,
                                  axisRange = myYRange,
                                  interactive = interactive) +
                coord_cartesian(
                    xlim = xLinearRange,
                    ylim = myYRange
                )
        } else {
            p <- p +
                coord_cartesian(
                    xlim = xLinearRange,
                    ylim = yLinearRange
                )
        }
    }
    return(p)
}

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).



# obtain OMIP021UTSamples, light-weight version used specifically for these
# unit tests
path <- system.file("scripts",
                    package = "CytoPipeline"
)

source(file.path(path,"MakeOMIP021UTSamples.R"))

# TO DO: reactivate tests when ggplot version is fixed
# error message (issue with vdiffr::expect_doppelganger())
# Backtrace:
# 1. └─vdiffr::expect_doppelganger(...) at test-ggplots.R:58:4
# 2.   └─vdiffr (local) writer(fig, testcase, title)
# 3.     └─vdiffr:::svglite(file)
# 4.       └─vdiffr:::svglite_(...)
# ── Error ('test-ggplots.R:175'): ggplotEvents with 2D works ────────────────────
# Error in `svglite_(filename, bg, width, height, pointsize, standalone,
#                    always_valid)`: Graphics API version mismatch

# test_that("ggplotFlowRate works", {
#     p <- ggplotFlowRate(OMIP021UTSamples[[1]])
#     vdiffr::expect_doppelganger("ggplotFlowRate single", fig = p)
#
#     p <- ggplotFlowRate(OMIP021UTSamples)
#     vdiffr::expect_doppelganger("ggplotFlowRate double", fig = p)
#
#     p <- ggplotFlowRate(OMIP021UTSamples[[1]], title = "Test Flow Rate plot")
#     vdiffr::expect_doppelganger("ggplotFlowRate single with title", fig = p)
#
#     p <- ggplotFlowRate(OMIP021UTSamples[[1]], timeUnit = 50)
#     vdiffr::expect_doppelganger(
#         "ggplotFlowRate single with time unit",
#         fig = p
#     )
# })
#
#
#
test_that("ggplotEvents with 1D works", {
    expect_error(ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "450/50Violet-A",
        xScale = "bi-exponential"
    ),
    regexp = "should be one of"
    )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "FSC-A",
        xScale = "linear"
    )
    #vdiffr::expect_doppelganger("ggplotEvents 1D linear single", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "FSC-A",
        xScale = "linear",
        xLinearRange = c(0, 250000)
    )
    # vdiffr::expect_doppelganger(
    #     "ggplotEvents 1D linear explicit range - single",
    #     fig = p
    # )

    p <- ggplotEvents(OMIP021UTSamples, xChannel = "FSC-A", xScale = "linear")
    #vdiffr::expect_doppelganger("ggplotEvents 1D linear double", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "450/50Violet-A",
        xScale = "logicle"
    )
    #vdiffr::expect_doppelganger("ggplotEvents 1D logicle single", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "450/50Violet-A",
        xScale = "logicle", xLogicleParams = list(
            a = 1,
            w = 2,
            m = 7,
            t = 270000
        )
    )

    # vdiffr::expect_doppelganger(
    #     "ggplotEvents 1D logicle with explicit params - single",
    #     fig = p
    # )

    p <- ggplotEvents(OMIP021UTSamples[[2]],
        xChannel = "450/50Violet-A",
        xScale = "logicle", nDisplayCells = 500, seed = 1
    )

    #vdiffr::expect_doppelganger("ggplotEvents 1D sub-sampling", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[2]],
        xChannel = "450/50Violet-A",
        xScale = "logicle", alpha = 0.5, fill = "red"
    )

    #vdiffr::expect_doppelganger("ggplotEvents 1D fill and color", fig = p)

    compensationMatrix <- flowCore::spillover(OMIP021UTSamples[[1]])$SPILL

    ffC <- runCompensation(OMIP021UTSamples[[1]],
        spillover = compensationMatrix,
        updateChannelNames = FALSE
    )

    transList <- flowCore::estimateLogicle(
        ffC,
        colnames(compensationMatrix)
    )

    transList <-
        c(
            transList,
            flowCore::transformList(
                "FSC-A",
                flowCore::linearTransform(a = 0.00001)
            )
        )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "450/50Violet-A",
        xScale = "linear", transList = transList,
        runTransforms = FALSE
    )
    # vdiffr::expect_doppelganger(
    #     "ggplotEvents 1D transformList logicle not run",
    #     fig = p
    # )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "450/50Violet-A",
        xScale = "linear", transList = transList,
        runTransforms = TRUE
    )
    # vdiffr::expect_doppelganger("ggplotEvents 1D transformList logicle run",
    #     fig = p
    # )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "FSC-A",
        xScale = "logicle", transList = transList,
        runTransforms = FALSE
    )
    # vdiffr::expect_doppelganger(
    #     "ggplotEvents 1D transformList linear not run",
    #     fig = p
    # )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "FSC-A",
        xScale = "logicle", transList = transList,
        runTransforms = TRUE
    )
    # vdiffr::expect_doppelganger("ggplotEvents 1D transformList linear run",
    #     fig = p
    # )
})

test_that("ggplotEvents with 2D works", {
    p <- ggplotEvents(OMIP021UTSamples,
        xChannel = "FSC-A",
        xScale = "linear",
        yChannel = "SSC-A",
        yScale = "linear"
    )
    vdiffr::expect_doppelganger("ggplotEvents 2D linear double", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "FSC-A",
        xScale = "linear",
        yChannel = "610/20Violet-A",
        yScale = "logicle"
    )
    vdiffr::expect_doppelganger("ggplotEvents 2D x linear y logicle", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "450/50Violet-A",
        xScale = "logicle",
        yChannel = "SSC-A",
        yScale = "linear"
    )
    vdiffr::expect_doppelganger("ggplotEvents 2D x logicle y linear", fig = p)

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "TETaGC",
        xScale = "logicle",
        yChannel = "CD27",
        yScale = "logicle"
    )
    vdiffr::expect_doppelganger(
        "ggplotEvents 2D x logicle y logicle by markers",
        fig = p
    )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "TETaGC",
        xScale = "logicle",
        yChannel = "CD27",
        yScale = "logicle",
        bins = 128
    )
    vdiffr::expect_doppelganger(
        "ggplotEvents 2D x logicle y logicle bins",
        fig = p
    )


    compensationMatrix <- flowCore::spillover(OMIP021UTSamples[[1]])$SPILL

    ffC <- runCompensation(OMIP021UTSamples[[1]],
        spillover = compensationMatrix,
        updateChannelNames = FALSE
    )

    transList <- flowCore::estimateLogicle(
        ffC,
        colnames(compensationMatrix)
    )

    transList <-
        c(
            transList,
            flowCore::transformList(
                "FSC-A",
                flowCore::linearTransform(a = 0.00001)
            )
        )

    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "TETaGC",
        xScale = "logicle",
        yChannel = "CD27",
        yScale = "logicle",
        transList = transList,
        runTransforms = FALSE
    )
    vdiffr::expect_doppelganger(
        "ggplotEvents 2D x logicle y logicle transList not run",
        fig = p
    )
    p <- ggplotEvents(OMIP021UTSamples[[1]],
        xChannel = "TETaGC",
        xScale = "logicle",
        yChannel = "CD27",
        yScale = "logicle",
        transList = transList,
        runTransforms = TRUE
    )
    vdiffr::expect_doppelganger(
        "ggplotEvents 2D x logicle y logicle transList run",
        fig = p
    )
})


test_that("ggplotFilterEvents works", {
    ffPre <- OMIP021UTSamples[[1]]

    LDMarker <- "L/D Aqua - Viability"

    LDChannel <- getChannelNamesFromMarkers(ffPre, markers = LDMarker)
    liveGateMatrix <- matrix(
        data = c(
            50000, 50000, 100000, 200000, 200000,
            100, 1000, 2000, 2000, 1
        ),
        ncol = 2,
        dimnames = list(
            c(),
            c("FSC-A", LDChannel)
        )
    )

    liveGate <- flowCore::polygonGate(
        filterId = "Live",
        .gate = liveGateMatrix
    )

    selectedLive <- flowCore::filter(ffPre, liveGate)
    ffL <- flowCore::Subset(ffPre, selectedLive)
    #ffL <- appendCellID(ffL, which(selectedLive@subSet))

    p <- ggplotFilterEvents(
        ffPre = ffPre,
        ffPost = ffL,
        nDisplayCells = 1000,
        xChannel = "FSC-A", xScale = "linear",
        yChannel = LDMarker, yScale = "logicle"
    ) +
        ggtitle("Live gate filter")
    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - all points",
        fig = p
    )

    p <- ggplotFilterEvents(
        ffPre = ffPre,
        ffPost = ffL,
        nDisplayCells = 1000,
        xChannel = "FSC-A", xScale = "linear",
        yChannel = LDMarker, yScale = "logicle",
        interactive = TRUE
    ) +
        ggtitle("Live gate filter")
    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - all points - interactive",
        fig = p
    )

    p <- ggplotFilterEvents(
        ffPre = ffPre,
        ffPost = ffL,
        nDisplayCells = 500,
        seed = 1,
        xChannel = "FSC-A", xScale = "linear",
        yChannel = LDMarker, yScale = "logicle"
    ) +
        ggtitle("Live gate filter")
    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - 5000 points",
        fig = p
    )

    p <- ggplotFilterEvents(
        ffPre = ffPre,
        ffPost = ffL,
        nDisplayCells = Inf,
        xChannel = "FSC-A", xScale = "linear",
        yChannel = LDMarker, yScale = "logicle"
    ) +
        ggtitle("Live gate filter")
    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - Infinite nb of points",
        fig = p
    )

    p <- ggplotFilterEvents(
        size = 0.1,
        ffPre = ffPre,
        ffPost = ffL,
        seed = 1,
        xChannel = "FSC-A", xScale = "linear",
        yChannel = LDMarker, yScale = "logicle"
    ) +
        ggtitle("Live gate filter")
    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - small size",
        fig = p
    )

    # calculate transformation list for the next steps
    compensationMatrix <- flowCore::spillover(OMIP021UTSamples[[1]])$SPILL

    ffC <- runCompensation(OMIP021UTSamples[[1]],
                           spillover = compensationMatrix,
                           updateChannelNames = FALSE
    )

    transList <- flowCore::estimateLogicle(
        ffC,
        colnames(compensationMatrix)
    )

    transList <-
        c(
            transList,
            flowCore::transformList(
                "FSC-A",
                flowCore::linearTransform(a = 0.00001)
            )
        )

    p <- ggplotFilterEvents(
        ffPre = ffPre,
        ffPost = ffL,
        seed = 1,
        xChannel = "FSC-A",
        yChannel = LDMarker,
        transList = transList,
        runTransforms = FALSE
    ) +
        ggtitle("Live gate filter")

    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - transList - not run",
        fig = p
    )

    p <- ggplotFilterEvents(
        ffPre = ffPre,
        ffPost = ffL,
        seed = 1,
        xChannel = "FSC-A",
        yChannel = LDMarker,
        transList = transList,
        runTransforms = TRUE
    ) +
        ggtitle("Live gate filter")

    # temporary deactivation of one unit test to have the test passed on GHA
    vdiffr::expect_doppelganger(
        "ggplotFilterEvents 2D - transList - run",
        fig = p
    )
})

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

library(testthat)
library(CytoPipeline)

# launch unit tests
test_check("CytoPipeline")

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

#' @title CytoPipeline package
#' 
#' @name CytoPipeline
#'
#' @rdname CytoPipeline
#' 
#' @seealso [CytoPipeline::CytoPipelineClass], [CytoPipeline::CytoProcessingStep]
#'
#' @description
#'
#' `CytoPipeline` is a package that provides support for automation and 
#' visualization of flow cytometry data analysis pipelines. In the current 
#' state, the package focuses on the preprocessing and quality control part.   
#' 
#' The framework is based on two main S4 classes, i.e. `CytoPipeline` and 
#' `CytoProcessingStep`. The `CytoProcessingStep` defines the link between
#' pipeline step names and corresponding R functions that are either provided 
#' in the CytoPipeline package itself, or exported from a third party package,
#' or coded by the user her/himself. The processing steps need to be specified 
#' centrally and explicitly using either a json input file or through 
#' step by step creation of a `CytoPipeline` object with dedicated methods.  
#' 
#' After having run the pipeline, obtained results at all steps 
#' can be retrieved and visualized thanks to file caching 
#' (the running facility uses a `BiocFileCache` implementation). 
#' The package provides also specific visualization tools like 
#' pipeline workflow summary display, and 1D/2D comparison plots 
#' of obtained flowFrames at various steps of the pipeline.
#' 
#' For a step by step example using `CytoPipeline`, please have a look 
#' at the vignette!
#' 
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

## ----style, echo = FALSE, results = 'asis'------------------------------------
BiocStyle::markdown()

## -----------------------------------------------------------------------------
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("CytoPipeline")

## ----scaleTransformQueueDisplay, results='markup', fig.cap="Scale transform processing queue", echo=FALSE, out.width='75%', fig.align='center', fig.wide = TRUE----
knitr::include_graphics("figs/scaleTransformQueue.png", error = FALSE)

## ----preProcessingQueueDisplay, results='markup', fig.cap="Pre-processing queue for two different pipeline settings", echo=FALSE, out.width='100%', fig.align='center', fig.wide = TRUE----
knitr::include_graphics("figs/preProcessingQueues.png", error = FALSE)

## ----pathsDef-----------------------------------------------------------------
library(CytoPipeline)

# raw data
rawDataDir <- system.file("extdata", package = "CytoPipeline")
# output files
workDir <- suppressMessages(base::tempdir())

## ----CytoPipelineSteps--------------------------------------------------------

# main parameters : sample files and output files

experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                pattern = "Donor"))

pipL_PeacoQC <- CytoPipeline(experimentName = experimentName,
                             sampleFiles = sampleFiles)

### SCALE TRANSFORMATION STEPS ###

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "flowframe_read",
            FUN = "readSampleFiles",
            ARGS = list(
                whichSamples = "all",
                truncate_max_range = FALSE,
                min.limit = NULL
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "remove_margins",
            FUN = "removeMarginsPeacoQC",
            ARGS = list()
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "compensate",
            FUN = "compensateFromMatrix",
            ARGS = list(matrixSource = "fcs")
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "flowframe_aggregate",
            FUN = "aggregateAndSample",
            ARGS = list(
                nTotalEvents = 10000,
                seed = 0
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "scale_transform_estimate",
            FUN = "estimateScaleTransforms",
            ARGS = list(
                fluoMethod = "estimateLogicle",
                scatterMethod = "linear",
                scatterRefMarker = "BV785 - CD3"
            )
        )
    )

### FLOW FRAME PRE-PROCESSING STEPS ###

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "flowframe_read",
            FUN = "readSampleFiles",
            ARGS = list(
                truncate_max_range = FALSE,
                min.limit = NULL
            )
        )
    )


pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_margins",
            FUN = "removeMarginsPeacoQC",
            ARGS = list()
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "compensate",
            FUN = "compensateFromMatrix",
            ARGS = list(matrixSource = "fcs")
        )
    )

pipL_PeacoQC <-
    addProcessingStep(
        pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "perform_QC",
            FUN = "qualityControlPeacoQC",
            ARGS = list(
                preTransform = TRUE,
                min_cells = 150, # default
                max_bins = 500, # default
                step = 500, # default,
                MAD = 6, # default
                IT_limit = 0.55, # default
                force_IT = 150, # default
                peak_removal = 0.3333, # default
                min_nr_bins_peakdetection = 10 # default
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(
        pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_doublets",
            FUN = "removeDoubletsCytoPipeline",
            ARGS = list(
                areaChannels = c("FSC-A", "SSC-A"),
                heightChannels = c("FSC-H", "SSC-H"),
                nmads = c(3, 5))
            )
    )
                    

                
pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_debris",
            FUN = "removeDebrisManualGate",
            ARGS = list(
                FSCChannel = "FSC-A",
                SSCChannel = "SSC-A",
                gateData =  c(73615, 110174, 213000, 201000, 126000,
                              47679, 260500, 260500, 113000, 35000)
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_dead_cells",
            FUN = "removeDeadCellsManualGate",
            ARGS = list(
                FSCChannel = "FSC-A",
                LDMarker = "L/D Aqua - Viability",
                gateData = c(0, 0, 250000, 250000,
                             0, 650, 650, 0)
            )
        )
    )



## ----CytoPipelineJson---------------------------------------------------------

jsonDir <- rawDataDir

# creation on CytoPipeline object,
# using json file as input
pipL_flowAI <-
  CytoPipeline(file.path(jsonDir, "OMIP021_flowAI_pipeline.json"),
               experimentName = "OMIP021_flowAI",
               sampleFiles = sampleFiles)



## ----executePeacoQC-----------------------------------------------------------
# execute PeacoQC pipeline
execute(pipL_PeacoQC, path = workDir)

## ----executeFlowAI------------------------------------------------------------
# execute flowAI pipeline
execute(pipL_flowAI, path = workDir)

## ----plotWorkFlows1, fig.cap = "PeacoQC pipeline - scale transformList processing queue"----

# plot work flow graph - PeacoQC - scale transformList
plotCytoPipelineProcessingQueue(
  pipL_PeacoQC,
  whichQueue = "scale transform",
  path = workDir)


## ----plotWorkFlows2, fig.cap = "PeacoQC pipeline - file pre-processing queue"----

# plot work flow graph - PeacoQC - pre-processing
plotCytoPipelineProcessingQueue(
  pipL_PeacoQC,
  whichQueue = "pre-processing",
  sampleFile = 1,
  path = workDir)

## ----plotWorkFlows3, fig.cap = "flowAI pipeline - scale transformList processing queue"----

# plot work flow graph - flowAI - scale transformList
plotCytoPipelineProcessingQueue(
  pipL_flowAI,
  whichQueue = "scale transform",
  path = workDir)

## ----plotWorkFlows4, fig.cap = "flowAI pipeline - file pre-processing queue"----

# plot work flow graph - flowAI - pre-processing

plotCytoPipelineProcessingQueue(
  pipL_flowAI,
  whichQueue = "pre-processing",
  sampleFile = 1,
  path = workDir)

## ----gettingObjectInfos-------------------------------------------------------
getCytoPipelineObjectInfos(pipL_PeacoQC, 
                           path = workDir,
                           whichQueue = "scale transform")
                                  
getCytoPipelineObjectInfos(pipL_PeacoQC, 
                           path = workDir,
                           whichQueue = "pre-processing",
                           sampleFile = sampleFiles(pipL_PeacoQC)[1])

## ----gettingFlowFrames--------------------------------------------------------
# example of retrieving a flow frame
# at a given step
ff <- getCytoPipelineFlowFrame(
  pipL_PeacoQC,
  whichQueue = "pre-processing",
  sampleFile = 1,
  objectName = "remove_doublets_obj",
  path = workDir)

#
ff2 <- getCytoPipelineFlowFrame(
  pipL_PeacoQC,
  whichQueue = "pre-processing",
  sampleFile = 1,
  objectName = "remove_debris_obj",
  path = workDir)


## ----ggplotEvents1, fig.cap = "1-dimensional distribution plot (forward scatter channel)"----
ggplotEvents(ff, xChannel = "FSC-A")

## ----ggplotEvents2, fig.cap = "2-dimensional distribution plot (forward scatter vs. side scatter channels)"----
ggplotEvents(ff, xChannel = "FSC-A", yChannel = "SSC-A")

## ----ggplotEvents3, fig.cap = "2-dimensional difference plot between remove_doublets and remove_debris steps"----
ggplotFilterEvents(ff, ff2, xChannel = "FSC-A", yChannel = "SSC-A")

## ----cacheObjectRetrieve------------------------------------------------------
obj <- getCytoPipelineObjectFromCache(pipL_PeacoQC,
                                      path = workDir,
                                      whichQueue = "scale transform",
                                      objectName = "compensate_obj")
show(obj)

## ----getNbEvents--------------------------------------------------------------
ret <- CytoPipeline::collectNbOfRetainedEvents(
    experimentName = "OMIP021_PeacoQC",
    path = workDir
)
ret

## ----getRetainedProp----------------------------------------------------------
retainedProp <- 
    as.data.frame(t(apply(
        ret,
        MARGIN = 1,
        FUN = function(line) {
            if (length(line) == 0 || is.na(line[1])) {
                as.numeric(rep(NA, length(line)))
            } else {
                round(line/line[1], 3)
            }
        }
    )))

retainedProp <- retainedProp[-1]

retainedProp

## ----getStepRemovedProp-------------------------------------------------------
stepRemovedProp <- 
    as.data.frame(t(apply(
        ret,
        MARGIN = 1,
        FUN = function(line) {
            if (length(line) == 0) {
                as.numeric(rep(NA, length(line)))
            } else {
                round(1-line/dplyr::lag(line), 3)
            }
        }
    )))

stepRemovedProp <- stepRemovedProp[-1]

stepRemovedProp

## ----loadAddPackages----------------------------------------------------------
library("reshape2")
library("ggplot2")

## ----plotRetainedProp---------------------------------------------------------
myGGPlot <- function(DF, title){
    stepNames = colnames(DF)
    rowNames = rownames(DF)
    DFLongFmt <- reshape(DF,
                         direction = "long",
                         v.names = "proportion",
                         varying = stepNames,
                         timevar = "step",
                         time = stepNames,
                         ids = rowNames)
    
    DFLongFmt$step <- factor(DFLongFmt$step, levels = stepNames)
    
    
    ggplot(data = DFLongFmt,
                 mapping = aes(x = step, y = proportion, text = id)) +
        geom_point(col = "blue") + 
        ggtitle(title) +
        theme(axis.text.x = element_text(angle = 90))
    
}

p1 <- myGGPlot(DF = retainedProp, 
               title = "Retained event proportion at each step")
p1

## ----plotStepRemovedProp------------------------------------------------------
p2 <- myGGPlot(DF = stepRemovedProp,
               title = "Event proportion removed by each step")
p2

## ----launchShinyApp-----------------------------------------------------------
#devtools::install_github("https://github.com/UCLouvain-CBIO/CytoPipelineGUI")
#CytoPipelineGUI::CytoPipelineCheckApp(dir = workDir)

## ----sessioninfo, echo=FALSE--------------------------------------------------
sessionInfo()

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).


#' @title find flow frame columns that represent true signal
#' @description : find flow frame columns that represent true signal
#' @param x a flowCore::flowFrame or a flowCore::flowSet
#' @param toRemovePatterns a vector of string patterns that are to be considered
#' as non signal
#'
#' @return a vector of booleans of which the dimension is equal to the number of
#' columns in ff
#' @export
#' 
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' areSignalCols(OMIP021Samples)
#'
areSignalCols <- function(x,
                          toRemovePatterns = c(
                              "Time", "Original_ID",
                              "File", "SampleID"
                          )) {
    if (!inherits(x, "flowFrame") && !inherits(x, "flowSet")) {
        stop("x type not recognized, should be a flowFrame or a flowSet")
    }


    retCols <- vapply(flowCore::colnames(x),
        FUN.VALUE = logical(1),
        FUN = function(ch, toRemovePatterns) {
            res <- TRUE
            for (pat in toRemovePatterns) {
                res <- res & !grepl(pat, ch, ignore.case = TRUE)
            }
            res
        },
        toRemovePatterns = toRemovePatterns
    )
    return(retCols)
}
#' @title find flow frame columns that represent fluorochrome channel
#' @description : find flow frame columns that represent fluorochrome channel
#' @param x a flowCore::flowFrame or a flowCore::flowSet
#' @param toRemovePatterns a vector of string patterns that are to be considered
#' as non fluorochrome
#'
#' @return a vector of booleans of which the dimension is equal to the number of
#' columns in ff
#' @export
#' 
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' areFluoCols(OMIP021Samples)
#'
areFluoCols <- function(x,
                        toRemovePatterns = c(
                            "FSC", "SSC",
                            "Time", "Original_ID",
                            "File", "SampleID"
                        )) {
    return(areSignalCols(x,
        toRemovePatterns = toRemovePatterns
    ))
}



#' @title sub-sampling of a flowFrame
#' @description : sub-samples a flowFrame
#' with the specified number of samples, without replacement.
#' adds also a column 'Original_ID' if not already present in flowFrame.
#' @param ff a flowCore::flowFrame
#' @param nEvents number of events to be obtained using sub-sampling
#' @param seed  can be set for reproducibility of event sub-sampling
#' @param keepOriginalCellIDs if TRUE, adds (if not already present)
#' a 'OriginalID' column containing the initial IDs of the cell 
#' (from 1 to nrow prior to subsampling).
#' if FALSE, does the same, but takes as IDs (1 to nrow after subsampling)
#' @param ... additional parameters (currently not used)
#'
#' @return new flowCore::flowFrame with the obtained subset of samples
#' @export
#' 
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' # take first sample of dataset, subsample 100 events and create new flowFrame
#' ff <- subsample(OMIP021Samples[[1]], nEvents = 100)
#' 
#'
subsample <- function(ff, 
                      nEvents, 
                      seed = NULL, 
                      keepOriginalCellIDs = TRUE, ...) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }

    eventCounts <- length(flowCore::exprs(ff)[, 1])
    nEvents <- min(eventCounts, nEvents)

    if (!is.null(seed)) {
        withr::with_seed(
            seed,
            keep <- sample(seq_len(eventCounts),
                size = nEvents,
                replace = FALSE
            )
        )
    } else {
        keep <- sample(seq_len(eventCounts),
            size = nEvents,
            replace = FALSE
        )
    }
    
    # order the kept events to maintain chronology of events ! 
    keep <- sort(keep)

    # add Original_ID as a new column if necessary
    if (keepOriginalCellIDs) {
        if (!("Original_ID" %in% colnames(flowCore::exprs(ff)))) {
            ff <- appendCellID(ff)         
        }
        ff <- ff[keep,]
    } else {
        ff <- ff[keep,]
        ff <- resetCellIDs(ff)
    }
    
    ff
}

#' @title compensate with additional options
#' @description : this is a simple wrapper around the flowCore::compensate()
#' utility, allowing to trigger an update of the fluo channel names
#' with a prefix 'comp-' (as in FlowJo)
#'
#' @param obj a flowCore::flowFrame or flowCore::flowSet
#' @param spillover compensation object or spillover matrix or a list of
#' compensation objects
#' @param updateChannelNames if TRUE, add a 'comp-' prefix to all fluorochrome
#' channels (hence does not impact the columns related to FSC, SSC, or other
#' specific keyword like TIME, Original_ID, File,...)
#' Default TRUE.
#'
#' @return a new object with compensated data, and possibly updated column names
#' @export
#' 
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' ff <- OMIP021Samples[[1]]
#' compMatrix <- flowCore::spillover(ff)$SPILL
#' ff <- runCompensation(ff, 
#'                       spillover = compMatrix, 
#'                       updateChannelNames = TRUE)
#'
runCompensation <- function(obj, spillover, updateChannelNames = TRUE) {
    isFlowSet <- FALSE
    if (inherits(obj, "flowSet")) {
        isFlowSet <- TRUE
    } else if (!inherits(obj, "flowFrame")) {
        stop("obj type not recognized, should be a flowFrame or flowSet")
    }

    res <- flowCore::compensate(obj, spillover)
    if (updateChannelNames) {
        if (isFlowSet) {
            res <-
                flowCore::fsApply(res, FUN = .addCompensation2FluoChannelNames)
        } else {
            res <- .addCompensation2FluoChannelNames(res)
        }
    }
    return(res)
}

#' @title Aggregate and sample multiple flow frames of a flow set together
#' @description Aggregate multiple flow frames in order to analyze them
#' simultaneously.
#' A new FF, which contains about nTotalEvents cells, nTotalEvents/nFiles
#' cells from each file.
#' Two new columns are added:
#' a column indicating the original file by index, and
#' a noisy version of this, for better plotting opportunities,
#' This function is based on PeacoQC::AggregateFlowframes() where file names
#' inputs have been replaced by a flowSet input.
#'
#' @param fs a flowCore::flowset
#' @param nTotalEvents Total number of cells to select from the input flow
#' frames
#' @param setup How to proceed when nTotalEvents/nFiles is too high 
#' for some of the flow frames: 
#' - forceBalance (default): compute the minimum nb of events per flow frame, 
#' and keep that amount of events from each flow frame.      
#' - forceNEvents: try to be as balanced as possible, but force a 
#' total of nTotalEvents if possible, i.e. takes all events from the flow frame 
#' with too low nb of events, and then fill in the total with events 
#' from the bigger flow frames in a balanced way.
#' However, if nTotalEvents is greater than the sum of all events, 
#' take all events only once.    
#' @param seed seed to be set before sampling for reproducibility.
#' Default NULL does not set any seed.
#' @param channels Channels/markers to keep in the aggregate.
#' Default NULL takes all channels of the first file.
#' @param writeOutput Whether to write the resulting flowframe to a file.
#' Default FALSE
#' @param outputFile Full path to output file. Default "aggregate.fcs"
#' @param keepOrder If TRUE, the random subsample will be ordered in the same
#' way as they were originally ordered in the file. Default = FALSE.
#'
#' @return returns a new flowCore::flowFrame
#' @export
#' 
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' nCells <- 1000
#' agg <- aggregateAndSample(
#'     fs = OMIP021Samples,
#'     nTotalEvents = nCells)
aggregateAndSample <- function(fs,
                               nTotalEvents,
                               setup = c("forceNEvent", "forceBalance"),
                               seed = NULL,
                               channels = NULL,
                               writeOutput = FALSE,
                               outputFile = "aggregate.fcs",
                               keepOrder = FALSE) {
    # browser()
    if (inherits(fs, "flowFrame")) {
        fs <- flowCore::flowSet(fs)
    }
    if (!inherits(fs, "flowSet")) {
        stop("fs object type not recognized, should be flowCore::flowSet,",
             " or flowCore::flowFrame")
    }

    setup <- match.arg(setup)

    if (!is.null(seed)) {
        # set the seed locally in the execution environment,
        # restore it afterward
        withr::local_seed(seed)
    }

    nFrames <- length(fs)
    
    
    flowFrame <- NULL
    diffNumberChannels <- FALSE
    diffMarkers <- FALSE
    
    # decide, depending on setup, what is the number of rows to select by ff
    ffNRows <- flowCore::fsApply(fs, 
                                 FUN = function(ff) {flowCore::nrow(ff)},
                                 simplify = TRUE)
    nRowsToSelect <- rep(0, nFrames)
    
    minNRow <- min(ffNRows) 
    sumNRow  <- sum(ffNRows)
    
    cFrame <- ceiling(nTotalEvents / nFrames)
    if (cFrame <= minNRow){
        nRowsToSelect <- rep(cFrame, nFrames)
    } else {
        if (setup == "forceBalance"){
            nRowsToSelect <- rep(minNRow, nFrames)
        } else { #setup == forceNEvent
            nRowsToSelect <- rep(0, nFrames)
            trialRows <- cFrame
            allocated <- 0.
            stillToAllocate <- nTotalEvents
            while(stillToAllocate > 0 && allocated < sumNRow) {
                for(i in seq_len(nFrames)) {
                     nRowsToSelect[i] <- min(ffNRows[i], trialRows)
                }
                allocated <- sum(nRowsToSelect)
                stillToAllocate <- nTotalEvents - allocated
                nSpareFF <- nFrames - sum(nRowsToSelect[i] == ffNRows[i])
                if (nSpareFF > 0)
                {
                    trialRows <- trialRows + ceiling(stillToAllocate / nSpareFF)
                }
            }
        }
    }
    
    nExcess <- sum(nRowsToSelect) - nTotalEvents
    if(nExcess < 0){
        warning("Could not choose as much as ",  
                nTotalEvents, 
                " events for subsampling, sampled number of events = ", 
                sum(nRowsToSelect))
    }
    while(nExcess > 0){
        nToWithdraw <- min(nExcess, nFrames)
        whereToWithdraw <- sample.int(nFrames, 
                                      size = nToWithdraw, 
                                      replace = FALSE)
        for(j in seq_along(whereToWithdraw)) {
            nRowsToSelect[whereToWithdraw[j]] <- 
                nRowsToSelect[whereToWithdraw[j]] - 1
        }
        nExcess <- nExcess - nToWithdraw
    }
    
    for (i in seq_len(nFrames)) {
        current_ff <- fs[[i]]
        ids <- sample(seq_len(nrow(current_ff)), nRowsToSelect[i])
        if (keepOrder) {
            ids <- sort(ids)
        }
        new_col_names <- c("File", "File_scattered", "Original_ID")
        prev_agg <- length(grep("File[0-9]*$", colnames(current_ff)))
        if (prev_agg > 0) {
            new_col_names[c(1, 2)] <- paste0(new_col_names[c(1, 2)], prev_agg +
                1)
        }
        prev_ids <- length(grep(
            "Original_ID[0-9]*$",
            flowCore::colnames(current_ff)
        ))
        if (prev_ids > 0) {
            new_col_names[3] <- paste0(new_col_names[3], prev_ids + 1)
        }
        file_ids <- rep(i, nRowsToSelect[i])
        m <- cbind(file_ids, file_ids + stats::rnorm(
            length(file_ids),
            0, 0.1
        ), ids)
        colnames(m) <- new_col_names
        current_ff <- flowCore::fr_append_cols(current_ff[ids, ], m)
        if (is.null(flowFrame)) {
            if (is.null(channels)) {
                channels <- flowCore::colnames(current_ff)
                flowFrame <- current_ff
            } else {
                channels <- getChannelNamesFromMarkers(current_ff, channels)
                flowFrame <- current_ff[, c(channels, colnames(m)),
                    drop = FALSE
                ]
            }
            flowCore::keyword(flowFrame)[["$FIL"]] <- basename(outputFile)
            flowCore::keyword(flowFrame)[["FILENAME"]] <- basename(outputFile)
            flowCore::identifier(flowFrame) <- basename(outputFile)
        } else {
            cols_f <- flowCore::colnames(current_ff)
            cols_flowFrame <- flowCore::colnames(flowFrame)
            commonCols <- intersect(cols_f, cols_flowFrame)
            if (length(commonCols) == 0) {
                stop("No common channels between flow frames")
            }
            if (!diffNumberChannels && length(cols_flowFrame) !=
                length(commonCols)) {
                diffNumberChannels <- TRUE
            }
            if (!diffMarkers &&
                any(!flowCore::markernames(current_ff)[commonCols] %in%
                    flowCore::markernames(flowFrame)[commonCols])) {
                diffMarkers <- TRUE
            }
            flowCore::exprs(flowFrame) <-
                rbind(
                    flowCore::exprs(flowFrame)[, commonCols, drop = FALSE],
                    flowCore::exprs(current_ff)[, commonCols, drop = FALSE]
                )
        }
    }

    if (diffNumberChannels) {
        warning(
            "Flow frames do not contain the same number of channels/markers"
        )
    }
    if (diffMarkers) {
        warning("Flow frames do not contain the same markers")
    }
    if (writeOutput) {
        flowCore::write.FCS(flowFrame, filename = outputFile)
    }
    return(flowFrame)
}

#' @title get tranformation parameters for a specific channel
#' @description investigates a flowCore::tranformList object to get the type
#' and parameters of the transformation applying to a specific channel
#' @param transList a flowCore::transformList
#' @param channel channel name
#'
#' @return If the transformation exists for the specified channel, and is either
#' recognized as a logicle transfo or a linear transfo, a list with two slots:
#' - $type a character containing the transfo type ('logicle' or 'linear')
#' - $params_list a list of named numeric, according to transfo type
#'
#' Otherwise, NULL is returned.
#' @export
#' 
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' # set-up a hybrid transformation list :
#' # - two channels are logicle-ly transformed with automatic param estimates
#' # - one channel has explicit logicle transfo with default parameters
#' # - one channel has linear transformation
#' # - other channels have no transformation
#' translist <- flowCore::estimateLogicle(
#'     OMIP021Samples[[1]],
#'     c("450/50Violet-A", "525/50Violet-A")
#' )
#' translist <- c(
#'     translist,
#'     flowCore::transformList(
#'         "FSC-A",
#'         flowCore::linearTransform(
#'             a = 0.1,
#'             b = 0
#'        )
#'     ),
#'     flowCore::transformList(
#'         "540/30Violet-A",
#'         flowCore::logicleTransform()
#'     )
#' )
#' 
#' ret1 <- getTransfoParams(translist, channel = "FSC-A")
#' ret1$type # "linear"
#' ret1$paramsList # a = 0.1, b = 0.
#' 
#' ret2 <- getTransfoParams(translist, channel = "525/50Violet-A")
#' ret2$type # "logicle"
#' ret2$paramsList # a = 0., w = 0.2834, m = 4.5, t = 262143
#' 
#' ret3 <- getTransfoParams(translist, channel = "540/30Violet-A")
#' ret3$type # "logicle
#' ret3$paramsList # a = 0., w = 0.5, m = 4.5, t = 262144
getTransfoParams <- function(transList,
                             channel) {
    if (!inherits(transList, "transformList")) {
        stop("transList parameter should be a flowCore::transformList!")
    }
    # browser()
    transMap <- transList@transforms[[channel]]
    if (is.null(transMap)) {
        return(NULL)
    } else {
        if (methods::.hasSlot(transMap, "f")) {
            tf <- methods::new("transform", .Data = transMap@f)
        } else if (methods::.hasSlot(transMap, ".Data")) {
            tf <- methods::new("transform", .Data = transMap@.Data)
        } else {
            stop(
                "transfo on channel does not have 'f' or '.Data' slot ",
                "=> not handled"
            )
        }

        sm <- flowCore::summary(tf)
        ret <- list()
        if (!is.null(sm$k) && methods::is(sm$k, "transform")) {
            ret$type <- "logicle"
            ret$paramsList <- list(
                a = unname(sm$a),
                w = unname(sm$w),
                m = unname(sm$m),
                t = unname(sm$t)
            )
        } else if (!is.null(sm$t) && methods::is(sm$t, "transform")) {
            ret$type <- "linear"
            ret$paramsList <- list(
                a = unname(sm$a),
                b = unname(sm$b)
            )
        } else {
            stop(
                "transformation type not recognized, ",
                "currently this function only ",
                "works with linear or logicle transforms"
            )
        }
        return(ret)
    }
}


#' @title compute linear transformation of scatter channels found in ff, based
#' on 5% and 95% of referenceChannel, set as target. If there is a
#' transformation defined in transList for referenceChannel, it is applied
#' first, before computing quantiles.
#' Then the computed linear transformations (or each scatter channel) are added
#' into the transfo_list. -A channels are computed, and same linear
#' transformation is then applied to corresponding -W and -H channels
#' (if they exist in ff).
#' @description based on a referenceChannel
#' @param ff a flowCore::flowFrame
#' @param transList an initial flowCore::transformList
#' @param referenceChannel the reference channel to take target quantile values
#' from. Can be defined as marker or channel name.
#' @param silent if FALSE, will output some information on the computed
#' linear transformations
#'
#' @return the transList with added linear scale transformations
#' @export
#'
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' ff <- OMIP021Samples[[1]]
#' refMarker <- "APCCy7 - CD4"
#' refChannel <- "780/60Red-A"

#' transList <- flowCore::estimateLogicle(ff,
#'                                        channels = refChannel)
#' retTransList <-
#'     computeScatterChannelsLinearScale(ff,
#'                                       transList = transList,
#'                                       referenceChannel = refMarker,
#'                                       silent = TRUE
#'     )
computeScatterChannelsLinearScale <- function(ff,
                                              transList = NULL,
                                              referenceChannel,
                                              silent = TRUE) {
    # check inputs

    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }
    fluoChannels <- flowCore::colnames(ff)[areFluoCols(ff)]
    scatterChannels <-
        flowCore::colnames(ff)[!areFluoCols(ff) & areSignalCols(ff)]
    if (length(scatterChannels) == 0) {
        warning("no scatter channel to scale")
        return(transList)
    }

    referenceChannel <- flowCore::getChannelMarker(ff, referenceChannel)$name

    if (!(referenceChannel %in% fluoChannels)) {
        stop("referenceChannel should be a fluorochrome channel of ff")
    }

    if (is.null(transList)) {
        if (!silent) {
            message(
                "NULL transList found...\n",
                "Continued with no transfo applied on reference channel"
            )
        }
        ff_t <- ff[, referenceChannel]
    } else {
        if (!inherits(transList, "transformList")) {
            stop("transList parameter should be a flowCore::transformList!")
        }
        if (is.null(transList@transforms[[referenceChannel]])) {
            if (!silent) {
                message(
                    "No transformation found for referenceChannel ",
                    "in transList\n",
                    "Continued with no transfo applied on reference ",
                    "channel"
                )
            }
            ff_t <- ff[, referenceChannel]
        } else {
            transfoList <-
                flowCore::transformList(
                    from = referenceChannel,
                    tfun = transList@transforms[[referenceChannel]]@f
                )
            ff_t <- flowCore::transform(ff[, referenceChannel], transfoList)
        }
    }


    q5Goal <- stats::quantile(flowCore::exprs(ff_t)[, referenceChannel], 0.05)
    q95Goal <- stats::quantile(flowCore::exprs(ff_t)[, referenceChannel], 0.95)

    # adapt scatter channels to have the same percentiles
    # -A channels are modified independently,
    #  and the SAME transfo are applied to corresponding -H and -W channels
    # (if they exist)
    foundAreaScatter <- FALSE
    if ("FSC-A" %in% scatterChannels) {
        ch <- "FSC-A"
        q5FSCA <- stats::quantile(flowCore::exprs(ff)[, ch], 0.05)
        q95FSCA <- stats::quantile(flowCore::exprs(ff)[, ch], 0.95)
        FSCAa <- (q95Goal - q5Goal) / (q95FSCA - q5FSCA)
        FSCAb <- q5Goal - q5FSCA * (q95Goal - q5Goal) / (q95FSCA - q5FSCA)
        if (!silent) {
            message(
                "applying specific linear transformation ",
                "for FSC-A channel..."
            )
            message(
                "initial quantiles : q5 = ", round(q5FSCA, 4), " ; q95 = ",
                round(q95FSCA, 4)
            )
            message(
                "target quantiles : q5 = ", round(q5Goal, 4), " ; q95 = ",
                round(q95Goal, 4)
            )
            message(
                "a = ",
                formatC(FSCAa, format = "e", digits = 2),
                " ; b = ",
                formatC(FSCAb, format = "e", digits = 2)
            )
        }
        tf <- flowCore::linearTransform(a = FSCAa, b = FSCAb)
        if (is.null(transList)) {
            transList <-
                flowCore::transformList(
                    from = ch,
                    tfun = tf
                )
        } else {
            transList@transforms[[ch]] <- NULL
            transList <- c(transList, flowCore::transformList(ch, tf))
            # transList@transforms[[ch]] <- tf
        }

        foundAreaScatter <- TRUE

        if ("FSC-W" %in% scatterChannels) {
            ch <- "FSC-W"
            if (!silent) {
                message(
                    "applying FSC-A linear transformation ",
                    "for FSC-W channel..."
                )
            }
            transList@transforms[[ch]] <- NULL
            transList <- c(transList, flowCore::transformList(ch, tf))
            # transList@transforms[[ch]] <- tf
        }
        if ("FSC-H" %in% scatterChannels) {
            ch <- "FSC-H"
            if (!silent) {
                message(
                    "applying FSC-A linear transformation ",
                    "for FSC-H channel..."
                )
            }
            transList@transforms[[ch]] <- NULL
            transList <- c(transList, flowCore::transformList(ch, tf))
            # transList@transforms[[ch]] <- tf
        }
    }
    if ("SSC-A" %in% scatterChannels) {
        ch <- "SSC-A"
        q5SSCA <- stats::quantile(flowCore::exprs(ff)[, ch], 0.05)
        q95SSCA <- stats::quantile(flowCore::exprs(ff)[, ch], 0.95)
        SSCAa <- (q95Goal - q5Goal) / (q95SSCA - q5SSCA)
        SSCAb <- q5Goal - q5SSCA * (q95Goal - q5Goal) / (q95SSCA - q5SSCA)
        if (!silent) {
            message(
                "applying specific linear transformation ",
                "for SSC-A channel..."
            )
            message(
                "initial quantiles : q5 = ", round(q5SSCA, 4), " ; q95 = ",
                round(q95SSCA, 4)
            )
            message(
                "target quantiles : q5 = ", round(q5Goal, 4), " ; q95 = ",
                round(q95Goal, 4)
            )
            message(
                "a = ",
                formatC(SSCAa, format = "e", digits = 2),
                " ; b = ",
                formatC(SSCAb, format = "e", digits = 2)
            )
        }
        tf <- flowCore::linearTransform(a = SSCAa, b = SSCAb)
        if (is.null(transList)) {
            transList <-
                flowCore::transformList(
                    from = ch,
                    tfun = tf
                )
        } else {
            transList@transforms[[ch]] <- NULL
            transList <- c(transList, flowCore::transformList(ch, tf))
            # transList@transforms[[ch]] <- tf
        }

        foundAreaScatter <- TRUE

        if ("SSC-W" %in% scatterChannels) {
            ch <- "SSC-W"
            if (!silent) {
                message(
                    "applying SSC-A linear transformation ",
                    "for SSC-W channel..."
                )
            }
            transList@transforms[[ch]] <- NULL
            transList <- c(transList, flowCore::transformList(ch, tf))
            # transList@transforms[[ch]] <- tf
        }
        if ("SSC-H" %in% scatterChannels) {
            ch <- "SSC-H"
            if (!silent) {
                message(
                    "applying SSC-A linear transformation ",
                    "for SSC-H channel..."
                )
            }
            transList@transforms[[ch]] <- NULL
            transList <- c(transList, flowCore::transformList(ch, tf))
            # transList@transforms[[ch]] <- tf
        }
    }

    if (!foundAreaScatter) {
        warning(
            "did not find any -A scatters in channels => ",
            "did not transform anything\n",
            "The following scatter channels were found: ",
            scatterChannels
        )
    }

    return(transList)
}

#' @title find time channel in flowSet/flowFrame
#' @description tries to find a channel in a flowSet/flowFrame that could
#' be the time channel. First tries to identify a channel name containing the
#' 'time' string, then tries to identify a single monotonically increasing
#' channel.
#'
#' @param obj a flowCore::flowFrame or flowCore::flowSet
#' @param excludeChannels vector of column names to exclude in the search
#'
#' @return a character, name of the found channel that should be representing
#' time. If not found, returns NULL.
#' @export
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' ret <- findTimeChannel(OMIP021Samples[[1]])
#' ret # "Time"
#'
findTimeChannel <- function(obj, excludeChannels = c()) {
    isFlowSet <- FALSE
    if (inherits(obj, "flowSet")) {
        isFlowSet <- TRUE
    } else if (inherits(obj, "flowFrame")) {
    } else {
        stop("obj type not recognized, should be a flowFrame or flowSet")
    }

    includedChannels <-
        flowCore::colnames(obj)[!(flowCore::colnames(obj) %in% excludeChannels)]

    time <- grep("^Time$", includedChannels,
        value = TRUE,
        ignore.case = TRUE
    )[1]
    if (is.na(time)) {
        if (isFlowSet) {
            xx <- flowCore::exprs(obj[[1]])[, includedChannels]
        } else if (methods::is(obj, "flowFrame")) {
            xx <- flowCore::exprs(obj)[, includedChannels]
        }
        cont <- apply(xx, 2, function(y) {
            all(sign(diff(y)) >=
                0)
        })
        cont <- apply(xx, 2, function(y) all(y == cummax(y)))
        time <- names(which(cont))
    }
    if (!length(time) || length(time) > 1) {
        time <- NULL
    }
    return(time)
}

#' @title get channel names from markers
#' @description finds name of channels corresponding to user provided markers
#'
#' @param ff a flowCore::flowFrame
#' @param markers a vector of markers, either provided as :
#' - an array of booleans (referring to flowFrame columns)
#' - an array of integers (indices in flowFrame columns)
#' - an array of characters (exact markers or channel patterns)
#'
#' @return a character vector, containing the names of the corresponding
#' channels
#' @export
#'
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' # with existing markers
#' ret <- getChannelNamesFromMarkers(
#'     OMIP021Samples[[1]],
#'     c(
#'         "FSC-A",
#'         "L/D Aqua - Viability",
#'         "FITC - gdTCR",
#'         "PECy5 - CD28"
#'     ))
#'     
#' ret # c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")
#' 
#' # with boolean vector
#' indices <- c(1, 6, 14, 18)
#' boolInput <- rep(FALSE, 21)
#' boolInput[indices] <- TRUE
#' ret2 <- getChannelNamesFromMarkers(
#'     OMIP021Samples[[1]],
#'     boolInput)
#'     
#' ret2 # c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")
#' 
#' # with indices vector
#' ret3 <- getChannelNamesFromMarkers(
#'     OMIP021Samples[[1]],
#'     indices
#' )
#' ret3 # c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")
#' 
#' 
getChannelNamesFromMarkers <- function(ff, markers) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }

    frameChannels <- unname(flowCore::parameters(ff)@data[["name"]])
    frameMarkers <- unname(flowCore::parameters(ff)@data[["desc"]])

    if (is.logical(markers)) {
        markers <- which(markers)
    }
    channelNames <- c()
    for (marker in markers) {
        if (is.numeric(marker)) {
            iChannel <- marker
        } else {
            marker <- paste0("^\\Q", marker, "\\E$")
            iChannel <- grep(marker, frameMarkers)
        }
        if (length(iChannel) != 0) {
            for (i in iChannel) {
                channel <- frameChannels[iChannel]
                names(channel) <- frameMarkers[iChannel]
                channelNames <- c(channelNames, channel)
            }
        } else {
            iChannel <- grep(marker, frameChannels)
            if (length(iChannel) != 0) {
                channel <- frameChannels[iChannel]
                names(channel) <- channel
                channelNames <- c(channelNames, channel)
            } else {
                stop("Marker", marker, "could not be found")
            }
        }
    }
    return(unname(channelNames))
}

#' @title update marker name of a given flowFrame channel
#' @description : in a flowCore::flowFrame, update the marker name (stored in
#' 'desc' of parameters data) of a given channel. 
#' Also update the corresponding keyword in the flowFrame.
#' @param ff a flowCore::flowFrame
#' @param channel the channel for which to update the marker name
#' @param newMarkerName the new marker name to be given to the selected channel
#'
#' @return a new flowCore::flowFrame with the updated marker name
#' @export
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' retFF <- updateMarkerName(OMIP021Samples[[1]],
#'                           channel = "FSC-A",
#'                           newMarkerName = "Fwd Scatter-A")
#'
updateMarkerName <- function(ff, channel, newMarkerName) {
    
    # 0. check inputs
    
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame!")
    }
    
    if (is.numeric(channel)) {
        channelIndex <- channel[1]
    } else if (is.character(channel)){
        channelIndex <- which(flowCore::colnames(ff) == channel)
        if (length(channelIndex) == 0) {
            # try to find corresponding marker
            chMk <- try(flowCore::getChannelMarker(ff, channel),
                        silent = TRUE)
            if (inherits(chMk, "try-error")) {
                stop("channel not found in flowFrame!")
            }
            channelIndex <- which(flowCore::colnames(ff) == chMk$name)
            
        }
    } else {
        stop("channel shoud be a numeric index of a character!")
    }
    
    # 1. update flowFrame parameters pheno data

    param <- flowCore::parameters(ff)
    paramData <- flowCore::pData(param)
    paramData[channelIndex, "desc"] <- newMarkerName
    flowCore::pData(param) <- paramData
    flowCore::parameters(ff) <- param
    
    # 2. update flowFrame keyword
    targetChannelName <- flowCore::colnames(ff)[channelIndex]
    nChannels <- length(flowCore::colnames(ff))
    potentialKeywordChannels <- paste0("$P", seq_len(nChannels), "N")
    potentialKeywordMarkers <- paste0("$P", seq_len(nChannels), "S")
    kch <- flowCore::keyword(ff, potentialKeywordChannels)
    #kmk <- flowCore::keyword(ff, potentialKeywordMarkers)
    
    # if channel name exists as a keyword in flowFrame
    # => update corresponding marker name
    for (k in seq_along(kch)) {
        if (kch[k][[potentialKeywordChannels[k]]] == targetChannelName) {
            flowCore::keyword(ff)[[potentialKeywordMarkers[k]]] <- 
                newMarkerName
            break
        }
    }
    return(ff)
}

#' @title remove channels from a flowFrame
#' @description : in a flowCore::flowFrame, remove the channels of the given
#' names.
#' @param ff a flowCore::flowFrame
#' @param channels the channel names to be removed
#'
#' @return a new flowCore::flowFrame with the removed channels
#' @export
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' retFF <- removeChannels(OMIP021Samples[[1]],
#'                         channel = "FSC-A")
#'
removeChannels <- function(ff, channels) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame!")
    }    
    keptCols <- rep(TRUE, length(flowCore::colnames(ff)))
    for (ch in channels) {
        channelIndex <- which(flowCore::colnames(ff) == ch)
        if (length(channelIndex) == 0) {
            warning("channel ", ch, " not found in flowFrame => ignoring...")
        } else {
            channelIndex <- channelIndex[1]
        }
        keptCols[channelIndex] <- FALSE
    }
    ff <- ff[, keptCols]
    return(ff)
}

#' @title append 'Original_ID' column to a flowframe
#' @description : on a flowCore::flowFrame, append a 'Original_ID' column.
#' This column can be used in plots comparing the events pre and post gating.
#' If the 'Original_ID' column already exists, the function does nothing
#' @param ff a flowCore::flowFrame
#' @param eventIDs an integer vector containing the values to be added 
#' in expression matrix, as Original ID's.
#' 
#' @return new flowCore::flowFrame containing the added 'Original_ID' column
#' @export
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' retFF <- appendCellID(OMIP021Samples[[1]])
#'
appendCellID <- function(ff, eventIDs = seq_len(flowCore::nrow(ff))) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }
    if (!("Original_ID" %in% colnames(flowCore::exprs(ff)))) {
        matrixCellIds <- matrix(
            data = eventIDs, ncol = 1,
            dimnames = list(c(), list("Original_ID"))
        )
        ff <- flowCore::fr_append_cols(ff, matrixCellIds)
    } 
    return(ff)
}

#' @title reset 'Original_ID' column in a flowframe
#' @description : on a flowCore::flowFrame, reset 'Original_ID' column.
#' This column can be used in plots comparing the events pre and post gating.
#' If the 'Original_ID' column already exists, the function replaces 
#' the existing IDs by the user provided ones.
#' If not, an `appendCellID()` is called.
#' @param ff a flowCore::flowFrame
#' @param eventIDs an integer vector containing the values to be set 
#' in expression matrix, as Original ID's.
#' 
#' @return new flowCore::flowFrame containing the amended (or added) 
#' 'Original_ID' column
#' @export
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' ff <- appendCellID(OMIP021Samples[[1]])
#' 
#' subsample_ff <- subsample(ff, 100, keepOriginalCellIDs = TRUE)
#' 
#' # re-create a sequence of IDs, ignoring the ones before subsampling
#' reset_ff <- resetCellIDs(subsample_ff)
#'
resetCellIDs <- function(ff, eventIDs = seq_len(flowCore::nrow(ff))) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }
    if (!("Original_ID" %in% colnames(flowCore::exprs(ff)))) {
        ff <- appendCellID(ff, eventIDs)
    } else {
        if (length(eventIDs) != flowCore::nrow(ff)) {
            stop("length of provided eventIDs does not match nb of rows")
        }
        exprs <- flowCore::exprs(ff)
        exprs[,"Original_ID"] <- eventIDs
        ff@exprs <- exprs
    }
    return(ff)
}

# #' @title add compensation to fluo column names
# #' @description : technical utility to add "Comp-" prefix to all
# #' column names for fluorochrom channels
# #' @param ff a flowCore::flowFrame
# #' @return new flowCore::flowFrame with the new column names
.addCompensation2FluoChannelNames <- function(ff) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }
    areFluoCols <- areFluoCols(ff)
    newColNames <- flowCore::colnames(ff)
    newColNames <-
        mapply(
            FUN = function(theName, need2Do) {
                newName <- theName
                if (need2Do) newName <- paste0("Comp-", theName)
                newName
            },
            newColNames, areFluoCols
        )
    flowCore::colnames(ff) <- newColNames
    return(ff)
}

#' @title get fcs file name 
#' @description get basename of $FILENAME keyword if exists
#' @param ff a flowCore::flowFrame
#' @return the basename of $FILENAME keyword
#' @export
#' @examples 
#' 
#' data(OMIP021Samples)
#' 
#' fName <- getFCSFileName(OMIP021Samples[[1]])
getFCSFileName <- function(ff) {
    if (!inherits(ff, "flowFrame")) {
        stop("ff type not recognized, should be a flowFrame")
    }
    fName <- basename(flowCore::keyword(ff, keyword = "FILENAME")$FILENAME)
    if (is.null(fName)) stop("No FILENAME keyword for flowFrame")
    return(fName)
}

#' @title write flowFrame to disk
#' @description wrapper around flowCore::write.FCS() or utils::write.csv
#' that discards any additional parameter passed in (...)
#' @param ff a flowCore::flowFrame
#' @param dir an existing directory to store the flowFrame, 
#' @param useFCSFileName if TRUE filename used will be based on original fcs 
#' filename
#' @param prefix file name prefix
#' @param suffix file name suffix
#' @param format either fcs or csv
#' @param csvUseChannelMarker if TRUE (default), converts the channels to the
#' corresponding marker names (where the Marker is not NA). This setting is
#' only applicable to export in csv format.
#' @param ... other arguments (not used)
#' @return nothing
#' @export
#' 
#' @examples
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' res <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#'     
#' ff_c <- compensateFromMatrix(res[[2]], matrixSource = "fcs") 
#' outputDir <- base::tempdir()
#' writeFlowFrame(ff_c, 
#'                dir = outputDir,
#'                suffix = "_fcs_export",
#'                format = "csv")
#' 
writeFlowFrame <- function(ff, dir = ".", 
                           useFCSFileName = TRUE,
                           prefix = "", suffix ="",
                           format = c("fcs", "csv"), 
                           csvUseChannelMarker = TRUE,
                           ...) {
    format <- match.arg(format)
    if (!useFCSFileName && prefix == "" && suffix == "") {
        stop ("No file name provided! If useFCSFileName == FALSE, ",
              "then either prefix or suffix should be provided.")
    }
    if(!dir.exists(dir)) {
        stop ("Provided directory does not exist!")
    }
    
    baseFileName <- prefix
    if (useFCSFileName) {
        fileNameSkeleton <- getFCSFileName(ff)
        fileNameSkeleton <- 
            substr(fileNameSkeleton, 1, nchar(fileNameSkeleton) - 4)
        baseFileName <- paste0(baseFileName, fileNameSkeleton)
    }
    baseFileName <- paste0(baseFileName, suffix, ".", format)
    
    fileName <- file.path(dir, baseFileName)
    
    if (format == "fcs") {
        flowCore::write.FCS(ff, filename = fileName)
    } else {
        exprs <- flowCore::exprs(ff)
        if (csvUseChannelMarker) {
            oldColNames <- colnames(exprs)
            newColNames <- oldColNames
            for (j in seq_along(oldColNames)) {
                ret <- flowCore::getChannelMarker(ff, name = oldColNames[j])
                if (!is.na(ret$desc)) {
                    newColNames[j] <- ret$desc
                }
            }
            colnames(exprs) <- newColNames
        }
        utils::write.csv(exprs, file = fileName, 
                         row.names = FALSE)
    }
}

# update compensation matrix labels, replace by bold channel name
# if label == concatenation of channel, " :: ", marker (as in FlowJo export)
# if label == marker
.updateCompMatrixLabels <- function(mat, ff) {
    stopifnot (inherits(ff, "flowFrame"))
    
    matLabels <- colnames(mat)
    for(j in seq_along(matLabels)) {
        # if label if the concatenation of channel & marker (as in FJ export),
        # keep only the part before " :: "
        ind <- regexpr(" :: ", matLabels[j])
        if (ind > 1) {
            matLabels[j] <- substr(matLabels[j], 1, ind-1)
        }
        
        # match channel or marker to channel
        matLabels[j] <- flowCore::getChannelMarker(ff, matLabels[j])[,"name"]
    }
    
    colnames(mat) <- matLabels
    rownames(mat) <- matLabels
    
    mat
}

# update marker names
# the new set of marker names should be ordered and have the same length
# as the channel names
.updateMarkersName <- function(ff, newMarkerNames) {
    # check inputs
    stopifnot (inherits(ff, "flowFrame"))
    if (!is.character(newMarkerNames) || length(newMarkerNames) == 0) {
        stop("new marker names should be provided as ",
             "a character of non zero length")
    }
    
    # 1. update fcs parameters pheno data
    newpData <- flowCore::pData(flowCore::parameters(ff))
    channels <- newpData$names
    nChannels <- length(channels)
    
    if (is.null(channels)) {
        stop("did not find channel names in flow frame pheno data",
             " => inconsistency!")
    }
    if (length(newMarkerNames) != nChannels) {
        stop("length of new marker names differs from length of channel names",
             " => inconsistency!")
    }
    newpData$desc <- newMarkerNames
    flowCore::pData(ff) <- newpData
    
    # 2. update fcs keywords
    potentialKeywordChannels <- paste0("$P", seq_len(nChannels), "N")
    potentialKeywordMarkers <- paste0("$P", seq_len(nChannels), "S")
    kch <- flowCore::keyword(ff, potentialKeywordChannels)
    kmk <- flowCore::keyword(ff, potentialKeywordMarkers)
    for (k in kch) {
        # if channel name exists as a keyword in fcs
        # => update corresponding marker name
        if (!is.null(kch[k][[potentialKeywordChannels[k]]])) {
            flowCore::keyword(ff)[[potentialKeywordMarkers[k]]] <- 
                newMarkerNames[k]
        }
    }
}

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

data("OMIP021Samples")

test_that("areSignalCols works", {
    ff <- OMIP021Samples[[1]]

    expectedRes <- c(rep(TRUE, times = 20), FALSE, FALSE)

    res <- areSignalCols(ff)
    expect_equal(res, expectedRes, ignore_attr = TRUE)

    # same with flowSet
    res <- areSignalCols(OMIP021Samples)
    expect_equal(res, expectedRes, ignore_attr = TRUE)
})

test_that("areFluoCols works", {
    ff <- OMIP021Samples[[1]]

    expectedRes <- c(
        rep(FALSE, times = 4),
        rep(TRUE, times = 16),
        FALSE, FALSE
    )

    res <- areFluoCols(ff)
    expect_equal(res, expectedRes, ignore_attr = TRUE)

    # same with flowSet
    res <- areFluoCols(OMIP021Samples)
    expect_equal(res, expectedRes, ignore_attr = TRUE)
})



test_that("subsample works", {
    ff <- OMIP021Samples[[1]]

    nEvents <- 50
    seed <- 0
    ffSub <- subsample(ff, nEvents, seed = seed)
    expect_equal(flowCore::nrow(ffSub), nEvents)
    # withr::with_seed(
    #     seed,
    #     {
    #         tgtLines <- sample(seq_len(flowCore::nrow(ff)),
    #                        size = nEvents,
    #                        replace = FALSE)
    #         tgtIDs <- flowCore::exprs(ff)[tgtLines,"Original_ID"]
    #         expect_true(all.equal(tgtIDs,
    #                               flowCore::exprs(ffSub)[,"Original_ID"]))
    #     }
    # )

    # same but reset the original IDs
    ffSub <- subsample(ff, nEvents, seed = seed, keepOriginalCellIDs = FALSE)
    expect_equal(flowCore::nrow(ffSub), nEvents)
    tgtIDs <- 1:nEvents
    expect_true(all.equal(tgtIDs, flowCore::exprs(ffSub)[,"Original_ID"]))


    # subsample with more samples than original nrow (5000)
    ffSub <- subsample(ff, 10000)

    nOriginal <- flowCore::nrow(ff)
    expect_equal(flowCore::nrow(ffSub), nOriginal)
})


test_that(".addCompensation2FluoChannelNames works", {
    ff <- OMIP021Samples[[1]]

    ff2 <- .addCompensation2FluoChannelNames(ff)

    expect_equal(
        flowCore::colnames(ff2),
        c(
            "FSC-A", "FSC-H", "SSC-A", "SSC-H",
            "Comp-450/50Violet-A", "Comp-525/50Violet-A", "Comp-540/30Violet-A",
            "Comp-585/15Violet-A", "Comp-610/20Violet-A", "Comp-670/30Violet-A",
            "Comp-670/14Red-A", "Comp-730//45Red-A", "Comp-780/60Red-A",
            "Comp-530/30Blue-A", "Comp-710/50Blue-A", "Comp-582/15Yellow-A",
            "Comp-610/20Yellow-A", "Comp-670/30Yellow-A", "Comp-710/50Yellow-A",
            "Comp-780/60Yellow-A", "Time", "Original_ID"
        )
    )

    expect_error(.addCompensation2FluoChannelNames(OMIP021Samples),
        regexp = "type not recognized"
    )
})

test_that("runCompensation works", {
    ff <- OMIP021Samples[[1]]

    compMatrix <- flowCore::spillover(ff)$SPILL
    ff1 <- flowCore::compensate(ff, spillover = compMatrix)
    ff2 <- runCompensation(ff,
        spillover = compMatrix,
        updateChannelNames = FALSE
    )

    expect_equal(ff2, ff1)

    ff2 <- runCompensation(ff, spillover = compMatrix)
    # the following avoids comparing name attributes
    expect_true(all(flowCore::exprs(ff2) == flowCore::exprs(ff1)))
    expect_equal(
        flowCore::colnames(ff2),
        c(
            "FSC-A", "FSC-H", "SSC-A", "SSC-H",
            "Comp-450/50Violet-A", "Comp-525/50Violet-A", "Comp-540/30Violet-A",
            "Comp-585/15Violet-A", "Comp-610/20Violet-A", "Comp-670/30Violet-A",
            "Comp-670/14Red-A", "Comp-730//45Red-A", "Comp-780/60Red-A",
            "Comp-530/30Blue-A", "Comp-710/50Blue-A", "Comp-582/15Yellow-A",
            "Comp-610/20Yellow-A", "Comp-670/30Yellow-A", "Comp-710/50Yellow-A",
            "Comp-780/60Yellow-A", "Time", "Original_ID"
        )
    )

    fs <- runCompensation(OMIP021Samples,
        spillover = compMatrix
    )
    ff3 <- flowCore::compensate(OMIP021Samples[[1]],
        spillover = compMatrix
    )
    # the following avoids comparing name attributes
    expect_true(all(flowCore::exprs(fs[[1]]) == flowCore::exprs(ff3)))
    expect_equal(
        flowCore::colnames(fs),
        c(
            "FSC-A", "FSC-H", "SSC-A", "SSC-H",
            "Comp-450/50Violet-A", "Comp-525/50Violet-A", "Comp-540/30Violet-A",
            "Comp-585/15Violet-A", "Comp-610/20Violet-A", "Comp-670/30Violet-A",
            "Comp-670/14Red-A", "Comp-730//45Red-A", "Comp-780/60Red-A",
            "Comp-530/30Blue-A", "Comp-710/50Blue-A", "Comp-582/15Yellow-A",
            "Comp-610/20Yellow-A", "Comp-670/30Yellow-A", "Comp-710/50Yellow-A",
            "Comp-780/60Yellow-A", "Time", "Original_ID"
        )
    )
})

test_that("aggregateAndSample works", {
    nCells <- 100
    agg <- aggregateAndSample(
        fs = OMIP021Samples,
        nTotalEvents = nCells,
        seed = 1
    )

    expect_equal(nrow(flowCore::exprs(agg)), nCells)

    ind1 <- which(flowCore::exprs(agg)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg)[ind1, ]), nCells / 2)
    ind2 <- which(flowCore::exprs(agg)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg)[ind2, ]), nCells / 2)
    
    nCells <- 99
    agg <- aggregateAndSample(
        fs = OMIP021Samples,
        nTotalEvents = nCells,
        seed = 1
    )
    
    expect_equal(nrow(flowCore::exprs(agg)), nCells)
    
    ind1 <- which(flowCore::exprs(agg)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg)[ind1, ]), (nCells-1) / 2)
    ind2 <- which(flowCore::exprs(agg)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg)[ind2, ]), (nCells+1) / 2)
    
    # too high total nb of cells
    nCells<- 10001
    expect_warning(
        {agg <- aggregateAndSample(
            fs = OMIP021Samples,
            nTotalEvents = nCells,
            seed = 1
        )},
        "Could not choose as much"
    )
    
    expect_equal(nrow(flowCore::exprs(agg)), 10000)
})

test_that("aggregateAndSample still works with flow frame", {
    nCells <- 100
    agg <- aggregateAndSample(fs = OMIP021Samples[[1]],
                              nTotalEvents = nCells,
                              seed = 1)
    expect_equal(nrow(flowCore::exprs(agg)), nCells)
    
    # too high total nb of cells
    nCells<- 10001
    expect_warning(
        {agg <- aggregateAndSample(
            fs = OMIP021Samples[[1]],
            nTotalEvents = nCells,
            seed = 1
        )},
        "Could not choose as much"
    )
    
    expect_equal(nrow(flowCore::exprs(agg)), 5000)
})

test_that("aggregateAndSample works with too few cells", {
    ffList <- list()
    ffList[[1]] <- OMIP021Samples[[1]]
    exprs1 <- flowCore::exprs(ffList[[1]])
    exprs1 <- exprs1[1:500,]
    flowCore::exprs(ffList[[1]]) <- exprs1
    ffList[[2]] <- OMIP021Samples[[2]]
    exprs2 <- flowCore::exprs(ffList[[2]])
    exprs2 <- exprs2[1:700,]
    flowCore::exprs(ffList[[2]]) <- exprs2
    ffList[[3]] <- OMIP021Samples[[2]]
    exprs3 <- flowCore::exprs(ffList[[3]])
    exprs3 <- exprs3[1:3000,]
    flowCore::exprs(ffList[[3]]) <- exprs3
    fsNew <- as(ffList,"flowSet")
    
    #(nCell/nFrame < min)
    nCells <- 600 
    agg1 <- aggregateAndSample(fs = fsNew,
                               setup = "forceBalance",
                               nTotalEvents = nCells,
                               seed = 1)
    expect_equal(nrow(flowCore::exprs(agg1)), nCells)
    ind1 <- which(flowCore::exprs(agg1)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg1)[ind1, ]), nCells / 3)
    ind2 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind2, ]), nCells / 3)
    ind3 <- which(flowCore::exprs(agg1)[, "File"] == 3)
    expect_equal(nrow(flowCore::exprs(agg1)[ind3, ]), nCells / 3)
    
    agg2 <- aggregateAndSample(fs = fsNew,
                               setup = "forceNEvent",
                               nTotalEvents = nCells,
                               seed = 1)
    expect_equal(nrow(flowCore::exprs(agg2)), nCells)
    ind4 <- which(flowCore::exprs(agg2)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg2)[ind4, ]), nCells / 3)
    ind5 <- which(flowCore::exprs(agg2)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg2)[ind5, ]), nCells / 3)
    ind6 <- which(flowCore::exprs(agg2)[, "File"] == 3)
    expect_equal(nrow(flowCore::exprs(agg2)[ind6, ]), nCells / 3)
    
    #(min < nCells/nFrame < nCells < max)
    nCells <- 1800
    expect_warning(
        {agg1 <- aggregateAndSample(fs = fsNew,
                               setup = "forceBalance",
                               nTotalEvents = nCells,
                               seed = 1)},
        "Could not choose as much")
    expect_equal(nrow(flowCore::exprs(agg1)), 1500)
    ind1 <- which(flowCore::exprs(agg1)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg1)[ind1, ]), 500)
    ind2 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind2, ]), 500)
    ind3 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind3, ]), 500)
    
    agg2 <- aggregateAndSample(fs = fsNew,
                               setup = "forceNEvent",
                               nTotalEvents = nCells,
                               seed = 1)
    expect_equal(nrow(flowCore::exprs(agg2)), nCells)
    ind4 <- which(flowCore::exprs(agg2)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg2)[ind4, ]), 500)
    ind5 <- which(flowCore::exprs(agg2)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg2)[ind5, ]), 650)
    ind6 <- which(flowCore::exprs(agg2)[, "File"] == 3)
    expect_equal(nrow(flowCore::exprs(agg2)[ind6, ]), 650)
    
    #(min < nCells/nFrame < max < nCells < sum)
    nCells <- 3300
    expect_warning(
        {agg1 <- aggregateAndSample(fs = fsNew,
                                    setup = "forceBalance",
                                    nTotalEvents = nCells,
                                    seed = 1)},
        "Could not choose as much")
    expect_equal(nrow(flowCore::exprs(agg1)), 1500)
    ind1 <- which(flowCore::exprs(agg1)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg1)[ind1, ]), 500)
    ind2 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind2, ]), 500)
    ind3 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind3, ]), 500)
    
    agg2 <- aggregateAndSample(fs = fsNew,
                               setup = "forceNEvent",
                               nTotalEvents = nCells,
                               seed = 1)
    expect_equal(nrow(flowCore::exprs(agg2)), nCells)
    ind4 <- which(flowCore::exprs(agg2)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg2)[ind4, ]), 500)
    ind5 <- which(flowCore::exprs(agg2)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg2)[ind5, ]), 700)
    ind6 <- which(flowCore::exprs(agg2)[, "File"] == 3)
    expect_equal(nrow(flowCore::exprs(agg2)[ind6, ]), 2100)
    
    # too high total nb of cells (nCells > sum)
    nCells <- 5000
    expect_warning(
        {agg1 <- aggregateAndSample(fs = fsNew,
                                    setup = "forceBalance",
                                    nTotalEvents = nCells,
                                    seed = 1)},
        "Could not choose as much")
    expect_equal(nrow(flowCore::exprs(agg1)), 1500)
    ind1 <- which(flowCore::exprs(agg1)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg1)[ind1, ]), 500)
    ind2 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind2, ]), 500)
    ind3 <- which(flowCore::exprs(agg1)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg1)[ind3, ]), 500)
    
    expect_warning(
        {agg2 <- aggregateAndSample(fs = fsNew,
                                    setup = "forceNEvent",
                                    nTotalEvents = nCells,
                                    seed = 1)},
        "Could not choose as much")
    expect_equal(nrow(flowCore::exprs(agg2)), 4200)
    ind4 <- which(flowCore::exprs(agg2)[, "File"] == 1)
    expect_equal(nrow(flowCore::exprs(agg2)[ind4, ]), 500)
    ind5 <- which(flowCore::exprs(agg2)[, "File"] == 2)
    expect_equal(nrow(flowCore::exprs(agg2)[ind5, ]), 700)
    ind6 <- which(flowCore::exprs(agg2)[, "File"] == 3)
    expect_equal(nrow(flowCore::exprs(agg2)[ind6, ]), 3000)
})

test_that("getTransfoParams works", {
    # hybrid transformation list :
    # - two channels are logicle-ly transformed with automatic param estimates
    # - one channel has explicit logicle transfo with default parameters
    # - one channel has linear transformation
    # - other channels have no transformation
    translist <- flowCore::estimateLogicle(
        OMIP021Samples[[1]],
        c("450/50Violet-A", "525/50Violet-A")
    )
    translist <- c(
        translist,
        flowCore::transformList(
            "FSC-A",
            flowCore::linearTransform(
                a = 0.1,
                b = 0
            )
        ),
        flowCore::transformList(
            "540/30Violet-A",
            flowCore::logicleTransform()
        )
    )

    ret <- getTransfoParams(translist, channel = "SSC-A")
    expect_equal(is.null(ret), TRUE)

    ret <- getTransfoParams(translist, channel = "FSC-A")
    expect_equal(ret$type, "linear")
    expect_equal(ret$paramsList$a, 0.1)
    expect_equal(ret$paramsList$b, 0.)

    ret <- getTransfoParams(translist, channel = "525/50Violet-A")
    expect_equal(ret$type, "logicle")
    expect_equal(ret$paramsList$a, 0.)
    myW <- 0.26281575
    expect_equal(ret$paramsList$w, myW)
    expect_equal(ret$paramsList$m, 4.5)
    expect_equal(ret$paramsList$t, 262143)

    ret <- getTransfoParams(translist, channel = "540/30Violet-A")
    expect_equal(ret$type, "logicle")
    expect_equal(ret$paramsList$a, 0.)
    expect_equal(ret$paramsList$w, 0.5)
    expect_equal(ret$paramsList$m, 4.5)
    expect_equal(ret$paramsList$t, 262144)

    tf <- flowCore::linearTransform(a = 1.1, b = 0.2)
    otherTransList <- flowCore::transformList(
        from = "FSC-A",
        tfun = tf
    )
    ret <- getTransfoParams(otherTransList,
        channel = "FSC-A"
    )
    expect_equal(ret$type, "linear")
    expect_equal(ret$paramsList$a, 1.1)
    expect_equal(ret$paramsList$b, 0.2)
})

test_that("computeScatterChannelsLinearScale works", {
    ff <- OMIP021Samples[[1]]
    refMarker <- "APCCy7 - CD4"
    refChannel <- "780/60Red-A"

    targetFSCA <- list()
    targetFSCA$type <- "linear"
    targetFSCA$paramsList <- list()
    targetFSCA$paramsList$a <- 1.3504601e-05
    targetFSCA$paramsList$b <- 0.41438898

    targetSSCA <- list()
    targetSSCA$type <- "linear"
    targetSSCA$paramsList <- list()
    targetSSCA$paramsList$a <- 1.07290554e-05
    targetSSCA$paramsList$b <- 0.38032844

    transList <- flowCore::estimateLogicle(ff,
        channels = refChannel
    )

    # base case, reference channel exists, transList pre-filled with logicle
    # transfo for ref channel
    retTransList <-
        computeScatterChannelsLinearScale(ff,
            transList = transList,
            referenceChannel = refMarker,
            silent = TRUE
        )

    retFSCA <- getTransfoParams(retTransList, channel = "FSC-A")
    expect_equal(retFSCA, targetFSCA)

    retFSCH <- getTransfoParams(retTransList, channel = "FSC-H")
    expect_equal(retFSCH, targetFSCA)

    retSSCA <- getTransfoParams(retTransList, channel = "SSC-A")
    expect_equal(retSSCA, targetSSCA)

    retSSCH <- getTransfoParams(retTransList, channel = "SSC-H")
    expect_equal(retSSCH, targetSSCA)

    # test with a reference channel that does not exist
    expect_error(computeScatterChannelsLinearScale(
        ff,
        transList = transList,
        referenceChannel = "Yipee",
        silent = TRUE
    ), regexp = "can't find")

    # test with a reference channel that is not a fluo channel
    expect_error(computeScatterChannelsLinearScale(
        ff,
        transList = transList,
        referenceChannel = "SSC-A",
        silent = TRUE
    ), regexp = "should be a fluorochrome channel")

    # test with a NULL transList

    targetFSCA$paramsList$a <- 0.047610702
    targetFSCA$paramsList$b <- -491.62068
    targetSSCA$paramsList$a <- 0.037825469
    targetSSCA$paramsList$b <- -611.70173
    retTransList <-
        computeScatterChannelsLinearScale(ff,
            transList = NULL,
            referenceChannel = refMarker,
            silent = TRUE
        )

    retFSCA <- getTransfoParams(retTransList, channel = "FSC-A")
    expect_equal(retFSCA, targetFSCA)

    retFSCH <- getTransfoParams(retTransList, channel = "FSC-H")
    expect_equal(retFSCH, targetFSCA)

    retSSCA <- getTransfoParams(retTransList, channel = "SSC-A")
    expect_equal(retSSCA, targetSSCA)

    retSSCH <- getTransfoParams(retTransList, channel = "SSC-H")
    expect_equal(retSSCH, targetSSCA)

    # test with a transList that does not contain the reference channel,
    # but well an already existing SSC-A transformation
    linTrans <- flowCore::linearTransform()
    stupidTransList <- flowCore::transformList(
        from = "SSC-A",
        tfun = linTrans
    )
    retTransList <-
        computeScatterChannelsLinearScale(ff,
            transList = stupidTransList,
            referenceChannel = refMarker,
            silent = TRUE
        )


    retFSCA <- getTransfoParams(retTransList, channel = "FSC-A")
    expect_equal(retFSCA, targetFSCA)

    retFSCH <- getTransfoParams(retTransList, channel = "FSC-H")
    expect_equal(retFSCH, targetFSCA)

    retSSCA <- getTransfoParams(retTransList, channel = "SSC-A")
    expect_equal(retSSCA, targetSSCA)

    retSSCH <- getTransfoParams(retTransList, channel = "SSC-H")
    expect_equal(retSSCH, targetSSCA)
})

test_that("findTimeChannel works", {
    # with flow set
    ret <- findTimeChannel(OMIP021Samples)
    expect_equal(ret, "Time")

    # with flow frame
    ret2 <- findTimeChannel(OMIP021Samples[[1]])
    expect_equal(ret2, "Time")

    # test exclude channels parameter
    ret3 <- findTimeChannel(OMIP021Samples[[1]],
        excludeChannels = "Time"
    )
    expect_null(ret3)
})

test_that("getChannelNamesFromMarkers works", {
    # with existing markers
    ret <- getChannelNamesFromMarkers(
        OMIP021Samples[[1]],
        c(
            "FSC-A",
            "L/D Aqua - Viability",
            "FITC - gdTCR",
            "PECy5 - CD28"
        )
    )

    expected <- c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")
    expect_equal(ret, expected)

    # with boolean vector
    indices <- c(1, 6, 14, 18)
    boolInput <- rep(FALSE, 21)
    boolInput[indices] <- TRUE
    ret <- getChannelNamesFromMarkers(
        OMIP021Samples[[1]],
        boolInput
    )

    expect_equal(ret, expected)


    # with indices vector
    ret <- getChannelNamesFromMarkers(
        OMIP021Samples[[1]],
        indices
    )
    expect_equal(ret, expected)

    # missing channel
    missingInput <- c("CD4")

    expect_error(getChannelNamesFromMarkers(
        OMIP021Samples[[1]],
        missingInput
    ),
    regexp = "could not be found"
    )
})

test_that("updateMarkerName works", {
    retFF <- updateMarkerName(OMIP021Samples[[1]],
                            channel = "FSC-A",
                            newMarkerName = "Fwd Scatter-A")

    checkMkName <- getChannelNamesFromMarkers(retFF, markers = "Fwd Scatter-A")
    expect_equal(checkMkName, "FSC-A")
    expect_equal(flowCore::keyword(retFF, "$P1S")[["$P1S"]], "Fwd Scatter-A")

    # same with channel as index
    retFF <- updateMarkerName(OMIP021Samples[[1]],
                              channel = 1,
                              newMarkerName = "Fwd Scatter-A")
    checkMkName <- getChannelNamesFromMarkers(retFF, markers = "Fwd Scatter-A")
    expect_equal(checkMkName, "FSC-A")
    expect_equal(flowCore::keyword(retFF, "$P1S")[["$P1S"]], "Fwd Scatter-A")

    # now channel provided as marker (desc), not channel name
    retFF <- updateMarkerName(OMIP021Samples[[1]],
                              channel = "BV785 - CD3",
                              newMarkerName = "fancy CD3")
    checkMkName <- getChannelNamesFromMarkers(retFF, markers = "fancy CD3")
    expect_equal(checkMkName, "670/30Violet-A")
    expect_equal(flowCore::keyword(retFF, "$P10S")[["$P10S"]], "fancy CD3")



    expect_error(updateMarkerName(OMIP021Samples[[1]], channel = "FFF-A",
                                  newMarkerName = "whatever"),
                 regexp = "channel not found")
})

test_that("removeChannels works", {
    retFF <- removeChannels(OMIP021Samples[[1]],
                            channels = c("FSC-A", "SSC-A"))
    nRemainingChannels <- length(flowCore::colnames(retFF))
    expect_equal(nRemainingChannels, 20)

    expect_warning(removeChannels(OMIP021Samples[[1]],
                                  channels = c("FSC-A", "SSC-S")),
                   regexp = "not found in flowFrame")


})

test_that("getFCSFileName works", {
    fName <- getFCSFileName(OMIP021Samples[[1]])
    expect_equal(fName, "Donor1.fcs")
})

test_that("writeFlowFrame works", {
    outputDir <- base::tempdir()

    ff_c <- readRDS(test_path("fixtures", "ff_c.rds"))

    expect_error(writeFlowFrame(ff, dir = file.path(outputDir, "notThere")),
                 regexp = "Provided directory does not exist")

    prefix <- "File_"
    suffix <- "_export"

    writeFlowFrame(ff_c, dir = outputDir,
                   useFCSFileName = TRUE,
                   prefix = prefix,
                   suffix = suffix,
                   format = "fcs")

    outputFile <- file.path(outputDir,
                            paste0(prefix, "Donor1", suffix, ".fcs"))

    thisFF <- flowCore::read.FCS(outputFile, transform = FALSE)
    expect_true(all(round(flowCore::exprs(thisFF), 0)
                    == round(flowCore::exprs(ff_c), 0)))

    writeFlowFrame(ff_c, dir = outputDir,
                   useFCSFileName = FALSE,
                   prefix = prefix,
                   suffix = suffix,
                   format = "csv")

    outputCSV <- file.path(outputDir,
                           paste0(prefix, suffix, ".csv"))

    thisExpr <- read.csv(file = outputCSV)
    expect_true(all(round(thisExpr,4) == round(flowCore::exprs(ff_c), 4)))
})


test_that("updateCompMatrixLabels works", {
    myMat <- diag(nrow = 3)
    names <- c("450/50Violet-A", "BV605 - CD161",
               "670/30Yellow-A :: BV785 - CD3")
    colnames(myMat) <- rownames(myMat) <- names

    myNewMat <- .updateCompMatrixLabels(myMat, OMIP021Samples[[1]])

    newColNames <- colnames(myNewMat)
    newRowNames <- rownames(myNewMat)

    expectedNewNames <- c("450/50Violet-A",
                          "540/30Violet-A",
                          "670/30Yellow-A")

    expect_equal(newColNames, expectedNewNames)
    expect_equal(newRowNames, expectedNewNames)
})

# CytoPipeline - Copyright (C) <2022>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
#   Public License as published by the Free Software Foundation, either
#   version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).


## code to prepare `OMIP021UT` dataset (light version of `OMIP021` used in
## unit testing (testthat))

library(CytoPipeline)

data(OMIP021Samples)

# sub-sample equal nb of events in each fcs
sampleSize <- 100
OMIP021UTSamples <- flowCore::fsApply(
    x = OMIP021Samples,
    FUN = function(ff) {
        subsample(ff, 
                  nEvents = sampleSize, 
                  seed = 1,
                  keepOriginalCellIDs = FALSE)
    }
)

# CytoPipeline - Copyright (C) <2022>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
#   Public License as published by the Free Software Foundation, either
#   version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).


## code to prepare `OMIP021` dataset

# flow cytometry core librairies
library(flowCore)
library(CytoPipeline)

# in the following statement, specify the folder containing the dataset 
# manually downloaded from FlowRepository.org
# note that automatic download API for FlowRepository does not work anymore 
# for the time being (May 2022)
datasetPath <- "/Datasets/OMIP-021/FlowRepository_FR-FCM-ZZ9H_files"

files <- list.files(datasetPath, pattern = "Donor", recursive = TRUE)

nDonors <- length(files)

# init Flow Set

OMIP021Samples <- read.flowSet(
    paste0(datasetPath, "/", files),
    transformation = "linearize",
    alter.names = FALSE,
    name = "OMIP-21",
    truncate_max_range = TRUE,
    min.limit = NULL
)

flowCore::pData(OMIP021Samples)$Donor <- paste0("Donor_", seq(nDonors))

# sub-sample equal nb of events in each fcs
# set 
sampleSize <- 5000
OMIP021Samples <- fsApply(
    x = OMIP021Samples,
    FUN = function(ff) {
        subsample(ff, 
                  nEvents = sampleSize, 
                  seed = 1,
                  keepOriginalCellIDs = FALSE)
    }
)

fsApply(
    x = OMIP021Samples,
    FUN = function(ff) {
        basefilename <- flowCore::identifier(ff)
        path <- system.file("extdata",
            package = "CytoPipeline"
        )
        flowCore::keyword(ff)[["$FIL"]] <- basefilename
        flowCore::write.FCS(ff,
            filename = file.path(path, basefilename)
        )
    }
)

usethis::use_data(OMIP021Samples, overwrite = TRUE)

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

setClassUnion("characterOrFunction", c("character", "function"))

#' @title Cyto Processing step
#'
#' @aliases CytoProcessingStep-class characterOrFunction-class
#'
#' @description
#'
#' Class containing the function and arguments to be applied in a lazy-execution
#' framework.
#'
#' Objects of this class are created using the `CytoProcessingStep()`    
#' function.
#' The processing step is executed with the `executeProcessingStep()`     
#' function.
#'
#' @details
#'
#' This object contains all relevant information of a data analysis processing
#' step, i.e. the function and all of its arguments to be applied to the data.
#'
#' @return The `CytoProcessingStep` function returns and object of type
#'     `CytoProcessingStep`.
#'
#' @exportClass CytoProcessingStep
#'
#' @examples
#'
#' ## Create a simple processing step object
#' ps1 <- CytoProcessingStep("summing step", sum)
#'
#' getCPSName(ps1)
#' 
#' getCPSFUN(ps1)
#' 
#' getCPSARGS(ps1)
#'
#' executeProcessingStep(ps1, 1:10)
#'
#' as.list(ps1)
#' 
#' js_str <- as.json.CytoProcessingStep(ps1)
#' 
#' ps2 <- from.json.CytoProcessingStep(js_str)
#' 
#' identical(ps1, ps2)
#'
#' @name CytoProcessingStep
NULL

setClass("CytoProcessingStep",
    representation = representation(
        name = "character",
        FUN = "characterOrFunction",
        ARGS = "list"
    ),
    prototype = prototype(
        name = character(),
        ARGS = list(),
        FUN = character()
    ),
    validity = function(object) {
        # msg <- character()
        # ## Fails with un-exported functions.
        # if (length(object@FUN)) {
        #     if (!is.function(object@FUN)) {
        #         res <- try(match.fun(object@FUN), silent = TRUE)
        #         if (methods::is(res, "try-error")) {
        #             msg <- c(msg, paste0(
        #                 "Function '", object@FUN,
        #                 "' not found."
        #             ))
        #         }
        #     }
        # }
        # if (length(msg)) {
        #     msg
        # } else {
            TRUE
    #     }
    }
)

#' @param name `character` denoting a name to the step, which can be different
#' from the function name
#'
#' @param FUN `function` or `character` representing a function name.
#'
#' @param ARGS `list` of arguments to be passed along to `FUN`.
#'
#' @rdname CytoProcessingStep
#'
#' @export
#' 
CytoProcessingStep <- function(name = character(), FUN = character(),
                               ARGS = list()) {
    if (missing(FUN)) {
        FUN <- character()
    }
    methods::new("CytoProcessingStep", name = name, FUN = FUN, ARGS = ARGS)
}

.cat_fun <- function(x) {
    if (is.function(x)) {
        "user-provided function"
    } else {
        paste(x, collapse = ", ")
    }
}

#' @rdname CytoProcessingStep
#'
#' @param object a `CytoProcessingStep` object.
#'
#' @exportMethod show
setMethod("show", "CytoProcessingStep", function(object) {
    cat("Object of class \"", class(object), "\"\n", sep = "")
    cat(" Name: ", object@name, "\n", sep = "")
    cat(" Function: ", .cat_fun(object@FUN), "\n", sep = "")
    args <- object@ARGS
    if (length(args) > 0) {
        cat(" Arguments:\n")
        for (i in seq_along(args)) {
            cat("  o ", names(args)[i], " = ", .cat_fun(args[[i]]),
                "\n",
                sep = ""
            )
        }
    }
})

#' @param x                 a `CytoProcessingStep` object.
#' @param ...               optional additional arguments to be passed along.
#'
#' @rdname CytoProcessingStep
#'
#' @md
#'
#' @export
executeProcessingStep <- function(x, ...) {
    
    if (!methods::is(x, "CytoProcessingStep")) {
        stop("'x' should be a 'CytoProcessingStep' object!")
    }
    
    # first validate processing step before execution!
    msg <- character()
    ## Fails with un-exported functions.
    if (length(x@FUN)) {
        if (!is.function(x@FUN)) {
            res <- try(match.fun(x@FUN), silent = TRUE)
            if (methods::is(res, "try-error")) {
                msg <- c(msg, paste0(
                    "Function '", x@FUN,
                    "' not found."
                ))
            }
        }
    }
    if (length(msg)) {
        stop(msg)
    }
    
    # extract samplePhenoData from ...
    dot3List <- list(...)
    pData <- dot3List[["pData"]]
    if (!is.null(pData)) {
        if (!inherits(pData, "data.frame")) {
            stop("pData must be a data frame")
        }
    }
    dot3List[["pData"]] <- NULL
    
    # match specific `$` arguments if pData
    theArgs <- x@ARGS 
    
    theArgs <- lapply(
        theArgs,
        FUN = function(arg, pData){
            if(is.character(arg) && length(arg) == 1) {
                if(substr(arg, 1, 1) == "$") {
                    if (is.null(pData)) {
                        stop("'$' argument needs not null pData")
                    } else {
                        pDataColName <- substr(arg, 2, nchar(arg))
                        arg <- pData[, pDataColName, drop = TRUE]
                    }
                }
            }
            arg
        },
        pData = pData
    )
    
    # execute processing step
    ret <- do.call(x@FUN, args = c(dot3List, theArgs))
    ret
}

#' @param x a `CytoProcessingStep` object.
#'
#' @rdname CytoProcessingStep
#'
#' @export
getCPSName <- function(x) {
    if (!methods::is(x, "CytoProcessingStep")) {
        stop("'x' should be a 'CytoProcessingStep' object!")
    }
    return(x@name)
}

#' @param x a `CytoProcessingStep` object.
#'
#' @rdname CytoProcessingStep
#'
#' @export
getCPSFUN <- function(x) {
    if (!methods::is(x, "CytoProcessingStep")) {
        stop("'x' should be a 'CytoProcessingStep' object!")
    }
    return(x@FUN)
}

#' @param x a `CytoProcessingStep` object.
#'
#' @rdname CytoProcessingStep
#'
#' @export
getCPSARGS <- function(x) {
    if (!methods::is(x, "CytoProcessingStep")) {
        stop("'x' should be a 'CytoProcessingStep' object!")
    }
    return(x@ARGS)
}

#' @param x a `CytoProcessingStep` object.
#' @param ... other arguments (not used)
#'
#' @rdname CytoProcessingStep
#'
#' @export
as.list.CytoProcessingStep <- function(x, ...) {
    if (!methods::is(x, "CytoProcessingStep")) {
        stop("'x' should be a 'CytoProcessingStep' object!")
    }
    
    ll <- list()
    ll$name <- x@name
    
    if (is.character(x@FUN)) {
        ll$FUN <- x@FUN
    } else if (utils::isS3stdGeneric(x@FUN) ||
        is.primitive(x@FUN)) {
        getFun <- function(fun){
            fun <- deparse(fun)
            chunk <- utils::tail(fun, 1)
            words <- strsplit(chunk, "\"")[[1]]
            return(words[2])
        }
        ll$FUN <- getFun(x@FUN)
    } else {
        stop("export of CytoProcessingStep as a list or json does not work ",
             "(yet) if FUN is a non primitive, non generic function\n",
             "=> use function name as character instead of function object")
    }
    
    ll$ARGS <- x@ARGS
    return(ll)
}

#' @param x a `CytoProcessingStep` object.
#' @param pretty formatting set-up (see jsonlite::toJSON doc)
#'
#' @rdname CytoProcessingStep
#'
#' @export
as.json.CytoProcessingStep <- function(x, pretty = FALSE) {
    myList <- as.list.CytoProcessingStep(x)
    ret <- jsonlite::toJSON(myList, pretty = pretty, null = "null")
    return(as.character(ret))
}

#' @param jsonString a `character()` containing a JSON string.
#'
#' @rdname CytoProcessingStep
#'
#' @export
from.json.CytoProcessingStep <- function(jsonString) {
    myList <- jsonlite::fromJSON(jsonString, simplifyDataFrame = FALSE)
    if (is.null(myList$name)) {
        stop("name not found in Processing Step json string")
    }
    if (is.null(myList$FUN)) {
        stop("FUN not found in Processing Step json string")
    }
    if (is.null(myList$ARGS)) {
        stop("FUN not found in Processing Step json string")
    }
    object <- CytoProcessingStep(myList$name, myList$FUN, myList$ARGS)
    return(object)
}

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).


# obtain OMIP021UTSamples, light-weight version used specifically for these
# unit tests
path <- system.file("scripts",
                    package = "CytoPipeline"
)

source(file.path(path,"MakeOMIP021UTSamples.R"))

test_that("CytoProcessingStep basics works", {
    ps <- CytoProcessingStep("summing step", sum)

    psName <- getCPSName(ps)
    expect_equal(psName, "summing step")

    psFUN <- getCPSFUN(ps)
    expect_true(is.primitive(psFUN))

    psARGS <- getCPSARGS(ps)
    expect_identical(psARGS, list())

    expect_error(show(ps), NA)

    res <- executeProcessingStep(ps, 1:10)
    expect_equal(res, 55)
})

test_that("CytoProcessingStep works with pData", {
    sumWithBounds <- function(low, high) {
        if (low > high) stop("low > high !")
        sum(seq(from = low, to = high))
    }
    ps <- CytoProcessingStep("summing step",
                             sumWithBounds,
                             ARGS = list(low = 1, high = 10))
    res <- executeProcessingStep(ps)
    expect_equal(res, 55)

    sPD <- data.frame(VAR1 = 5, VAR2 = 8)

    ps2 <- CytoProcessingStep("summing step",
                              sumWithBounds,
                              ARGS = list(low = 1,
                                          high = "$VAR2"))
    res2 <- executeProcessingStep(ps2,
                                  pData = sPD)
    expect_equal(res2, 36)

    ps3 <- CytoProcessingStep("summing step",
                             sumWithBounds,
                             ARGS = list(low = "$VAR1",
                                         high = "$VAR2"))

    res3 <- executeProcessingStep(ps3,
                                  pData = sPD)

    expect_equal(res3, 26)

})

test_that("CytoProcessingStep wrong function works", {
    ps <- CytoProcessingStep("dummy step", "mistake_fun")

    psName <- getCPSName(ps)
    expect_equal(psName, "dummy step")

    expect_error(executeProcessingStep(ps, 1:10),
                 regexp = "not found")
})

test_that("CytoProcessingStep exports and imports work", {
    # case of a primitive
    ps <- CytoProcessingStep("summing step", sum)
    js_str <- as.json.CytoProcessingStep(ps)

    ps2 <- from.json.CytoProcessingStep(js_str)

    res <- executeProcessingStep(ps2, 1:10)
    expect_equal(res, 55)

    # case of a generic function
    ps <- CytoProcessingStep("median step", stats::median)

    js_str <- as.json.CytoProcessingStep(ps)

    ps2 <- from.json.CytoProcessingStep(js_str)

    res <- executeProcessingStep(ps2, 1:10)
    expect_equal(res, 5.5)

    # other case
    ps <- CytoProcessingStep("compensate step", "compensateFromMatrix")

    ff <- executeProcessingStep(ps, OMIP021UTSamples[[1]])
    res <- sum(flowCore::exprs(ff)[,"FSC-A"])
    expect_equal(res, 12553542.8)

    js_str <- as.json.CytoProcessingStep(ps)
    ps2 <- from.json.CytoProcessingStep(js_str)

    ff <- executeProcessingStep(ps2, OMIP021UTSamples[[1]])
    res <- sum(flowCore::exprs(ff)[,"FSC-A"])
    expect_equal(res, 12553542.8)

    # not yet implemented case (non generic, non primitive function as object)
    ps <- CytoProcessingStep("compensate step", compensateFromMatrix)
    expect_error(as.json.CytoProcessingStep(ps),
                 regexp = "does not work")

})

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

#' @title estimates scale tranformations
#' @description this function estimates the scale transformations to be applied
#' on a flowFrame to obtain 'good behaving' distributions, i.e. the best
#' possible separation between + population and - population.
#' It distinguishes between scatter channels, where either linear, or no
#' transform is applied, and fluo channels, where either logicle transform
#' - using flowCore::estimateLogicle - is estimated, or no transform   
#' is applied.   
#' 
#' The idea of linear transform of scatter channels is as follows: a reference
#' channel (not a scatter one) is selected and a linear transform (Y = AX + B)
#' is applied to all scatter channel, as to align their 5 and 95 percentiles to
#' those of the reference channel
#' For the estimateLogicle function, see flowCore documentation.

#' @param ff a flowCore::flowFrame
#' @param fluoMethod method to be applied to all fluo channels
#' @param scatterMethod method to be applied to all scatter channels
#' @param scatterRefMarker the reference channel that is used to align the
#' @param specificScatterChannels vector of scatter channels for which we 
#' still want to apply the fluo method (and not the scatter Method)
#' @param verbose if TRUE, send messages to the user at each step
#' @param ... additional parameters passed to flowCore::estimateLogicle()
#'
#' @return a flowCore::flowFrame with removed low quality events from the input
#' @export
#'
#' @examples
#' 
#' data(OMIP021Samples)
#' 
#' compMatrix <- flowCore::spillover(OMIP021Samples[[1]])$SPILL
#' ff_c <- runCompensation(OMIP021Samples[[1]], spillover = compMatrix)
#' 
#' transList <- 
#'     estimateScaleTransforms(        
#'         ff = ff_c,
#'         fluoMethod = "estimateLogicle",
#'         scatterMethod = "linear",
#'         scatterRefMarker = "BV785 - CD3")
#' 
estimateScaleTransforms <- function(ff,
                                    fluoMethod = c("estimateLogicle",
                                                   "none"),
                                    scatterMethod = c("none",
                                                      "linearQuantile"),
                                    scatterRefMarker = NULL,
                                    specificScatterChannels = NULL,
                                    verbose = FALSE,
                                    ...){
    fluoMethod <- match.arg(fluoMethod)
    scatterMethod <- match.arg(scatterMethod)
    
    if (fluoMethod == "estimateLogicle") {
        if (verbose) {
            message(
                "estimating logicle transformations ",
                "for fluorochrome channels..."
            ) 
        }
        fluoCols <- flowCore::colnames(ff)[areFluoCols(ff)]
        transList <- flowCore::estimateLogicle(ff, fluoCols, ...)
    } # else do nothing
    
    if (scatterMethod == "linearQuantile") {
        if (is.null(scatterRefMarker)) {
            stop("linear scatter method requires a scatterRefMarker")
        }
        
        if (verbose) {
            message(
                "Estimating linear transformation for scatter channels : ",
                "reference marker = ",
                scatterRefMarker,
                "..."
            )
        }
        
        transList <-
            computeScatterChannelsLinearScale(
                ff,
                transList = transList,
                referenceChannel = scatterRefMarker,
                silent = !verbose
            )
    } # else do nothing
    
    # handle specific cases of scatter channels that still need fluo method
    if (!is.null(specificScatterChannels) && fluoMethod == "estimateLogicle") {
        effectiveScatterChannels <- NULL
        for (ch in specificScatterChannels) {
            scatterChannels <-
                flowCore::colnames(ff)[!areFluoCols(ff) & areSignalCols(ff)]
            if (!(ch %in% scatterChannels)) {
                if (verbose) {
                    message("Specific channel [", ch, "] is not a scatter channel",
                            " => no correction of scale transformation done")
                }
            } else {
                transList@transforms[[ch]] <- NULL
                effectiveScatterChannels <- c(effectiveScatterChannels, ch)
                if (verbose) {
                    message("Specific scatter channel found: [", ch, "] ",
                            "=> correcting scale transformation to logicle...")
                }
            } 
        }
        
        transList <- c(transList, 
                       flowCore::estimateLogicle(ff, 
                                                 effectiveScatterChannels,
                                                 ...))
    }
    
    return(transList)
}


#' @title Read fcs sample files
#' @description Wrapper around flowCore::read.fcs() or flowCore::read.flowSet().
#' Also adds a "Cell_ID" additional column, used in flowFrames comparison
#' @param sampleFiles a vector of character path to sample files
#' @param whichSamples one of:
#' - 'all' if all sample files need to be read
#' - 'random' if some samples need to be chosen randomly 
#' (in that case, using `nSamples` and `seed`)
#' - a vector of indexes pointing to the sampleFiles vector
#' @param nSamples number of samples to randomly select 
#' (if `whichSamples == "random"`). 
#' If `nSamples` is higher than nb of available samples, 
#' the output will be all samples
#' @param seed an optional seed parameters (provided to ease reproducibility).
#' @param channelMarkerFile an optional path to a csv file which provides the 
#' mapping between channels and markers. If provided, this csv file should 
#' contain a `Channel` column, and a `Marker` column. Optionally a 'Used' 
#' column can be provided as well (TRUE/FALSE). Channels for which the 'Used' 
#' column is set to FALSE will not be incorporated in the created flowFrame. 
#' @param ... additional parameters passed to flowCore file reading functions.
#'
#' @return either a flowCore::flowSet or a flowCore::flowFrame if
#' length(sampleFiles) == 1
#' @export
#' 
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' res <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' #res
#' 
#' # create a flowCore::flowFrame with one single sample
#' res2 <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = 2,
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' #res2
#' 
readSampleFiles <- function(sampleFiles,
                            whichSamples = "all", 
                            nSamples = NULL,
                            seed = NULL,
                            channelMarkerFile = NULL,
                            ...) {
    
    if (whichSamples == "all") {
        # do nothing : sampleFiles should contain all the input sample files
        # already
    } else if (whichSamples == "random") {
        if (!is.numeric(nSamples) || nSamples < 1) {
            stop("[nSamples] should be a numeric >= 1")
        }
        
        nAvailableSamples <- length(sampleFiles)
        
        if (nSamples > nAvailableSamples) nSamples <- nAvailableSamples
        
        if (!is.null(seed)) {
            # set the seed locally in the execution environment,
            # restore it afterward
            withr::local_seed(seed)
        }
        
        whichSamples <- sample(seq_along(sampleFiles), nSamples)
        sampleFiles <- sampleFiles[whichSamples]
    } else if (is.numeric(whichSamples)) {
        sampleFiles <- sampleFiles[whichSamples]
    } else {
        stop("'whichSamples' should be either 'all', or a vector of indexes")
    }
    

    if (length(sampleFiles) == 0) {
        stop("no sample files to read")
    } else if (length(sampleFiles) == 1) {
        res <- flowCore::read.FCS(sampleFiles, ...)
        
        # Add a column with Cell ID
        res <- appendCellID(res)
    } else {
        res <- flowCore::read.flowSet(sampleFiles,
                                      ...)
        
        # work around to have the full file paths in sampleNames
        # as apparently this is not done automatically by
        # flowCore::read.flowSet - only the base filename is stored
        flowCore::sampleNames(res) <- sampleFiles
        
        # Add a column with Cell ID
        res <- flowCore::fsApply(
            x = res,
            FUN = function(ff) {
                appendCellID(ff)
            }
        )
        
        
    } 
    
    
    
    # do we need to do any post processing to the files ?
    # => remove channels or update marker names ? 
    
    if (!is.null(channelMarkerFile)) {
        if (!file.exists(channelMarkerFile)) {
            stop("channelMarkerFile [", channelMarkerFile, "] not found!")
        }
        channelMarkerMapping <- utils::read.csv(channelMarkerFile)
        
        message("COL NAMES MARKER MAPPING: ", 
                paste(colnames(channelMarkerMapping), collapse = ", "))
        
        if (!("Channel" %in% colnames(channelMarkerMapping))) {
            stop("channelMarkersMapping should contain [Channel] column!")
        }
        if (!("Marker" %in% colnames(channelMarkerMapping))) {
            stop("channelMarkersMapping should contain [Marker] column!")
        }
        if ("Used" %in% colnames(channelMarkerMapping)) {
            channelMarkerMapping$Used <- as.logical(channelMarkerMapping$Used)
            # if (!is.logical(channelMarkerMapping$Used)) {
            #     stop("channelMarkerMapping [Used] column ",
            #          "should be of logical type!")
            # }
            notUsedChannels <- 
                channelMarkerMapping[!channelMarkerMapping$Used, "Channel"]
            
        } 
        
        postProcessing <- function(ff, channelMarkerMapping, notUsedChannels){
            ff <- removeChannels(ff, notUsedChannels)
            
            channels <- 
                channelMarkerMapping[channelMarkerMapping$Used, "Channel"]
            markers <- 
                channelMarkerMapping[channelMarkerMapping$Used, "Marker"]
                
            for (i in seq_along(channels)) {
                if (markers[i] != "") {
                    ff <- updateMarkerName(ff, channel = channels[i], 
                                           newMarkerName = markers[i])
                }
            }
            return(ff)
        }
        
        if (length(sampleFiles) == 1) {
            res <- postProcessing(res, channelMarkerMapping, notUsedChannels)
        } else {
            res <- flowCore::fsApply(
                x = res,
                FUN = function(ff) {
                    postProcessing(ff, channelMarkerMapping, notUsedChannels)
                }
            )
        } 
    }
    
    return (res)
}



#' @title remove margin events using PeacoQC
#' @description Wrapper around PeacoQC::RemoveMargins().
#' Also pre-selects the channels to be handled (=> all signal channels)
#' If input is a flowSet, it applies removeMargins() to each flowFrame of the
#' flowSet.
#' @param x a flowCore::flowSet or a flowCore::flowFrame
#' @param channelSpecifications A list of lists with parameter specifications 
#' for certain channels. This parameter should only be used if the values in 
#' the internal parameters description is too strict or wrong for a number or 
#' all channels. This should be one list per channel with first a minRange 
#' and then a maxRange value. This list should have the channel name found back 
#' in colnames(flowCore::exprs(ff)), or the corresponding marker name (found in
#' flowCore::pData(flowCore::description(ff)) ) . 
#' If a channel is not listed in this parameter, its default internal values 
#' will be used. The default of this parameter is NULL.
#' If the name of one list is set to `AllFluoChannels`, then the  
#' `minRange` and `maxRange` specified there will be taken as default   
#' for all fluorescent channels (not scatter)
#' @param ... additional parameters passed to PeacoQC::RemoveMargins()
#'
#' @return either a flowCore::flowSet or a flowCore::flowFrame depending on
#' the input.
#' 
#' @export
#'
#' @examples
#' 
#' rawDataDir <- 
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <- 
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' fsRaw <- readSampleFiles(sampleFiles, 
#'                          truncate_max_range = truncateMaxRange,
#'                          min.limit = minLimit)
#' suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
#' ggplotFilterEvents(ffPre = fsRaw[[2]],
#'                    ffPost = ff_m,
#'                    xChannel = "FSC-A",
#'                    yChannel = "SSC-A")
removeMarginsPeacoQC <- function(x, channelSpecifications = NULL, ...) {
    PQCChannelSpecs <- channelSpecifications
    
    if (!is.null(channelSpecifications)) {
        if (!methods::is(channelSpecifications, "list")) 
            stop("channelSpecifications should be a list of lists.")
        if (!all(lengths(channelSpecifications) == 2)) 
            stop("channel_specifications should be a list of lists. \n",
                 "Every list should have the channel name and should contain\n",
                 "a minRange and maxRange value.")
    }
    
    
    myFunc <- function(ff, channelSpecifications) {
        message("Removing margins from file : ", getFCSFileName(ff))
        channel4Margins <-
            flowCore::colnames(ff)[areSignalCols(ff)]
        
        markers4Margins <- 
            flowCore::pData(flowCore::parameters(ff))$desc[areSignalCols(ff)]
        
        PQCChannelSpecs <- channelSpecifications
        
        
        if (!is.null(PQCChannelSpecs)) {
            
            #store default fluo parameters (if any)
            defaultFluoChannelList <- PQCChannelSpecs[["AllFluoChannels"]]
            if (!is.null(defaultFluoChannelList)) {
                PQCChannelSpecs[["AllFluoChannels"]] <- NULL
            }
            
            newNames <- names(PQCChannelSpecs)
            for (l in seq_along(newNames)) {
                chName <- newNames[l]
                if (!(chName %in% flowCore::colnames(ff))) {
                    # try as marker
                    whichMarker <- which(markers4Margins == chName)
                    if (length(whichMarker) == 0)
                        stop("channelSpecifications names: ",
                             "could not find [", chName, "], neither as ",
                             "channel, nor as marker")
                    else {
                        newNames[l] <- channel4Margins[whichMarker[1]]
                    }
                }
            }
            names(PQCChannelSpecs) <- newNames
            
            # apply default fluo parameters, if any
            if (!is.null(defaultFluoChannelList)) {
                for (ch in flowCore::colnames(ff)
                     [CytoPipeline::areFluoCols(ff)]) {
                    if (!(ch %in% newNames)) {
                        PQCChannelSpecs[[ch]] <- defaultFluoChannelList
                    }
                }
            }
        }
        
        ffOut <- PeacoQC::RemoveMargins(ff, channels = channel4Margins,
                                        channel_specifications = 
                                            PQCChannelSpecs)
        return(ffOut)
    }
    if (inherits(x, "flowFrame")) {
        return(myFunc(x, channelSpecifications = channelSpecifications))
    } else if (inherits(x, "flowSet")) {
        fsOut <- flowCore::fsApply(x, FUN = myFunc, simplify = TRUE,
                                   channelSpecifications = 
                                       channelSpecifications)
        return(fsOut)
    } else {
        stop("x should be a flowCore::flowFrame or a flowCore::flowSet")
    }
}

### FUNCTIONS for pre-processing / compensation ###

#' @title extract compensation matrix from a flowCore::flowFrame
#' @description helper function retrieving the compensation matrix stored
#' in fcs file (if any). It scans the following keywords: $SPILL, $spillover
#' and $SPILLOVER
#' @param ff a flowCore::flowFrame
#'
#' @return the found compensation matrix
#' @export
#'
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' compensationMatrix <- getAcquiredCompensationMatrix(fsRaw[[2]])
getAcquiredCompensationMatrix <- function(ff) {
    
    stopifnot (inherits(ff, "flowFrame"))
    
    res <- flowCore::spillover(ff)
    if (!is.null(res$SPILL)) {
        compensationMatrix <- res$SPILL
    } else if (!is.null(res$spillover)) {
        compensationMatrix <- res$spillover
    } else if (!is.null(res$`$SPILLOVER`)) {
        compensationMatrix <- res$`$SPILLOVER`
    } else {
        fileId <- getFCSFileName(ff)
        stop(
            "Issue retrieving compensation matrix for file ",
            fileId, " : slot is NULL!"
        )
    }
    return(compensationMatrix)
}

#' @title compensation of fcs file(s) from matrix
#' @description executes the classical compensation function on a flowSet or
#' flowFrame, given a compensation matrix. The matrix can be either retrieved
#' in the fcs files themselves or provided as a csv file.
#' @param x a `flowCore::flowFrame` or `flowCore::flowSet`
#' @param matrixSource if "fcs", the compensation matrix will be fetched from
#' the fcs files (different compensation matrices can then be applied by fcs
#' file)
#' if "import", uses `matrixPath` to read the matrix (should be a csv file)
#' @param matrixPath if matrixSource == "import", will be used as the input csv
#' file path
#' @param updateChannelNames if TRUE, updates the fluo channel names by
#' prefixing them with "comp-"
#' @param verbose if TRUE, displays information messages
#' @param ... additional arguments (not used)
#' @return the compensated flowSet or flowFrame
#' @export
#'
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
#'     
#' ff_c <-
#'     compensateFromMatrix(ff_m,
#'                          matrixSource = "fcs")        
compensateFromMatrix <- function(x,
                                 matrixSource = c("fcs", "import"),
                                 matrixPath = NULL,
                                 updateChannelNames = TRUE,
                                 verbose = FALSE,
                                 ...) {

    matrixSource <- match.arg(matrixSource)
    
    if (matrixSource == "import"){
        if (is.null(matrixPath)) {
            stop("matrixPath can't be NULL if matrixSource == 'import'")
        }
    }
    
    importCompensationMatrix <- function(ff, matrixPath) {
        # import matrix from file (path)
        if (is.null(matrixPath)) {
            stop("No path specified for compensation matrix!")
        }
        if (!file.exists(matrixPath)) {
            stop("Compensation matrix file [", matrixPath,
                 "] not found!")
        }
        compensationMatrix <-
            utils::read.csv(matrixPath,
                            check.names = FALSE,
                            row.names = 1
            )
        compensationMatrix <- 
            .updateCompMatrixLabels(compensationMatrix, ff)
        
        return(compensationMatrix)
    }
    
    compensateOneFFWithSource <- function(
        ff, matrixSource, matrixPath, verbose){
        
        if (verbose) {
            message("Compensating file : ", 
                    getFCSFileName(ff),
                    "; matrixSource = ",
                    matrixSource,
                    "; matrixPath = ",
                    matrixPath)
        }
        
        if (matrixSource == "fcs") {
            # obtains compensation matrix
            compensationMatrix <-
                getAcquiredCompensationMatrix(ff)
        } else {
            # find correct matrix path
            compensationMatrix <- importCompensationMatrix(ff, matrixPath)
        }
        
        ffOut <- runCompensation(ff,
                                 compensationMatrix,
                                 updateChannelNames = updateChannelNames
        )
        
        return(ffOut)
    }
            
    if (inherits(x, "flowFrame")) {
        res <- compensateOneFFWithSource(
            ff = x, 
            matrixSource = matrixSource, 
            matrixPath = matrixPath,
            verbose = verbose)
    } else if (inherits(x, "flowSet")) {
        if (matrixSource == "fcs" || 
            matrixSource == "import" && length(matrixPath) <= 1) {
            res <- flowCore::fsApply(x,
                                     FUN = compensateOneFFWithSource,
                                     simplify = TRUE,
                                     matrixSource = matrixSource,
                                     matrixPath = matrixPath,
                                     verbose = verbose)
        } else {
            # matrix path is different from flowFrame to flowFrame
            # => need to use mapply() instead of fsApply()
            res <- structure(
                mapply(x, 
                       matrixPath,
                       FUN = function(ff, matrixPath, matrixSource, verbose) {
                           ff <- compensateOneFFWithSource(
                               ff = ff,
                               matrixSource = matrixSource,
                               matrixPath = matrixPath,
                               verbose = verbose
                           )
                           ff
                       },
                       MoreArgs = list(matrixSource = matrixSource,
                                       verbose = verbose)), 
                names = flowCore::sampleNames(x))
            res <- methods::as(res, "flowSet")
            flowCore::phenoData(res) <- 
                flowCore::phenoData(x)[flowCore::sampleNames(x), , drop = FALSE]
        }
        
    } else {
        stop("x should be a flowCore::flowFrame or a flowCore::flowSet")
    }
    
    res

}



### FUNCTIONS FOR DOUBLETS REMOVAL ###



#' @title remove doublets from a flowFrame, using CytoPipeline custom algorithm
#' @description Wrapper around CytoPipeline::singletGate().
#' Can apply the flowStats function subsequently on several channel pairs,
#' e.g. (FSC-A, FSC-H) and (SSC-A, SSC-H)
#' @param ff a flowCore::flowFrame
#' @param areaChannels a character vector containing the name of the 'area type'
#' channels one wants to use
#' @param heightChannels a character vector containing the name of the
#' 'height type' channels one wants to use
#' @param nmads a numeric vector with the bandwidth above the ratio allowed, per
#' channels pair (cells are kept if the ratio between -A channel\[i\] and
#' -H channel\[i\] is smaller than the median ratio + nmad\[i\] times the median
#' absolute deviation of the ratios). Default is 4, for all channel pairs.
#' @param ... additional parameters passed to CytoPipeline::singletGate()
#'
#' @return a flowCore::flowFrame with removed doublets events from the input
#' @export
#'
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
#'     
#' ff_c <-
#'     compensateFromMatrix(ff_m,
#'                          matrixSource = "fcs")        
#' 
#' ff_s <-
#'     removeDoubletsCytoPipeline(ff_c,
#'                                areaChannels = c("FSC-A", "SSC-A"),
#'                                heightChannels = c("FSC-H", "SSC-H"),
#'                                nmads = c(3, 5))
#'                             
removeDoubletsCytoPipeline <- function(ff,
                                       areaChannels,
                                       heightChannels,
                                       nmads,
                                       ...) {
    # if not present already, add a column with Cell ID
    ff <- appendCellID(ff)

    # validate common scatter channel parameters
    nScatterFilters <- length(areaChannels)
    if (nScatterFilters < 1 || nScatterFilters > 2) {
        stop(
            "nb of scatter channels for doublets removal ",
            "should be either 1 or 2!"
        )
    }
    if (length(heightChannels) != nScatterFilters) {
        stop(
            "inconsistency between length of area ",
            "and height channel vectors!"
        )
    }

    if (length(nmads) != nScatterFilters) {
        stop("inconsistency between length of area channel and nMAD vectors!")
    }
    for (i in seq_len(nScatterFilters)) {
        currentSingletGate <-
            singletsGate(ff,
                filterId = paste0(
                    "Singlets_",
                    areaChannels[i]
                ),
                channel1 = areaChannels[i],
                channel2 = heightChannels[i],
                nmad = nmads[i]
            )


        if (i == 1) {
            singletGateCombined <- currentSingletGate
        } else {
            singletGateCombined <- singletGateCombined & currentSingletGate
        }
    }

    fltSinglet <- flowCore::filter(ff, singletGateCombined)

    ff <- flowCore::Subset(ff, fltSinglet)
    #ff <- ff[fltSinglet@subSet, ]

    return(ff)
}


### FUNCTIONS FOR DEBRIS REMOVAL ###

#' @title remove debris from a flowFrame using manual gating
#' @description remove debris from a flowFrame, using manual gating in the
#' FSC-A, SSC-A 2D representation. The function internally uses
#' flowCore::polygonGate()
#' @param ff a flowCore::flowFrame
#' @param FSCChannel a character containing the exact name of the forward
#' scatter channel
#' @param SSCChannel a character containing the exact name of the side scatter
#' channel
#' @param gateData a numerical vector containing the polygon gate coordinates
#' first the `FSCChannel` channel coordinates   
#' of each points of the polygon gate,   
#' then the `SSCChannel` channel coordinates of each points.
#' @param ... additional parameters passed to flowCore::polygonGate()
#'
#' @return a flowCore::flowFrame with removed debris events from the input
#' @export
#' 
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
#'     
#' ff_c <-
#'     compensateFromMatrix(ff_m,
#'                          matrixSource = "fcs")        
#' 
#' 
#' remDebrisGateData <- c(73615, 110174, 213000, 201000, 126000,
#'                        47679, 260500, 260500, 113000, 35000)
#' 
#' ff_cells <-
#'     removeDebrisManualGate(ff_c,
#'                            FSCChannel = "FSC-A",
#'                            SSCChannel = "SSC-A",
#'                            gateData = remDebrisGateData)
#' 
#'
removeDebrisManualGate <- function(ff,
                                   FSCChannel,
                                   SSCChannel,
                                   gateData,
                                   ...) {
    # if not present already, add a column with Cell ID
    ff <- appendCellID(ff)

    cellsGateMatrix <- matrix(
        data = gateData, ncol = 2,
        dimnames = list(c(), c(FSCChannel, SSCChannel))
    )

    cellsGate <- flowCore::polygonGate(
        filterId = "Cells",
        .gate = cellsGateMatrix,
        ...
    )
    selectedCells <- flowCore::filter(ff, cellsGate)

    ff <- flowCore::Subset(ff, selectedCells)
    #ff <- ff[selectedCells@subSet, ]
}




### FUNCTIONS FOR DEAD CELLS REMOVAL ###

#' @title remove dead cells from a flowFrame using manual gating
#' @description remove dead cells from a flowFrame, using manual gating in the
#' FSC-A, '(a)Live/Dead' 2D representation. The function uses
#' flowCore::polygonGate()
#' @param ff a flowCore::flowFrame
#' @param preTransform boolean, if TRUE: the transList list of scale transforms
#' will be applied first on the LD channel.
#' @param transList applied in conjunction with preTransform == TRUE
#' @param FSCChannel a character containing the exact name of the forward
#' scatter channel
#' @param LDMarker a character containing the exact name of the marker
#' corresponding to (a)Live/Dead channel, or the Live/Dead channel name itself
#' @param gateData a numerical vector containing the polygon gate coordinates
#' first the `FSCChannel` channel coordinates    
#' of each points of the polygon gate,   
#' then the LD channel coordinates of each points   
#' (prior to scale transform)
#' @param ... additional parameters passed to flowCore::polygonGate()
#'
#' @return a flowCore::flowFrame with removed dead cells from the input
#' @export
#'
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
#'     
#' ff_c <-
#'     compensateFromMatrix(ff_m,
#'                          matrixSource = "fcs")    
#'                          
#' remDeadCellsGateData <- c(0, 0, 250000, 250000,
#'                           0, 650, 650, 0)  
#'
#' ff_lcells <-
#'     removeDeadCellsManualGate(ff_c,
#'                               FSCChannel = "FSC-A",
#'                               LDMarker = "L/D Aqua - Viability",
#'                               gateData = remDeadCellsGateData)
#'    
removeDeadCellsManualGate <- function(ff,
                                      preTransform = FALSE,
                                      transList = NULL,
                                      FSCChannel,
                                      LDMarker,
                                      gateData,
                                      ...) {
    # if not present already, add a column with Cell ID
    ff <- appendCellID(ff)

    if (preTransform) {
        if (is.null(transList)) {
            stop(
                "tranformation list needs to be provided ",
                "if preTransform = TRUE!"
            )
        }
        ffIn <- flowCore::transform(ff, transList)
    } else {
        ffIn <- ff
    }

    
    if (LDMarker %in% flowCore::colnames(ff)) {
        LDChannel <- LDMarker
    } else {
        LDChannel <- getChannelNamesFromMarkers(ffIn, markers = LDMarker)
    }

    liveGateMatrix <- matrix(
        data = gateData, ncol = 2,
        dimnames = list(c(), c(
            FSCChannel,
            LDChannel
        ))
    )

    liveGate <- flowCore::polygonGate(
        filterId = "Live_Cells",
        .gate = liveGateMatrix
    )



    selectedLive <- flowCore::filter(ffIn, liveGate)

    # note we take ff and not ffIn (no transfo)
    ff <- flowCore::Subset(ff, selectedLive)
    #ff <- ff[selectedLive@subSet, ] 
}


### FUNCTIONS for Quality Control ###

#' @title perform QC with flowAI
#' @description this function is a wrapper around flowAI::flow_auto_qc()
#' function.
#' It also pre-selects the channels to be handled (=> all signal channels)
#' @param ff a flowCore::flowFrame
#' @param preTransform if TRUE, apply the transList scale transform prior to
#' running the gating algorithm
#' @param transList applied in conjunction with preTransform
#' @param outputDiagnostic if TRUE, stores diagnostic files generated by
#' flowAI in outputDir directory
#' @param outputDir used in conjunction with outputDiagnostic
#' @param ... additional parameters passed to flowAI::flow_auto_qc()
#'
#' @return a flowCore::flowFrame with removed low quality events from the input
#' @export
#'
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' suppressWarnings(ff_QualityControl <- 
#'     qualityControlFlowAI(fsRaw[[2]],
#'                          remove_from = "all", # all default
#'                          second_fractionFR = 0.1,
#'                          deviationFR = "MAD",
#'                          alphaFR = 0.01,
#'                          decompFR = TRUE,
#'                          outlier_binsFS = FALSE,
#'                          pen_valueFS = 500,
#'                          max_cptFS = 3,
#'                          sideFM = "both",
#'                          neg_valuesFM = 1))
#' 
qualityControlFlowAI <- function(ff,
                                 preTransform = FALSE,
                                 transList = NULL,
                                 outputDiagnostic = FALSE,
                                 outputDir = NULL,
                                 ...) {
    # if not present already, add a column with Cell ID
    ff <- appendCellID(ff)

    if (preTransform) {
        if (is.null(transList)) {
            stop(
                "tranformation list needs to be provided ",
                "if preTransform = TRUE!"
            )
        }
        ffIn <- flowCore::transform(ff, transList)
    } else {
        ffIn <- ff
    }

    channel2Exclude <-
        flowCore::colnames(ffIn)[!areSignalCols(ffIn)]
    message("Applying flowAI method...")
    if (outputDiagnostic) {
        html_report <- "_QC"
        mini_report <- "QCmini"
        if (!is.null(outputDir)) {
            folder_results <- outputDir
        } else {
            folder_results <- "resultsQC"
        }
    } else {
        html_report <- FALSE
        mini_report <- FALSE
        folder_results <- FALSE
    }

    badEventIDs <-
        flowAI::flow_auto_qc(
            fcsfiles = ffIn,
            output = 3,
            timeCh = NULL,
            timestep = NULL,
            ChExcludeFS = channel2Exclude,
            ChExcludeFM = channel2Exclude,
            html_report = html_report,
            mini_report = mini_report,
            fcs_QC = FALSE,
            fcs_highQ = FALSE,
            fcs_lowQ = FALSE,
            folder_results = folder_results,
            ...
        )

    goodEvents <- !(seq_len(flowCore::nrow(ffIn)) %in% badEventIDs[[1]])
    ff <- ff[goodEvents, ] # note we take ff and not ffIn (no transfo)

    return(ff)
}

#' @title perform QC with PeacoQC
#' @description this function is a wrapper around PeacoQC::PeacoQC()
#' function.
#' It also pre-selects the channels to be handled (=> all signal channels)
#' @param ff a flowCore::flowFrame
#' @param preTransform if TRUE, apply the transList scale transform prior to
#' running the gating algorithm
#' @param transList applied in conjunction with preTransform
#' @param outputDiagnostic if TRUE, stores diagnostic files generated by
#' PeacoQC in outputDir directory
#' @param outputDir used in conjunction with outputDiagnostic
#' @param ... additional parameters passed to PeacoQC::PeacoQC()
#'
#' @return a flowCore::flowFrame with removed low quality events from the input
#' @export
#'
#' @examples
#'
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' sampleFiles <-
#'     file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#' 
#' truncateMaxRange <- FALSE
#' minLimit <- NULL
#' 
#' # create flowCore::flowSet with all samples of a dataset
#' fsRaw <- readSampleFiles(
#'     sampleFiles = sampleFiles,
#'     whichSamples = "all",
#'     truncate_max_range = truncateMaxRange,
#'     min.limit = minLimit)
#' 
#' suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
#'     
#' ff_c <-
#'     compensateFromMatrix(ff_m,
#'                          matrixSource = "fcs")        
#'
#' transList <- 
#'     estimateScaleTransforms(        
#'         ff = ff_c,
#'         fluoMethod = "estimateLogicle",
#'         scatterMethod = "linear",
#'         scatterRefMarker = "BV785 - CD3")
#'
#'
#' ff_QualityControl <- suppressWarnings(
#'     qualityControlPeacoQC(
#'         ff_c,
#'         preTransform = TRUE,
#'         transList = transList,
#'         min_cells = 150,
#'         max_bins = 500,
#'         MAD = 6,
#'         IT_limit = 0.55,
#'         force_IT = 150, 
#'         peak_removal = (1/3),
#'         min_nr_bins_peakdetection = 10))
#'         
qualityControlPeacoQC <- function(ff,
                                  preTransform = FALSE,
                                  transList = NULL,
                                  outputDiagnostic = FALSE,
                                  outputDir = NULL,
                                  ...) {
    # if not present already, add a column with Cell ID
    ff <- appendCellID(ff)

    if (preTransform) {
        if (is.null(transList)) {
            stop(
                "tranformation list needs to be provided ",
                "if preTransform = TRUE!"
            )
        }
        ffIn <- flowCore::transform(ff, transList)
    } else {
        ffIn <- ff
    }

    # qualityControl with PeacoQC
    message("Applying PeacoQC method...")
    channel4QualityControl <-
        flowCore::colnames(ffIn)[areSignalCols(ffIn)]
    if (outputDiagnostic) {
        plot <- TRUE
        report <- TRUE
        if (!is.null(outputDir)) {
            output_directory <- outputDir
        } else {
            output_directory <- "."
        }
    } else {
        plot <- FALSE
        report <- FALSE
        output_directory <- NULL # not used
    }

    res <- PeacoQC::PeacoQC(
        ff = ffIn,
        channels = channel4QualityControl,
        report = report,
        plot = plot,
        save_fcs = FALSE,
        output_directory = output_directory,
        ...
    )

    ff <- ff[res$GoodCells, ] # note we take ff and not ffIn (no transfo)
    return(ff)
}


##' @title read RDS object 
##' @description wrapper around readRDS, which discards any additional 
##' parameters passed in (...)
##' @param RDSFile a RDS file containing a R object
##' object
##' @param ... other arguments (not used)
##' @return the read R object
##' @export
##' 
##' @examples
##' 
##' data(OMIP021Samples)
##' 
##' transListPath <- file.path(system.file("extdata", 
##'                                         package = "CytoPipeline"),
##'                            "OMIP021_TransList.rds") 
##' 
##' transList <- readRDSObject(transListPath)
##' 
##' ff_c <- compensateFromMatrix(OMIP021Samples[[1]],
##'                              matrixSource = "fcs")  
##' 
##' ff_t <- applyScaleTransforms(ff_c, transList = transList)
##' 
readRDSObject <- function(RDSFile, ...) {
    if (!file.exists(RDSFile)) {
        stop("file ", RDSFile, " does not exist")
    }
    obj <- readRDS(file = RDSFile)
}

##' @title apply scale transforms
##' @description wrapper around flowCore::transform() that discards any 
##' additional parameter passed in (...)
##' Additionally, some checks regarding channels correspondance is done:  
##' if `transList` contains transformations for channels that are not present 
##' in `x`, then these transformations are first removed.
##' @param x a flowCore::flowSet or a flowCore::flowFrame
##' @param transList a flowCore::transformList
##' @param verbose if TRUE, send a message per flowFrame transformed
##' @param ... other arguments (not used)
##' @return the transformed flowFrame
##' @export
##' 
##' @examples
##' 
##' data(OMIP021Samples)
##' 
##' transListPath <- file.path(system.file("extdata", 
##'                                        package = "CytoPipeline"),
##'                            "OMIP021_TransList.rds") 
##' 
##' transList <- readRDSObject(transListPath)
##' 
##' ff_c <- compensateFromMatrix(OMIP021Samples[[1]],
##'                              matrixSource = "fcs")  
##' 
##' ff_t <- applyScaleTransforms(ff_c, transList = transList)
##' 
applyScaleTransforms <- function(x, transList, verbose = FALSE, ...) {
    
    if (!inherits(transList, "transformList")) {
        stop("transList should inherit from transformList class")
    }
    
    if (!inherits(x, "flowFrame") && !inherits(x, "flowSet")) {
        stop("x type not recognized, should be a flowFrame or a flowSet")
    }
    
    # check on channels/markers present in transList
    signalChannels <- flowCore::colnames(x)[areSignalCols(x)]
    transfoChannels <- names(transList@transforms)
    commonChannels <- intersect(signalChannels, transfoChannels)
    
    transList@transforms <- transList@transforms[commonChannels]
    
    if (inherits(x, "flowFrame")) {
        if (verbose) {
            message("scale transforming ff: ", 
                    flowCore::identifier(x))
        }
        ret <- flowCore::transform(x, transList)
    } else if (inherits(x, "flowSet")) {
        ret <- flowCore::fsApply(
            x,
            FUN = function(ff){
                if (verbose) {
                    message("scale transforming ff: ", 
                            flowCore::identifier(ff))
                }
                flowCore::transform(ff, transList)
            })
    } else {
        stop("x should be a flowCore::flowFrame or a flowCore::flowSet")
    }
    
    
    
    return(ret)
}

# CytoPipeline - Copyright (C) <2022-2025>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

# obtain OMIP021UTSamples, light-weight version used specifically for these 
# unit tests
path <- system.file("scripts",
                    package = "CytoPipeline"
)

source(file.path(path,"MakeOMIP021UTSamples.R"))


test_that("estimateScaleTransforms work", {
    compMatrix <- flowCore::spillover(OMIP021UTSamples[[1]])$SPILL
    ff_c <- runCompensation(OMIP021UTSamples[[1]], spillover = compMatrix)

    transList <-
        suppressMessages(estimateScaleTransforms(
            ff = ff_c,
            fluoMethod = "estimateLogicle",
            scatterMethod = "linearQuantile",
            scatterRefMarker = "BV785 - CD3"
        ))

    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList.rds")
    refTransList <- readRDS(transListPath)

    # saveRDS(transList, transListPath)

    refFF <- flowCore::transform(ff_c, refTransList)
    thisFF <- flowCore::transform(ff_c, transList)

    expect_equal(
        flowCore::exprs(thisFF),
        flowCore::exprs(refFF)
    )

    transList2 <- suppressMessages(estimateScaleTransforms(
        ff = ff_c,
        fluoMethod = "estimateLogicle",
        scatterMethod = "linearQuantile",
        scatterRefMarker = "BV785 - CD3",
        specificScatterChannels = c("SSC-A", "SSC-H")
    ))

    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList2.rds")

    refTransList2 <- readRDS(transListPath)

    # saveRDS(transList2, transListPath)

    refFF2 <- flowCore::transform(ff_c, refTransList2)
    thisFF2 <- flowCore::transform(ff_c, transList2)

    expect_equal(
        flowCore::exprs(thisFF2),
        flowCore::exprs(refFF2)
    )

    transList3 <- suppressMessages(estimateScaleTransforms(
        ff = ff_c,
        fluoMethod = "estimateLogicle",
        scatterMethod = "none",
        scatterRefMarker = "BV785 - CD3"
    ))

    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList3.rds")

    refTransList3 <- readRDS(transListPath)

    # saveRDS(transList3, transListPath)

    refFF3 <- flowCore::transform(ff_c, refTransList3)
    thisFF3 <- flowCore::transform(ff_c, transList3)

    expect_equal(
        flowCore::exprs(thisFF3),
        flowCore::exprs(refFF3)
    )
})


test_that("readSampleFiles with no sample files gives an error", {
    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    sampleFiles <-
        file.path(rawDataDir, list.files(rawDataDir, pattern = "NoGood"))

    expect_error(
        readSampleFiles(
            sampleFiles = sampleFiles,
            whichSamples = "all",
            truncate_max_range = truncateMaxRange,
            min.limit = minLimit
        ), regexp = "no sample files to read")
})



test_that("readSampleFiles works", {
    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    sampleFiles <-
        file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

    truncateMaxRange <- FALSE
    minLimit <- NULL

    fs_raw <-
        flowCore::read.flowSet(sampleFiles,
            truncate_max_range = truncateMaxRange,
            min.limit = minLimit
        )
    fs_raw <- flowCore::fsApply(fs_raw, FUN = appendCellID)

    res <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = "all",
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit
    )
    expect_equal(
        flowCore::exprs(res[[1]]),
        flowCore::exprs(fs_raw[[1]])
    )
    expect_equal(
        flowCore::exprs(res[[2]]),
        flowCore::exprs(fs_raw[[2]])
    )

    res2 <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = 2,
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit
    )

    expect_equal(
        flowCore::exprs(res2),
        flowCore::exprs(fs_raw[[2]])
    )
})

test_that("readSampleFiles with random samples works", {
    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    sampleFiles <-
        file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

    truncateMaxRange <- FALSE
    minLimit <- NULL

    fs_raw <-
        flowCore::read.flowSet(sampleFiles,
                               truncate_max_range = truncateMaxRange,
                               min.limit = minLimit
        )
    fs_raw <- flowCore::fsApply(fs_raw, FUN = appendCellID)

    seed <- 2
    nSamples <- 1

    res <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = "random",
        nSamples = nSamples,
        seed = seed,
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit
    )

    expect_equal(
        flowCore::exprs(res),
        flowCore::exprs(fs_raw[[1]])
    )

    res <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = "random",
        nSamples = 3,
        seed = seed,
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit
    )

    expect_equal(
        flowCore::exprs(res[[1]]),
        flowCore::exprs(fs_raw[[1]])
    )

    expect_equal(
        flowCore::exprs(res[[2]]),
        flowCore::exprs(fs_raw[[2]])
    )


    expect_error(readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = "random",
        nSamples = 0,
        seed = seed,
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit),
        regexp = "should be a numeric >= 1")
})

test_that("readSampleFiles with post-processing works", {
    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    sampleFiles <-
        file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
    
    channelMarkerFile <-
        system.file("extdata/ChannelMarkerUsed.csv", package = "CytoPipeline")
    
    truncateMaxRange <- FALSE
    minLimit <- NULL
    
    fs_raw <-
        flowCore::read.flowSet(sampleFiles,
                               truncate_max_range = truncateMaxRange,
                               min.limit = minLimit
        )
    fs_raw <- flowCore::fsApply(fs_raw, FUN = appendCellID)
    
    res <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = "all",
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit,
        channelMarkerFile = channelMarkerFile)
    
    expectedMarkerNames <- 
        flowCore::pData(flowCore::parameters(fs_raw[[1]]))$desc
    expectedMarkerNames[6] <- "Viability"
    expectedMarkerNames <- expectedMarkerNames[-5]
    
    expect_equal(
        flowCore::pData(flowCore::parameters(res[[1]]))$desc,
        expectedMarkerNames)
    expect_equal(
        flowCore::pData(flowCore::parameters(res[[2]]))$desc,
        expectedMarkerNames)
    
    res2 <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = 2,
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit,
        channelMarkerFile = channelMarkerFile)
    
    expect_equal(
        flowCore::pData(flowCore::parameters(res2))$desc,
        expectedMarkerNames
    )
})

# test_that("readSampleFiles with phenoData works", {
#     rawDataDir <-
#         system.file("extdata", package = "CytoPipeline")
#     sampleFiles <-
#         file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#     
#     truncateMaxRange <- FALSE
#     minLimit <- NULL
#     
#     phenoData <- data.frame(Donor = c(1, 2),
#                             Group = c("G1", "G2"),
#                             row.names = basename(sampleFiles))
#     
#     res <- readSampleFiles(
#         sampleFiles = sampleFiles,
#         whichSamples = "all",
#         truncate_max_range = truncateMaxRange,
#         min.limit = minLimit,
#         pData = phenoData)
#     
#     retPData <- flowCore::pData(res)
#     expectedPData <- phenoData
#     expectedPData$name <- c("Donor1.fcs", "Donor2.fcs")
#     
#     expect_equal(retPData, expectedPData)
#     
#     expect_equal(
#         flowCore::keyword(res[[1]], "CytoPipeline_Donor")[["CytoPipeline_Donor"]],
#         1)
#     
#     expect_equal(
#         flowCore::keyword(res[[1]], "CytoPipeline_Group")[["CytoPipeline_Group"]],
#         "G1")
#     
#     expect_equal(
#         flowCore::keyword(res[[2]], "CytoPipeline_Donor")[["CytoPipeline_Donor"]],
#         2)
#     
#     expect_equal(
#         flowCore::keyword(res[[2]], "CytoPipeline_Group")[["CytoPipeline_Group"]],
#         "G2")
#     
#     res2 <- readSampleFiles(
#         sampleFiles = sampleFiles,
#         whichSamples = 2,
#         truncate_max_range = truncateMaxRange,
#         min.limit = minLimit,
#         pData = phenoData)
#     
#     expect_equal(
#         flowCore::keyword(res2, "CytoPipeline_Donor")[["CytoPipeline_Donor"]],
#         2)
#     
#     expect_equal(
#         flowCore::keyword(res2, "CytoPipeline_Group")[["CytoPipeline_Group"]],
#         "G2")
# })

test_that("removeMarginsPeacoQC works", {
    fs_raw <- OMIP021UTSamples

    ff_m <-
        suppressWarnings(removeMarginsPeacoQC(x = fs_raw[[1]]))

    ref_ff_m <- readRDS(test_path("fixtures", "ff_m.rds"))

    # saveRDS(ff_m, test_path("fixtures", "ff_m.rds"))

    expect_equal(
        flowCore::exprs(ff_m),
        flowCore::exprs(ref_ff_m)
    )

    channelSpecifications <-
        list( "450/50Violet-A" = c(0, 262143),
              "PECy5 - CD28" = c(0, 20000))

    ff_m2 <-
        suppressWarnings(removeMarginsPeacoQC(x = fs_raw[[1]],
                         channelSpecifications = channelSpecifications))

    ref_ff_m2 <- readRDS(test_path("fixtures", "ff_m2.rds"))

    # saveRDS(ff_m2, test_path("fixtures", "ff_m2.rds"))

    expect_equal(
        flowCore::exprs(ff_m2),
        flowCore::exprs(ref_ff_m2)
    )

    channelSpecifications <-
        list( "AllFluoChannels" = c(0, 40000))

    ff_m3 <-
        suppressWarnings(removeMarginsPeacoQC(x = fs_raw[[1]],
                         channelSpecifications = channelSpecifications))

    ref_ff_m3 <- readRDS(test_path("fixtures", "ff_m3.rds"))

    # saveRDS(ff_m3, test_path("fixtures", "ff_m3.rds"))

    expect_equal(
        flowCore::exprs(ff_m3),
        flowCore::exprs(ref_ff_m3)
    )

    channelSpecifications <-
        list( "AllFluoChannels" = c(0, 40000),
              "450/50Violet-A" = c(0, 262143),
              "PECy5 - CD28" = c(0, 20000))

    ff_m4 <-
        suppressWarnings(removeMarginsPeacoQC(x = fs_raw[[1]],
                         channelSpecifications = channelSpecifications))

    ref_ff_m4 <- readRDS(test_path("fixtures", "ff_m4.rds"))

    # saveRDS(ff_m4, test_path("fixtures", "ff_m4.rds"))

    expect_equal(
        flowCore::exprs(ff_m4),
        flowCore::exprs(ref_ff_m4)
    )
})


test_that("compensateFromMatrix works", {
    ref_ff_m <- readRDS(test_path("fixtures", "ff_m.rds"))

    ff_c <-
        compensateFromMatrix(ref_ff_m,
            matrixSource = "fcs"
        )

    ref_ff_c <- readRDS(test_path("fixtures", "ff_c.rds"))

    # saveRDS(ff_c, test_path("fixtures", "ff_c.rds"))

    expect_equal(
        flowCore::exprs(ff_c),
        flowCore::exprs(ref_ff_c)
    )
})


# test_that("compensateFromMatrix works with pData", {
#     rawDataDir <-
#         system.file("extdata", package = "CytoPipeline")
#     sampleFiles <-
#         file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))
#
#     truncateMaxRange <- FALSE
#     minLimit <- NULL
#
#     phenoData <- data.frame(Donor = c(1, 2),
#                             Group = c("G1", "G2"),
#                             row.names = basename(sampleFiles))
#
#     fs <- readSampleFiles(
#         sampleFiles = sampleFiles,
#         whichSamples = "all",
#         truncate_max_range = truncateMaxRange,
#         min.limit = minLimit,
#         pData = phenoData)
#
#     pDataVar <- "Group"
#     pDataPathMapping <-
#         list(G1 = "compMatForG1.csv",
#              G2 = "compMatForG2.csv")
#
#     expect_error(
#         compensateFromMatrix(fs[[1]],
#                              matrixSource = "pData",
#                              pDataVar = pDataVar,
#                              pDataPathMapping = pDataPathMapping),
#         regexp = "compMatForG1.csv")
#
#     expect_error(
#         compensateFromMatrix(fs[[2]],
#                              matrixSource = "pData",
#                              pDataVar = pDataVar,
#                              pDataPathMapping = pDataPathMapping),
#         regexp = "compMatForG2.csv")
#
#     expect_error(
#         compensateFromMatrix(fs,
#                              matrixSource = "pData",
#                              pDataVar = pDataVar,
#                              pDataPathMapping = pDataPathMapping),
#         regexp = "compMatForG1.csv")
#
#     pDataPathMapping <-
#         list(W1 = "compMatForG1.csv",
#              G2 = "compMatForG2.csv")
#
#     expect_error(
#         compensateFromMatrix(fs[[1]],
#                              matrixSource = "pData",
#                              pDataVar = pDataVar,
#                              pDataPathMapping = pDataPathMapping),
#         regexp = "No mapping found for variable")
#
#     pDataVar <- "NonExistentVar"
#     expect_error(
#         compensateFromMatrix(fs[[1]],
#                              matrixSource = "pData",
#                              pDataVar = pDataVar,
#                              pDataPathMapping = pDataPathMapping),
#         regexp = "not found in flowFrame")
# })

test_that("compensateFromMatrix works with imported matrices", {
    rawDataDir <-
        system.file("extdata", package = "CytoPipeline")
    sampleFiles <-
        file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

    truncateMaxRange <- FALSE
    minLimit <- NULL

    fs <- readSampleFiles(
        sampleFiles = sampleFiles,
        whichSamples = "all",
        truncate_max_range = truncateMaxRange,
        min.limit = minLimit)

    # here we use errors to check that the code is properly handled
    # the error message should correspond to the regexp, not to another type
    # of error...

    expect_error(
        compensateFromMatrix(fs[[1]],
                             matrixSource = "import",
                             matrixPath = "compMatForG1.csv"),
        regexp = "compMatForG1.csv")

    expect_error(
        compensateFromMatrix(fs[[2]],
                             matrixSource = "import",
                             matrixPath = "compMatForG2.csv"),
        regexp = "compMatForG2.csv")

    expect_error(
        compensateFromMatrix(fs,
                             matrixSource = "import",
                             matrixPath = "compMatForG1.csv"),
        regexp = "compMatForG1.csv")

    expect_error(
        compensateFromMatrix(fs,
                             matrixSource = "import",
                             matrixPath = c("compMatForG1.csv",
                                            "compMatForG2.csv")),
        regexp = "compMatForG1.csv")
})


test_that("removeDoubletsCytoPipeline works", {
    ref_ff_c <- readRDS(test_path("fixtures", "ff_c.rds"))

    ff_s3 <-
        removeDoubletsCytoPipeline(ref_ff_c,
            areaChannels = c("FSC-A", "SSC-A"),
            heightChannels = c("FSC-H", "SSC-H"),
            nmads = c(3, 5)
        )


    ref_ff_s3 <- readRDS(test_path("fixtures", "ff_s3.rds"))

    # saveRDS(ff_s3, test_path("fixtures", "ff_s3.rds"))

    expect_equal(
        flowCore::exprs(ff_s3),
        flowCore::exprs(ref_ff_s3)
    )
})


test_that("removeDebrisManualGate works", {
    ref_ff_s <- readRDS(test_path("fixtures", "ff_s3.rds"))

    remDebrisGateData <- c(73615, 110174, 213000, 201000, 126000,
                           47679, 260500, 260500, 113000, 35000)

    ff_cells <-
        removeDebrisManualGate(ref_ff_s,
                               FSCChannel = "FSC-A",
                               SSCChannel = "SSC-A",
                               gateData = remDebrisGateData
        )

    ref_ff_cells <- readRDS(test_path("fixtures", "ff_cells_manual_gate.rds"))

    # saveRDS(ff_cells, test_path("fixtures", "ff_cells_manual_gate.rds"))

    expect_equal(
        flowCore::exprs(ff_cells),
        flowCore::exprs(ref_ff_cells)
    )
})



test_that("removeDeadCellsManualGate works", {
    ref_ff_cells <- readRDS(test_path("fixtures", "ff_cells_manual_gate.rds"))

    remDeadCellsGateData <- c(0, 0, 250000, 250000,
                              0, 650, 650, 0)

    ff_lcells <-
        removeDeadCellsManualGate(ref_ff_cells,
                                  FSCChannel = "FSC-A",
                                  LDMarker = "L/D Aqua - Viability",
                                  gateData = remDeadCellsGateData)

    ref_ff_lcells <- readRDS(test_path("fixtures", "ff_lcells_manual_gate.rds"))

    # saveRDS(ff_lcells, test_path("fixtures", "ff_lcells_manual_gate.rds"))

    expect_equal(
        flowCore::exprs(ff_lcells),
        flowCore::exprs(ref_ff_lcells)
    )

    # same with channel name instead of marker for L/D
    ff_lcells2 <-
        removeDeadCellsManualGate(ref_ff_cells,
                                  FSCChannel = "FSC-A",
                                  LDMarker = "Comp-525/50Violet-A",
                                  gateData = remDeadCellsGateData)

    expect_equal(
        flowCore::exprs(ff_lcells2),
        flowCore::exprs(ref_ff_lcells)
    )

})

test_that("qualityControlPeacoQC", {
    ref_ff_lcells <- readRDS(test_path("fixtures", "ff_lcells_manual_gate.rds"))

    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList.rds")
    refTransList <- readRDS(transListPath)

    expect_error(suppressWarnings(qualityControlPeacoQC(
            ref_ff_lcells,
            preTransform = TRUE,
            transList = refTransList,
            min_cells = 30, # decreased from the 150 default
            max_bins = 500, # default
            MAD = 6, # default
            IT_limit = 0.55, # default
            force_IT = 150, # default
            peak_removal = (1 / 3), # default
            min_nr_bins_peakdetection = 10 # default
    )), regexp = "must be strictly positive and finite")


    # ref_ff_qualityControl <-
    #     readRDS(test_path("fixtures", "ff_QC_PeacoQC.rds"))
    #
    # # saveRDS(ff_QualityControl, test_path("fixtures", "ff_QC_PeacoQC.rds"))
    #
    # expect_equal(
    #     flowCore::exprs(ff_QualityControl),
    #     flowCore::exprs(ref_ff_qualityControl)
    # )
})




test_that("qualityControlFlowAI works", {
    fs_raw <- OMIP021UTSamples

    ff_QualityControl <- suppressWarnings(
        qualityControlFlowAI(fs_raw[[1]],
            remove_from = "all", # all default
            second_fractionFR = 0.1,
            deviationFR = "MAD",
            alphaFR = 0.01,
            decompFR = TRUE,
            outlier_binsFS = FALSE,
            pen_valueFS = 500,
            max_cptFS = 3,
            sideFM = "both",
            neg_valuesFM = 1
        )
    )

    ref_ff_qualityControl_flowAI <-
        readRDS(test_path("fixtures", "ff_QC_flowAI.rds"))

    # saveRDS(ff_QualityControl, test_path("fixtures", "ff_QC_flowAI.rds"))

    expect_equal(
        flowCore::exprs(ff_QualityControl),
        flowCore::exprs(ref_ff_qualityControl_flowAI)
    )
})


test_that("readRDSObject works", {
    expect_error(readRDSObject("dummyPath.rds"),
                 regexp = "file dummyPath.rds does not exist")
    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList.rds")

    obj <- readRDSObject(transListPath)

    refTransList <- readRDS(transListPath)

    expect_identical(obj, refTransList)
})

test_that("applyScaleTransform works", {
    transListPath <- file.path(system.file("extdata",
                                           package = "CytoPipeline"),
                               "OMIP021_TransList.rds")

    ff_c <- readRDS(test_path("fixtures", "ff_c.rds"))
    transList <- readRDS(transListPath)

    refFF <- flowCore::transform(ff_c, transList)
    thisFF <- applyScaleTransforms(ff_c, transList)

    expect_equal(
        flowCore::exprs(thisFF),
        flowCore::exprs(refFF)
    )

    # with flowSets
    fs_c <- compensateFromMatrix(OMIP021Samples, matrixSource = "fcs")
    thisFS <- applyScaleTransforms(fs_c, transList)
    refFF1 <- flowCore::transform(fs_c[[1]], transList)
    refFF2 <- flowCore::transform(fs_c[[2]], transList)

    expect_equal(
        flowCore::exprs(thisFS[[1]]),
        flowCore::exprs(refFF1)
    )

    expect_equal(
        flowCore::exprs(thisFS[[2]]),
        flowCore::exprs(refFF2)
    )
})

# CytoPipeline - Copyright (C) <2022-2026>
# <Université catholique de Louvain (UCLouvain), Belgique>
#
#   Description and complete License: see LICENSE file.
#
# This program (CytoPipeline) is free software:
#   you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details (<http://www.gnu.org/licenses/>).

setClassUnion("DataFrameOrNull",
              members=c("data.frame", "NULL"))

#' @title CytoPipeline class
#'
#' @aliases CytoPipeline-class, CytoPipelineClass
#'
#' @name CytoPipeline-class
#'
#' @rdname CytoPipelineClass
#'
#' @description
#'
#' Class representing a flow cytometry pipeline, and composed of two processing
#' queues, i.e. lists of CytoProcessingStep objects :
#' - a list of CytoProcessingStep(s) for pre-calculation of scale
#' transformations per channel
#' - a list of CytoProcessingStep(s) for the pre-processing of flow frames
#'
#' @slot scaleTransformProcessingQueue A `list` of    
#' CytoProcessingStep objects containing the steps   
#' for obtaining the scale transformations per channel
#'
#' @slot flowFramesPreProcessingQueue A `list` of     
#' CytoProcessingStep objects containing the steps   
#' for pre-processing of the samples flow frames
#'
#' @slot experimentName A `character` containing     
#' the experiment (run) name
#'
#' @slot sampleFiles A `character` vector storing   
#' all fcs files to be run into the pipeline
#'
#' @slot pData An optional `data.frame` containing
#' additional information for each sample file. 
#' The `pData` raw names should correspond to the sample files
#' (using full paths or base paths).  
#' If the `pData` contains a columns with name 'displayName', this will have
#' an impact in the `sampleDisplayNames()` function, i.e. sample display
#' names will be the one mentioned in `pData`, instead of typically 
#' base file names (or larger paths if base file names are not unique)
#'
#' @exportClass CytoPipeline
#' 
#' @return nothing
#' 
#' @examples
#' 
#' ### *** EXAMPLE 1: building CytoPipeline step by step *** ###
#' 
#' rawDataDir <-
#'     system.file("extdata", package = "CytoPipeline")
#' experimentName <- "OMIP021_PeacoQC"
#' sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
#'                                              pattern = "Donor"))
#'                                              
#' outputDir <- base::tempdir()
#' 
#' # main parameters : sample files and output files
#' pipL <- CytoPipeline(experimentName = experimentName,
#'                      sampleFiles = sampleFiles)
#' 
#' ### SCALE TRANSFORMATION STEPS ###
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "flowframe_read",
#'                           FUN = "readSampleFiles",
#'                           ARGS = list(
#'                               whichSamples = "all",
#'                               truncate_max_range = FALSE,
#'                               min.limit = NULL
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "remove_margins",
#'                           FUN = "removeMarginsPeacoQC",
#'                           ARGS = list()
#'                      )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "compensate",
#'                           FUN = "compensateFromMatrix",
#'                           ARGS = list(matrixSource = "fcs")
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "flowframe_aggregate",
#'                           FUN = "aggregateAndSample",
#'                           ARGS = list(
#'                               nTotalEvents = 10000,
#'                               seed = 0
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "scale transform",
#'                       CytoProcessingStep(
#'                           name = "scale_transform_estimate",
#'                           FUN = "estimateScaleTransforms",
#'                           ARGS = list(
#'                               fluoMethod = "estimateLogicle",
#'                               scatterMethod = "linear",
#'                               scatterRefMarker = "BV785 - CD3"
#'                           )
#'                       )
#'     )
#' 
#' ### PRE-PROCESSING STEPS ###
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "flowframe_read",
#'                           FUN = "readSampleFiles",
#'                           ARGS = list(
#'                               truncate_max_range = FALSE,
#'                               min.limit = NULL
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "remove_margins",
#'                           FUN = "removeMarginsPeacoQC",
#'                           ARGS = list()
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "compensate",
#'                           FUN = "compensateFromMatrix",
#'                           ARGS = list(matrixSource = "fcs")
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(
#'         pipL,
#'         whichQueue = "pre-processing",
#'         CytoProcessingStep(
#'             name = "remove_debris",
#'             FUN = "removeDebrisManualGate",
#'             ARGS = list(
#'                 FSCChannel = "FSC-A",
#'                 SSCChannel = "SSC-A",
#'                 gateData =  c(73615, 110174, 213000, 201000, 126000,
#'                               47679, 260500, 260500, 113000, 35000)))
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "remove_dead_cells",
#'                           FUN = "removeDeadCellsManualGate",
#'                           ARGS = list(
#'                               FSCChannel = "FSC-A",
#'                               LDMarker = "L/D Aqua - Viability",
#'                               gateData = c(0, 0, 250000, 250000,
#'                                            0, 650, 650, 0)
#'                           )
#'                       )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(
#'         pipL,
#'         whichQueue = "pre-processing",
#'         CytoProcessingStep(
#'             name = "perform_QC",
#'             FUN = "qualityControlPeacoQC",
#'             ARGS = list(
#'                 preTransform = TRUE,
#'                 min_cells = 150, # default
#'                 max_bins = 500, # default
#'                 step = 500, # default,
#'                 MAD = 6, # default
#'                 IT_limit = 0.55, # default
#'                 force_IT = 150, # default
#'                 peak_removal = 0.3333, # default
#'                 min_nr_bins_peakdetection = 10 # default
#'             )
#'         )
#'     )
#' 
#' pipL <-
#'     addProcessingStep(pipL,
#'                       whichQueue = "pre-processing",
#'                       CytoProcessingStep(
#'                           name = "transform",
#'                           FUN = "applyScaleTransforms",
#'                           ARGS = list()
#'                       )
#'     )
#' 
#' ### *** EXAMPLE 2: building CytoPipeline from JSON file *** ###
#' 
#' jsonDir <- system.file("extdata", package = "CytoPipeline")
#' jsonPath <- file.path(jsonDir, "pipelineParams.json")
#' 
#' pipL2 <- CytoPipeline(jsonPath,
#'                       experimentName = experimentName,
#'                       sampleFiles = sampleFiles)
#' 
setClass("CytoPipeline",
    slots = c(
        experimentName = "character",
        scaleTransformProcessingQueue = "list",
        flowFramesPreProcessingQueue = "list",
        sampleFiles = "character",
        pData = "DataFrameOrNull"
    ),
    prototype = list(
        experimentName = "default_experiment",
        scaleTransformProcessingQueue = list(),
        flowFramesPreProcessingQueue = list(),
        sampleFiles = character(),
        pData = NULL
    )
)

setValidity("CytoPipeline", function(object) {
    msg <- .validPData(object)
    
    if (length(msg))
        return(msg)
    
    msg1 <- .validProcessingQueue(
        object@scaleTransformProcessingQueue,
        "scaleTransformProcessingQueue"
    )
    msg2 <- .validProcessingQueue(
        object@flowFramesPreProcessingQueue,
        "flowFramesPreProcessingQueue"
    )

    msg <- NULL
    if (length(msg1)) {
        msg <- msg1
        if (length(msg2)) {
            msg <- paste0(msg, " ; ", msg2)
        }
    } else {
        if (length(msg2)) {
            msg <- msg2
        }
    }
    if (length(msg)) {
        msg
    } else {
        TRUE
    }
})

#' @rdname CytoPipelineClass
#'
#' @param object a `CytoPipeline` object
#'
#' @importMethodsFrom methods show
#' @importFrom utils head
#' 
#'
setMethod(
    "show", "CytoPipeline",
    function(object) {
        cat(
            "Pipeline object for flow cytometry experiment:",
            object@experimentName, "\n"
        )
        if (length(object@sampleFiles) > 0) {
            cat(
                "Sample files:", length(object@sampleFiles),
                "sample file(s)\n"
            )
            sampleInfo <- data.frame(
                displayName = sampleDisplayNames(object),
                sampleFile = sampleFiles(object))
            cat("head(samples):\n")
            show(head(sampleInfo))
            
        } else {
            cat("No sample file\n")
        }
        if (!is.null(object@pData)) {
            cat("pheno data (head):\n")
            show(head(object@pData))
        } else {
            cat("No pheno data\n")
        }
        showProcessingSteps(object, whichQueue = "scale transform")
        showProcessingSteps(object, whichQueue = "pre-processing")
    }
)


setGeneric("CytoPipeline", function(object, ...) {
    standardGeneric("CytoPipeline")
})

#' @rdname CytoPipelineClass
#' @param experimentName the experiment name
#' @param sampleFiles the sample files
#' @param pData the pheno data (data.frame or NULL)
#' @export
#'
setMethod(
    "CytoPipeline", "missing",
    function(object,
             experimentName = "default_experiment",
             sampleFiles = character(),
             pData = NULL) {
        x <- methods::new("CytoPipeline",
            experimentName = experimentName,
            scaleTransformProcessingQueue = list(),
            flowFramesPreProcessingQueue = list(),
            sampleFiles = sampleFiles,
            pData = NULL
        )
        if (!is.null(pData)) {
            pData(x) <- pData
        }
        
        return(x)
    }
)

#' @rdname CytoPipelineClass
#'
#' @param object a `list()`
#' @param experimentName the experiment name
#' @param sampleFiles the sample files
#' @param pData the phenoData (data.frame or NULL)
#'
#' @export
#'
setMethod(
    "CytoPipeline", "list",
    function(object,
             experimentName = "default_experiment",
             sampleFiles = character(),
             pData = NULL) {
        x <- methods::new("CytoPipeline",
            experimentName = experimentName,
            scaleTransformProcessingQueue = list(),
            flowFramesPreProcessingQueue = list(),
            sampleFiles = sampleFiles,
            pData = NULL # will be created later on
        )
        x <- .makeSlots(x, object)
        x <- .makeProcessingQueues(x, object)
        if (!is.null(pData)) {
            # with validation and possible sample order change
            pData(x) <- pData
        }
        return(x)
    }
)

#' @rdname CytoPipelineClass
#'
#' @param object a `character()` containing a JSON input
#' @param experimentName the experiment name
#' @param sampleFiles the sample files
#' @param pData the pheno Data (data.frame or NULL)
#'
#' @export
#'
setMethod(
    "CytoPipeline", "character",
    function(object,
             experimentName = "default_experiment",
             sampleFiles = character(),
             pData = NULL) {
        
        pipelineParams <- jsonlite::read_json(
            object,
            simplifyVector = TRUE,
            simplifyDataFrame = FALSE)
        
        x <- CytoPipeline(pipelineParams,
                          experimentName = experimentName,
                          sampleFiles = sampleFiles,
                          pData = pData)
        return(x)
    }
)

#' @rdname CytoPipelineClass
#'
#' @param x a `CytoPipeline` object
#' @param ... additional arguments (not used here)
#'
#' @return - for `as.list.CytoPipeline`: the obtained list
#'
#' @export
#'
as.list.CytoPipeline <- function(x, ...) {
    stopifnot(inherits(x, "CytoPipeline"))

    slots <- methods::slotNames(x)
    slots <- slots[!slots %in% c(
        "scaleTransformProcessingQueue",
        "flowFramesPreProcessingQueue"
    )]

    outputList <- list()

    for (sl in slots) {
        outputList[[sl]] <- methods::slot(x, sl)
    }

    # now handles both processing queue
    pQueue2List <- function(pQueue) {
        pList <- list()
        for (p in seq_along(pQueue)) {
            pList[[p]] <- as.list.CytoProcessingStep(pQueue[[p]])
        }
        pList
    }

    pList1 <- pQueue2List(x@scaleTransformProcessingQueue)
    outputList[["scaleTransformProcessingSteps"]] <- pList1

    pList2 <- pQueue2List(x@flowFramesPreProcessingQueue)
    outputList[["flowFramesPreProcessingSteps"]] <- pList2

    return(outputList)
}

##' @rdname CytoPipelineClass
##'
##' @param x a `CytoPipeline` object
##'
##' @export
##'
experimentName <- function(x) {
    stopifnot(inherits(x, "CytoPipeline"))
    return(x@experimentName)
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @param value the new value to be assigned
##' @export
##'
"experimentName<-" <- function(x, value) {
    stopifnot(inherits(x, "CytoPipeline"))
    x@experimentName <- value
    return(x)
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @export
##'
sampleFiles <- function(x) {
    stopifnot(inherits(x, "CytoPipeline"))
    return(x@sampleFiles)
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @param value the new value to be assigned
##' @export
##'
"sampleFiles<-" <- function(x, value) {
    stopifnot(inherits(x, "CytoPipeline"))
    if (!is.character(value)) {
        stop("cannot assign sampleFiles: not a character vector passed.")
    }    
    if (length(unique(value)) < length(value)){
        stop("cannot assign sampleFiles: character vector does not ",
             "contain unique values.")
    }
    x@sampleFiles <- value
    
    if (!is.null(pData(x))) {
        # maintain consistency between order of x@pData and order of x@sampleFile
        # x@pData is driving
        x@sampleFiles <- x@sampleFiles[order(match(
            x@sampleFiles, rownames(x@pData)))]
    }
    
    if (isTRUE(methods::validObject(x))) return(x)
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @export
##'
pData <- function(x) {
    stopifnot(inherits(x, "CytoPipeline"))
    return(x@pData)
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @param value the new value to be assigned. 
##' the `pData<-` setter is a bit more liberal than it used to be:
##' 1. It can accept new pData containing more rows than existing sample names 
##' (the corresponding subset of pData is taken).
##' 2. It can accept pData with row names pointing to either 
##' sample file full paths or base file names
##' 3. It can accept pData with no row names 
##' provided the number of rows correspond to the number of sample files. 
##' Row names are then set by default to sample file base names (if unique), 
##' or sample file full paths.
##' @export
##'
`pData<-` <- function(x, value) {
    stopifnot(inherits(x, "CytoPipeline"))
    stopifnot(inherits(value, "data.frame")) 
    if (length(x@sampleFiles) == 0) {
        stop(paste0("Cannot assign pData to empty CytoPipeline object ",
                    "(no sample files)"))
    }
    
    if (length(x@sampleFiles) != nrow(value)){
        # are full names of sample files used as pData row names ? 
        # (or) are base names of sample files used as pData row names ? 
        # if so, take the subset of pData that correspond to sample names
        if (all(x@sampleFiles %in% rownames(value))) {
            subset <- which(rownames(value) %in% x@sampleFiles)
        } else if (all(basename(x@sampleFiles) %in% rownames(value))){
            subset <- which(rownames(value) %in% basename(x@sampleFiles))
        } else {
            stop(paste0("Cannot assign pData to CytoPipeline object: ",
                        "mismatch between pData number of rows ",
                        "and number of samples, AND ",
                        "could not find all sample names in pData row names"))
        }
        message(paste0("pData row subset corresponding ",
                       "to existing sample names has been applied."))
        value <- value[subset,]
    }
    
    #if(is.null(rownames(value))){
    if(isTRUE(all.equal(rownames(value), as.character(1:nrow(value))))){
        # row names cannot be null, when not specified the data frame
        # contains by default numbers converted in strings
        
        # in that case, we automatically assign row names
        # if basename of sample files are unique => use base sample filename
        # otherwise, use full name (= full path)
        if(length(unique(basename(x@sampleFiles))) == length(x@sampleFiles)){
            rownames(value) <- basename(x@sampleFiles)
        } else {
            rownames(value) <- x@sampleFiles
        }
        message(paste0("pData row names has been set by default ",
                       "to sample names."))
    }
    
    x@pData <- value
    
    retMsg <- methods::validObject((x))
    
    if (isTRUE(retMsg)){
        x <- .alignSampleOrderWithPData(x)
    } else {
        stop(retMsg)
    }
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @param sampleFiles a character (e.g. sampleFileNames) or a numeric vector
##' (e.g. indices of sample files). If NULL, all samples will be displayed.
##' @return - for `sampleDisplayNames`: a character vector
##' of sample display names
##' @export
##'
sampleDisplayNames <- function(x, sampleFiles = NULL) {
    stopifnot(inherits(x, "CytoPipeline"))
    if (length(x@sampleFiles) == 0)
        return(NULL)
    
    sampleFileIndices <- NULL
    if (is.null(sampleFiles)) {
        sampleFileIndices <- seq_along(sampleFiles(x))
    } else if (is.numeric(sampleFiles)) {
        if (all(sampleFiles > 0) && 
            length(sampleFiles) <= length(sampleFiles(x))) {
            sampleFileIndices <- sampleFiles
        } else {
            stop("sampleFiles out of bounds")
        }
    } else {
        sampleFileIndices <-
            which(sampleFiles(x) %in% sampleFiles)
        if (length(sampleFileIndices) < length(sampleFiles)) {
            stop("could not find all sampleFiles in CytoPipeline")
        } 
    }
    
    if (!is.null(x@pData)) {
        if("displayname" %in% tolower(gsub("[[:punct:]]", 
                                           "",
                                           colnames(x@pData)))) {
            colIndex <- which(
                tolower(gsub("[[:punct:]]", 
                             "",
                             colnames(x@pData))) == "displayname")
            
            return(x@pData[sampleFileIndices, colIndex, drop = TRUE])
        } 
    }
    
    # display names not found in pData => build them based on sampleFiles
    .getMinimumUniqueFileNames(
        sampleFiles(x))[sampleFileIndices, drop = TRUE]
}

##' @rdname CytoPipelineClass
##' @param x a `CytoPipeline` object
##' @param displayName a character
##' @return - for `sampleNameFromDisplayName`: the sample name corresponding 
##' to the specified display name.
##' of sample display names
##' @export
##'
sampleNameFromDisplayName <- function(x, displayName) {
    stopifnot(inherits(x, "CytoPipeline"))
    displayNames <- sampleDisplayNames(x)
    index <- which(displayNames == displayName)
    if (length(index) == 0) {
        stop("could not find sample display name in CytoPipeline")
    }
    sampleFiles(x)[index]
}

.validProcessingQueue <- function(x, queueName) {
    msg <- NULL
    if (length(x) && !all(vapply(
        x, inherits, what = "CytoProcessingStep", FUN.VALUE = TRUE))) {
        msg <- paste0(
            "'",
            queueName,
            "' should only contain CytoProcessingStep objects."
        )
    }
    return(msg)
}

.validPData <- function(x) {
    msg <- NULL
    if (!is.null(x@pData)) {
        if (!inherits(x@pData, "data.frame")) {
            return("Non-null @pData slot should be a data.frame")
        }
        
        # check that nrow of @pData correspond to number of samples
        nSamples <- length(x@sampleFiles)
        if (nrow(x@pData) != nSamples) {
            return("@pData number of rows should be equal to number of samples")
        }
        
        # are full names of sample files used as pData row names ? 
        if (all(x@sampleFiles %in% rownames(x@pData)) && 
            all(rownames(x@pData) %in% x@sampleFiles)) {
            return(msg) #NULL
        }
        
        # are base names of sample files used as pData row names ? 
        if (all(basename(x@sampleFiles) %in% rownames(x@pData)) &&
            all(rownames(x@pData) %in% basename(x@sampleFiles))) {
            return(msg) #NULL
        }
        
        msg <- paste0("Row names of non-null @pData slot should correspond ",
                      "to sample file names (full path or basename)")
        return(msg)
    }
    return(msg)
}

.alignSampleOrderWithPData <- function(x) {
    # maintain consistency between order of x@pData and order x@sampleFiles;
    # x@pData is driving
    matching <- match(x@sampleFiles, rownames(x@pData))
    if (sum(is.na(matching)) > 0) {
        matching <- match(basename(x@sampleFiles), rownames(x@pData))
    }
    if (sum(is.na(matching)) > 0) {
        stop(paste0("mismatch between sample file names ",
                    "and pData row names (= unexpected inconsistency)"))
    }
    x@sampleFiles <- x@sampleFiles[order(matching)]
    return(x)
}

.makeSlots <- function(x, params) {
    stopifnot(c(
        inherits(x, "CytoPipeline"),
        is.list(params)
    ))

    # experimentName no more explicitly mandatory from the params list
    # because populated by default (default_experiment)
    #mandatory <- c("experimentName")
    mandatory <- c() 
    #optional <- c("sampleFiles")
    optional <- c("experimentName",
                  "sampleFiles",
                  "pData")

    for (m in mandatory) {
        if (is.null(params[[m]])) {
            stop("No ", m, " provided")
        }

        if (length(params[[m]]) == 0) {
            methods::slot(x, m) <-
                vector(
                    mode = mode(methods::slot(x, m)),
                    length = 0
                )
        } else {
            methods::slot(x, m) <- params[[m]]
        }
    }

    for (o in optional) {
        if (!is.null(params[[o]])) {
            if (length(params[[o]]) == 0) {
                methods::slot(x, o) <-
                    vector(
                        mode = mode(methods::slot(x, o)),
                        length = 0
                    )
            } else {
                methods::slot(x, o) <- params[[o]]
            }
        }
    }

    return(x)
}

.makeProcessingQueues <- function(x, params) {
    stopifnot(c(
        inherits(x, "CytoPipeline"),
        is.list(params)
    ))

    # generate scale transform processing queue
    if (!is.null(params$scaleTransformProcessingSteps)) {
        for (s in seq_along(params$scaleTransformProcessingSteps)) {
            prStep <- params$scaleTransformProcessingSteps[[s]]
            if (is.null(prStep$name) || is.null(prStep$FUN) ||
                is.null(prStep$ARGS)) {
                stop(
                    "step ", prStep,
                    " not well defined in input data, needs 'name',",
                    "'FUN' and 'ARGS'"
                )
            }
            x <- addProcessingStep(x,
                whichQueue = "scale transform",
                CytoProcessingStep(
                    name = prStep$name,
                    FUN = prStep$FUN,
                    ARGS = prStep$ARGS
                )
            )
        }
    }

    # generate flow frame pre-processing queue
    if (!is.null(params$flowFramesPreProcessingSteps)) {
        for (s in seq_along(params$flowFramesPreProcessingSteps)) {
            prStep <- params$flowFramesPreProcessingSteps[[s]]
            if (is.null(prStep$name) || is.null(prStep$FUN) ||
                is.null(prStep$ARGS)) {
                stop(
                    "step ", prStep,
                    " not well defined in input data, needs 'name',",
                    "'FUN' and 'ARGS'"
                )
            }
            x <- addProcessingStep(x,
                whichQueue = "pre-processing",
                CytoProcessingStep(
                    name = prStep$name,
                    FUN = prStep$FUN,
                    ARGS = prStep$ARGS
                )
            )
        }
    }
    return(x)
}

# find minimum extension of file basename which keeps the vector
# of file names with unique values
.getMinimumUniqueFileNames <- function(fileNames){
    stopifnot(is.character(fileNames))
    
    foundUnique <- FALSE
    trialNames <- basename(fileNames)
    stopCriterium <- FALSE
    while (!foundUnique && !stopCriterium){
        if (length(unique(trialNames)) == length(trialNames)){
            foundUnique <- TRUE
        } else {
            prefixes <- mapply(
                FUN = function(x, y){gsub(x, "", y)}, 
                trialNames, 
                fileNames, 
                SIMPLIFY = TRUE, 
                USE.NAMES = FALSE)
            candidateTrialNames <- mapply(
                FUN = function(x, y){
                    if(x != ""){
                        file.path(basename(x), y)
                    } else {
                        y
                    }
                },
                prefixes,
                trialNames,
                SIMPLIFY = TRUE,
                USE.NAMES = FALSE)
            
            if(all(candidateTrialNames == trialNames))
            {
                # not possible to go further in extending name
                stopCriterium <- TRUE
            }
            
            trialNames <- candidateTrialNames
        }
    }
    if (!foundUnique){
        stop("it seems provided file names are not unique.")
    }
    return(trialNames)
}

---
title: "Automation and Visualization of Flow Cytometry Data Analysis Pipelines"
author:
    - name: Philippe Hauchamps
    - name: Laurent Gatto
package: CytoPipeline
abstract: >
 This vignette describes the functionality implemented in the `CytoPipeline`
 package. `CytoPipeline` provides support for automation and visualization of 
 flow cytometry data analysis pipelines. In the current state, the package 
 focuses on the pre-processing and quality control part. 
 This vignette is distributed under a CC BY-SA 4.0 license.
output:
  BiocStyle::html_document:
    toc_float: true
bibliography: CytoPipeline.bib
vignette: >
  %\VignetteIndexEntry{Automation and Visualization of Flow Cytometry Data Analysis Pipelines}
  %\VignetteEngine{knitr::rmarkdown}
  %%\VignetteKeywords{FlowCytometry, Preprocessing, QualityControl, WorkflowStep, ImmunoOncology, Software, Visualization}
  %\VignetteEncoding{UTF-8}
---

```{r style, echo = FALSE, results = 'asis'}
BiocStyle::markdown()
```
# Installation

To install this package, start R and enter (uncommented):

```{r}
# if (!require("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("CytoPipeline")
```

Note that CytoPipeline imports ggplot2 (>= 3.4.1).  
The version requirement is due to a bug in version 3.4.0., 
affecting `ggplot2::geom_hex()`.

# Introduction

The `CytoPipeline` package provides infrastructure to support the definition, 
run and standardized visualization of pre-processing and quality control 
pipelines for flow cytometry data. This infrastructure consists of two main S4 
classes, i.e. `CytoPipeline` and `CytoProcessingStep`, as well as dedicated 
wrapper functions around selected third-party package methods often used to 
implement these pre-processing steps.

In the following sections, we demonstrate how to create a `CytoPipeline` 
object implementing a simple pre-processing pipeline, how to run it and 
how to retrieve and visualize the results after each step. 

# Example dataset
The example dataset that will be used throughout this vignette is derived from 
a reference public dataset accompanying the OMIP-021 (Optimized Multicolor 
Immunofluorescence Panel 021) article [@Gherardin2014-pj].  

A sub-sample of this public dataset is built-in in the `CytoPipeline` 
package, as the OMIP021 dataset. 
See the `MakeOMIP021Samples.R` script for more details 
on how the `OMIP021` dataset was created. This script is to be found 
in the `script` subdirectory in the `CytoPipeline` package installation path.

Note that in the `CytoPipeline`package, as in the current vignette, 
matrices of flow cytometry events intensities are stored as 
`flowCore::flowFrame` objects [@flowCore]. 

# Example of pre-processing and QC pipelines
Let's assume that we want to pre-process the two samples of the `OMIP021` 
dataset, and let's assume that we want to compare what we would obtain when
pre-processing these files using two different QC methods.    

In the first pre-processing pipeline, we will use the flowAI QC method 
[@Monaco2016-vo], while in the second pipeline, we will use the PeacoQC method 
[@Emmaneel2021-xy]. Note that when we here refer to QC method, we mean the 
algorithm used to ensure stability (stationarity) of the channel signals 
in time.

In both pipelines, the first part consists in estimating appropriate scale
transformation functions for all channels present in the sample `flowFrame`.
In order to do this, we propose the following *scale transformation processing 
queue* (Fig. 1):   

- reading the two samples `.fcs` files
- removing the margin events from each file
- applying compensation for each file
- aggregating and sub-sampling from each file
- estimating the scale transformations from the aggregated 
and sub-sampled data   

```{r scaleTransformQueueDisplay, results='markup', fig.cap="Scale transform processing queue", echo=FALSE, out.width='75%', fig.align='center', fig.wide = TRUE}
knitr::include_graphics("figs/scaleTransformQueue.png", error = FALSE)
```

When this first part is done, one can apply pre-processing for each file 
one by one. However, depending on the choice of QC method, the order of steps 
needs to be slightly different:

- when using flowAI, it is advised to eliminate the 'bad events' starting from
raw data (see [@Monaco2016-vo])
- when using PeacoQC, it is advised to eliminate the 'bad events' from already
compensated and scale transformed data (see [@Emmaneel2021-xy])

Therefore, we propose the following *pre-processing queues* represented in 
Fig. 2.

```{r preProcessingQueueDisplay, results='markup', fig.cap="Pre-processing queue for two different pipeline settings", echo=FALSE, out.width='100%', fig.align='center', fig.wide = TRUE}
knitr::include_graphics("figs/preProcessingQueues.png", error = FALSE)
```

# Building the CytoPipeline

`CytoPipeline` is the central S4 class used in the `CytoPipeline` package to 
represent a flow cytometry pre-processing pipeline. The main slots of
`CytoPipeline` objects are :    

- an `experimentName`, which gives a name to a particular user definition 
of a pre-processing pipeline. The *experiment* here, is not related to an assay
experiment, but refers to a specific way to design a pipeline. For example, in
the current use case, we will define two `experimentName`s, one to refer to the
flowAI pipeline, and another one to refer to the PeacoQC pipeline 
(see previous section);   

- a vector of `sampleFiles`, which are `.fcs` raw data files on which one need 
to run the pre-processing pipeline;  

- two processing queues, i.e. a `scaleTransformProcessingQueue`, and a 
`flowFramesPreProcessingQueue`, which correspond to the two parts described in 
previous section. Each of these queues are composed of one or several 
`CytoProcessingStep` objects, will be processed in linear sequence, the output 
of one step being the input of the next step.   

Note there are important differences between the two processing queues. 
On the one hand, the `scaleTransformProcessingQueue` takes the vector of all 
sample files as an input, and will be executed first, and only once. 
On the other hand, the `flowFramesPreProcessingQueue` will be run after the 
scale transformation processing queue, on each sample file one after the other, 
within a loop. The final output of the `scaleTransformProcessingQueue`, which 
should be a `flowCore::tranformList`, is also provided as input to the 
`flowFramesPreProcessingQueue`, by convention.      

In the next subsections, we show the different steps involved in creating a
`CytoPipeline` object.

## preliminaries: paths definition

In the following code, `rawDataDir` refers to the directory in which the `.fcs`
raw data files are stored. `workDir` will be used as root directory to store 
the disk cache. Indeed, when running the `CytoPipeline` objects, all the 
different step outputs will be stored in a `BiocFileCache` instance, in a 
sub-directory that will be created in `workDir`and of which the name will be 
set to the pipeline `experimentName`.

```{r pathsDef}
library(CytoPipeline)

# raw data
rawDataDir <- system.file("extdata", package = "CytoPipeline")
# output files
workDir <- suppressMessages(base::tempdir())
```

## first method: step by step, using CytoPipeline methods

In this sub-section, we build a `CytoPipeline` object and successively add 
`CytoProcessingStep` objects to the two different processing queues. We do this
for the PeacoQC pipeline.

```{r CytoPipelineSteps}

# main parameters : sample files and output files

experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                                pattern = "Donor"))

pipL_PeacoQC <- CytoPipeline(experimentName = experimentName,
                             sampleFiles = sampleFiles)

### SCALE TRANSFORMATION STEPS ###

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "flowframe_read",
            FUN = "readSampleFiles",
            ARGS = list(
                whichSamples = "all",
                truncate_max_range = FALSE,
                min.limit = NULL
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "remove_margins",
            FUN = "removeMarginsPeacoQC",
            ARGS = list()
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "compensate",
            FUN = "compensateFromMatrix",
            ARGS = list(matrixSource = "fcs")
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "flowframe_aggregate",
            FUN = "aggregateAndSample",
            ARGS = list(
                nTotalEvents = 10000,
                seed = 0
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "scale transform",
        CytoProcessingStep(
            name = "scale_transform_estimate",
            FUN = "estimateScaleTransforms",
            ARGS = list(
                fluoMethod = "estimateLogicle",
                scatterMethod = "linear",
                scatterRefMarker = "BV785 - CD3"
            )
        )
    )

### FLOW FRAME PRE-PROCESSING STEPS ###

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "flowframe_read",
            FUN = "readSampleFiles",
            ARGS = list(
                truncate_max_range = FALSE,
                min.limit = NULL
            )
        )
    )


pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_margins",
            FUN = "removeMarginsPeacoQC",
            ARGS = list()
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "compensate",
            FUN = "compensateFromMatrix",
            ARGS = list(matrixSource = "fcs")
        )
    )

pipL_PeacoQC <-
    addProcessingStep(
        pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "perform_QC",
            FUN = "qualityControlPeacoQC",
            ARGS = list(
                preTransform = TRUE,
                min_cells = 150, # default
                max_bins = 500, # default
                step = 500, # default,
                MAD = 6, # default
                IT_limit = 0.55, # default
                force_IT = 150, # default
                peak_removal = 0.3333, # default
                min_nr_bins_peakdetection = 10 # default
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(
        pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_doublets",
            FUN = "removeDoubletsCytoPipeline",
            ARGS = list(
                areaChannels = c("FSC-A", "SSC-A"),
                heightChannels = c("FSC-H", "SSC-H"),
                nmads = c(3, 5))
            )
    )
                    

                
pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_debris",
            FUN = "removeDebrisManualGate",
            ARGS = list(
                FSCChannel = "FSC-A",
                SSCChannel = "SSC-A",
                gateData =  c(73615, 110174, 213000, 201000, 126000,
                              47679, 260500, 260500, 113000, 35000)
            )
        )
    )

pipL_PeacoQC <-
    addProcessingStep(pipL_PeacoQC,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_dead_cells",
            FUN = "removeDeadCellsManualGate",
            ARGS = list(
                FSCChannel = "FSC-A",
                LDMarker = "L/D Aqua - Viability",
                gateData = c(0, 0, 250000, 250000,
                             0, 650, 650, 0)
            )
        )
    )


```


## second method: in one go, using JSON file input 

In this sub-section, we build the flowAI pipeline, this time using a JSON
file as an input. Note that the `experimentName` and `sampleFiles` are here 
specified in the JSON file itself. This is not necessary, as 
one could well specify the processing steps only in the JSON file, and 
pass the `experimentName` and `sampleFiles` directly 
in the `CytoPipeline` constructor.

```{r CytoPipelineJson}

jsonDir <- rawDataDir

# creation on CytoPipeline object,
# using json file as input
pipL_flowAI <-
  CytoPipeline(file.path(jsonDir, "OMIP021_flowAI_pipeline.json"),
               experimentName = "OMIP021_flowAI",
               sampleFiles = sampleFiles)


```

# Executing pipelines

## Executing PeacoQC pipeline

Note: executing the next statement might generate some warnings.  
These are generated by the `PeacoQC method`, are highly dependent 
on the shape of the data investigated, and can safely be ignored here.

```{r executePeacoQC}
# execute PeacoQC pipeline
execute(pipL_PeacoQC, path = workDir)
```


## Executing flowAI pipeline

Note: again this might generate some warnings, due to flowAI.  
These are highly dependent on the shape of the data investigated,
and can safely be ignored here.

```{r executeFlowAI }
# execute flowAI pipeline
execute(pipL_flowAI, path = workDir)
```




# Inspecting results and visualization

## Plotting processing queues as workflow graphs

```{r plotWorkFlows1, fig.cap = "PeacoQC pipeline - scale transformList processing queue"}

# plot work flow graph - PeacoQC - scale transformList
plotCytoPipelineProcessingQueue(
  pipL_PeacoQC,
  whichQueue = "scale transform",
  path = workDir)

```

```{r plotWorkFlows2, fig.cap = "PeacoQC pipeline - file pre-processing queue"}

# plot work flow graph - PeacoQC - pre-processing
plotCytoPipelineProcessingQueue(
  pipL_PeacoQC,
  whichQueue = "pre-processing",
  sampleFile = 1,
  path = workDir)
```

```{r plotWorkFlows3, fig.cap = "flowAI pipeline - scale transformList processing queue"}

# plot work flow graph - flowAI - scale transformList
plotCytoPipelineProcessingQueue(
  pipL_flowAI,
  whichQueue = "scale transform",
  path = workDir)
```

```{r plotWorkFlows4, fig.cap = "flowAI pipeline - file pre-processing queue"}

# plot work flow graph - flowAI - pre-processing

plotCytoPipelineProcessingQueue(
  pipL_flowAI,
  whichQueue = "pre-processing",
  sampleFile = 1,
  path = workDir)
```


## Obtaining information about pipeline generated objects

```{r gettingObjectInfos}
getCytoPipelineObjectInfos(pipL_PeacoQC, 
                           path = workDir,
                           whichQueue = "scale transform")
                                  
getCytoPipelineObjectInfos(pipL_PeacoQC, 
                           path = workDir,
                           whichQueue = "pre-processing",
                           sampleFile = sampleFiles(pipL_PeacoQC)[1])
```

## Retrieving flow frames at different steps and plotting them


```{r gettingFlowFrames}
# example of retrieving a flow frame
# at a given step
ff <- getCytoPipelineFlowFrame(
  pipL_PeacoQC,
  whichQueue = "pre-processing",
  sampleFile = 1,
  objectName = "remove_doublets_obj",
  path = workDir)

#
ff2 <- getCytoPipelineFlowFrame(
  pipL_PeacoQC,
  whichQueue = "pre-processing",
  sampleFile = 1,
  objectName = "remove_debris_obj",
  path = workDir)

```


```{r ggplotEvents1, fig.cap = "1-dimensional distribution plot (forward scatter channel)"}
ggplotEvents(ff, xChannel = "FSC-A")
```

```{r ggplotEvents2, fig.cap = "2-dimensional distribution plot (forward scatter vs. side scatter channels)"}
ggplotEvents(ff, xChannel = "FSC-A", yChannel = "SSC-A")
```

```{r ggplotEvents3, fig.cap = "2-dimensional difference plot between remove_doublets and remove_debris steps"}
ggplotFilterEvents(ff, ff2, xChannel = "FSC-A", yChannel = "SSC-A")
```


## Example of retrieving another type of object

We now provide an example on how to retrieve an object from the cache, that
is not specifically a `flowCore::flowFrame`.  

Here we retrieve a `flowCore::flowSet` object, which represents a set of  
`flowCore::flowFrame`objects, that was obtained after the compensation step 
of the scale transformation processing queue, prior to aggregating the 
two samples.

```{r cacheObjectRetrieve}
obj <- getCytoPipelineObjectFromCache(pipL_PeacoQC,
                                      path = workDir,
                                      whichQueue = "scale transform",
                                      objectName = "compensate_obj")
show(obj)
```

## Getting and plotting the nb of retained events are each step

Getting the number of retained events at each pre-processing step, and tracking 
these changes throughout the pre-processing steps of a pipeline 
for different samples is a useful quality control.  

This can be implemented using *CytoPipeline* `collectNbOfRetainedEvents()` 
function. Examples of using this function in quality control plots are shown 
in this section.

```{r getNbEvents}
ret <- CytoPipeline::collectNbOfRetainedEvents(
    experimentName = "OMIP021_PeacoQC",
    path = workDir
)
ret
```
```{r getRetainedProp}
retainedProp <- 
    as.data.frame(t(apply(
        ret,
        MARGIN = 1,
        FUN = function(line) {
            if (length(line) == 0 || is.na(line[1])) {
                as.numeric(rep(NA, length(line)))
            } else {
                round(line/line[1], 3)
            }
        }
    )))

retainedProp <- retainedProp[-1]

retainedProp
```

```{r getStepRemovedProp}
stepRemovedProp <- 
    as.data.frame(t(apply(
        ret,
        MARGIN = 1,
        FUN = function(line) {
            if (length(line) == 0) {
                as.numeric(rep(NA, length(line)))
            } else {
                round(1-line/dplyr::lag(line), 3)
            }
        }
    )))

stepRemovedProp <- stepRemovedProp[-1]

stepRemovedProp
```

```{r loadAddPackages}
library("reshape2")
library("ggplot2")
```



```{r plotRetainedProp}
#| fig.alt: >
#|   Scatterplot of proportion of retained events as a function of the 
#|   cumulated preprocessing steps, for the two fcs files.
#|   As expected, this proportion decreases at each step.
myGGPlot <- function(DF, title){
    stepNames = colnames(DF)
    rowNames = rownames(DF)
    DFLongFmt <- reshape(DF,
                         direction = "long",
                         v.names = "proportion",
                         varying = stepNames,
                         timevar = "step",
                         time = stepNames,
                         ids = rowNames)
    
    DFLongFmt$step <- factor(DFLongFmt$step, levels = stepNames)
    
    
    ggplot(data = DFLongFmt,
                 mapping = aes(x = step, y = proportion, text = id)) +
        geom_point(col = "blue") + 
        ggtitle(title) +
        theme(axis.text.x = element_text(angle = 90))
    
}

p1 <- myGGPlot(DF = retainedProp, 
               title = "Retained event proportion at each step")
p1
```


```{r plotStepRemovedProp}
#| fig.alt: >
#|   Scatterplot of number of events removed as a function of the 
#|   preprocessing step, for the two fcs files.
#|   The proportion varies between 0.0 and 0.4.
p2 <- myGGPlot(DF = stepRemovedProp,
               title = "Event proportion removed by each step")
p2
```



## Interactive visualization

Using the `CytoPipelineGUI` package, it is possible to interactively inspect
results at the different steps of the pipeline, either in the form of 
`flowCore::flowFrame` objects, or `flowCore::transformList`. 
To do this, install the `CytoPipelineGUI` package, and uncomment 
the following code: 

```{r launchShinyApp}
#devtools::install_github("https://github.com/UCLouvain-CBIO/CytoPipelineGUI")
#CytoPipelineGUI::CytoPipelineCheckApp(dir = workDir)
```

# Adding function wrappers - note on the CytoPipelineUtils package

As was described in the previous sections, `CytoPipeline` requires the user to 
provide wrappers to pre-processing functions, as `FUN` parameter of 
`CytoProcessingSteps`. These can be coded by the user themself, or 
come from a built-in function provided in `CytoPipeline` itself.

However, in order to avoid having too many external dependencies for
`CytoPipeline`, another package `CytoPipelineUtils`, is also
[available](https://github.com/UCLouvain-CBIO/CytoPipelineUtils)
`CytoPipelineUtils` is meant to be used in conjunction with `CytoPipeline` 
package. It is a helper package, which is aimed at hosting wrapper 
implementations of various functions of various packages.

`CytoPipelineUtils` is open to contributions. If you want to implement your 
own wrapper of your favourite pre-processing function and use it in a 
`CytoPipeline` object, this is the place to do it! 

# Session information {-}

```{r sessioninfo, echo=FALSE}
sessionInfo()
```


# References {-}

---
title: "Demonstration of the CytoPipeline R package suite functionalities"
author:
    - name: Philippe Hauchamps
    - name: Laurent Gatto
package: CytoPipeline
abstract: >
 This vignette provides links to videos demonstrating some of the functionalities of CytoPipeline R package suite. It is distributed, as well as the accompanying videos, under a CC BY-SA license.
output:
  BiocStyle::html_document:
    toc_float: true
bibliography: CytoPipeline.bib
vignette: >
  %\VignetteIndexEntry{Demonstration of the CytoPipeline R package suite functionalities}
  %\VignetteEngine{knitr::rmarkdown}
  %%\VignetteKeywords{FlowCytometry, Preprocessing, QualityControl, WorkflowStep, Software, Visualization}
  %\VignetteEncoding{UTF-8}
---

# Introduction

<iframe src="https://www.youtube.com/embed/IB832cQkXdg" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Background information

<iframe src="https://www.youtube.com/embed/0AmO9ZaxJoE" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Illustrating dataset

<iframe src="https://www.youtube.com/embed/D8Jkf_bhrfQ" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Specifying the pipeline

<iframe src="https://www.youtube.com/embed/z25TcU39Bqc" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Running the pipeline

<iframe src="https://www.youtube.com/embed/6q4MTZi_Hvw" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Visualizing the results

<iframe src="https://www.youtube.com/embed/NAtwhr7UQW4" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Comparing pipelines

<iframe src="https://www.youtube.com/embed/F2gdPr-_zi0" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Example with two different QC methods

<iframe src="https://www.youtube.com/embed/uLGeARGGxR8" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Visualizing scale transformations

<iframe src="https://www.youtube.com/embed/b3htJsbn3Ag" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Defining technical run parameters

<iframe src="https://www.youtube.com/embed/-jPp_aNnA_U" width="533" height="300" frameborder="0" allowfullscreen data-external="1"></iframe>

# Session information {-}

```{r sessioninfo, echo=FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplots.R
\name{ggplotEvents}
\alias{ggplotEvents}
\title{plot events in 1D or 2D, using ggplot2}
\usage{
ggplotEvents(
  obj,
  xChannel,
  yChannel = NULL,
  nDisplayCells = Inf,
  seed = NULL,
  bins = 216,
  fill = "lightblue",
  alpha = 0.2,
  xScale = c("linear", "logicle"),
  yScale = c("linear", "logicle"),
  xLogicleParams = NULL,
  yLogicleParams = NULL,
  xLinearRange = NULL,
  yLinearRange = NULL,
  transList = NULL,
  runTransforms = FALSE
)
}
\arguments{
\item{obj}{a flowCore::flowFrame or flowCore::flowSet}

\item{xChannel}{channel (name or index) or marker name to be displayed
on x axis}

\item{yChannel}{channel (name or index) or marker name to be displayed
on y axis}

\item{nDisplayCells}{maximum number of events that will be plotted. If
the number of events exceed this number, a sub-sampling will be performed}

\item{seed}{seed used for sub-sampling (if any)}

\item{bins}{used in geom_hex}

\item{fill}{used in geom_density}

\item{alpha}{used in geom_density}

\item{xScale}{scale to be used for the x axis
(note "linear" corresponds to no transformation)}

\item{yScale}{scale to be used for the y axis
(note "linear" corresponds to no transformation)}

\item{xLogicleParams}{if (xScale == "logicle"), the parameters of the logicle
transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...).
If NULL, these parameters will be estimated by flowCore::estimateLogicle()}

\item{yLogicleParams}{if (yScale == "logicle"), the parameters of the logicle
transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...).
If NULL, these parameters will be estimated by flowCore::estimateLogicle()}

\item{xLinearRange}{if (xScale == "linear"), the x axis range to be used}

\item{yLinearRange}{if (yScale == "linear"), the y axis range to be used}

\item{transList}{optional list of scale transformations to be applied to each
channel. If it is non null, 'x/yScale', 'x/yLogicleParams' and
'x/yLinear_range' will be discarded.}

\item{runTransforms}{(TRUE/FALSE) Will the application of non linear scale
result in data being effectively transformed ?
\itemize{
\item If TRUE, than the data will undergo transformations prior to
visualization.
\item If FALSE, the axis will be scaled but the data themselves will not be
transformed.
}}
}
\value{
a list of ggplot objects
}
\description{
plot events of specific channels of either :
flowCore::flowFrame, or flowCore::flowSet
in 2D or 1D, mimicking FlowJo type of graph. \cr
if 1D : geom_density will be used \cr
if 2D : geom_hex will be used \cr
}
\examples{

data(OMIP021Samples)

### 1D Examples ###

# simple linear scale example
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "FSC-A",
             xScale = "linear")

# with explicit linear range
ggplotEvents(OMIP021Samples[[1]],
                  xChannel = "FSC-A",
                  xScale = "linear",
                  xLinearRange = c(0, 250000))

# with linear scale, several flow frames
ggplotEvents(OMIP021Samples, xChannel = "FSC-A", xScale = "linear")

# simple logicle scale example
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "450/50Violet-A",
             xScale = "logicle")

# logicle scale, explicit parameters
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "450/50Violet-A",
             xScale = "logicle", xLogicleParams = list(
                 a = 1,
                 w = 2,
                 m = 7,
                 t = 270000))

# with sub-sampling
ggplotEvents(OMIP021Samples[[2]],
             xChannel = "450/50Violet-A",
             xScale = "logicle", nDisplayCells = 5000)

# tuning some plot parameters
ggplotEvents(OMIP021Samples[[2]],
             xChannel = "450/50Violet-A",
             xScale = "logicle", alpha = 0.5, fill = "red")

# examples that use a transformation list, estimated after compensation
compensationMatrix <- flowCore::spillover(OMIP021Samples[[1]])$SPILL

ffC <- runCompensation(OMIP021Samples[[1]],
                       spillover = compensationMatrix,
                       updateChannelNames = FALSE)

transList <- flowCore::estimateLogicle(
    ffC,
    colnames(compensationMatrix))

transList <-
    c(transList,
      flowCore::transformList(
          "FSC-A",
          flowCore::linearTransform(a = 0.00001)))

# linear example, without running the transformations on data
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "450/50Violet-A",
             xScale = "linear", 
             transList = transList,
             runTransforms = FALSE)

# linear example, now running the transformations on data
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "450/50Violet-A",
             xScale = "linear", 
             transList = transList,
             runTransforms = TRUE)

# logicle example, without running the transformations on data
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "FSC-A",
             xScale = "logicle", 
             transList = transList,
             runTransforms = FALSE)

# logicle example, now running the transformations on data
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "FSC-A",
             xScale = "logicle", 
             transList = transList,
             runTransforms = TRUE)

### 2D examples ###


# simple linear example
ggplotEvents(OMIP021Samples[[1]],
                  xChannel = "FSC-A",
                  xScale = "linear",
                  yChannel = "610/20Violet-A",
                  yScale = "logicle")

# simple linear example, 2 flow frames
ggplotEvents(OMIP021Samples,
             xChannel = "FSC-A",
             xScale = "linear",
             yChannel = "SSC-A",
             yScale = "linear")

# logicle vs linear example
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "450/50Violet-A",
             xScale = "logicle",
             yChannel = "SSC-A",
             yScale = "linear")

# 2X logicle example
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "TETaGC",
             xScale = "logicle",
             yChannel = "CD27",
             yScale = "logicle")

# tuning nb of bins
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "TETaGC",
             xScale = "logicle",
             yChannel = "CD27",
             yScale = "logicle",
             bins = 128)

# using transformation list, not run on data
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "TETaGC",
             xScale = "logicle",
             yChannel = "CD27",
             yScale = "logicle",
             transList = transList,
             runTransforms = FALSE)

# using transformation list, run on data                  
ggplotEvents(OMIP021Samples[[1]],
             xChannel = "TETaGC",
             xScale = "logicle",
             yChannel = "CD27",
             yScale = "logicle",
             transList = transList,
             runTransforms = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{qualityControlFlowAI}
\alias{qualityControlFlowAI}
\title{perform QC with flowAI}
\usage{
qualityControlFlowAI(
  ff,
  preTransform = FALSE,
  transList = NULL,
  outputDiagnostic = FALSE,
  outputDir = NULL,
  ...
)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{preTransform}{if TRUE, apply the transList scale transform prior to
running the gating algorithm}

\item{transList}{applied in conjunction with preTransform}

\item{outputDiagnostic}{if TRUE, stores diagnostic files generated by
flowAI in outputDir directory}

\item{outputDir}{used in conjunction with outputDiagnostic}

\item{...}{additional parameters passed to flowAI::flow_auto_qc()}
}
\value{
a flowCore::flowFrame with removed low quality events from the input
}
\description{
this function is a wrapper around flowAI::flow_auto_qc()
function.
It also pre-selects the channels to be handled (=> all signal channels)
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

suppressWarnings(ff_QualityControl <- 
    qualityControlFlowAI(fsRaw[[2]],
                         remove_from = "all", # all default
                         second_fractionFR = 0.1,
                         deviationFR = "MAD",
                         alphaFR = 0.01,
                         decompFR = TRUE,
                         outlier_binsFS = FALSE,
                         pen_valueFS = 500,
                         max_cptFS = 3,
                         sideFM = "both",
                         neg_valuesFM = 1))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{applyScaleTransforms}
\alias{applyScaleTransforms}
\title{apply scale transforms}
\usage{
applyScaleTransforms(x, transList, verbose = FALSE, ...)
}
\arguments{
\item{x}{a flowCore::flowSet or a flowCore::flowFrame}

\item{transList}{a flowCore::transformList}

\item{verbose}{if TRUE, send a message per flowFrame transformed}

\item{...}{other arguments (not used)}
}
\value{
the transformed flowFrame
}
\description{
wrapper around flowCore::transform() that discards any
additional parameter passed in (...)
Additionally, some checks regarding channels correspondance is done:
if \code{transList} contains transformations for channels that are not present
in \code{x}, then these transformations are first removed.
}
\examples{

data(OMIP021Samples)

transListPath <- file.path(system.file("extdata", 
                                       package = "CytoPipeline"),
                           "OMIP021_TransList.rds") 

transList <- readRDSObject(transListPath)

ff_c <- compensateFromMatrix(OMIP021Samples[[1]],
                             matrixSource = "fcs")  

ff_t <- applyScaleTransforms(ff_c, transList = transList)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipeline-functions.R
\name{interactingWithCytoPipelineCache}
\alias{interactingWithCytoPipelineCache}
\alias{deleteCytoPipelineCache}
\alias{buildCytoPipelineFromCache}
\alias{checkCytoPipelineConsistencyWithCache}
\title{interaction between CytoPipeline object and disk cache}
\usage{
deleteCytoPipelineCache(x, path = ".")

buildCytoPipelineFromCache(experimentName, path = ".")

checkCytoPipelineConsistencyWithCache(
  x,
  path = ".",
  whichQueue = c("both", "scale transform", "pre-processing"),
  sampleFile = NULL
)
}
\arguments{
\item{x}{a CytoPipeline object}

\item{path}{the full path to the experiment storage on disk
(without the /.cache)}

\item{experimentName}{the experimentName used to select the file cache on
disk}

\item{whichQueue}{which processing queue to check the consistency of}

\item{sampleFile}{if whichQueue == "pre-processing" or "both": which sample
file(s) to check on the disk cache}
}
\value{
for \code{deleteCytoPipelineCache}: TRUE if successfully removed\cr
for \code{buildCytoPipelineFromCache}: the built CytoPipeline object\cr
for \code{checkCytoPipelineConsistencyWithCache}: a list with the following
values:
\itemize{
\item \code{isConsistent} (TRUE/FALSE)
\item \code{inconsistencyMsg}: character filled in
by an inconsistency message in case the cache and
CytoPipeline object are not consistent with each other
\item \code{scaleTransformStepStatus}: a character vector,
containing, for each scale transform step, a status
from c("run", "not run", "inconsistent")
\item \code{preProcessingStepStatus}: a character matrix,
containing, for each pre-processing step (rows),
for each sample file (columns), a status from
c("run", "not run", "inconsistent")
}
}
\description{
functions supporting the interaction between a CytoPipeline
object and the file cache on disk
}
\section{Functions}{
\itemize{
\item \code{deleteCytoPipelineCache()}: delete the whole disk cache
corresponding to the experiment of a CytoPipeline object

\item \code{buildCytoPipelineFromCache()}: builds a new CytoPipeline
object,
based on the information stored in the file cache

\item \code{checkCytoPipelineConsistencyWithCache()}: check the consistency
between the processing steps described in a CytoPipeline object,
and what is stored in the file cache

}}
\examples{

# preliminary run:
# build CytoPipeline object using json input, run and store results in cache
rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                             pattern = "Donor"))
                                             
jsonDir <- system.file("extdata", package = "CytoPipeline")
jsonPath <- file.path(jsonDir, "pipelineParams.json")
outputDir <- base::tempdir()
pipL <- CytoPipeline(jsonPath,
                     experimentName = experimentName,
                     sampleFiles = sampleFiles)

# note we temporarily set working directory into package root directory
# needed as json path mentions "./" path for sample files
suppressWarnings(execute(pipL, rmCache = TRUE, path = outputDir))
     

# rebuild CytoPipeline from stored results in cache, for a specific 
# experiment

experimentName <- "OMIP021_PeacoQC"
pipL2 <- buildCytoPipelineFromCache(
    experimentName = experimentName,
    path = outputDir)


# checking consistency between CytoPipeline object and cache
res <- checkCytoPipelineConsistencyWithCache(pipL2)
#res

suppressWarnings(execute(pipL2, rmCache = FALSE, path = outputDir))
# (everything is already stored in cache)

# deleting cache related to a specific experiment
pipL3 <- CytoPipeline(experimentName = experimentName)
deleteCytoPipelineCache(pipL3, path = outputDir)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipeline-functions.R
\name{inspectCytoPipelineObjects}
\alias{inspectCytoPipelineObjects}
\alias{getCytoPipelineExperimentNames}
\alias{getCytoPipelineObjectFromCache}
\alias{getCytoPipelineObjectInfos}
\alias{getCytoPipelineFlowFrame}
\alias{getCytoPipelineScaleTransform}
\alias{plotCytoPipelineProcessingQueue}
\alias{collectNbOfRetainedEvents}
\title{inspect CytoPipeline results objects}
\usage{
getCytoPipelineExperimentNames(
  path = ".",
  pattern = NULL,
  ignore.case = FALSE,
  fixed = FALSE
)

getCytoPipelineObjectFromCache(
  x,
  path = ".",
  whichQueue = c("scale transform", "pre-processing"),
  sampleFile = NULL,
  objectName
)

getCytoPipelineObjectInfos(
  x,
  path = ".",
  whichQueue = c("scale transform", "pre-processing"),
  sampleFile = NULL
)

getCytoPipelineFlowFrame(
  x,
  path = ".",
  whichQueue = c("scale transform", "pre-processing"),
  sampleFile,
  objectName
)

getCytoPipelineScaleTransform(
  x,
  path = ".",
  whichQueue = c("scale transform", "pre-processing"),
  sampleFile = NULL,
  objectName
)

plotCytoPipelineProcessingQueue(
  x,
  whichQueue = c("pre-processing", "scale transform"),
  purpose = c("run status", "description"),
  sampleFile = NULL,
  path = ".",
  title = TRUE,
  box.type = "ellipse",
  lwd = 1,
  box.prop = 0.5,
  box.cex = 0.7,
  cex.txt = 0.7,
  box.size = 0.1,
  dtext = 0.15,
  ...
)

collectNbOfRetainedEvents(experimentName, path = ".", whichSampleFiles)
}
\arguments{
\item{path}{root path to locate the search for file caches}

\item{pattern}{optional pattern limiting the search for experiment
names}

\item{ignore.case}{(TRUE/FALSE) used in pattern matching (grepl)}

\item{fixed}{(TRUE/FALSE) used in pattern matching (grepl)}

\item{x}{a CytoPipeline object}

\item{whichQueue}{which queue to look into}

\item{sampleFile}{which sampleFile is looked for:
\itemize{
\item if whichQueue == "scale transform", the sampleFile is ignored
\item if NULL and whichQueue == "pre-processing", the sampleFile is
defaulted to the first one belonging to the experiment
}}

\item{objectName}{(character) which object name to look for}

\item{purpose}{purpose of the workflow plot
\itemize{
\item if "run status" (default), the disk cache will be inspected and the
box colours will be set according to run status (green = run,
orange = not run, red = definition not consistent with cache). Moreover, the
object classes and names will be filled in if found in the cache.
\item if "description", the workflow will be obtained from the step definition
in the \code{x} object, not from the disk cache. As a result, all boxes will be
coloured in black, and no object class and name will be provided.
}}

\item{title}{if TRUE, adds a title to the plot}

\item{box.type}{shape of label box (rect, ellipse, diamond,
round, hexa, multi)}

\item{lwd}{default line width of arrow and box (one numeric value)}

\item{box.prop}{length/width ratio of label box (one numeric value)}

\item{box.cex}{relative size of text in boxes (one numeric value)}

\item{cex.txt}{relative size of arrow text (one numeric value)}

\item{box.size}{size of label box (one numeric value)}

\item{dtext}{controls the position of arrow text relative to arrowhead
(one numeric value)}

\item{...}{other arguments passed to diagram::plotmat()}

\item{experimentName}{the experimentName used to select the file cache on
disk}

\item{whichSampleFiles}{indicates for which sample files the number
of retained events are to be collected.
If missing, all sample files will be used.}
}
\value{
\itemize{
\item for \code{getCytoPipelineExperimentNames}:
a vector of character containing found experiment names
}

\itemize{
\item for \code{getCytoPipelineObjectFromCache}:
the found object (or stops with an error message
if the target object is not found)
}

\itemize{
\item for \code{getCytoPipelineObjectInfos}:
a dataframe with the collected information about the
found objects (or stops with an error message if no target object
was found)
}

\itemize{
\item for \code{getCytoPipelineFlowFrame}:
the found flowFrame (or stops with an error message if the
target object is not found, or if the object is no flowFrame)
}

\itemize{
\item for \code{getCytoPipelineScaleTransform}: the found flowFrame
(or stops with an error message if the
target object is not found, or if the object is no transformList)
}

\itemize{
\item for \code{plotCytoPipelineProcessingQueue}:
nothing
}

\itemize{
\item for \code{collectNbOfRetainedEvents}:
a dataframe with the collected number of events
columns refer to pre-processing steps
rows refer to samples
}
}
\description{
functions to obtain results objects
formats
}
\section{Functions}{
\itemize{
\item \code{getCytoPipelineExperimentNames()}: This function
looks into a path for stored file caches
and gets the corresponding experiment names

\item \code{getCytoPipelineObjectFromCache()}: Given a CytoPipeline object,
this function retrieves
a specific object in the corresponding file cache

\item \code{getCytoPipelineObjectInfos()}: Given a CytoPipeline object,
this function retrieves
the information related to a specific object name,
i.e. object name and object class

\item \code{getCytoPipelineFlowFrame()}: Given a CytoPipeline object,
this function retrieves
a specific flowCore::flowFrame object in the corresponding
file cache object name and object class

\item \code{getCytoPipelineScaleTransform()}: Given a CytoPipeline object,
this function retrieves
a specific flowCore::transformList object in the corresponding
file cache

\item \code{plotCytoPipelineProcessingQueue()}: This functions displays
a plot of a processing queue of a CytoPipeline object,
using diagram::plotmat().
\itemize{
\item If a step is in run state for all sample files, the
corresponding box appears in green
\item If a step is in non run state for at least one sample file,
the corresponding box appears in orange
\item If at least one step is not consistent with cache, the whole
set of boxes appears in red
}

\item \code{collectNbOfRetainedEvents()}: Given a CytoPipeline object,
this function retrieves, for all pre-processing steps,
given the output is a flowFrame,
the number of retained event.

}}
\examples{


# preliminary run:
# build CytoPipeline object using json input, run and store results in cache
rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                             pattern = "Donor"))

jsonDir <- system.file("extdata", package = "CytoPipeline")
jsonPath <- file.path(jsonDir, "pipelineParams.json")
outputDir <- base::tempdir()
pipL <- CytoPipeline(jsonPath,
                     experimentName = experimentName,
                     sampleFiles = sampleFiles)

# note we temporarily set working directory into package root directory
# needed as json path mentions "./" path for sample files
suppressWarnings(execute(pipL, rmCache = TRUE, path = outputDir))
     

# get a list of all stored experiments in a specific path taken as root dir
experimentNames <- getCytoPipelineExperimentNames(path = outputDir)

# rebuilding Cytopipeline object from cache
pipL2 <- buildCytoPipelineFromCache(experimentName = experimentNames[1],
                                    path = outputDir)

# plot scale transformation queue
plotCytoPipelineProcessingQueue(pipL2, whichQueue = "pre-processing",
                                path = outputDir)

# plot pre-processing queue
plotCytoPipelineProcessingQueue(pipL2, whichQueue = "scale transform",
                                path = outputDir)
                                
# get object infos for a specific queue
df <- getCytoPipelineObjectInfos(pipL2, whichQueue = "pre-processing",
                                 path = outputDir,
                                 sampleFile = sampleFiles(pipL2)[1]) 
                                
# get transform list (output of one step)
trans <-
    getCytoPipelineScaleTransform(pipL2, whichQueue = "scale transform",
                                  objectName =
                                      "scale_transform_estimate_obj",
                                  path = outputDir)

# get flowFrame (output of one step)
ff <- getCytoPipelineFlowFrame(pipL2, whichQueue = "pre-processing",
                               objectName = "remove_doublets_obj",
                               path = outputDir,
                               sampleFile = sampleFiles(pipL2)[1])

# get any object (output of one step)
obj <-
    getCytoPipelineObjectFromCache(pipL2, whichQueue = "scale transform",
                                   objectName = "compensate_obj",
                                   path = outputDir)
class(obj) # flowCore::flowSet 

# collect number of retained events at each step
nbEventsDF <- collectNbOfRetainedEvents( 
        experimentName = experimentNames[1],
        path = outputDir) 

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getChannelNamesFromMarkers}
\alias{getChannelNamesFromMarkers}
\title{get channel names from markers}
\usage{
getChannelNamesFromMarkers(ff, markers)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{markers}{a vector of markers, either provided as :
\itemize{
\item an array of booleans (referring to flowFrame columns)
\item an array of integers (indices in flowFrame columns)
\item an array of characters (exact markers or channel patterns)
}}
}
\value{
a character vector, containing the names of the corresponding
channels
}
\description{
finds name of channels corresponding to user provided markers
}
\examples{

data(OMIP021Samples)

# with existing markers
ret <- getChannelNamesFromMarkers(
    OMIP021Samples[[1]],
    c(
        "FSC-A",
        "L/D Aqua - Viability",
        "FITC - gdTCR",
        "PECy5 - CD28"
    ))
    
ret # c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")

# with boolean vector
indices <- c(1, 6, 14, 18)
boolInput <- rep(FALSE, 21)
boolInput[indices] <- TRUE
ret2 <- getChannelNamesFromMarkers(
    OMIP021Samples[[1]],
    boolInput)
    
ret2 # c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")

# with indices vector
ret3 <- getChannelNamesFromMarkers(
    OMIP021Samples[[1]],
    indices
)
ret3 # c("FSC-A", "525/50Violet-A", "530/30Blue-A", "670/30Yellow-A")


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{estimateScaleTransforms}
\alias{estimateScaleTransforms}
\title{estimates scale tranformations}
\usage{
estimateScaleTransforms(
  ff,
  fluoMethod = c("estimateLogicle", "none"),
  scatterMethod = c("none", "linearQuantile"),
  scatterRefMarker = NULL,
  specificScatterChannels = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{fluoMethod}{method to be applied to all fluo channels}

\item{scatterMethod}{method to be applied to all scatter channels}

\item{scatterRefMarker}{the reference channel that is used to align the}

\item{specificScatterChannels}{vector of scatter channels for which we
still want to apply the fluo method (and not the scatter Method)}

\item{verbose}{if TRUE, send messages to the user at each step}

\item{...}{additional parameters passed to flowCore::estimateLogicle()}
}
\value{
a flowCore::flowFrame with removed low quality events from the input
}
\description{
this function estimates the scale transformations to be applied
on a flowFrame to obtain 'good behaving' distributions, i.e. the best
possible separation between + population and - population.
It distinguishes between scatter channels, where either linear, or no
transform is applied, and fluo channels, where either logicle transform
\itemize{
\item using flowCore::estimateLogicle - is estimated, or no transform
is applied.
}

The idea of linear transform of scatter channels is as follows: a reference
channel (not a scatter one) is selected and a linear transform (Y = AX + B)
is applied to all scatter channel, as to align their 5 and 95 percentiles to
those of the reference channel
For the estimateLogicle function, see flowCore documentation.
}
\examples{

data(OMIP021Samples)

compMatrix <- flowCore::spillover(OMIP021Samples[[1]])$SPILL
ff_c <- runCompensation(OMIP021Samples[[1]], spillover = compMatrix)

transList <- 
    estimateScaleTransforms(        
        ff = ff_c,
        fluoMethod = "estimateLogicle",
        scatterMethod = "linear",
        scatterRefMarker = "BV785 - CD3")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{removeMarginsPeacoQC}
\alias{removeMarginsPeacoQC}
\title{remove margin events using PeacoQC}
\usage{
removeMarginsPeacoQC(x, channelSpecifications = NULL, ...)
}
\arguments{
\item{x}{a flowCore::flowSet or a flowCore::flowFrame}

\item{channelSpecifications}{A list of lists with parameter specifications
for certain channels. This parameter should only be used if the values in
the internal parameters description is too strict or wrong for a number or
all channels. This should be one list per channel with first a minRange
and then a maxRange value. This list should have the channel name found back
in colnames(flowCore::exprs(ff)), or the corresponding marker name (found in
flowCore::pData(flowCore::description(ff)) ) .
If a channel is not listed in this parameter, its default internal values
will be used. The default of this parameter is NULL.
If the name of one list is set to \code{AllFluoChannels}, then the
\code{minRange} and \code{maxRange} specified there will be taken as default
for all fluorescent channels (not scatter)}

\item{...}{additional parameters passed to PeacoQC::RemoveMargins()}
}
\value{
either a flowCore::flowSet or a flowCore::flowFrame depending on
the input.
}
\description{
Wrapper around PeacoQC::RemoveMargins().
Also pre-selects the channels to be handled (=> all signal channels)
If input is a flowSet, it applies removeMargins() to each flowFrame of the
flowSet.
}
\examples{

rawDataDir <- 
    system.file("extdata", package = "CytoPipeline")
sampleFiles <- 
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL
fsRaw <- readSampleFiles(sampleFiles, 
                         truncate_max_range = truncateMaxRange,
                         min.limit = minLimit)
suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
ggplotFilterEvents(ffPre = fsRaw[[2]],
                   ffPost = ff_m,
                   xChannel = "FSC-A",
                   yChannel = "SSC-A")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{qualityControlPeacoQC}
\alias{qualityControlPeacoQC}
\title{perform QC with PeacoQC}
\usage{
qualityControlPeacoQC(
  ff,
  preTransform = FALSE,
  transList = NULL,
  outputDiagnostic = FALSE,
  outputDir = NULL,
  ...
)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{preTransform}{if TRUE, apply the transList scale transform prior to
running the gating algorithm}

\item{transList}{applied in conjunction with preTransform}

\item{outputDiagnostic}{if TRUE, stores diagnostic files generated by
PeacoQC in outputDir directory}

\item{outputDir}{used in conjunction with outputDiagnostic}

\item{...}{additional parameters passed to PeacoQC::PeacoQC()}
}
\value{
a flowCore::flowFrame with removed low quality events from the input
}
\description{
this function is a wrapper around PeacoQC::PeacoQC()
function.
It also pre-selects the channels to be handled (=> all signal channels)
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
    
ff_c <-
    compensateFromMatrix(ff_m,
                         matrixSource = "fcs")        

transList <- 
    estimateScaleTransforms(        
        ff = ff_c,
        fluoMethod = "estimateLogicle",
        scatterMethod = "linear",
        scatterRefMarker = "BV785 - CD3")


ff_QualityControl <- suppressWarnings(
    qualityControlPeacoQC(
        ff_c,
        preTransform = TRUE,
        transList = transList,
        min_cells = 150,
        max_bins = 500,
        MAD = 6,
        IT_limit = 0.55,
        force_IT = 150, 
        peak_removal = (1/3),
        min_nr_bins_peakdetection = 10))
        
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipeline-functions.R
\name{exportCytoPipeline}
\alias{exportCytoPipeline}
\alias{export2JSONFile}
\title{exporting CytoPipeline objects}
\usage{
export2JSONFile(x, path)
}
\arguments{
\item{x}{a CytoPipeline object}

\item{path}{the full path to the name of the file to be created}
}
\value{
\itemize{
\item for \code{export2JSONFile}: nothing
}
}
\description{
functions to export CytoPipeline objects in various
formats
}
\section{Functions}{
\itemize{
\item \code{export2JSONFile()}: exports a CytoPipeline object
to a JSON file (writing the file = side effect)

}}
\examples{

outputDir <- base::tempdir()

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                             pattern = "Donor"))

# build CytoPipeline object using json input
jsonPath <- file.path(system.file("extdata", package = "CytoPipeline"), 
                      "pipelineParams.json")
  
pipL <- CytoPipeline(jsonPath,
                     experimentName = experimentName,
                     sampleFiles = sampleFiles)

# remove the last pre-processing step
nPreProcessing <- getNbProcessingSteps(pipL, whichQueue = "pre-processing")
pipL <- removeProcessingStep(pipL, whichQueue = "pre-processing", 
                                   index = nPreProcessing)

# export back to json file    
export2JSONFile(pipL, path = file.path(outputDir, "newFile.json")) 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{readSampleFiles}
\alias{readSampleFiles}
\title{Read fcs sample files}
\usage{
readSampleFiles(
  sampleFiles,
  whichSamples = "all",
  nSamples = NULL,
  seed = NULL,
  channelMarkerFile = NULL,
  ...
)
}
\arguments{
\item{sampleFiles}{a vector of character path to sample files}

\item{whichSamples}{one of:
\itemize{
\item 'all' if all sample files need to be read
\item 'random' if some samples need to be chosen randomly
(in that case, using \code{nSamples} and \code{seed})
\item a vector of indexes pointing to the sampleFiles vector
}}

\item{nSamples}{number of samples to randomly select
(if \code{whichSamples == "random"}).
If \code{nSamples} is higher than nb of available samples,
the output will be all samples}

\item{seed}{an optional seed parameters (provided to ease reproducibility).}

\item{channelMarkerFile}{an optional path to a csv file which provides the
mapping between channels and markers. If provided, this csv file should
contain a \code{Channel} column, and a \code{Marker} column. Optionally a 'Used'
column can be provided as well (TRUE/FALSE). Channels for which the 'Used'
column is set to FALSE will not be incorporated in the created flowFrame.}

\item{...}{additional parameters passed to flowCore file reading functions.}
}
\value{
either a flowCore::flowSet or a flowCore::flowFrame if
length(sampleFiles) == 1
}
\description{
Wrapper around flowCore::read.fcs() or flowCore::read.flowSet().
Also adds a "Cell_ID" additional column, used in flowFrames comparison
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
res <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

#res

# create a flowCore::flowFrame with one single sample
res2 <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = 2,
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

#res2

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{appendCellID}
\alias{appendCellID}
\title{append 'Original_ID' column to a flowframe}
\usage{
appendCellID(ff, eventIDs = seq_len(flowCore::nrow(ff)))
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{eventIDs}{an integer vector containing the values to be added
in expression matrix, as Original ID's.}
}
\value{
new flowCore::flowFrame containing the added 'Original_ID' column
}
\description{
: on a flowCore::flowFrame, append a 'Original_ID' column.
This column can be used in plots comparing the events pre and post gating.
If the 'Original_ID' column already exists, the function does nothing
}
\examples{

data(OMIP021Samples)

retFF <- appendCellID(OMIP021Samples[[1]])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipeline-package.R
\docType{package}
\name{CytoPipeline}
\alias{CytoPipeline-package}
\alias{CytoPipeline}
\title{CytoPipeline package}
\description{
\code{CytoPipeline} is a package that provides support for automation and
visualization of flow cytometry data analysis pipelines. In the current
state, the package focuses on the preprocessing and quality control part.

The framework is based on two main S4 classes, i.e. \code{CytoPipeline} and
\code{CytoProcessingStep}. The \code{CytoProcessingStep} defines the link between
pipeline step names and corresponding R functions that are either provided
in the CytoPipeline package itself, or exported from a third party package,
or coded by the user her/himself. The processing steps need to be specified
centrally and explicitly using either a json input file or through
step by step creation of a \code{CytoPipeline} object with dedicated methods.

After having run the pipeline, obtained results at all steps
can be retrieved and visualized thanks to file caching
(the running facility uses a \code{BiocFileCache} implementation).
The package provides also specific visualization tools like
pipeline workflow summary display, and 1D/2D comparison plots
of obtained flowFrames at various steps of the pipeline.

For a step by step example using \code{CytoPipeline}, please have a look
at the vignette!
}
\seealso{
\link{CytoPipelineClass}, \link{CytoProcessingStep}
}
\author{
\strong{Maintainer}: Philippe Hauchamps \email{philippe.hauchamps@uclouvain.be} (\href{https://orcid.org/0000-0003-2865-1852}{ORCID})

Authors:
\itemize{
  \item Laurent Gatto \email{laurent.gatto@uclouvain.be} (\href{https://orcid.org/0000-0002-1520-2268}{ORCID})
}

Other contributors:
\itemize{
  \item Dan Lin \email{dan.8.lin@gsk.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{OMIP021Samples}
\alias{OMIP021Samples}
\title{OMIP021Samples dataset}
\format{
a flowCore::flowSet with two different flowFrames
each one contains one flow cytometry sample
corresponding to Donor1.fcs and Donor2.fcs in following source.
A subsampling of 5,000 events has been performed on each file.
}
\source{
\url{https://flowrepository.org/experiments/305}
}
\value{
nothing
}
\description{
OMIP021Samples dataset
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{computeScatterChannelsLinearScale}
\alias{computeScatterChannelsLinearScale}
\title{compute linear transformation of scatter channels found in ff, based
on 5\% and 95\% of referenceChannel, set as target. If there is a
transformation defined in transList for referenceChannel, it is applied
first, before computing quantiles.
Then the computed linear transformations (or each scatter channel) are added
into the transfo_list. -A channels are computed, and same linear
transformation is then applied to corresponding -W and -H channels
(if they exist in ff).}
\usage{
computeScatterChannelsLinearScale(
  ff,
  transList = NULL,
  referenceChannel,
  silent = TRUE
)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{transList}{an initial flowCore::transformList}

\item{referenceChannel}{the reference channel to take target quantile values
from. Can be defined as marker or channel name.}

\item{silent}{if FALSE, will output some information on the computed
linear transformations}
}
\value{
the transList with added linear scale transformations
}
\description{
based on a referenceChannel
}
\examples{

data(OMIP021Samples)

ff <- OMIP021Samples[[1]]
refMarker <- "APCCy7 - CD4"
refChannel <- "780/60Red-A"
transList <- flowCore::estimateLogicle(ff,
                                       channels = refChannel)
retTransList <-
    computeScatterChannelsLinearScale(ff,
                                      transList = transList,
                                      referenceChannel = refMarker,
                                      silent = TRUE
    )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{aggregateAndSample}
\alias{aggregateAndSample}
\title{Aggregate and sample multiple flow frames of a flow set together}
\usage{
aggregateAndSample(
  fs,
  nTotalEvents,
  setup = c("forceNEvent", "forceBalance"),
  seed = NULL,
  channels = NULL,
  writeOutput = FALSE,
  outputFile = "aggregate.fcs",
  keepOrder = FALSE
)
}
\arguments{
\item{fs}{a flowCore::flowset}

\item{nTotalEvents}{Total number of cells to select from the input flow
frames}

\item{setup}{How to proceed when nTotalEvents/nFiles is too high
for some of the flow frames:
\itemize{
\item forceBalance (default): compute the minimum nb of events per flow frame,
and keep that amount of events from each flow frame.
\item forceNEvents: try to be as balanced as possible, but force a
total of nTotalEvents if possible, i.e. takes all events from the flow frame
with too low nb of events, and then fill in the total with events
from the bigger flow frames in a balanced way.
However, if nTotalEvents is greater than the sum of all events,
take all events only once.
}}

\item{seed}{seed to be set before sampling for reproducibility.
Default NULL does not set any seed.}

\item{channels}{Channels/markers to keep in the aggregate.
Default NULL takes all channels of the first file.}

\item{writeOutput}{Whether to write the resulting flowframe to a file.
Default FALSE}

\item{outputFile}{Full path to output file. Default "aggregate.fcs"}

\item{keepOrder}{If TRUE, the random subsample will be ordered in the same
way as they were originally ordered in the file. Default = FALSE.}
}
\value{
returns a new flowCore::flowFrame
}
\description{
Aggregate multiple flow frames in order to analyze them
simultaneously.
A new FF, which contains about nTotalEvents cells, nTotalEvents/nFiles
cells from each file.
Two new columns are added:
a column indicating the original file by index, and
a noisy version of this, for better plotting opportunities,
This function is based on PeacoQC::AggregateFlowframes() where file names
inputs have been replaced by a flowSet input.
}
\examples{

data(OMIP021Samples)

nCells <- 1000
agg <- aggregateAndSample(
    fs = OMIP021Samples,
    nTotalEvents = nCells)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipeline-functions.R
\name{handlingProcessingSteps}
\alias{handlingProcessingSteps}
\alias{addProcessingStep}
\alias{removeProcessingStep}
\alias{getNbProcessingSteps}
\alias{getProcessingStep}
\alias{getProcessingStepNames}
\alias{cleanProcessingSteps}
\alias{showProcessingSteps}
\title{handling processing steps in CytoPipeline objects}
\usage{
addProcessingStep(
  x,
  whichQueue = c("scale transform", "pre-processing"),
  newPS
)

removeProcessingStep(
  x,
  whichQueue = c("scale transform", "pre-processing"),
  index
)

getNbProcessingSteps(x, whichQueue = c("scale transform", "pre-processing"))

getProcessingStep(
  x,
  whichQueue = c("scale transform", "pre-processing"),
  index
)

getProcessingStepNames(x, whichQueue = c("scale transform", "pre-processing"))

cleanProcessingSteps(
  x,
  whichQueue = c("both", "scale transform", "pre-processing")
)

showProcessingSteps(x, whichQueue = c("scale transform", "pre-processing"))
}
\arguments{
\item{x}{a CytoPipeline object}

\item{whichQueue}{selects the processing queue for which we manage the
processing steps}

\item{newPS}{the new processing step to be added (CytoProcessingStep object)}

\item{index}{index of the processing step to remove}
}
\value{
\itemize{
\item for \code{addProcessingStep}: the updated CytoPipeline object
}

\itemize{
\item for \code{removeProcessingStep}: the updated CytoPipeline object
}

\itemize{
\item for \code{getNbProcessingSteps}: the number of processing steps present
in the target queue
}

\itemize{
\item for \code{getProcessingStep}: the obtained CytoProcessingStep object
}

\itemize{
\item for \code{getProcessingStepNames}: the vector of step names
}

\itemize{
\item for \code{cleanProcessingSteps}: the updated CytoPipeline object
}

\itemize{
\item for \code{showProcessingSteps}: nothing (only console display side
effect is required)
}
}
\description{
functions to manipulate processing steps in processing queues
of CytoPipeline objects
}
\section{Functions}{
\itemize{
\item \code{addProcessingStep()}: adds a processing step in one of the
processing queues (at the end), returns the modified CytoPipeline object

\item \code{removeProcessingStep()}: removes a processing step from one of
the processing queues, returns the modified CytoPipeline object

\item \code{getNbProcessingSteps()}: gets the number of processing
steps in a processing queue

\item \code{getProcessingStep()}: gets a processing step at a
specific index of a processing queue

\item \code{getProcessingStepNames()}: gets a character vector of all
processing step names of a specific processing queue

\item \code{cleanProcessingSteps()}: deletes all processing steps in one
or both processing queues, returns the modified CytoPipeline object

\item \code{showProcessingSteps()}: shows all processing steps in a
processing queue

}}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                             pattern = "Donor"))
transListPath <- 
    file.path(system.file("extdata", package = "CytoPipeline"), 
              "OMIP021_TransList.rds")

# main parameters : sample files and experiment name
pipelineParams <- list()
pipelineParams$experimentName <- experimentName
pipelineParams$sampleFiles <- sampleFiles

# create CytoPipeline object (no step defined yet)
pipL <- CytoPipeline(pipelineParams)

# add a processing step in scale tranformation queue
pipL <- addProcessingStep(pipL,
                          whichQueue = "scale transform",
                          CytoProcessingStep(
                              name = "scale_transform_read",
                              FUN = "readRDS",
                              ARGS = list(file = transListPath)
                          ))

getNbProcessingSteps(pipL, "scale transform") # returns 1

# add another processing step in scale transformation queue
pipL <- addProcessingStep(pipL,
                          whichQueue = "scale transform",
                          CytoProcessingStep(
                              name = "scale_transform_sum",
                              FUN = "sum",
                              ARGS = list()
                          )
)

getNbProcessingSteps(pipL, "scale transform") # returns 2

getProcessingStepNames(pipL, whichQueue = "scale transform")

# removes second processing step in scale transformation queue
pipL <- removeProcessingStep(pipL,
                             whichQueue = "scale transform",
                             index = 2)

# get processing step object
pS <- getProcessingStep(pipL, whichQueue = "scale transform", index = 1)
getCPSName(pS) #"scale_transform_read"

# add a processing step in pre-processing queue
pipL <- addProcessingStep(pipL,
                          whichQueue = "pre-processing",
                          CytoProcessingStep(
                              name = "pre-processing_sum",
                              FUN = "sum",
                              ARGS = list()
                          ))
getNbProcessingSteps(pipL, "scale transform") # returns 1
getNbProcessingSteps(pipL, "pre-processing") # returns also 1

showProcessingSteps(pipL, whichQueue = "scale transform")
showProcessingSteps(pipL, whichQueue = "pre-processing")

# cleans both processing queues
pipL <- cleanProcessingSteps(pipL)
pipL
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{removeDeadCellsManualGate}
\alias{removeDeadCellsManualGate}
\title{remove dead cells from a flowFrame using manual gating}
\usage{
removeDeadCellsManualGate(
  ff,
  preTransform = FALSE,
  transList = NULL,
  FSCChannel,
  LDMarker,
  gateData,
  ...
)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{preTransform}{boolean, if TRUE: the transList list of scale transforms
will be applied first on the LD channel.}

\item{transList}{applied in conjunction with preTransform == TRUE}

\item{FSCChannel}{a character containing the exact name of the forward
scatter channel}

\item{LDMarker}{a character containing the exact name of the marker
corresponding to (a)Live/Dead channel, or the Live/Dead channel name itself}

\item{gateData}{a numerical vector containing the polygon gate coordinates
first the \code{FSCChannel} channel coordinates
of each points of the polygon gate,
then the LD channel coordinates of each points
(prior to scale transform)}

\item{...}{additional parameters passed to flowCore::polygonGate()}
}
\value{
a flowCore::flowFrame with removed dead cells from the input
}
\description{
remove dead cells from a flowFrame, using manual gating in the
FSC-A, '(a)Live/Dead' 2D representation. The function uses
flowCore::polygonGate()
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
    
ff_c <-
    compensateFromMatrix(ff_m,
                         matrixSource = "fcs")    
                         
remDeadCellsGateData <- c(0, 0, 250000, 250000,
                          0, 650, 650, 0)  

ff_lcells <-
    removeDeadCellsManualGate(ff_c,
                              FSCChannel = "FSC-A",
                              LDMarker = "L/D Aqua - Viability",
                              gateData = remDeadCellsGateData)
   
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplots.R
\name{ggplotFilterEvents}
\alias{ggplotFilterEvents}
\title{plot filtered events in 2D, using ggplot}
\usage{
ggplotFilterEvents(
  ffPre,
  ffPost,
  xChannel,
  yChannel,
  nDisplayCells = 10000,
  seed = NULL,
  size = 0.5,
  xScale = c("linear", "logicle"),
  yScale = c("linear", "logicle"),
  xLogicleParams = NULL,
  yLogicleParams = NULL,
  xLinearRange = NULL,
  yLinearRange = NULL,
  transList = NULL,
  runTransforms = FALSE,
  interactive = FALSE
)
}
\arguments{
\item{ffPre}{a flowCore::flowFrame, before applying filter}

\item{ffPost}{a flowCore::flowFrame, after applying filter}

\item{xChannel}{channel (name or index) or marker name to be displayed
on x axis}

\item{yChannel}{channel (name or index) or marker name to be displayed
on y axis}

\item{nDisplayCells}{maximum number of events that will be plotted. If
the number of events exceed this number, a subsampling will be performed}

\item{seed}{seed used for sub-sampling (if any)}

\item{size}{used by geom_point()}

\item{xScale}{scale to be used for the x axis
(note "linear" corresponds to no transformation)}

\item{yScale}{scale to be used for the y axis
(note "linear" corresponds to no transformation)}

\item{xLogicleParams}{if (xScale == "logicle"), the parameters of the logicle
transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...)
If NULL, these parameters will be estimated by flowCore::estimateLogicle()}

\item{yLogicleParams}{if (yScale == "logicle"), the parameters of the logicle
transformation to be used, as a list(w = ..., m = ..., a = ..., t = ...)
If NULL, these parameters will be estimated by flowCore::estimateLogicle()}

\item{xLinearRange}{if (xScale == "linear"), linear range to be used}

\item{yLinearRange}{if (yScale == "linear"), linear range to be used}

\item{transList}{optional list of scale transformations to be applied to each
channel. If it is non null, 'x/yScale', 'x/yLogicleParams' and
'x/yLinear_range' will be discarded.}

\item{runTransforms}{(TRUE/FALSE) Will the application of non linear scale
result in data being effectively transformed ?
\itemize{
\item If TRUE, than the data will undergo transformations prior to
visualization.
\item If FALSE, the axis will be scaled but the data themselves are
not transformed.
}}

\item{interactive}{if TRUE, transform the scaling formats such that the
ggcyto::x_scale_logicle() and ggcyto::y_scale_logicle() do work with
plotly::ggplotly()}
}
\value{
a ggplot object
}
\description{
plot events of specific channels of either :
flowCore::flowFrame, or
flowCore::flowSet
in 2D, showing the impact of applying a filter between :
\itemize{
\item a 'pre' flowframe
}
}
\examples{

data(OMIP021Samples)

ffPre <- OMIP021Samples[[1]]

# creating a manual polygon gate filter based on channels L/D and FSC-A

LDMarker <- "L/D Aqua - Viability"

LDChannel <- getChannelNamesFromMarkers(ffPre, markers = LDMarker)
liveGateMatrix <- matrix(
    data = c(
        50000, 50000, 100000, 200000, 200000,
        100, 1000, 2000, 2000, 1
    ),
    ncol = 2,
    dimnames = list(
        c(),
        c("FSC-A", LDChannel)
    )
)

liveGate <- flowCore::polygonGate(
    filterId = "Live",
    .gate = liveGateMatrix
)

selectedLive <- flowCore::filter(ffPre, liveGate)
ffL <- flowCore::Subset(ffPre, selectedLive)


# show the results

# subsample 5000 points    
ggplotFilterEvents(
    ffPre = ffPre,
    ffPost = ffL,
    nDisplayCells = 5000,
    xChannel = "FSC-A", xScale = "linear",
    yChannel = LDMarker, yScale = "logicle") +
    ggplot2::ggtitle("Live gate filter - 5000 points")

# with all points
ggplotFilterEvents(
    ffPre = ffPre,
    ffPost = ffL,
    nDisplayCells = Inf,
    xChannel = "FSC-A", xScale = "linear",
    yChannel = LDMarker, yScale = "logicle") +
    ggplot2::ggtitle("Live gate filter - all points")


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{removeChannels}
\alias{removeChannels}
\title{remove channels from a flowFrame}
\usage{
removeChannels(ff, channels)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{channels}{the channel names to be removed}
}
\value{
a new flowCore::flowFrame with the removed channels
}
\description{
: in a flowCore::flowFrame, remove the channels of the given
names.
}
\examples{

data(OMIP021Samples)

retFF <- removeChannels(OMIP021Samples[[1]],
                        channel = "FSC-A")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{removeDebrisManualGate}
\alias{removeDebrisManualGate}
\title{remove debris from a flowFrame using manual gating}
\usage{
removeDebrisManualGate(ff, FSCChannel, SSCChannel, gateData, ...)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{FSCChannel}{a character containing the exact name of the forward
scatter channel}

\item{SSCChannel}{a character containing the exact name of the side scatter
channel}

\item{gateData}{a numerical vector containing the polygon gate coordinates
first the \code{FSCChannel} channel coordinates
of each points of the polygon gate,
then the \code{SSCChannel} channel coordinates of each points.}

\item{...}{additional parameters passed to flowCore::polygonGate()}
}
\value{
a flowCore::flowFrame with removed debris events from the input
}
\description{
remove debris from a flowFrame, using manual gating in the
FSC-A, SSC-A 2D representation. The function internally uses
flowCore::polygonGate()
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
    
ff_c <-
    compensateFromMatrix(ff_m,
                         matrixSource = "fcs")        


remDebrisGateData <- c(73615, 110174, 213000, 201000, 126000,
                       47679, 260500, 260500, 113000, 35000)

ff_cells <-
    removeDebrisManualGate(ff_c,
                           FSCChannel = "FSC-A",
                           SSCChannel = "SSC-A",
                           gateData = remDebrisGateData)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{subsample}
\alias{subsample}
\title{sub-sampling of a flowFrame}
\usage{
subsample(ff, nEvents, seed = NULL, keepOriginalCellIDs = TRUE, ...)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{nEvents}{number of events to be obtained using sub-sampling}

\item{seed}{can be set for reproducibility of event sub-sampling}

\item{keepOriginalCellIDs}{if TRUE, adds (if not already present)
a 'OriginalID' column containing the initial IDs of the cell
(from 1 to nrow prior to subsampling).
if FALSE, does the same, but takes as IDs (1 to nrow after subsampling)}

\item{...}{additional parameters (currently not used)}
}
\value{
new flowCore::flowFrame with the obtained subset of samples
}
\description{
: sub-samples a flowFrame
with the specified number of samples, without replacement.
adds also a column 'Original_ID' if not already present in flowFrame.
}
\examples{

data(OMIP021Samples)

# take first sample of dataset, subsample 100 events and create new flowFrame
ff <- subsample(OMIP021Samples[[1]], nEvents = 100)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipeline-functions.R
\name{execute}
\alias{execute}
\title{executing CytoPipeline object}
\usage{
execute(
  x,
  path = ".",
  rmCache = FALSE,
  useBiocParallel = FALSE,
  BPPARAM = BiocParallel::bpparam(),
  BPOPTIONS = BiocParallel::bpoptions(packages = c("flowCore")),
  saveLastStepFF = TRUE,
  saveFFSuffix = "_preprocessed",
  saveFFFormat = c("fcs", "csv"),
  saveFFCsvUseChannelMarker = TRUE,
  saveScaleTransforms = FALSE
)
}
\arguments{
\item{x}{CytoPipeline object}

\item{path}{base path, a subdirectory with name equal to the experiment will
be created to store the output data, in particular the experiment cache}

\item{rmCache}{if TRUE, starts by removing the already existing cache
directory corresponding to the experiment}

\item{useBiocParallel}{if TRUE, use BiocParallel for computation of the
sample file pre-processing in parallel (one file per worker at a time).
Note the BiocParallel function used is \code{bplapply()}}

\item{BPPARAM}{if \code{useBiocParallel} is TRUE, sets the BPPARAM back-end to
be used for the computation. If not provided, will use the top back-end on
the \code{BiocParallel::registered()} stack.}

\item{BPOPTIONS}{if \code{useBiocParallel} is TRUE, sets the BPOPTIONS to be
passed to \code{bplapply()} function.
Note that if you use a \code{SnowParams} back-end, you need to specify all
the packages that need to be loaded for the different CytoProcessingStep
to work properly (visibility of functions). As a minimum,
the \code{flowCore} package needs to be loaded.
(hence the default \code{BPOPTIONS = bpoptions(packages = c("flowCore"))} )}

\item{saveLastStepFF}{if TRUE, save the final result of the pre-processing,
for each file.
By convention, these output files are stored in
\code{path}/\code{x@experimentName}/output/, the file names used are the same
as the initial fcs file basenames, concatenated with \code{saveFFSuffix},
and with file extension corresponding to \code{saveFFFormat}.}

\item{saveFFSuffix}{FF file name suffix}

\item{saveFFFormat}{either \code{fcs} or \code{csv}}

\item{saveFFCsvUseChannelMarker}{if TRUE (default), converts the channels
to the corresponding marker names (where the Marker is not NA).
This setting is only applicable to export in csv format.}

\item{saveScaleTransforms}{if TRUE (default FALSE), save on disk
(in RDS format) the \code{flowCore::transformList} object obtained after running
the scaleTransform processing queue. The file name is hardcoded to
\code{path}/\code{experimentName}/\code{RDS}/\code{scaleTransformList.rds}}
}
\value{
nothing
}
\description{
this function triggers the execution of the processing queues of
a CytoPipeline object.
First, the scale tranform processing queue is run, taking
the set of sample names as an implicit first input. At the end of the queue,
a scale transform List is assumed to be created.
Second, the flowFrame pre-processing queue, reapeatedly for each sample file.
The scale transform list generated in the previous step is taken as implicit
input, together with the initial sample file. At the end of the queue run, a
pre-processed flowFrame is assumed to be generated.
No change is made on the input CytoPipeline object, all results are stored in
the cache.
}
\examples{

### *** EXAMPLE 1: building CytoPipeline step by step *** ###

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                             pattern = "Donor"))
                                             
outputDir <- base::tempdir()

# main parameters : sample files and output files
pipelineParams <- list()
pipelineParams$experimentName <- experimentName
pipelineParams$sampleFiles <- sampleFiles
pipL <- CytoPipeline(pipelineParams)

### SCALE TRANSFORMATION STEPS ###

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "flowframe_read",
                          FUN = "readSampleFiles",
                          ARGS = list(
                              whichSamples = "all",
                              truncate_max_range = FALSE,
                              min.limit = NULL
                          )
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "remove_margins",
                          FUN = "removeMarginsPeacoQC",
                          ARGS = list()
                     )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "compensate",
                          FUN = "compensateFromMatrix",
                          ARGS = list(matrixSource = "fcs")
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "flowframe_aggregate",
                          FUN = "aggregateAndSample",
                          ARGS = list(
                              nTotalEvents = 10000,
                              seed = 0
                          )
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "scale_transform_estimate",
                          FUN = "estimateScaleTransforms",
                          ARGS = list(
                              fluoMethod = "estimateLogicle",
                              scatterMethod = "linear",
                              scatterRefMarker = "BV785 - CD3"
                          )
                      )
    )

### PRE-PROCESSING STEPS ###

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "flowframe_read",
                          FUN = "readSampleFiles",
                          ARGS = list(
                              truncate_max_range = FALSE,
                              min.limit = NULL
                          )
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "remove_margins",
                          FUN = "removeMarginsPeacoQC",
                          ARGS = list()
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "compensate",
                          FUN = "compensateFromMatrix",
                          ARGS = list(matrixSource = "fcs")
                      )
    )

pipL <-
addProcessingStep(
    pipL,
    whichQueue = "pre-processing",
    CytoProcessingStep(
        name = "remove_debris",
        FUN = "removeDebrisManualGate",
        ARGS = list(
            FSCChannel = "FSC-A",
            SSCChannel = "SSC-A",
            gateData =  c(73615, 110174, 213000, 201000, 126000,
                          47679, 260500, 260500, 113000, 35000)
                   )
   )
)

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "remove_dead_cells",
                          FUN = "removeDeadCellsManualGate",
                          ARGS = list(
                              FSCChannel = "FSC-A",
                              LDMarker = "L/D Aqua - Viability",
                              gateData = c(0, 0, 250000, 250000,
                                           0, 650, 650, 0)
                          )
                      )
    )

pipL <-
    addProcessingStep(
        pipL,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "perform_QC",
            FUN = "qualityControlPeacoQC",
            ARGS = list(
                preTransform = TRUE,
                min_cells = 150, # default
                max_bins = 500, # default
                step = 500, # default,
                MAD = 6, # default
                IT_limit = 0.55, # default
                force_IT = 150, # default
                peak_removal = 0.3333, # default
                min_nr_bins_peakdetection = 10 # default
            )
        )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "transform",
                          FUN = "applyScaleTransforms",
                          ARGS = list()
                      )
    )

# execute pipeline, remove cache if existing with the same experiment name
suppressWarnings(execute(pipL, rmCache = TRUE, path = outputDir))

# re-execute as is without removing cache => all results found in cache!
suppressWarnings(execute(pipL, rmCache = FALSE, path = outputDir))

### *** EXAMPLE 2: building CytoPipeline from JSON file *** ###

jsonDir <- system.file("extdata", package = "CytoPipeline")
jsonPath <- file.path(jsonDir, "pipelineParams.json")

pipL2 <- CytoPipeline(jsonPath, 
                      experimentName = experimentName,
                      sampleFiles = sampleFiles)

# note we temporarily set working directory into package root directory
# needed as json path mentions "./" path for sample files
suppressWarnings(execute(pipL2, rmCache = TRUE, path = outputDir))

### *** EXAMPLE 3: building CytoPipeline from cache (previously run) *** ###

experimentName <- "OMIP021_PeacoQC"
pipL3 <- buildCytoPipelineFromCache(
    experimentName = experimentName,
    path = outputDir)

suppressWarnings(execute(pipL3,
        rmCache = FALSE,
        path = outputDir))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoPipelineClass.R
\docType{class}
\name{CytoPipeline-class}
\alias{CytoPipeline-class}
\alias{CytoPipeline-class,}
\alias{CytoPipelineClass}
\alias{show,CytoPipeline-method}
\alias{CytoPipeline,missing-method}
\alias{CytoPipeline,list-method}
\alias{CytoPipeline,character-method}
\alias{as.list.CytoPipeline}
\alias{experimentName}
\alias{experimentName<-}
\alias{sampleFiles}
\alias{sampleFiles<-}
\alias{pData}
\alias{pData<-}
\alias{sampleDisplayNames}
\alias{sampleNameFromDisplayName}
\title{CytoPipeline class}
\usage{
\S4method{show}{CytoPipeline}(object)

\S4method{CytoPipeline}{missing}(
  object,
  experimentName = "default_experiment",
  sampleFiles = character(),
  pData = NULL
)

\S4method{CytoPipeline}{list}(
  object,
  experimentName = "default_experiment",
  sampleFiles = character(),
  pData = NULL
)

\S4method{CytoPipeline}{character}(
  object,
  experimentName = "default_experiment",
  sampleFiles = character(),
  pData = NULL
)

\method{as.list}{CytoPipeline}(x, ...)

experimentName(x)

experimentName(x) <- value

sampleFiles(x)

sampleFiles(x) <- value

pData(x)

pData(x) <- value

sampleDisplayNames(x, sampleFiles = NULL)

sampleNameFromDisplayName(x, displayName)
}
\arguments{
\item{object}{a \code{character()} containing a JSON input}

\item{experimentName}{the experiment name}

\item{sampleFiles}{a character (e.g. sampleFileNames) or a numeric vector
(e.g. indices of sample files). If NULL, all samples will be displayed.}

\item{pData}{the pheno Data (data.frame or NULL)}

\item{x}{a \code{CytoPipeline} object}

\item{...}{additional arguments (not used here)}

\item{value}{the new value to be assigned.
the \verb{pData<-} setter is a bit more liberal than it used to be:
\enumerate{
\item It can accept new pData containing more rows than existing sample names
(the corresponding subset of pData is taken).
\item It can accept pData with row names pointing to either
sample file full paths or base file names
\item It can accept pData with no row names
provided the number of rows correspond to the number of sample files.
Row names are then set by default to sample file base names (if unique),
or sample file full paths.
}}

\item{displayName}{a character}
}
\value{
nothing

\itemize{
\item for \code{as.list.CytoPipeline}: the obtained list
}

\itemize{
\item for \code{sampleDisplayNames}: a character vector
of sample display names
}

\itemize{
\item for \code{sampleNameFromDisplayName}: the sample name corresponding
to the specified display name.
of sample display names
}
}
\description{
Class representing a flow cytometry pipeline, and composed of two processing
queues, i.e. lists of CytoProcessingStep objects :
\itemize{
\item a list of CytoProcessingStep(s) for pre-calculation of scale
transformations per channel
\item a list of CytoProcessingStep(s) for the pre-processing of flow frames
}
}
\section{Slots}{

\describe{
\item{\code{scaleTransformProcessingQueue}}{A \code{list} of
CytoProcessingStep objects containing the steps
for obtaining the scale transformations per channel}

\item{\code{flowFramesPreProcessingQueue}}{A \code{list} of
CytoProcessingStep objects containing the steps
for pre-processing of the samples flow frames}

\item{\code{experimentName}}{A \code{character} containing
the experiment (run) name}

\item{\code{sampleFiles}}{A \code{character} vector storing
all fcs files to be run into the pipeline}

\item{\code{pData}}{An optional \code{data.frame} containing
additional information for each sample file.
The \code{pData} raw names should correspond to the sample files
(using full paths or base paths).
If the \code{pData} contains a columns with name 'displayName', this will have
an impact in the \code{sampleDisplayNames()} function, i.e. sample display
names will be the one mentioned in \code{pData}, instead of typically
base file names (or larger paths if base file names are not unique)}
}}

\examples{

### *** EXAMPLE 1: building CytoPipeline step by step *** ###

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
experimentName <- "OMIP021_PeacoQC"
sampleFiles <- file.path(rawDataDir, list.files(rawDataDir,
                                             pattern = "Donor"))
                                             
outputDir <- base::tempdir()

# main parameters : sample files and output files
pipL <- CytoPipeline(experimentName = experimentName,
                     sampleFiles = sampleFiles)

### SCALE TRANSFORMATION STEPS ###

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "flowframe_read",
                          FUN = "readSampleFiles",
                          ARGS = list(
                              whichSamples = "all",
                              truncate_max_range = FALSE,
                              min.limit = NULL
                          )
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "remove_margins",
                          FUN = "removeMarginsPeacoQC",
                          ARGS = list()
                     )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "compensate",
                          FUN = "compensateFromMatrix",
                          ARGS = list(matrixSource = "fcs")
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "flowframe_aggregate",
                          FUN = "aggregateAndSample",
                          ARGS = list(
                              nTotalEvents = 10000,
                              seed = 0
                          )
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "scale transform",
                      CytoProcessingStep(
                          name = "scale_transform_estimate",
                          FUN = "estimateScaleTransforms",
                          ARGS = list(
                              fluoMethod = "estimateLogicle",
                              scatterMethod = "linear",
                              scatterRefMarker = "BV785 - CD3"
                          )
                      )
    )

### PRE-PROCESSING STEPS ###

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "flowframe_read",
                          FUN = "readSampleFiles",
                          ARGS = list(
                              truncate_max_range = FALSE,
                              min.limit = NULL
                          )
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "remove_margins",
                          FUN = "removeMarginsPeacoQC",
                          ARGS = list()
                      )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "compensate",
                          FUN = "compensateFromMatrix",
                          ARGS = list(matrixSource = "fcs")
                      )
    )

pipL <-
    addProcessingStep(
        pipL,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "remove_debris",
            FUN = "removeDebrisManualGate",
            ARGS = list(
                FSCChannel = "FSC-A",
                SSCChannel = "SSC-A",
                gateData =  c(73615, 110174, 213000, 201000, 126000,
                              47679, 260500, 260500, 113000, 35000)))
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "remove_dead_cells",
                          FUN = "removeDeadCellsManualGate",
                          ARGS = list(
                              FSCChannel = "FSC-A",
                              LDMarker = "L/D Aqua - Viability",
                              gateData = c(0, 0, 250000, 250000,
                                           0, 650, 650, 0)
                          )
                      )
    )

pipL <-
    addProcessingStep(
        pipL,
        whichQueue = "pre-processing",
        CytoProcessingStep(
            name = "perform_QC",
            FUN = "qualityControlPeacoQC",
            ARGS = list(
                preTransform = TRUE,
                min_cells = 150, # default
                max_bins = 500, # default
                step = 500, # default,
                MAD = 6, # default
                IT_limit = 0.55, # default
                force_IT = 150, # default
                peak_removal = 0.3333, # default
                min_nr_bins_peakdetection = 10 # default
            )
        )
    )

pipL <-
    addProcessingStep(pipL,
                      whichQueue = "pre-processing",
                      CytoProcessingStep(
                          name = "transform",
                          FUN = "applyScaleTransforms",
                          ARGS = list()
                      )
    )

### *** EXAMPLE 2: building CytoPipeline from JSON file *** ###

jsonDir <- system.file("extdata", package = "CytoPipeline")
jsonPath <- file.path(jsonDir, "pipelineParams.json")

pipL2 <- CytoPipeline(jsonPath,
                      experimentName = experimentName,
                      sampleFiles = sampleFiles)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{areFluoCols}
\alias{areFluoCols}
\title{find flow frame columns that represent fluorochrome channel}
\usage{
areFluoCols(
  x,
  toRemovePatterns = c("FSC", "SSC", "Time", "Original_ID", "File", "SampleID")
)
}
\arguments{
\item{x}{a flowCore::flowFrame or a flowCore::flowSet}

\item{toRemovePatterns}{a vector of string patterns that are to be considered
as non fluorochrome}
}
\value{
a vector of booleans of which the dimension is equal to the number of
columns in ff
}
\description{
: find flow frame columns that represent fluorochrome channel
}
\examples{

data(OMIP021Samples)

areFluoCols(OMIP021Samples)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{runCompensation}
\alias{runCompensation}
\title{compensate with additional options}
\usage{
runCompensation(obj, spillover, updateChannelNames = TRUE)
}
\arguments{
\item{obj}{a flowCore::flowFrame or flowCore::flowSet}

\item{spillover}{compensation object or spillover matrix or a list of
compensation objects}

\item{updateChannelNames}{if TRUE, add a 'comp-' prefix to all fluorochrome
channels (hence does not impact the columns related to FSC, SSC, or other
specific keyword like TIME, Original_ID, File,...)
Default TRUE.}
}
\value{
a new object with compensated data, and possibly updated column names
}
\description{
: this is a simple wrapper around the flowCore::compensate()
utility, allowing to trigger an update of the fluo channel names
with a prefix 'comp-' (as in FlowJo)
}
\examples{

data(OMIP021Samples)

ff <- OMIP021Samples[[1]]
compMatrix <- flowCore::spillover(ff)$SPILL
ff <- runCompensation(ff, 
                      spillover = compMatrix, 
                      updateChannelNames = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getFCSFileName}
\alias{getFCSFileName}
\title{get fcs file name}
\usage{
getFCSFileName(ff)
}
\arguments{
\item{ff}{a flowCore::flowFrame}
}
\value{
the basename of $FILENAME keyword
}
\description{
get basename of $FILENAME keyword if exists
}
\examples{

data(OMIP021Samples)

fName <- getFCSFileName(OMIP021Samples[[1]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{writeFlowFrame}
\alias{writeFlowFrame}
\title{write flowFrame to disk}
\usage{
writeFlowFrame(
  ff,
  dir = ".",
  useFCSFileName = TRUE,
  prefix = "",
  suffix = "",
  format = c("fcs", "csv"),
  csvUseChannelMarker = TRUE,
  ...
)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{dir}{an existing directory to store the flowFrame,}

\item{useFCSFileName}{if TRUE filename used will be based on original fcs
filename}

\item{prefix}{file name prefix}

\item{suffix}{file name suffix}

\item{format}{either fcs or csv}

\item{csvUseChannelMarker}{if TRUE (default), converts the channels to the
corresponding marker names (where the Marker is not NA). This setting is
only applicable to export in csv format.}

\item{...}{other arguments (not used)}
}
\value{
nothing
}
\description{
wrapper around flowCore::write.FCS() or utils::write.csv
that discards any additional parameter passed in (...)
}
\examples{
rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
res <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)
    
ff_c <- compensateFromMatrix(res[[2]], matrixSource = "fcs") 
outputDir <- base::tempdir()
writeFlowFrame(ff_c, 
               dir = outputDir,
               suffix = "_fcs_export",
               format = "csv")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStep.R
\name{CytoProcessingStep}
\alias{CytoProcessingStep}
\alias{CytoProcessingStep-class}
\alias{characterOrFunction-class}
\alias{show,CytoProcessingStep-method}
\alias{executeProcessingStep}
\alias{getCPSName}
\alias{getCPSFUN}
\alias{getCPSARGS}
\alias{as.list.CytoProcessingStep}
\alias{as.json.CytoProcessingStep}
\alias{from.json.CytoProcessingStep}
\title{Cyto Processing step}
\usage{
CytoProcessingStep(name = character(), FUN = character(), ARGS = list())

\S4method{show}{CytoProcessingStep}(object)

executeProcessingStep(x, ...)

getCPSName(x)

getCPSFUN(x)

getCPSARGS(x)

\method{as.list}{CytoProcessingStep}(x, ...)

as.json.CytoProcessingStep(x, pretty = FALSE)

from.json.CytoProcessingStep(jsonString)
}
\arguments{
\item{name}{\code{character} denoting a name to the step, which can be different
from the function name}

\item{FUN}{\code{function} or \code{character} representing a function name.}

\item{ARGS}{\code{list} of arguments to be passed along to \code{FUN}.}

\item{object}{a \code{CytoProcessingStep} object.}

\item{x}{a \code{CytoProcessingStep} object.}

\item{...}{other arguments (not used)}

\item{pretty}{formatting set-up (see jsonlite::toJSON doc)}

\item{jsonString}{a \code{character()} containing a JSON string.}
}
\value{
The \code{CytoProcessingStep} function returns and object of type
\code{CytoProcessingStep}.
}
\description{
Class containing the function and arguments to be applied in a lazy-execution
framework.

Objects of this class are created using the \code{CytoProcessingStep()}
function.
The processing step is executed with the \code{executeProcessingStep()}
function.
}
\details{
This object contains all relevant information of a data analysis processing
step, i.e. the function and all of its arguments to be applied to the data.
}
\examples{

## Create a simple processing step object
ps1 <- CytoProcessingStep("summing step", sum)

getCPSName(ps1)

getCPSFUN(ps1)

getCPSARGS(ps1)

executeProcessingStep(ps1, 1:10)

as.list(ps1)

js_str <- as.json.CytoProcessingStep(ps1)

ps2 <- from.json.CytoProcessingStep(js_str)

identical(ps1, ps2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{compensateFromMatrix}
\alias{compensateFromMatrix}
\title{compensation of fcs file(s) from matrix}
\usage{
compensateFromMatrix(
  x,
  matrixSource = c("fcs", "import"),
  matrixPath = NULL,
  updateChannelNames = TRUE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{x}{a \code{flowCore::flowFrame} or \code{flowCore::flowSet}}

\item{matrixSource}{if "fcs", the compensation matrix will be fetched from
the fcs files (different compensation matrices can then be applied by fcs
file)
if "import", uses \code{matrixPath} to read the matrix (should be a csv file)}

\item{matrixPath}{if matrixSource == "import", will be used as the input csv
file path}

\item{updateChannelNames}{if TRUE, updates the fluo channel names by
prefixing them with "comp-"}

\item{verbose}{if TRUE, displays information messages}

\item{...}{additional arguments (not used)}
}
\value{
the compensated flowSet or flowFrame
}
\description{
executes the classical compensation function on a flowSet or
flowFrame, given a compensation matrix. The matrix can be either retrieved
in the fcs files themselves or provided as a csv file.
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
    
ff_c <-
    compensateFromMatrix(ff_m,
                         matrixSource = "fcs")        
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{findTimeChannel}
\alias{findTimeChannel}
\title{find time channel in flowSet/flowFrame}
\usage{
findTimeChannel(obj, excludeChannels = c())
}
\arguments{
\item{obj}{a flowCore::flowFrame or flowCore::flowSet}

\item{excludeChannels}{vector of column names to exclude in the search}
}
\value{
a character, name of the found channel that should be representing
time. If not found, returns NULL.
}
\description{
tries to find a channel in a flowSet/flowFrame that could
be the time channel. First tries to identify a channel name containing the
'time' string, then tries to identify a single monotonically increasing
channel.
}
\examples{

data(OMIP021Samples)

ret <- findTimeChannel(OMIP021Samples[[1]])
ret # "Time"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{removeDoubletsCytoPipeline}
\alias{removeDoubletsCytoPipeline}
\title{remove doublets from a flowFrame, using CytoPipeline custom algorithm}
\usage{
removeDoubletsCytoPipeline(ff, areaChannels, heightChannels, nmads, ...)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{areaChannels}{a character vector containing the name of the 'area type'
channels one wants to use}

\item{heightChannels}{a character vector containing the name of the
'height type' channels one wants to use}

\item{nmads}{a numeric vector with the bandwidth above the ratio allowed, per
channels pair (cells are kept if the ratio between -A channel[i] and
-H channel[i] is smaller than the median ratio + nmad[i] times the median
absolute deviation of the ratios). Default is 4, for all channel pairs.}

\item{...}{additional parameters passed to CytoPipeline::singletGate()}
}
\value{
a flowCore::flowFrame with removed doublets events from the input
}
\description{
Wrapper around CytoPipeline::singletGate().
Can apply the flowStats function subsequently on several channel pairs,
e.g. (FSC-A, FSC-H) and (SSC-A, SSC-H)
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)

suppressWarnings(ff_m <- removeMarginsPeacoQC(x = fsRaw[[2]]))
    
ff_c <-
    compensateFromMatrix(ff_m,
                         matrixSource = "fcs")        

ff_s <-
    removeDoubletsCytoPipeline(ff_c,
                               areaChannels = c("FSC-A", "SSC-A"),
                               heightChannels = c("FSC-H", "SSC-H"),
                               nmads = c(3, 5))
                            
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating.R
\name{singletsGate}
\alias{singletsGate}
\title{Clean doublet events from flow cytometry data}
\usage{
singletsGate(
  ff,
  filterId = "Singlets",
  channel1 = "FSC-A",
  channel2 = "FSC-H",
  nmad = 4,
  verbose = FALSE
)
}
\arguments{
\item{ff}{A flowCore::flowframe that contains flow cytometry data.}

\item{filterId}{the name for the filter that is returned}

\item{channel1}{The first channel that will be used to determine the doublet
events. Default is "FSC-A"}

\item{channel2}{The second channels that will be used to determine the
doublet events. Default is "FSC-H"}

\item{nmad}{Bandwidth above the ratio allowed (cells are kept if their ratio
is smaller than the median ratio + nmad times the median absolute deviation
of the ratios). Default is 4.}

\item{verbose}{If set to TRUE, the median ratio and width will be printed.
Default is FALSE.}
}
\value{
This function returns a flowCore::polygonGate.
}
\description{
will adjust a polygon gate aimed at cleaning doublet events
from the flowFrame.
The main idea is to use the ratio between the two indicated channel as an
indicator and select only the events for which this ratio is 'not too far'
from the median ratio. More specifically, the computed ratio is ch1/(1+ch2).
However, instead of looking at a constant range of this ratio, as is done in
PeacoQC::removeDoublets(), which leads to a semi-conic gate,
we apply a parallelogram shaped gate, by keeping a constant range of channel
2 intensity, based on the target ratio range at the mid value of channel 1.
}
\examples{

data(OMIP021Samples)

# simple example with one single singlets gate filter 
# FSC-A and FSC-H channels are used by default

mySingletsGate <- singletsGate(OMIP021Samples[[1]], nmad = 3)

selectedSinglets <- flowCore::filter(
    OMIP021Samples[[1]],
    mySingletsGate)

ff_l <- flowCore::Subset(OMIP021Samples[[1]], selectedSinglets)

linRange <- c(0, 250000)

ggplotFilterEvents(
    ffPre = OMIP021Samples[[1]],
    ffPost = ff_l,
    xChannel = "FSC-A", xLinearRange = linRange,
    yChannel = "FSC-H", yLinearRange = linRange)

# application of two singlets gates one after the other

singletsGate1 <- singletsGate(OMIP021Samples[[1]], nmad = 3)
singletsGate2 <- singletsGate(OMIP021Samples[[1]],
                              channel1 = "SSC-A",
                              channel2 = "SSC-H",
                              filterId = "Singlets2")

singletCombinedGate <- singletsGate1 & singletsGate2

selectedSinglets <- flowCore::filter(
    OMIP021Samples[[1]],
    singletCombinedGate)

ff_l <- flowCore::Subset(OMIP021Samples[[1]], selectedSinglets)

ggplotFilterEvents(
    ffPre = OMIP021Samples[[1]],
    ffPost = ff_l,
    xChannel = "FSC-A", xLinearRange = linRange,
    yChannel = "FSC-H", yLinearRange = linRange)

ggplotFilterEvents(
    ffPre = OMIP021Samples[[1]],
    ffPost = ff_l,
    xChannel = "SSC-A", xLinearRange = linRange,
    yChannel = "SSC-H", yLinearRange = linRange)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{getAcquiredCompensationMatrix}
\alias{getAcquiredCompensationMatrix}
\title{extract compensation matrix from a flowCore::flowFrame}
\usage{
getAcquiredCompensationMatrix(ff)
}
\arguments{
\item{ff}{a flowCore::flowFrame}
}
\value{
the found compensation matrix
}
\description{
helper function retrieving the compensation matrix stored
in fcs file (if any). It scans the following keywords: $SPILL, $spillover
and $SPILLOVER
}
\examples{

rawDataDir <-
    system.file("extdata", package = "CytoPipeline")
sampleFiles <-
    file.path(rawDataDir, list.files(rawDataDir, pattern = "Donor"))

truncateMaxRange <- FALSE
minLimit <- NULL

# create flowCore::flowSet with all samples of a dataset
fsRaw <- readSampleFiles(
    sampleFiles = sampleFiles,
    whichSamples = "all",
    truncate_max_range = truncateMaxRange,
    min.limit = minLimit)
compensationMatrix <- getAcquiredCompensationMatrix(fsRaw[[2]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{updateMarkerName}
\alias{updateMarkerName}
\title{update marker name of a given flowFrame channel}
\usage{
updateMarkerName(ff, channel, newMarkerName)
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{channel}{the channel for which to update the marker name}

\item{newMarkerName}{the new marker name to be given to the selected channel}
}
\value{
a new flowCore::flowFrame with the updated marker name
}
\description{
: in a flowCore::flowFrame, update the marker name (stored in
'desc' of parameters data) of a given channel.
Also update the corresponding keyword in the flowFrame.
}
\examples{

data(OMIP021Samples)

retFF <- updateMarkerName(OMIP021Samples[[1]],
                          channel = "FSC-A",
                          newMarkerName = "Fwd Scatter-A")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getTransfoParams}
\alias{getTransfoParams}
\title{get tranformation parameters for a specific channel}
\usage{
getTransfoParams(transList, channel)
}
\arguments{
\item{transList}{a flowCore::transformList}

\item{channel}{channel name}
}
\value{
If the transformation exists for the specified channel, and is either
recognized as a logicle transfo or a linear transfo, a list with two slots:
\itemize{
\item $type a character containing the transfo type ('logicle' or 'linear')
\item $params_list a list of named numeric, according to transfo type
}

Otherwise, NULL is returned.
}
\description{
investigates a flowCore::tranformList object to get the type
and parameters of the transformation applying to a specific channel
}
\examples{

data(OMIP021Samples)

# set-up a hybrid transformation list :
# - two channels are logicle-ly transformed with automatic param estimates
# - one channel has explicit logicle transfo with default parameters
# - one channel has linear transformation
# - other channels have no transformation
translist <- flowCore::estimateLogicle(
    OMIP021Samples[[1]],
    c("450/50Violet-A", "525/50Violet-A")
)
translist <- c(
    translist,
    flowCore::transformList(
        "FSC-A",
        flowCore::linearTransform(
            a = 0.1,
            b = 0
       )
    ),
    flowCore::transformList(
        "540/30Violet-A",
        flowCore::logicleTransform()
    )
)

ret1 <- getTransfoParams(translist, channel = "FSC-A")
ret1$type # "linear"
ret1$paramsList # a = 0.1, b = 0.

ret2 <- getTransfoParams(translist, channel = "525/50Violet-A")
ret2$type # "logicle"
ret2$paramsList # a = 0., w = 0.2834, m = 4.5, t = 262143

ret3 <- getTransfoParams(translist, channel = "540/30Violet-A")
ret3$type # "logicle
ret3$paramsList # a = 0., w = 0.5, m = 4.5, t = 262144
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplots.R
\name{ggplotFlowRate}
\alias{ggplotFlowRate}
\title{plot flow rate as a function of time, using ggplot2}
\usage{
ggplotFlowRate(obj, title = "Flow Rate", timeUnit = 100)
}
\arguments{
\item{obj}{a flowCore::flowFrame or flowCore::flowSet}

\item{title}{a title for the graph}

\item{timeUnit}{which time interval is used to calculate "instant" flow rate
(default = 100 ms)}
}
\value{
a ggplot graph
}
\description{
plot flow rate as a function of time, using ggplot2
}
\examples{

data(OMIP021Samples)

# single flowFrame plot
ggplotFlowRate(OMIP021Samples[[1]])

# two flowFrames plot 
ggplotFlowRate(OMIP021Samples)

# single plot with title
ggplotFlowRate(OMIP021Samples[[1]], title = "Test Flow Rate plot")

# explicit time unit
ggplotFlowRate(OMIP021Samples[[1]], timeUnit = 50)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoProcessingStepImplementations.R
\name{readRDSObject}
\alias{readRDSObject}
\title{read RDS object}
\usage{
readRDSObject(RDSFile, ...)
}
\arguments{
\item{RDSFile}{a RDS file containing a R object
object}

\item{...}{other arguments (not used)}
}
\value{
the read R object
}
\description{
wrapper around readRDS, which discards any additional
parameters passed in (...)
}
\examples{

data(OMIP021Samples)

transListPath <- file.path(system.file("extdata", 
                                        package = "CytoPipeline"),
                           "OMIP021_TransList.rds") 

transList <- readRDSObject(transListPath)

ff_c <- compensateFromMatrix(OMIP021Samples[[1]],
                             matrixSource = "fcs")  

ff_t <- applyScaleTransforms(ff_c, transList = transList)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{resetCellIDs}
\alias{resetCellIDs}
\title{reset 'Original_ID' column in a flowframe}
\usage{
resetCellIDs(ff, eventIDs = seq_len(flowCore::nrow(ff)))
}
\arguments{
\item{ff}{a flowCore::flowFrame}

\item{eventIDs}{an integer vector containing the values to be set
in expression matrix, as Original ID's.}
}
\value{
new flowCore::flowFrame containing the amended (or added)
'Original_ID' column
}
\description{
: on a flowCore::flowFrame, reset 'Original_ID' column.
This column can be used in plots comparing the events pre and post gating.
If the 'Original_ID' column already exists, the function replaces
the existing IDs by the user provided ones.
If not, an \code{appendCellID()} is called.
}
\examples{

data(OMIP021Samples)

ff <- appendCellID(OMIP021Samples[[1]])

subsample_ff <- subsample(ff, 100, keepOriginalCellIDs = TRUE)

# re-create a sequence of IDs, ignoring the ones before subsampling
reset_ff <- resetCellIDs(subsample_ff)

}
